


#include "GLfunctions.h"
#include "Player.h"

#include "UtilFunctions.h"
#include "SimFunctions.h"
#include "Defaults.h"

#include "Group.h"
#include "Order.h"

#include "Strategies.h"

// how often the frame rate is recalculated
#define FRAME_RATE_UPDATE_MILLIS 500

// this determines how many objects at most can be selected, hopefully it will never get even close to this!!!!
#define HIT_BUF_SIZE 512


void resize(int width,int height);
void initGL();
void display();
void processKeyboard(unsigned char key, int x, int y);

void processMouse(int button, int state, int x, int y);
void processMouseActiveMotion(int x, int y);
void processMousePassiveMotion(int x, int y);
void processMouseEntry(int state);

void createMenus();
void menu(int value);
void quit();

// this version is timed, and used if fps > 0
void timerCallback(int value);

// this version is untimed, and is executed as often as possible
void idleCallback();

void updateObjects();
void deleteObjects();
void moveObjects(int timeElapsed);
void drawObjects();

void highlightObject(MapObject* objectToHighlight);
void drawDesiredDestination(MapObject* sourceObject);
void determineSelectedObjects(const Vector3D& selectionRectangleStart,const Vector3D& selectionRectangleEnd);

// stuff for selection objects with mouse
void drawObjectsInSelectionMode();

void processSelectionModeHits(GLuint numHits, GLuint buffer[],std::vector<unsigned int> *selectedObjectNames);

GLuint selectionBuffer[HIT_BUF_SIZE];

void initSim();

int windowID;

bool objectJustSelected = false; // true if an object has just been selected by the mouse
bool choosingDestination = false; // true if the mouse is being dragged after an object(s) has been selected
bool selectingObjects = false; // if holding ctrl and selecting multiple objects
bool multipleObjectSelectionMode = false; // if tryign to select many objects with rectangle

Vector3D selectionRectangleStart;
Vector3D selectionRectangleEnd;
Vector3D desiredDestination;

void calculateDesiredDestinationTarget(int x,int y,Vector3D& desiredPos);
unsigned short int linePlaneIntersection(const Vector3D &linePoint1, const Vector3D &linePoint2, const Vector3D &planePoint, const Vector3D &planeNormal, Vector3D &intersection);
	
double frameCounter = 0;
double currentMillisElapsed = 0;
double lastMillisElapsed = 0;
void updateFrameCounter();
double currUpdateMillis = 0;


int main(int argc,char **argv){

	initSim();
	
	if(argc == 2){	// if only one file is specified, it is assumed to be the config file
		configFile = argv[1];
	}else if(argc == 3){ // if two are specified
		configFile = argv[1];
		unitFile = argv[2];
	}else if(argc == 4){
		configFile = argv[1];
		unitFile = argv[2];
		groupFile = argv[3];
	}

	readConfigFile(configFile);
	readUnitFile(unitFile);
	readGroupFile(groupFile);
	
	int i=0;
	PlayerMap::iterator it;
	for(it = playerMap.begin();it != playerMap.end();++it){
		//Player newPlayer(*it->second);
		//print(newPlayer.toString());
		
		if(i==0){
			SDL_CreateThread(massAttack,(void*)it->second);
		}else if(i==1){
			SDL_CreateThread(spreadAttack,(void*)it->second);
		}
		++i;
		
	}
	
	glutInit(&argc, argv);
	initGL();
	
	if(fps == 0.0){
		glutIdleFunc(idleCallback);
	}else{
		glutTimerFunc((unsigned int)(1000.0/fps), timerCallback, 1);
	}
	
	glutMainLoop();

	return 0;
}

void processMouse(int button, int state, int x, int y) {
	objectJustSelected = false;
	
	// if a mouse button is pressed
	if (state == GLUT_DOWN){
		int specialKey = glutGetModifiers();
		
		if(button == GLUT_LEFT_BUTTON){
			enterSelectionMode(x,y,viewDistance,HIT_BUF_SIZE,selectionBuffer);
			
			drawObjectsInSelectionMode();
			
			unsigned int numHits = exitSelectionMode();
			
			// if there are hits, process them
			if (numHits != 0){
				std::vector<unsigned int> selectedObjectNames;
				processSelectionModeHits(numHits,selectionBuffer,&selectedObjectNames);
				
				// the first element of the returned vector is the type of the object, then playerID, then objectID
				Player* objectOwner = getPlayer(selectedObjectNames[1]);
				unsigned int objectID = selectedObjectNames[2];
				
				// if we are holding down CTRL, then keep track of the previously selected objects, else, get rid of the previous ones
				if(specialKey != GLUT_ACTIVE_CTRL){
					selectedObjects.clear();
				}
				selectedObjects.insert(objectOwner->getObject(objectID)); // now we add the selected object to the list
				
				objectJustSelected = true;
			}else if(specialKey != GLUT_ACTIVE_CTRL){ // if we don't select anything, and are not holding ctrl, then deselect everything
				selectedObjects.clear();
				calculateDesiredDestinationTarget(x,y,selectionRectangleStart);
				selectingObjects = true;
			}
		}else if(button == GLUT_RIGHT_BUTTON){
			calculateDesiredDestinationTarget(x,y,desiredDestination);
			MapObjectSet::iterator it;
			for(it=selectedObjects.begin(); it != selectedObjects.end(); ++it){
				Group* currGroup = dynamic_cast<Group*> (*it);
				if(currGroup != NULL){
					currGroup->setOrder(new Order("MOVE",Vector3D(desiredDestination.x,desiredDestination.y,desiredDestination.z),currGroup->getMaxSpeed()),NULL,NULL);
				}
			}
		}
	}else if(state == GLUT_UP){
		if(choosingDestination){ // if we were choosing a destination, process the target destination
			MapObjectSet::iterator it;
			for(it=selectedObjects.begin(); it != selectedObjects.end(); ++it){
				Group* currGroup = dynamic_cast<Group*> (*it);
				if(currGroup != NULL){
					currGroup->setOrder(new Order("MOVE",Vector3D(desiredDestination.x,desiredDestination.y,desiredDestination.z),currGroup->getMaxSpeed()),NULL,NULL);
					
					double time = getNextCollisionTime(playerMap,currUpdateMillis);
					print("estimated earliest collision time: " + convertToString(time));
					
					double time2 = getNextOrderCompletionTime(playerMap,currUpdateMillis);
					print("estimated earliest completion time: " + convertToString(time2));
				}
			}
			
		}
		choosingDestination = false;
		selectingObjects = false;
	}
}

void processMouseActiveMotion(int x,int y) {
	 
	if(objectJustSelected){ // this should only happen if an object was clicked on, and the mouse has yet to be released from that click
		choosingDestination = true;
		calculateDesiredDestinationTarget(x,y,desiredDestination);
	}else if(selectingObjects){
		calculateDesiredDestinationTarget(x,y,selectionRectangleEnd);
		determineSelectedObjects(selectionRectangleStart,selectionRectangleEnd);
		
	}
}
void processMousePassiveMotion(int /*x*/,int /*y*/) {
	
}
void processMouseEntry(int /*state*/) {

}

void calculateDesiredDestinationTarget(int x,int y,Vector3D& desiredPos){
	GLint viewport[4];
	GLdouble modelMatrix[16];
	GLdouble projMatrix[16];
	GLdouble posX,posY,posZ;
	Vector3D linePoint1, linePoint2;
	
	glGetIntegerv(GL_VIEWPORT,viewport);
	glGetDoublev(GL_PROJECTION_MATRIX,projMatrix);	
	glGetDoublev(GL_MODELVIEW_MATRIX,modelMatrix);
	
	gluUnProject(x,viewport[3] - y,0.5,modelMatrix,projMatrix,viewport,&posX,&posY,&posZ);
	linePoint1 = Vector3D(posX,posY,posZ);
	
	gluUnProject(x,viewport[3] - y,1.5,modelMatrix,projMatrix,viewport,&posX,&posY,&posZ);
	linePoint2 = Vector3D(posX,posY,posZ);
	
	Vector3D planeNormal = Vector3D(0,0,1);
	Vector3D planePoint = Vector3D(0,0,0);
	
	linePlaneIntersection(linePoint1,linePoint2,planePoint,planeNormal,desiredPos);
	
}

void determineSelectedObjects(const Vector3D& selectionRectangleStart,const Vector3D& selectionRectangleEnd){
	selectedObjects.clear();
	
	PlayerMap::iterator it;
	for(it = playerMap.begin();it != playerMap.end();++it){
		(it->second)->addContainedObjectsToList(selectionRectangleStart,selectionRectangleEnd,selectedObjects);	
	}
}


// this function is called whenever we need to resize the window...

void resize(int width, int height){
	if (height == 0) height = 1; // just to make sure we don't divide by zero

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	glViewport(0,0,width,height);

	// note we divide our width by our height to get the aspect ratio
	gluPerspective(45.0, (double)width / (double)height,1.0,viewDistance);

	glMatrixMode(GL_MODELVIEW);
}

void initGL(){
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH | GLUT_ALPHA); // more flags online if needed!!!!
	
	glutInitWindowSize(windowWidth,windowHeight);
	glutInitWindowPosition(windowX,windowY);
	windowID = glutCreateWindow("AI Simulator");
	
	glClearColor(0.0,0.0,0.0,0.0); // start with a black screen
	
	// the next two lines are used for alpha blending
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	
	glTranslatef(cameraPos.x,cameraPos.y,cameraPos.z);
	
	createMenus();
	
	glutDisplayFunc(display);
	glutKeyboardFunc(processKeyboard);
	
	glutReshapeFunc(resize);
	
	// mouse processing callbacks
	glutMouseFunc(processMouse);
	glutMotionFunc(processMouseActiveMotion);
	glutPassiveMotionFunc(processMousePassiveMotion);
	glutEntryFunc(processMouseEntry);
}

// this function is the main one used for animation...
// it is called as often as the FPS are set to... unless somethign is hogging the CPU ;)

void timerCallback(int /*value*/){
		
	updateObjects();

	glutTimerFunc ((unsigned int)(1000.0/fps), timerCallback, 1);
}

// all we do here is redraw as fast as possible...
// this function is only called if the FPS were set to 0
void idleCallback(){
	glutPostRedisplay();
	//updateObjects();
}

void updateObjects(){
	//int currUpdateMillis = lastMoveUpdateMillis;
	double lastMoveUpdateMillis = currUpdateMillis;
	currUpdateMillis = (double)glutGet(GLUT_ELAPSED_TIME);
	moveObjects(playerMap,currUpdateMillis - lastMoveUpdateMillis);
	blah(playerMap,currUpdateMillis);
	calculateCollisions(playerMap,currUpdateMillis);
	
	// now we delete any objects that have < 0 hitpoints
	PlayerMap::iterator it;
	for(it = playerMap.begin();it != playerMap.end(); ++it){
		std::vector<MapObject*> deletedObjects; // get the addresses of all the deleted objects...
		(it->second)->checkForDestroyedObjects(deletedObjects);
		
		for(unsigned int i=0;i<deletedObjects.size();i++){
			MapObjectSet::iterator it2 = selectedObjects.find(deletedObjects[i]);
			if(it2 != selectedObjects.end()){
				selectedObjects.erase(it2);
			}
			delete deletedObjects[i];
		}
	}	
	glutPostRedisplay();
}
void updateFrameCounter(){
	frameCounter++;
	currentMillisElapsed=(double)glutGet(GLUT_ELAPSED_TIME);
	if ((currentMillisElapsed-lastMillisElapsed) > FRAME_RATE_UPDATE_MILLIS) {
		sprintf(fpsString,"FPS:%4.2f",frameCounter*1000.0/(double)(currentMillisElapsed-lastMillisElapsed));
		lastMillisElapsed = currentMillisElapsed;
		frameCounter = 0;
	}
}

void display(void){
	glClear(GL_COLOR_BUFFER_BIT);
	
	updateFrameCounter();
	
	// draw all the objects...
	drawObjects();

	if(selectingObjects){
		drawRectangle(selectionRectangleStart,selectionRectangleEnd,yellow);
	}
	
	MapObjectSet::iterator it;
	for(it = selectedObjects.begin();it!=selectedObjects.end();++it){
		highlightObject((*it));
		drawDesiredDestination((*it));
	}
	
	draw2DText();
	
	glutSwapBuffers(); // finally switch buffers, since we were doing double buffering
}

void highlightObject(MapObject* objectToHighlight){
	if(objectToHighlight == NULL || unitSemaphore) return;
	
	double objectExtents = objectToHighlight->getExtents();
	Group* group = dynamic_cast<Group*> (objectToHighlight);
	if(group != NULL){
		objectExtents+= group->getMaxAttackRange();
	}
	drawSphere(objectToHighlight->getPosition(),objectExtents+(objectExtents*0.2),yellow.transparent(95));
}

void drawDesiredDestination(MapObject* sourceObject){
	if(sourceObject == NULL || choosingDestination == false || unitSemaphore) return;
	
	drawArrow(sourceObject->getPosition(),desiredDestination,3.0,yellow.transparent(30));
}

void processKeyboard(unsigned char key, int x, int y){
	switch (key){
		// 27 is the escape key...
		case 27:
			quit();
			break;
		case 'n':
			advanceTime(playerMap);
			glutPostRedisplay();
			break;
		case 'f':
			glRotatef(cameraSpeed, 1.0, 0.0, 0.0); // rotate up
			pitch+=cameraSpeed;
			break;
		case 'r':
			glRotatef(-cameraSpeed, 1.0, 0.0, 0.0); // rotate down
			pitch-=cameraSpeed;
			break;
		case 'q':
			glRotatef(cameraSpeed, 0.0, 1.0, 0.0); // rotate left
			yaw+=cameraSpeed;
			break;
		case 'e':
			glRotatef(-cameraSpeed, 0.0, 1.0, 0.0); // rotate right
			yaw-=cameraSpeed;
			break;
		case 'c':
		glRotatef(cameraSpeed,0.0,0.0,1.0); // roll left
			roll+=cameraSpeed;
			break;
		case 'v':
		glRotatef(-cameraSpeed,0.0,0.0,1.0); // roll right
			roll-=cameraSpeed;
			break;
		case 'w':
			glTranslatef(0.0,-cameraSpeed,0.0); //move up
			cameraPos.y-=cameraSpeed;
			break;
		case 's':
			glTranslatef(0.0,cameraSpeed,0.0); // move down
			cameraPos.y+=cameraSpeed;
			break;
		case 'd':
			glTranslatef(-cameraSpeed,0.0,0.0); // move right
			cameraPos.x-=cameraSpeed;
			break;
		case 'a':
			glTranslatef(cameraSpeed,0.0,0.0); //move left
			cameraPos.x+=cameraSpeed;
			break;
		case 'z':
			glTranslatef(0.0,0.0,cameraSpeed); //zoom in
			cameraPos.z+=cameraSpeed;
			break;
		case 'x':
			glTranslatef(0.0,0.0,-cameraSpeed); //zoom out
			cameraPos.z-=cameraSpeed;
			break;
		case '-':
			fps--;
			if(fps <= 0.0) fps = 1.0;
			break;
		case '+':
			fps++;
			break;
		case 'p':
			paused = !paused;
			break;
		case 'i':
			displayObjectInfo = !displayObjectInfo;
			break;
		case 't':
			draw3DTextInfo = !draw3DTextInfo;
			break;
	}
	processMouseActiveMotion(x,y); // this should be called, since technically, the mouse cursor is now in a different relative position worldwise
	
	display();
}

void createMenus(){
	
	// this is the submenu for objects...
	int subMenuID = glutCreateMenu(menu);

	glutAddMenuEntry("Save current state",1);
	glutAddMenuEntry("Load saved state",2);
	glutAddMenuEntry("Load starting state",3);
	
	// this is the main menu
	glutCreateMenu(menu);
	
	glutAddSubMenu("Game State",subMenuID);

	glutAddMenuEntry("Quit",0);
	
	glutAttachMenu(GLUT_MIDDLE_BUTTON); // right click brings up the menu
}
 
// here actions are performed, based on what menu item was selected
void menu(int value){
	if(value == 0){
		quit();
	}else if(value == 1){
		// save units to a file specified by unitSaveFile
		saveGroupFile(groupSaveFile);
	}else if(value == 2){
		// load the units specified in unitSaveFile
		deleteObjects(); // first get rid of our current ones
		readGroupFile(groupSaveFile);
	}else if(value == 3){
		// load the original units file,specified by unitFile
		deleteObjects();
		readGroupFile(groupFile);
	}
	glutPostRedisplay();
}

// cleanup and exit
void quit(){
	glutDestroyWindow(windowID);
	deleteObjects();
	if(debug){
		fprintf(stderr,"Program terminated normally.\n");
	}
	
	exit(0);
}

void initSim(){
	//createNewPlayer(0); // by default, create a player with ID 0
	
}

void drawObjects(){
	if(unitSemaphore) return;
	
	PlayerMap::iterator it;
	for(it = playerMap.begin(); it != playerMap.end(); ++it){
		(it->second)->drawObjects();
	}
	
}
void drawObjectsInSelectionMode(){
	if(unitSemaphore) return;

	PlayerMap::iterator it;
	for(it = playerMap.begin(); it != playerMap.end(); ++it){
		(it->second)->drawObjectsInSelectionMode();
	}
	
}


void processSelectionModeHits(GLuint numHits, GLuint buffer[],std::vector<unsigned int> *selectedObjectNames){
	GLuint names = 0xFFFFFFFF;
	GLuint numberOfNames = 0;
	
	GLuint minZ = 0xFFFFFFFF; // need to know the minimum Z value, in case objects are on top of each other
	
	GLuint *minZPtr = NULL;
	GLuint *minZPtrNames = NULL; // since an entry can have up to 64 names...in our case, each entry only has one name
	
	minZPtr = (GLuint *) buffer; // this is the buffer that stores the hits that were made(if any)
	
	for (unsigned int i = 0; i < numHits; i++){ // go through all the hits
		names = *minZPtr; // grab the first name pointer
		minZPtr++; // go to the next entry in buffer
		if (*minZPtr < minZ) { // check if the second entry in buffer(the z-value) was less than the minimum so far..
			numberOfNames = names; // if it was, record how many names this object has
			minZ = *minZPtr; // record the new minimum Z
			minZPtrNames = minZPtr+2; // record the location of the start of the names in the buffer
		}
		
		minZPtr += names+2; //go to the next entry in the buffer
	}
	
	minZPtr = minZPtrNames;
	for (unsigned int j = 0; j < numberOfNames; j++,minZPtr++) {
		selectedObjectNames->push_back(*minZPtr);
		
		//fprintf (stderr,"%d ", *minZPtr);
	}
	//fprintf (stderr,"\n");
}


void deleteObjects(){
	// just to be paranoid, make sure noone does anythign with these vectors until they are completely wiped
	unitSemaphore = true;
	
	PlayerMap::iterator it;
	for(it = playerMap.begin(); it != playerMap.end(); ++it){
		delete (it->second);
	}
	playerMap.clear();
	selectedObjects.clear();
	unitSemaphore = false;
}

unsigned short int linePlaneIntersection(const Vector3D &linePoint1, const Vector3D &linePoint2, const Vector3D &planePoint, const Vector3D &planeNormal, Vector3D &intersection){
	Vector3D u = linePoint2 - linePoint1;
	Vector3D w = linePoint1 - planePoint;
	
	double d = planeNormal.dot(u);
	double n = -planeNormal.dot(w);
	
	if (fabs(d) < 1e-6) { // segment parallel to plane
		if (n == 0) 
			return 2;         // segment in plane
		else
			return 1;         // no intersection
	}
	
	intersection = linePoint1 + (n/d) * u;    
	return 0;
}

// the end

