/** @file GfxEmitter.C
    @see GfxEmitter.H

    $Id: GfxEmitter.C,v 1.40 2006/04/03 21:32:15 orts_mburo Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/gfxclient/src/GfxEmitter.C,v $
*/

// This is an ORTS file
// (c) Keith Yerex
// (c) Tim Furtak
// (c) Erik Benner
// licensed under the GPL


#include "GfxEmitter.H"
#include "GfxModule.H"
#include "Image.H"

using namespace std;

//===================================================================

void GfxEmitter::beginDraw()
{
  glEnable(GL_BLEND);
  glDepthMask(GL_FALSE);
  glEnable(GL_CULL_FACE);
  glCullFace(GL_BACK);

  glActiveTextureARB(GL_TEXTURE0);
  glEnable(GL_TEXTURE_2D);
  
  glActiveTextureARB(GL_TEXTURE1);
  glEnable(GL_TEXTURE_2D);
  glDisable(GL_TEXTURE_GEN_S);
  glDisable(GL_TEXTURE_GEN_T);

  glColor4f(1,1,1,1);
}

//-------------------------------------------------------------------

void GfxEmitter::endDraw()
{
  glDisable(GL_BLEND);
  glDepthMask(GL_TRUE);

  glActiveTextureARB(GL_TEXTURE0);
}

//===================================================================

void Particle::update(real4 dt)
{
  pos += vel*dt+(acc+vel*fric)*dt*dt*0.5;
  vel += (acc+vel*fric)*dt;
  age += dt;
}

// Pass the global acceleration instead of storing it per particle
bool ParticleLite::update(real4 dt, const Vec3<real4> &acc)
{
  pos += vel*dt;
  vel += acc*dt;
  age += dt;
  if (age >= life) {
    life = 0.0f;
    return false;
  } else
    return true;
}

//===================================================================

GfxPListEmitter::GfxPListEmitter()
{
  first = last = 0;
}

//-------------------------------------------------------------------

void GfxPListEmitter::reorder()
{
  sint4 nmax = int(ceil(rate) * ceil(life));
  if (nmax <= max_particles) return;

  Vector<Particle> old_p;

  for (int i = first; i != last; i = i % max_particles) {
    int e = i < last ? last : max_particles;
    for (; i < e; ++i) {
      old_p.push_back(list[i]);
    }
  }

  list.resize(nmax);
  assert(list.size() >= old_p.size());
  FORU (i, old_p.size()) {
    list[i] = old_p[i];
  }
  first = 0;
  last = old_p.size() % nmax;
  if (last == first && old_p.size() != 0) first = (first+1) % nmax;
  max_particles = nmax;
}

//-------------------------------------------------------------------

void GfxPListEmitter::set_rate(real4 r)
{
  rate = r;
  reorder();
}

//-------------------------------------------------------------------

Particle& GfxPListEmitter::newparticle()
{
  Particle *p = &list[last];
  last = (last+1) % max_particles;
  if (last == first) first = (first+1) % max_particles;
  return *p;
}

//===================================================================

GfxBoxEmitter::GfxBoxEmitter(GfxModule &gfxm_, real4 rate_, real4 life_, real4 time_limit,
                             Vec3<real4> p, Vec3<real4> pr, Vec3<real4> v, Vec3<real4> vr, Vec3<real4> accel,
                             real4 f, real4 fr, real4 p_width, real4 p_height,
                             GLuint texid, GLuint colid)
{
  set_gfxm(gfxm_);
  dying = false;
  rate = rate_;
  life = life_;
  limit = time_limit;
  if (rate < 0) rate = 0;
  if (life < 0) life = 0;
  max_particles = int(ceil(rate) * ceil(life));

  if (max_particles < 1) max_particles = 1;

  pos = p;
  p_rand = pr;
  vel = v;
  v_rand = vr;
  acc = accel;

  fric = f;
  f_rand = fr;

  w = p_width;
  h = p_height;
  billboard_tex = texid;
  color_tex = colid;
  blend = STANDARD;

  pose.center = Vec3<real4>(0, 0, 0);

  list.resize(max_particles);

  unlimited = limit <= 0; // initial setting

  create = 0;
  enabled = true;

  radius = life * (vel + v_rand).length();

  glBindTexture(GL_TEXTURE_2D, colid);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
}

//-------------------------------------------------------------------

void GfxBoxEmitter::spawn_particle()
{
  Particle &p = newparticle();

  p.pos = pose.center + pos + Vec3<real4>::vrand(gfxm->rand)*p_rand;

#if 1
  // adjust particle velocity based on object rotation
  Vec3<real4> v = vel + Vec3<real4>::vrand(gfxm->rand) * v_rand;
  p.vel.x = (v.x * cos_ang) + (v.z * sin_ang);
  p.vel.y = v.y;
  p.vel.z = (v.x * -sin_ang) + (v.z * cos_ang);
#else
  p.vel = vel + Vec3<real4>::vrand(gfxm->rand) * v_rand;
#endif
  p.acc = acc;
  p.fric = fric + gfxm->rand.rand_real8() * f_rand - f_rand / 2;
  p.age = 0.0;
  p.life = life;
}

//-------------------------------------------------------------------

void GfxBoxEmitter::update(real4 dt)
{
  GfxObject::update(dt);

  if (first == last && dying == true) {
    dead = true;
    return;
  }

  aabb_o.x = pose.center.x - radius;
  aabb_o.y = pose.center.y - radius;
  aabb_o.z = pose.center.z - radius;
  aabb_d.x = aabb_d.y = aabb_d.z = 2*radius;
  
  for (int i = first; i != last; i = i % max_particles) {
    int e = i < last ? last : max_particles;
    for (; i < e; ++i) {
      list[i].update(dt);
      // assumes all particles live the same amount of time
      if (list[i].age >= list[i].life) {
        first = (i+1) % max_particles;
      }
    }
  }

  if ((unlimited || limit>0) && enabled) {
    if (limit > 0) limit -= dt;

    double ang = pose.y_rot * M_PI / 180.0;
    cos_ang = cos(ang);
    sin_ang = sin(ang);
    create += rate*dt;
    while (create >= 1) {
      spawn_particle();
      create--;
    }
  }
}

//===================================================================

GfxSprayEmitter::GfxSprayEmitter(GfxModule &gfxm, real4 rate_, real4 life_, real4 time_limit,
                                 const Vec3<real4> &p, real4 min_lat, real4 max_lat,
                                 real4 offset_, real4 o_rand_, real4 speed_, real4 s_rand_,
                                 const Vec3<real4> &accel, real4 f, real4 fr,
                                 real4 ww, real4 hh, GLuint texid, GLuint colid)
{
  set_gfxm(gfxm);
  dying = false;
  rate = rate_;
  life = life_;
  limit = time_limit;
  if (rate < 0) rate = 0;
  if (life < 0) life = 0;
  max_particles = int(ceil(rate) * ceil(life));

  if (max_particles < 1) max_particles = 1;

  pos = p;
  min_ang = min_lat * M_PI / 180.0;
  max_ang = max_lat * M_PI / 180.0;
  offset = offset_;
  o_rand = o_rand_;
  speed = speed_;
  s_rand = s_rand_;

  acc = accel;

  fric = f;
  f_rand = fr;

  w = ww;
  h = hh;
  billboard_tex = texid;
  color_tex = colid;
  blend = STANDARD;

  pose.center = Vec3<real4>(0, 0, 0);

  list.resize(max_particles);

  unlimited = limit <= 0; // initial setting

  create = 0;
  enabled = true;

  radius = life * (speed + s_rand);

  glBindTexture(GL_TEXTURE_2D, colid);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
}

//-------------------------------------------------------------------

void GfxSprayEmitter::spawn_particle()
{
  Particle &p = newparticle();

  // adjust particle velocity based on object rotation
  real8 lat = gfxm->rand.rand_real8() * (max_ang - min_ang) + min_ang;
  real8 theta = gfxm->rand.rand_real8() * 2 * M_PI;
  real8 s = speed + gfxm->rand.rand_real8() * s_rand;
  real8 d = offset + gfxm->rand.rand_real8() * o_rand;

  p.pos = pose.center + pos;
  p.pos.x += cos(lat) * cos(theta) * d;
  p.pos.y += sin(lat) * d;
  p.pos.z += cos(lat) * sin(theta) * d;

  p.vel.x = cos(lat) * cos(theta) * s;
  p.vel.y = sin(lat) * s;
  p.vel.z = cos(lat) * sin(theta) * s;

  p.acc = acc;
  p.fric = fric + gfxm->rand.rand_real8() * f_rand - f_rand / 2;
  p.age = 0.0;
  p.life = life;
}

//-------------------------------------------------------------------

void GfxSprayEmitter::update(real4 dt)
{
  GfxObject::update(dt);

  if (first == last && dying == true) {
    dead = true;
    return;
  }

  aabb_o.x = pose.center.x - radius;
  aabb_o.y = pose.center.y - radius;
  aabb_o.z = pose.center.z - radius;
  aabb_d.x = aabb_d.y = aabb_d.z = 2*radius;
  
  for (int i = first; i != last; i = i % max_particles) {
    int e = i < last ? last : max_particles;
    for (; i < e; ++i) {
      list[i].update(dt);
      // assumes all particles live the same amount of time
      if (list[i].age >= list[i].life) {
        first = (i+1) % max_particles;
      }
    }
  }

  if ((unlimited || limit>0) && enabled) {
    if (limit > 0) limit -= dt;

    create += rate*dt;
    while (create >= 1) {
      spawn_particle();
      create--;
    }
  }
}

//===================================================================

void GfxPListEmitter::draw_emitters()
{
  glPushMatrix();

  if (draw_type == STREAKS) draw_as_streaks();
  else draw_as_points();

  glTranslatef(pose.center.x, pose.center.y, pose.center.z);
  glRotatef(pose.y_rot, 0, 1, 0);

  //  FORALL (sub_objects, i) (*i)->draw_emitters();
  FORALL (emitters, i) (*i)->draw_emitters();

  glPopMatrix();
}

//-------------------------------------------------------------------

void GfxPListEmitter::draw_as_points()
{
  if (first == last) return;

  if (blend == ADDITIVE)
    glBlendFunc(GL_SRC_ALPHA,GL_ONE);

  else if (blend == STANDARD)
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

  Mat4<real4> m;
  glGetModelview(m);
  m = m.inv_quick();

  Vec3<real4> p1 = m * Vec4<real4>(-w, h,0,0);
  Vec3<real4> p2 = m * Vec4<real4>( w, h,0,0);
  Vec3<real4> p3 = m * Vec4<real4>( w,-h,0,0);
  Vec3<real4> p4 = m * Vec4<real4>(-w,-h,0,0);
  
  glActiveTextureARB(GL_TEXTURE0);
  glBindTexture(GL_TEXTURE_2D, billboard_tex);
  
  glActiveTextureARB(GL_TEXTURE1);
  glBindTexture(GL_TEXTURE_2D, color_tex);

  glBegin(GL_QUADS);

  for (int i = first; i != last; i = i % max_particles) {
    int e = i < last ? last : max_particles;
    for (; i < e; ++i) {
      real4 alpha_pos = list[i].age / list[i].life;
      glMultiTexCoord2fARB(GL_TEXTURE1,alpha_pos,0.5);
      glMultiTexCoord2fARB(GL_TEXTURE0, 1,0);  glVertex3fv((list[i].pos+p4).ptr());
      glMultiTexCoord2fARB(GL_TEXTURE0, 0,0);  glVertex3fv((list[i].pos+p3).ptr());
      glMultiTexCoord2fARB(GL_TEXTURE0, 0,1);  glVertex3fv((list[i].pos+p2).ptr());
      glMultiTexCoord2fARB(GL_TEXTURE0, 1,1);  glVertex3fv((list[i].pos+p1).ptr());
    }
  }

  glEnd();
}

//-------------------------------------------------------------------

// draw each particle as n triangles stretched along the velocity vector

void GfxPListEmitter::draw_as_streaks()
{
  if (first == last) return;

  glDisable(GL_CULL_FACE);

  if (blend == ADDITIVE)
    glBlendFunc(GL_SRC_ALPHA,GL_ONE);
  else if (blend == STANDARD)
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

  Mat4<real4> m;
  glGetModelview(m);
  m.invert();
  //m = m.inv_quick();

  Vec3<real4> p1 = m * Vec4<real4>(-w, 0, 0, 0);
  Vec3<real4> p2 = m * Vec4<real4>( w, 0, 0, 0);
  Vec3<real4> p3 = m * Vec4<real4>( 0, w, 0, 0);
  Vec3<real4> p4 = m * Vec4<real4>( 0,-w, 0, 0);

  real4 s = h; // scale factor

  glActiveTextureARB(GL_TEXTURE0);
  glBindTexture(GL_TEXTURE_2D,billboard_tex);

  glActiveTextureARB(GL_TEXTURE1);
  glBindTexture(GL_TEXTURE_2D,color_tex);

  glBegin(GL_TRIANGLES);

  for (int i = first; i != last; i = i % max_particles) {
    int e = i < last ? last : max_particles;
    for (; i < e; ++i) {
      real4 alpha_pos = list[i].age / list[i].life;
      glMultiTexCoord2fARB(GL_TEXTURE1,alpha_pos,0.5);
      glMultiTexCoord2fARB(GL_TEXTURE0, 0.5, 0.5);  glVertex3fv(list[i].pos.ptr());
      glMultiTexCoord2fARB(GL_TEXTURE0, 1.0, 1.0);  glVertex3fv((list[i].pos+list[i].vel*s+p1).ptr());
      glMultiTexCoord2fARB(GL_TEXTURE0, 1.0, 0.0);  glVertex3fv((list[i].pos+list[i].vel*s+p2).ptr());
      
      glMultiTexCoord2fARB(GL_TEXTURE1,alpha_pos,0.5);
      glMultiTexCoord2fARB(GL_TEXTURE0, 0.5, 0.5);  glVertex3fv(list[i].pos.ptr());
      glMultiTexCoord2fARB(GL_TEXTURE0, 1.0, 1.0);  glVertex3fv((list[i].pos+list[i].vel*s+p3).ptr());
      glMultiTexCoord2fARB(GL_TEXTURE0, 1.0, 0.0);  glVertex3fv((list[i].pos+list[i].vel*s+p4).ptr());
    }
  }
  glEnd();
  
  glEnable(GL_CULL_FACE);
}

//===================================================================

GfxLightCorona::GfxLightCorona(GfxModule &gfxm_, real4 size, real4 speed, GLuint texid, GLuint colid)
{
  set_gfxm(gfxm_);
	
  this->billboard_tex = texid;
  this->color_tex = colid;
  this->size = size;
  this->fade = speed;
  this->age = 0.0f;
  
  this->enabled = true;
}

void GfxLightCorona::draw()
{
  if(!enabled) return;

  Mat4<real4> m;
  glGetModelview(m);
  m = m.inv_quick();

  Vec3<real4> p1 = m * Vec4<real4>(-size, size,0,0);
  Vec3<real4> p2 = m * Vec4<real4>( size, size,0,0);
  Vec3<real4> p3 = m * Vec4<real4>( size,-size,0,0);
  Vec3<real4> p4 = m * Vec4<real4>(-size,-size,0,0);
  
  glColor4f(1, 1, 1, 1);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  
  glActiveTextureARB(GL_TEXTURE0);
  glBindTexture(GL_TEXTURE_2D, billboard_tex);
  
  glActiveTextureARB(GL_TEXTURE1);
  glBindTexture(GL_TEXTURE_2D, color_tex);
  
  real4 colour = age * fade;
  
  glBegin(GL_QUADS);
    glMultiTexCoord2fARB(GL_TEXTURE1, colour, 0.5);
    glMultiTexCoord2fARB(GL_TEXTURE0, 1,0); glVertex3fv(p4.ptr());
    glMultiTexCoord2fARB(GL_TEXTURE0, 0,0); glVertex3fv(p3.ptr());
    glMultiTexCoord2fARB(GL_TEXTURE0, 0,1); glVertex3fv(p2.ptr());
    glMultiTexCoord2fARB(GL_TEXTURE0, 1,1); glVertex3fv(p1.ptr());
  glEnd();
}

void GfxLightCorona::update(real4 dt)
{
  GfxObject::update(dt);
	
  age += dt;
}

//===================================================================

GLuint GfxFireEmitter::flame_tex = 0;
GLuint GfxFireEmitter::spark_tex = 0;
GLuint GfxFireEmitter::fire_tex  = 0;
GLuint GfxFireEmitter::cloud_tex = 0;
GLuint GfxFireEmitter::smoke_tex = 0;

GfxFireEmitter::GfxFireEmitter(GfxModule &gfxm_, real4 size_, real4 speed_)
{
  set_gfxm(gfxm_);
	
  size = max(size_, 0.001f);
  speed = max(speed_, 0.001f);
  target = static_cast<real4>(gfxm->rand.rand_real8());
  
  theta = 90.0f;
  phi = 90.0f;
  spread = 45.0f;
  acc = Vec3<real4>(0.15f, 0.0f, 0.0f);
  
  num_ptc = 0;
  new_ptc = 0;
  
  max_spark = static_cast<sint4>(speed * (1.f + (speed * 0.05f)));
  spark.resize(max_spark);
  FORALL(spark, i) i->life = 0.0f;
   
  // If we're the first to initialize, load the static fire textures
  string path = sanitize_path(gfxm->opt.path + PATH_SEP);
  Image texture;
  if (!flame_tex)
  {
    if (!texture.loadTGA(path + "explosion/fire_flame.tga"))
      cout << "Error loading flame texture" << endl;
    flame_tex = texture.createTexture(Image::BILINEAR);
  }
  if (!spark_tex)
  {
    if (!texture.loadTGA(path + "explosion/fire_spark.tga"))
      cout << "Error loading spark texture" << endl;
    spark_tex = texture.createTexture(Image::BILINEAR);
  }
  if (!fire_tex)
  {
    if (!texture.loadTGA(path + "explosion/fire_color.tga"))
      cout << "Error loading fire colour texture" << endl;
    fire_tex = texture.createTexture(Image::LINEAR);
  }
  
}

void GfxFireEmitter::draw()
{
  glPushMatrix();
  glTranslatef(pose.center.x, pose.center.y, pose.center.z);
  glRotatef(pose.y_rot, 0, 1, 0);
  glRotatef(pose.x_rot, 0, 0, 1);
  glRotatef(pose.y_off, 0, 1, 0);
  glScalef(pose.scale, pose.scale, pose.scale);


  Mat4<real4> m;
  glGetModelview(m);
  m = m.inv_quick();
 
  Vec3<real4> p1 = m * Vec4<real4>(-1, 1,0,0);
  Vec3<real4> p2 = m * Vec4<real4>( 1, 1,0,0);
  Vec3<real4> p3 = m * Vec4<real4>( 1,-1,0,0);
  Vec3<real4> p4 = m * Vec4<real4>(-1,-1,0,0);
  
  // Render pulsing flame - base of the fire
  float pulse = (size + (target*size/5.f)) / 3.f;
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  
  glActiveTextureARB(GL_TEXTURE1);
  glDisable(GL_TEXTURE_2D);
  
  glActiveTextureARB(GL_TEXTURE0);
  glBindTexture(GL_TEXTURE_2D, flame_tex);
  
  // Temporary translation to line up the fire with the extractor tower
  //glTranslatef(-0.135, -0.465, 0.30);
  
  glPushMatrix();
  glColor4f(1.f, 1.f, 1.f, 0.5f);
  glScalef(pulse, pulse, pulse);
  glBegin(GL_QUADS);
    glMultiTexCoord2fARB(GL_TEXTURE0, 1,0); glVertex3fv(p4.ptr());
    glMultiTexCoord2fARB(GL_TEXTURE0, 0,0); glVertex3fv(p3.ptr());
    glMultiTexCoord2fARB(GL_TEXTURE0, 0,1); glVertex3fv(p2.ptr());
    glMultiTexCoord2fARB(GL_TEXTURE0, 1,1); glVertex3fv(p1.ptr());
  glEnd();
  glPopMatrix();
  
  // Render emitting sparks 
  glBlendFunc(GL_SRC_ALPHA,GL_ONE); // Additive blending
  
  glBindTexture(GL_TEXTURE_2D, spark_tex);
  glActiveTextureARB(GL_TEXTURE1);
  
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, fire_tex);
  
  real4 percent;
  real4 scale;
  
  FORALL(spark, i) {
    if (i->life <= 0.0f) continue;
    percent = i->age / i->life;
    scale = (1.0f - percent) * size/1.5f;
    
    glPushMatrix();
    glTranslatef(i->pos.x, i->pos.y, i->pos.z);
    glScalef(scale, scale, scale);

    glBegin(GL_QUADS);
      glMultiTexCoord2fARB(GL_TEXTURE1, percent, 0.5);
      glMultiTexCoord2fARB(GL_TEXTURE0, 0,1); glVertex3fv(p4.ptr());
      glMultiTexCoord2fARB(GL_TEXTURE0, 1,1); glVertex3fv(p3.ptr());
      glMultiTexCoord2fARB(GL_TEXTURE0, 1,0); glVertex3fv(p2.ptr());
      glMultiTexCoord2fARB(GL_TEXTURE0, 0,0); glVertex3fv(p1.ptr());
    glEnd();
    glPopMatrix();
  }

  glPopMatrix();
}

void GfxFireEmitter::update(real4 dt)
{
  GfxObject::update(dt);

  target -= dt;
  if (target <= 0) target = static_cast<real4>(gfxm->rand.rand_real8());
  
  new_ptc += dt * speed;
  //cout << (dt * speed) << endl;
  
  if ((num_ptc += uint4(new_ptc)) > max_spark) num_ptc = max_spark;
  
  //cout  << "num_ptc = " << num_ptc << endl;
  
  // Update positions of each spark
  FORALL(spark, i) {
    if (i->life > 0.0f) {
      // If the particle just died
      if (!i->update(dt, acc)) num_ptc--;
    } else if (enabled && new_ptc >= 1.0f) {
      new_ptc--;
      spawn(*i);
    }
  }
}

void GfxFireEmitter::spawn(ParticleLite &p)
{
  const double M_PI_180 = 0.01745329251994329576;	  //!< \f$ \pi /180 \f$

  Vec3<real4> spawn_box(size/4.f, size/4.f, size/4.f);
  p.pos = Vec3<real4>::vrand(gfxm->rand) * spawn_box;
  
  real4 vel = 0.01f * speed; 
  real4 vel_spread = vel * 0.10f;
  
  real4 particle_theta = (theta + (spread * (gfxm->rand.rand_real8()*2.f-1.f)) ) * M_PI_180;
  real4 particle_phi   = (phi +   (spread * (gfxm->rand.rand_real8()*2.f-1.f)) ) * M_PI_180;
  
  p.vel.y = cosf(particle_phi) * (vel + (vel_spread * gfxm->rand.rand_real8()));
  p.vel.x = (sinf(particle_phi) * cosf(particle_theta)) * (vel + (vel_spread * gfxm->rand.rand_real8()));
  p.vel.z = (sinf(particle_phi) * sinf(particle_theta)) * (vel + (vel_spread * gfxm->rand.rand_real8()));
  
  p.age = 0.0f;
  p.life = static_cast<real4>(gfxm->rand.rand_real8()) + speed*0.05f;
}

//===================================================================

GLuint GfxShellEmitter::id = 0;

GfxShellEmitter::GfxShellEmitter(GfxModule &gfxm_, real4 size_, real4 speed_)
{
  set_gfxm(gfxm_);
	
	size = max(size_ * 0.1f, 0.0001f);
  speed = max(speed_ * 100.f, 0.0001f);
  
  theta = 0.0f;
  phi = 45.0f;
  spread = 15.0f;
  acc = Vec3<real4>(0.0f, -0.918f, 0.0f);
  
  num_ptc = 0;
  new_ptc = 0;
  
  max_shell = static_cast<sint4>(speed+2);
  shell.resize(max_shell);
  FORALL(shell, i) i->life = 0.0f;
  
  if(!id)
  {  
    const uint1 slices = 5; // Number of sides to the bullet shell
    GLUquadricObj *quad = gluNewQuadric();
  
    id = glGenLists(1);
    glNewList(id, GL_COMPILE);

    glRotatef(90.0, 0.0, 0.0, 1.0);
    glRotatef(90.0, 1.0, 0.0, 0.0);
    gluCylinder(quad, size, size, size*5.f, slices, 1);
    gluDisk(quad, 0.f, size, slices, 1);
    
    glEndList();
    
    gluDeleteQuadric(quad);
  }
}

void GfxShellEmitter::draw_at(const Mat4<real4> &mat)
{
  glPushMatrix();

/*
  Mat4<real4> result, model, tag;
  glGetModelview(model);  
  tag = mat.trans();
 
  // fixme: render the shells relative to the world not the 
  //        tag position + orientation
  
  
  result = model * tag;
  result = result.trans();
  
  glLoadMatrixf(result.ptr());
*/
  glMultMatrixf(mat.ptr());
    
  draw();
  glPopMatrix();
}

void GfxShellEmitter::draw()
{
  glDisable(GL_BLEND);
  glActiveTextureARB(GL_TEXTURE1);
  glDisable(GL_TEXTURE_2D);  
  glActiveTextureARB(GL_TEXTURE0);
  glDisable(GL_TEXTURE_2D);
   
  glColor3f(0.75f, 0.68f, 0.39f);
  
  glDisable(GL_CULL_FACE);
  glEnable(GL_COLOR_MATERIAL);
  glEnable(GL_LIGHTING);
  
  GLfloat mat_ambient[] = {0.2f,0.2f,0.2f,1.0f};
  GLfloat mat_diffuse[] = {0.9f,0.9f,0.9f,1.0f};
  GLfloat mat_specular[] = {0.97f,0.96f,0.67f,1.0f};
  GLfloat mat_shiniess[] = {100.0f};

  glMaterialfv(GL_FRONT,GL_AMBIENT,mat_ambient);
  glMaterialfv(GL_FRONT,GL_DIFFUSE,mat_diffuse);
  glMaterialfv(GL_FRONT,GL_SPECULAR,mat_specular);
  glMaterialfv(GL_FRONT,GL_SHININESS,mat_shiniess);
  
  FORALL(shell, i)
  {
    if(i->life <= 0.0f)
      continue;
    
    glPushMatrix();
    glTranslatef(i->pos.x, i->pos.y, i->pos.z);
    glCallList(id);
    glPopMatrix();
  }
  
  glDisable(GL_COLOR_MATERIAL);
  glDisable(GL_LIGHTING);
  glEnable(GL_CULL_FACE);
  
  glEnable(GL_BLEND); 
  glActiveTextureARB(GL_TEXTURE1);
  glEnable(GL_TEXTURE_2D); 
  glActiveTextureARB(GL_TEXTURE0);
  glEnable(GL_TEXTURE_2D);
  
  glColor4f(1,1,1,1);
}

void GfxShellEmitter::update(real4 dt)
{
  GfxObject::update(dt);
    
  if(enabled)
	  new_ptc += dt * speed;

	if((num_ptc += uint4(new_ptc)) > max_shell)
		num_ptc = max_shell;  
  
  // Update positions of each spark
  FORALL(shell, i)
  {
    if(i->life > 0.0f)
		{
      // If the particle just died
      if(!i->update(dt, acc))
        num_ptc--;
    }else if(enabled && new_ptc >= 1.0f)
    {
      new_ptc--;
      spawn(*i);
    }
  }
}

void GfxShellEmitter::spawn(ParticleLite &p)
{
  const double M_PI_180 = 0.01745329251994329576;	  //!< \f$ \pi /180 \f$
  
  p.pos = Vec3<real4>(0.0f, 0.0f, 0.0f);
  
  real4 vel = 0.02f * speed; 
  real4 vel_spread = vel * 0.30f;
  
	real4 particle_theta  = (theta + (spread * (gfxm->rand.rand_real8()*2.f-1.f)) ) * M_PI_180;
	real4 particle_phi	  = (phi + (spread * (gfxm->rand.rand_real8()*2.f-1.f)) ) * M_PI_180;
  
  
	
	p.vel.z = cosf(particle_phi) * (vel + (vel_spread * gfxm->rand.rand_real8()));
	p.vel.x = (sinf(particle_phi) * cosf(particle_theta)) * (vel + (vel_spread * gfxm->rand.rand_real8()));
	p.vel.y = (sinf(particle_phi) * sinf(particle_theta)) * (vel + (vel_spread * gfxm->rand.rand_real8()));

  p.age = 0.0f;
  p.life = 1.0f;
}

//===================================================================


