/** @file Blueprints.C
    @see Blueprints.H

    $Id: Blueprints.C,v 1.28 2006/04/25 03:08:38 orts_mburo Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/Blueprints.C,v $
*/

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Game.H"
#include "GameObj.H"
#include "ScriptParser.H"
#include "Token.H"
#include "PlayerInfo.H"
#include "Blueprints.H"

using namespace std;

#define PARSE_ERR(s) {\
  errstr << endl << ": parse error at line " << tokenizer.cur_line();\
  ERR(s);\
}

#define PARSE_ERR2(s1, s2) {\
  errstr << endl << ": parse error at line " << tokenizer.cur_line();\
  ERR2(s1,s2);\
}

Vector<ObjActionDefinition> GameObjBlueprint::ext_action_defs;
Vector<ObjAction*>          GameObjBlueprint::ext_actions;
bool GameObjBlueprint::all_locked = false;

//===================================================================

/** Read information about attributes and actions and append it to this blueprint. */

void GameObjBlueprint::parse_class_def(Tokenizer &tokenizer,
                                       GameObjFactory::Name2BlueprintMap &classes,
                                       GameObjFactory::Name2ConstructorMap &classes2,
                                       Game &game, PlayerInfo &pi,
                                       bool read_constructors)
{
  string token;
  enabled = true;

  FOREVER {
    token = tokenizer.next_token();

    if (token == "action" || token == "constructor" || token == "destructor") {
      string act_type = token;
      ObjActionDefinition actdef;

      token = tokenizer.next_token();
      if (token == "hidden") {
        actdef.hidden = true;
      } else {
        actdef.hidden = (act_type == "constructor" || act_type == "destructor");
        tokenizer.putback(token);
      }

      string act_name = tokenizer.next_token();

      GameObjBlueprint *blueprint = this;
      FOREVER { // check if the action name contains sub-components
        size_t i = act_name.find('.', 0);
        if (i == string::npos) break;
        string comp = act_name.substr(0, i);
        act_name = act_name.substr(i+1, act_name.size() - (i+1));
        FIND (blueprint->name2comp, it2, comp);
        if (it2 == blueprint->name2comp.end()) PARSE_ERR2("unknown component:", comp.c_str());
        BP_Const &bpc = blueprint->component[it2->second];
        if (bpc.bp == 0) PARSE_ERR2("component isn't a blueprint", comp.c_str());
        blueprint = bpc.bp;
      }
      actdef.name = act_name;
      
      token = tokenizer.next_token();
      if (token == "external") {
        bool found = false;
        string ext_name = tokenizer.next_token();
        FORALL (ext_action_defs, it) {
          if (ext_name == it->name) {
            actdef = *it;
            actdef.external = true;
            found = true;
            break;
          }
        }
        if (!found) PARSE_ERR2("Unknown action:", ext_name.c_str());
      } else {
        tokenizer.putback(token);
        actdef.file_name = tokenizer.file_name();
        actdef.line_num = tokenizer.cur_line();
        actdef.parse_action_def(tokenizer);
        actdef.external = false;
        actdef.act_type = ObjActionDefinition::SCRIPT;

        // Automatically updates actdef to point to the new ObjAction.
        ObjAction::create(actdef, pi, game, name);
      }

      // put the action in its proper bin
      if (act_type == "action") {
        blueprint->attach_action(actdef);
      } else if (act_type == "constructor") {
        if (actdef.format != ObjActionDefinition::NONE) PARSE_ERR2("Constructor has parameters:", token.c_str());
        if (read_constructors) blueprint->constructor_defs.push_back(actdef);
        // don't check if we're reading constructors (would skew client/server action ids)
        blueprint->attach_action(actdef);
      } else {
        if (actdef.format != ObjActionDefinition::NONE) PARSE_ERR2("Destructor has parameters:", token.c_str());
        if (read_constructors) blueprint->destructor_defs.push_back(actdef);
      }
    }
    
    else if (token == "var" || token == "set" || token == "setf" || token == "SET") {
      string cmd = token;
      bool emphatic = cmd == "SET";
      if (cmd == "SET") cmd = "set";
      
      string attr_id;
      bool stable = false, constant = false, visible = true, internal = false;
      //      bool is_array = false;
      
      if (cmd == "setf") constant = true;
      
      FOREVER {
        token = tokenizer.next_token();
        if (token == "hidden") {
          if (cmd == "set" || cmd == "setf") PARSE_ERR ("can only use 'hidden' with 'var'");
          if (!visible) PARSE_ERR ("duplicate descriptor 'hidden'");
          visible = false;
        }
        else if (token == "const") {
          if (cmd == "set" || cmd == "setf") PARSE_ERR ("can only use 'const' with 'var'");
          if (constant) PARSE_ERR ("duplicate descriptor 'const'");
          constant = true;
        }
        else if (token == "internal") {
          if (cmd == "set" || cmd == "setf") PARSE_ERR ("can only use 'internal' with 'var'");
          if (internal) PARSE_ERR ("duplicate descriptor 'internal'");
          internal = true;
        }
        else
          break;
      }
      attr_id = token;
      AttrProp attr_prop(stable, constant, visible, internal);

      GameObjBlueprint *blueprint = this;
      FOREVER { // check if the attribute name contains sub-components
        size_t i = attr_id.find('.', 0);
        if (i == string::npos) break;
        string comp = attr_id.substr(0, i);
        attr_id = attr_id.substr(i+1, attr_id.size() - (i+1));

        FIND(blueprint->name2comp, it2, comp);
        if (it2 == blueprint->name2comp.end()) PARSE_ERR2("unknown component:", comp.c_str());
        if (blueprint->component[it2->second].bp == 0) PARSE_ERR2("component isn't a blueprint", comp.c_str());
        blueprint = blueprint->component[it2->second].bp;
      }
      
      string line; // read the line this way to convert defined tokens
      
      sint4 f_line = tokenizer.cur_line();
      FOREVER {
        token = tokenizer.next_raw_token();
        if (tokenizer.cur_line() != f_line || token == "") break;
        if (tokenizer.token_type() == Tokenizer::SPACE) token = " ";
        line += token;
      }
      ScriptParser temp;
      ScriptNode *s = temp.eval_expression(line);
      stype val;
      if (s) {
        val = s->eval();
        init_vals.push_back(s); // fixme: clean up this value (not in the destructor, that would cause conflicts)
        // the script tree needs to be maintained because any string values are still being referenced by val
      }

      if (cmd == "var")
        blueprint->add_attr(attr_id, attr_prop, val);
      else if (cmd == "set")
        blueprint->change_val(attr_id, val, emphatic);
      else if (cmd == "setf")
        blueprint->change_val(attr_id, attr_prop, val);
    }
    
    
    else if (token == "has" || token == "is") {
      
      token = tokenizer.next_token();
      
      FIND (classes, from_it, token);
      if (from_it == classes.end()) {
        cerr << "in blueprint " << name << endl;
        PARSE_ERR2("unknown class/blueprint", token.c_str());
      }
      
      const GameObjBlueprint &from = *from_it->second;
      
      // variables
      FORT (i, from.attr_vals.size()) {
        AttrProp p =  from.attr_props[i];
        stype v =     from.attr_vals[i];
        string name = from.attr_names[i];
        
        add_attr(name, p, v);
      }
      
      // actions from included class
      FORU (i, from.root_action_count) attach_action(from.action_defs[i]);
      //      FORALL(from.action_defs, it) attach_action(*it);
      FORALL(from.constructor_defs, it) constructor_defs.push_back(*it);
      FORALL(from.destructor_defs, it) destructor_defs.push_back(*it);

      // Components
      FORALL (from.name2comp, it) {
        if (name2comp.find(it->first) != name2comp.end())
          ERR2("component of same name already attached:", it->first.c_str());
        const BP_Const &bpc = from.component[it->second];

        if (bpc.bp) {
          const GameObjBlueprint *f = bpc.bp;
          attach_component(it->first, f, f->enabled); // creates a new GameObjBlueprint copy
        } else {
          GameObjFactory::BPConstructorData b = bpc.c;
          name2comp[it->first] = component.size();
          action_count += b.num_actions;
          component.push_back(bpc);
        }

        comp_props.push_back(from.comp_props[it->second]);
      }
    }
    
    else if (token == "class") {

      bool constant = false, visible = true, internal = false;
      FOREVER {
        token = tokenizer.next_token();
        if (token == "hidden") {
          if (!visible) PARSE_ERR ("duplicate descriptor 'hidden'");
          visible = false;
        }
        else if (token == "const") {
          if (constant) PARSE_ERR ("duplicate descriptor 'const'");
          constant = true;
        }
        else if (token == "internal") {
          if (internal) PARSE_ERR ("duplicate descriptor 'internal'");
          internal = true;
        }
        else
          break;
      }
      CompProp comp_prop(constant, visible, internal);

      string comp_type_name = token;
      
      FOREVER {
        bool enabled = true;
        token = tokenizer.next_token();
        if (token == "disabled") enabled = false;
        else if (token == "enabled")  enabled = true;
        else tokenizer.putback(token);

        string comp_var_name = tokenizer.next_token();
        if (name2comp.find(comp_var_name) != name2comp.end())
          ERR2("component of same name already attached:", comp_var_name.c_str());

        FIND (classes, from_it, comp_type_name);
        FIND (classes2, from_it2, comp_type_name);

        if (from_it != classes.end()) {
          const GameObjBlueprint *from = from_it->second;
          attach_component(comp_var_name, from, enabled);
        } else if (from_it2 != classes2.end()) {
          GameObjFactory::BPConstructorData b = from_it2->second;
          name2comp[comp_var_name] = component.size();
          BP_Const bpc;
          bpc.bp = 0;
          bpc.c = b;
          component.push_back(bpc);
          action_count += b.num_actions;
        } else {
          cerr << "in blueprint " << name << endl;
          PARSE_ERR2("unknown class/blueprint", comp_type_name.c_str());
        }

        comp_props.push_back(comp_prop);

        token = tokenizer.next_token();
        if (token != ",") {
          tokenizer.putback(token);
          break;
        }
      }
    }
    else if (token == "end") {
      return;
    }
    else {
      cout << "BP = " << name << endl;
      PARSE_ERR2("Unknown token", token.c_str());
    }
  }
}

//===================================================================

GameObjBlueprint::GameObjBlueprint(bool can_instance_)
{
  code = 0;
  action_count = 0;
  root_action_count = 0;

  num_const = num_stable = num_variable = 0;

  num_comp = 0;

  locked = false;
  can_instance = can_instance_;
}

//-------------------------------------------------------------------

void copy_components(GameObjBlueprint *to, const GameObjBlueprint *from)
{
  map<string, sint4>::const_iterator from_it = from->name2comp.begin();
  map<string, sint4>::const_iterator to_it = to->name2comp.begin();

  while (from_it != from->name2comp.end() &&
         to_it   != to->name2comp.end()) {

    const GameObjBlueprint::BP_Const &f_bpc = from->component[from_it->second];
    GameObjBlueprint::BP_Const &t_bpc = to->component[to_it->second];

    if (f_bpc.bp) {
      t_bpc.bp = new GameObjBlueprint(*f_bpc.bp);
      copy_components(t_bpc.bp, f_bpc.bp);

    } else {
      t_bpc.c = f_bpc.c;
    }

    ++from_it;
    ++to_it;
  }
}

// copying should only be done before binding
GameObjBlueprint::GameObjBlueprint(const GameObjBlueprint &x)
{
  locked = false;
  can_instance =        x.can_instance;

  action_count =        x.action_count; // component actions can make this number higher
  root_action_count =   x.root_action_count;

  action_defs =         x.action_defs;
  constructor_defs =    x.constructor_defs;
  destructor_defs =     x.destructor_defs;


  code =                x.code;
  attributes =          x.attributes;
  attr_names =          x.attr_names;
  attr_vals =           x.attr_vals;
  attr_props =          x.attr_props;

  num_const =           x.num_const;
  num_stable =          x.num_stable;
  num_variable =        x.num_variable;

  num_comp =            x.num_comp;
  name2comp =           x.name2comp;
  component =           x.component;
  comp_props =          x.comp_props;

  // don't copy action instances (they are created during binding using action_defs)

  name =                x.name;
  enabled =             x.enabled;
  enable_count =        x.enable_count;

  copy_components(this, &x);
}

//-------------------------------------------------------------------

GameObjBlueprint::~GameObjBlueprint()
{
  /*
  FORALL (component_name2bp, it) {
    GameObjBlueprint *bp = it->second;
    delete bp;
  }
  */
}

//===================================================================

sint4 GameObjBlueprint::attr_index(const string &a)
{
  FIND (attributes, i, a);
  if (i == attributes.end()) return -1;
  return i->second;
}

//-------------------------------------------------------------------

void GameObjBlueprint::add_attr(const string &name,
                                const AttrProp &ap, stype val)
{
  FIND(attributes, it, name);
  if (it == attributes.end()) {
    AttrProp p = ap;
    p.index = attr_vals.size();
    attributes[name] = p.index;
    attr_vals.push_back(val);
    attr_props.push_back(p);
    attr_names.push_back(name);
    return;
  }

  attr_vals[(*it).second] = val;
  attr_props[(*it).second].constant = ap.constant;
  attr_props[(*it).second].stable = ap.stable;
  attr_props[(*it).second].visible_to_opponents = ap.visible_to_opponents;
  attr_props[(*it).second].internal = ap.internal;
}

//-------------------------------------------------------------------

#if 0
void GameObjBlueprint::add_str_attr(const string &name,
				    const AttrProp &ap, const string &val)
{
  AttrProp p = ap;
  p.index = attr_vals.size();
  p.type = AttrProp::STR;
  attributes[name] = p.index;

  stype v = val;
  attr_vals.push_back(v);
  attr_props.push_back(p);
  attr_names.push_back(name);
}
#endif

//-------------------------------------------------------------------

void GameObjBlueprint::change_val(const string &name, stype val, bool force)
{
  FIND(attributes, it, name);
  if (it == attributes.end()) {
    cout << this->name << endl;
    ERR2("unknown attribute ", name.c_str());
  }
  if (attr_props[(*it).second].constant && !force)
    ERR2("cannot change constant variable", name.c_str());

  attr_vals[(*it).second] = stype(val);
}

//-------------------------------------------------------------------

void GameObjBlueprint::change_val(const string &name, const AttrProp &ap,
                                  stype val, bool force)
{
  FIND(attributes, it, name);
  if (it == attributes.end()) {
    cout << this->name << endl;
    ERR2("unknown attribute ", name.c_str());
  }
  if (attr_props[(*it).second].constant && !force)
    ERR2("cannot change constant variable", name.c_str());

  attr_vals [(*it).second] = stype(val);
  attr_props[(*it).second].constant = ap.constant;
  attr_props[(*it).second].stable = ap.stable;
  attr_props[(*it).second].visible_to_opponents = ap.visible_to_opponents;
}

//-------------------------------------------------------------------

void GameObjBlueprint::add_ext_action(const ObjActionDefinition &def)
{
  FORALL (ext_action_defs, it) {
    if ((*it).name == def.name) {
      (*it) = def;
      return;
    }
  }
  ext_action_defs.push_back(def);
}

//-------------------------------------------------------------------

void GameObjBlueprint::attach_action(const ObjActionDefinition &def)
{
  FORU (i, root_action_count) {
    ObjActionDefinition &d = action_defs[i];
    if (d.name == def.name) {
      d = def;
      // start here
      /*
      FORU (j, min((uint4)root_action_count, (uint4)action_names.size())) {
        if (action_names[j] == def.name) {
          assert (def.func);
          actions[j] = def.func;
          break;
        }
      }
      */
      return;
    }
  }
  action_defs.push_back(def);
  ++action_count;
  ++root_action_count;
}

//-------------------------------------------------------------------

void GameObjBlueprint::attach_component(const string &var_name, const GameObjBlueprint *type, bool enabled)
{
  if (name2comp.find(var_name) != name2comp.end())
    ERR2("component of same name already attached:", var_name.c_str());

  BP_Const bpc;
  bpc.c.c = 0;
  bpc.bp = new GameObjBlueprint(*type);
  bpc.bp->enabled = enabled;

  name2comp[var_name] = component.size();
  component.push_back(bpc);

  action_count += type->action_count; // make room for the complete list
}

//-------------------------------------------------------------------

/** Sort attributes by property type (constant, stable, variable). */

void GameObjBlueprint::sort_attributes()
{
  assert (!locked); // structure shouldn't be changed at this point

  Vector<string>   _attr_names;
  Vector<stype>    _attr_vals;
  Vector<AttrProp> _attr_props;

  // sort attributes by property type (constant, stable, variable)

  uint4 aps = attr_props.size();

  FORU (i, aps) {
    if (!attr_props[i].constant && !attr_props[i].stable) {
      AttrProp ap = attr_props[i];
      ap.index = _attr_vals.size();
      attr_props[attributes[attr_names[i]]] = ap;
      attributes[attr_names[i]] = ap.index;
      _attr_props.push_back(ap);
      _attr_vals.push_back(attr_vals[i]);
      _attr_names.push_back(attr_names[i]);
      ++num_variable;
    }
  }
  FORU (i, aps) {
    if (!attr_props[i].constant && attr_props[i].stable) {
      AttrProp ap = attr_props[i];
      ap.index = _attr_vals.size();
      attr_props[attributes[attr_names[i]]] = ap;
      attributes[attr_names[i]] = ap.index;
      _attr_props.push_back(ap);
      _attr_vals.push_back(attr_vals[i]);
      _attr_names.push_back(attr_names[i]);
      ++num_stable;
    }
  }
  FORU (i, aps) {
    if (attr_props[i].constant) {
      AttrProp ap = attr_props[i];
      ap.index = _attr_vals.size();
      attr_props[attributes[attr_names[i]]] = ap;
      attributes[attr_names[i]] = ap.index;
      _attr_props.push_back(ap);
      _attr_vals.push_back(attr_vals[i]);
      _attr_names.push_back(attr_names[i]);
      ++num_const;
    }
  }

  attr_names = _attr_names;
  attr_vals  = _attr_vals;
  attr_props = _attr_props;
}

//-------------------------------------------------------------------

/** Reorder the attributes to improve transmission and create reverse action id mapping.
    @param id_start  Starting index to use for action ids.
*/

void GameObjBlueprint::lock(sint4 id_start)
{
  if (locked || all_locked) return;
  assert(actions.size() == root_action_count);
  assert(action_names.size() == root_action_count);

  sort_attributes(); // reorder attributes
  num_comp = component.size();
  
  FORU(i, root_action_count) {
    action_name2id[action_names[i]] = id_start + i; // reverse mapping for the client (outgoing)
  }
  id_start += root_action_count;
  
  // bind components, add their actions to the list of numbered actions here
  // note: this recurses over all (sub)component actions because each subcomponent builds
  //       their own list of component actions before it is added to the root list.

  // IMPORTANT: must traverse the components in the same mannner as GameObj::GameObj(...),
  //            this ensures that the action -> component maps are correct.
  FORALL (name2comp, it) {
    GameObjBlueprint *gob = component[it->second].bp; //(*it).bp;
    if (!gob) continue;

    gob->lock(id_start);

    FORU (i, gob->action_count) {
      actions.push_back(gob->actions[i]); // add component actions to the list
      action_defs.push_back(gob->action_defs[i]);
      //      action_names.push_back(gob->action_names[i]);
      action_names.push_back("");
    }
    id_start += gob->action_count;
  }

  locked = true;
}

//-------------------------------------------------------------------

/** Creates any new actions that have been defined.
    @param bp_name Name of this blueprint for printing out debug information;
                   includes the list of parent blueprints (if any).
*/

void GameObjBlueprint::bind(PlayerInfo &pi, Game &game, const string &bp_name)
{
  for (uint4 i = constructor.size(); i < constructor_defs.size(); ++i) {
    ObjActionDefinition &def = constructor_defs[i];
    if (!def.func) {
      cerr << bp_name << endl;
      ERR2("Constructor action not found when binding:", def.name.c_str());
    }
    constructor.push_back(def.func);
  }

  for (uint4 i = destructor.size(); i < destructor_defs.size(); ++i) {
    ObjActionDefinition &def = destructor_defs[i];
    if (!def.func) {
      cerr << bp_name << endl;
      ERR2("Destructor action not found when binding:", def.name.c_str());
    }
    destructor.push_back(def.func);
  }

  for (uint4 i = actions.size(); i < action_defs.size(); ++i) {
    ObjActionDefinition &def = action_defs[i];
    if (!def.func) {
      cerr << bp_name << endl;
      ERR2("Action not found when binding:", def.name.c_str());
    }
    actions.push_back(def.func);
    if (locked || all_locked) {
      action_names.push_back("");
      // the action map won't be set anywhere else since it is locked
      action_map[def.name] = def.func;
    } else {
      if (i >= root_action_count) REM("buh?");
      action_names.push_back(i < root_action_count ? def.name : "");
    }
  }

  assert(action_names.size() == actions.size());

  FORT (i, action_names.size()) {
    if (action_names[i] != "") {
      if (i >= root_action_count) cout << "WHOA - " << action_names[i] << endl;
      action_map[action_names[i]] = actions[i];
    }
  }

  FORALL (name2comp, it) {
    GameObjBlueprint *gob = component[it->second].bp;
    if (!gob) continue;
    gob->bind(pi, game, bp_name + "::" + it->first);
  }

  assert(action_names.size() == actions.size());

  hooks.resize(attr_names.size());
  FORT (i, attr_names.size()) {
    string n = "event_changed_" + attr_names[i];
    FIND(action_map, action_it, n);
    hooks[i] = action_it != action_map.end() ? action_it->second : 0;
  }
}

//===================================================================

/** For server use: clears the id for each action, thereby enabling all of them
    for local execution.  Actions with an id are set as outgoing (client). */

void GameObjBlueprint::clear_action_map()
{
  //  cout << "CLEARING ACTION MAP" << endl;
  action_name2id.clear();

  FORALL (component, it) {
    GameObjBlueprint *gob = (*it).bp;
    if (gob) gob->clear_action_map();
  }
}

//===================================================================
