#ifndef Game_H
#define Game_H

/** @file Game.H
    Main game class.

    $Id: Game.H,v 1.50 2006/06/09 14:46:08 orts_mburo Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/Game.H,v $
*/

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Global.H"
#include "Array.H"
#include "Coord.H"
#include "Action.H"
#include "Map.H"
#include "GameConst.H"
#include "PlayerInfo.H"
#include "Random.H"
#include "Tile.H"
#include "SaveLoad.H"
#include "GameTile.H"

class Game;
class GameObj;
class GameObjBlueprint;
class GameObjFactory;
class BoolMatrix;
class GameChanges;
class ActionChanges;
class Tokenizer;
class TileView;
class DualView;
class Options;
class ScriptFunctionDef;
class ScriptObj;

typedef Array<GameObj*> MotionSectorInfo;

struct MotionSectors {

  Vector<MotionSectorInfo> secs;
  Vector<sint4> sizes; // base (empty) size (# of boundary objects)
  Array<sint4, 600> active_sectors;
  uint1 *active;//[sec_num]; having moving objects in them

  MotionSectors() { active = 0; }

  void reset_sizes();
  void clear();
};

/** \nosubgrouping */

class Game : public SaveLoadAbs
{
public:

  struct Options {

    sint4 cpind; // >=0: client player index, <0: server
    bool nfog;
    bool stat;
    sint4 seed;
    std::string save_file;
    std::string load_file;
    std::string pack_file;
    
    Options();
    static void add();
  };
  
  typedef Vector<ScriptObj::Ptr> ObjCont;
  Random rand;

  Game(const std::string &world, const Game::Options &o);
  virtual ~Game();

  //-------------------------------------------------------------------
  /// @name Helper functions
  //-------------------------------------------------------------------
  //@{
  sint4 num_intersect(GameObj &a);
  Scalar fine_grid(sint4 x) const { return x << fine_grid_res_bits; }
  sint4 get_sec_wh()  const { return sec_wh;  }
  sint4 get_sec_nx()  const { return sec_nx;  }
  sint4 get_sec_ny()  const { return sec_ny;  }
  sint4 get_sec_num() const { return sec_num; }
  SimpleActions &get_simple_actions() { return simple_actions; } // fixme: unsafe
  const MotionSectors &get_motion_secs() const { return motion_secs; }
  void compute_view(sint4 player, Vector<sint4> *new_tiles = 0);
  //@}

  //-------------------------------------------------------------------
  /// @name Movement and position
  //-------------------------------------------------------------------
  //@{
  bool obj_move_to(GameObj &e, const ScalarPoint &tpos, sint4 speed) const;
  static real8 distance(const Object &e1, const Object &e2);
  void update_pos(GameObj &a);

  void compute_footprint(const GameObj &gob, Vector<sint4> &obj_tiles) const;
  void compute_footprint(const GameObj &gob, Vector<sint4> &obj_tiles_W, Vector<sint4> &obj_tiles_E) const;
  //@}
  
  //-------------------------------------------------------------------
  /// @name Visibility
  //-------------------------------------------------------------------
  //@{
  bool can_see_obj(const GameObj &a, const GameObj &b);
  //@}

  //-------------------------------------------------------------------
  /// @name Map state
  //-------------------------------------------------------------------
  //@{
  bool is_rotated() const { return rotated != 0; }

  sint4 get_gtiles_x() const { return gtiles_x; } ///< (Approx.) width of the game field.
  sint4 get_gtiles_y() const { return gtiles_y; } ///< (Approx.) height of the game field.

  sint4 get_tile_points() const { return tile_points; }
  real4 get_inv_tile_points() const { return inv_tile_points; }
  //@}

  //-------------------------------------------------------------------
  /// @name Script interface layer
  //-------------------------------------------------------------------
  //@{

  static stype script_printnum          (const SActParam&);

  static stype script_Distance          (const SActParam&);
  static stype script_Notify_Attack     (const SActParam&);
  static stype script_GOB_Kill          (const SActParam&);
  static stype script_GOB_Enable        (const SActParam&);
  static stype script_GOB_Disable       (const SActParam&);
  static stype script_GOB_Is_Enabled    (const SActParam&);
  static stype script_GOB_Is_Alive      (const SActParam&);
  static stype script_GOB_Type          (const SActParam&);
         stype script_GOB_Type2Name     (const SActParam&);
  static stype script_GOB_Visible_By    (const SActParam&);
  static stype script_GOB_Cancel_Action (const SActParam&);
  static stype script_GOB_Has_Attr      (const SActParam&);
  static stype script_GOB_Get_Attr      (const SActParam&);
  static stype script_GOB_Get_Comp      (const SActParam&);
  static stype script_GOB_SOD           (const SActParam&);

  stype script_GOB_Create               (const SActParam&);
  stype script_GOB_CreateId             (const SActParam&);
  stype script_GOB_Id                   (const SActParam&);
  stype script_GOB_Move                 (const SActParam&);
  stype script_GOB_Fix_Position         (const SActParam&);
  stype script_GOB_Can_See              (const SActParam&);
  stype script_GOB_Footprint            (const SActParam&);
  stype script_get_tile_type            (const SActParam&);
  stype script_set_tile_type            (const SActParam&);

  stype script_eval                     (const SActParam&);
  stype script_close_to                 (const SActParam&);

  stype script_end_game                 (const SActParam&);
  stype script_save_game                (const SActParam&);
  stype script_Tick                     (const SActParam&);

  stype script_map_width                (const SActParam&);
  stype script_map_height               (const SActParam&);

  stype script_sqrt                     (const SActParam&);

  stype script_cout                     (const SActParam&);
  //@}


  //-------------------------------------------------------------------
  /// @name Server functions
  //-------------------------------------------------------------------
  //@{
  void preprocess_actions(Vector<std::string> &actions);
  bool simulation_step(Vector<std::string> &actions, Vector<std::string> &views);
  void gen_view_diffs(Vector<std::string> &views);
  void gen_actions(Vector<std::string> &actions);
  void exec_actions(Vector<std::string> &commands);
  static std::string remove_secret_info(const std::string &world);
  std::string get_client_world();

  bool get_save_req() const { return save_req; }  ///< True iff there is a save and quit request.
  //@}

  //-------------------------------------------------------------------
  /// @name Client functions
  //-------------------------------------------------------------------
  //@{
  sint4 get_view_frame() const { return view_frame; }
  sint4 get_action_frame() const { return action_id; } ///< Frame id of last action message executed (-1 if none)
  sint4 get_skipped_actions() const { return skipped_actions; }
  sint4 get_merged_actions() const { return merged_actions; }
  real4 get_freq() const { return freq; }
  
  // objects for player ind (or non-player objects when ind=player_n)
  const ObjCont& get_objs(sint4 ind) const {
    assert(ind >= 0 && ind <= player_n);
    return objs[ind];
  }

  bool get_finished() const { return game_over; } ///< True iff the game is over.

  sint4 get_client_player() const { return opt.cpind; } ///< Id of the current player.
  std::string get_savefile() const { return opt.save_file; }

  PlayerInfo &get_cplayer_info() const { return *cplayer; }
  PlayerInfo &get_player_info(sint4 i) const { return *players[i]; }

  const ObjCont &get_boundaries() const { return boundaries; }
  const Map<GameTile> &get_map() const { return pf; }

  void gen_client_actions();
  void gen_action_message(sint4 action_frame, std::string &msg) const;
  void apply_view_diff(const std::string &view, GameChanges &changes, ActionChanges &act_changes);
  const char* read_header(const char *world);

  //@}

  //-------------------------------------------------------------------
  /// @name Common functions
  //-------------------------------------------------------------------
  //@{
  uint4 tick() const { return tick_count; }
  sint4 get_player_num() const { return player_n; } ///< Number of players.

  void read_tiledef_block(Tokenizer &tokenizer);
  void read_blueprint_block(Tokenizer &tokenizer, bool extend=false);
  void read_action_block(Tokenizer &tokenizer);
  void read_define_block(Tokenizer &tokenizer);
  void read_object_block(Tokenizer &tokenizer);
  void bind();

  void save(std::ostream &os) const;
  void load(std::istream &is);
  //@}
  

private:

  // prevent copying and assigning (not implemented yet)
  Game(const Game &);
  Game &operator=(const Game &);
  
  bool game_over;
  bool save_req;
  sint4 save_and_quit;
  real4 freq;  // server simulation frequency (HZ)
  uint4 tick_count;

  Vector<PlayerInfo*> players;
  PlayerInfo *NPC;             ///< Environment/unowned/temp objects

  Map<GameTile> pf;            ///< playfield

  sint4 gtiles_x, gtiles_y;    // fixme: should be part of Map?
  sint4 z_sight;
  bool rotated;                // fixme: should be part of Map?

  sint4 player_n;        ///< Number of players.

  ObjCont *objs;         ///< Player-owned objects and NPCs.


  // only used in client mode vvvvvvvvvvvvvvvvvvvvvvvv
  /// @name Client-mode variables
  //@{
  PlayerInfo *cplayer;   ///< Points to the same info as players[opt.cpind].

  // View info
  sint4 view_frame;      ///< Current simulation frame.
  sint4 action_id;       ///< Frame # of action just executed.
  sint4 skipped_actions; ///< # of skipped actions.
  sint4 merged_actions;  ///< # of merged actions.
  //@}
  // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

  std::string world;

  Game::Options opt;
  
  MoveActions   move_actions;
  SimpleActions simple_actions;

  Vector<BaseActions*> actions;     ///< Actions for each category (movement, simple(scripted), etc).
  Vector<ActionFunction> c_actions; ///< C++ functions to be used as object actions.

  sint4 tile_points;                ///< The granularity of the coordinate system relative to one tile.
  real4 inv_tile_points;

  // computed
  sint4 points_x, fine_points_x;    ///< # of fine points along the x-axis.
  sint4 points_y, fine_points_y;    ///< # of fine points along the y-axis.

  sint4 fine_grid_res_bits;         ///< An extra degree of accuracy above tile_points for collisions.
  sint4 fine_grid_res;              ///< 1 << fine_grid_res_bits

  BoolMatrix *x_allied_with_y;

  // motion sector information
  
  sint4 sec_wh;       ///< The side length of a motion sector in fine points.
  sint4 sec_nx;       ///< # of sectors along the x-axis.
  sint4 sec_ny;       ///< # of sectors along the y-axis.
  sint4 sec_num;      ///< # of sectors.

  uint4 magic; // used for incrementally resetting neighbor lists and visit flags
  uint4 visit_magic;
  ObjCont boundaries; // owns objects -> delete!

  // game field boundaries, fringe + terrain boundaries
  MotionSectors fringe_secs, boundary_secs;

  // contains all objects, collision mask determines intersections
  MotionSectors motion_secs;

  //-------------------------------------------------------------------
  /// @name Visibility
  //-------------------------------------------------------------------
  //@{

  // Maximum sight range of objects on each tile. (reset for each player)
  sint1 *v1w;  ///< Max. sight range on each tile; normal vision; West/full side
  sint1 *v1e;  ///< Max. sight range on each tile; normal vision; East side
  sint1 *v2;   ///< Max. sight range on each tile; normal vision; aerial
  sint1 *dv1w; ///< Max. sight range on each tile; detector vision; West/full side
  sint1 *dv1e; ///< Max. sight range on each tile; detector vision; East side
  sint1 *dv2;  ///< Max. sight range on each tile; detector vision; aerial

  /// Tile # -> Index i where the first block (upper left) of the tile's DualView matches a MapView crop[i].
  sint4 *tind2bitmap;

  /// Index i -> Smallest tile # stored in the T block MapView::crop[i]. (used as b2t in MapView::map_whatsnew)
  sint4 *bitmap2tind;

  /// Tile # -> The minimum lshift required to put a TileView's edge on a T boundary in a MapView.
  sint4 *lshift_amount;

  TileView full_tile_views[GameConst::VIEW_MAX+1]; ///< Visibility circles of each radius used for masking.
  DualView dfull_tile_views[GameConst::VIEW_MAX+1][sizeof(VisionT)*8]; ///< T-aligned versions of full_tile_views.

  DualView *edge_masks; ///< Edge masks for each x position.  Used to mask aerial views (prevent wrapping).

  /// \f$ fv\_yc[x] \approx MIN(x, \sqrt{VIEW\_MAX^2 - x^2}) + 1 \f$.
  /// In generate_tile_view, the maximum distance from the axis that tiles need to be checked.
  sint4 fv_yc[GameConst::VIEW_MAX+1];

  BitStrip *tile_known_range;      ///< The range of slopes that the entire tile occupies.
  BitStrip *tile_fshadow_range;    ///< Inverted version of the above.
  BitStrip *tile_f_visible_range;  ///< The visible part of the tile when it is unobscured.

  Vector<BitStrip> tile_shadows;
  Vector<BitStrip> tile_visibility;

  sint4 **tile_shadow_range[4];    ///< The area obscured by walls.
  sint4 **tile_visible_range_W[4]; ///< The visible portion of the West side when considering walls.
  sint4 **tile_visible_range_E[4]; ///< The visible portion of the East side when considering walls.

  BitStrip **edge_shadows[4];      ///< The little shadows that all the bigger ones are composed of.

  /// A reused strip for each quadrant; 1 marks visible portions between slopes.
  BitStrip qstrip;

  void allocate_vision_memory();
  void deallocate_vision_memory();
  void precompute_vision_data();

  static void generate_full_tile_view(sint4 r, TileView &v);
  void generate_tile_view(sint4 tind, GameTile &t, bool West);

  // vW and vE are the visibility of the West and East sides of the surrounding tiles.
  inline void gen_tile_view_E(sint4 tind, sint4 height, TileView &vW, TileView &vE, TileView &k);
  inline void gen_tile_view_S(sint4 tind, sint4 height, TileView &vW, TileView &vE, TileView &k);
  inline void gen_tile_view_W(sint4 tind, sint4 height, TileView &vW, TileView &vE, TileView &k);
  inline void gen_tile_view_N(sint4 tind, sint4 height, TileView &vW, TileView &vE, TileView &k);
  inline void check_tile(GameTile &tile, TileView &vW, TileView &vE, TileView &kv,
                         sint4 q, sint4 vx, sint4 vy, sint4 max_z);

  void compute_visibility(GameObj &gob) const;
  //  bool can_see_obj(const GameObj &a, const GameObj &b) const;

  //@}

  //-------------------------------------------------------------------
  /// @name Tiles and boundaries
  //-------------------------------------------------------------------
  //@{
  sint4 num_tile_types;
  Vector<sint4> tile_collision_mask; ///< Defines collision masks between different tile types.
  sint4 tile_height_diff_mask; ///< Collision mask when tiles have a height difference.
  void generate_boundaries();
  GameObj* add_tile_boundary(sint4 x, sint4 y, sint4 x2, sint4 y2,
                             Tile::Type t1, Tile::Type t2, bool height_diff, bool is_fringe = false);
  void remove_tile_boundary(sint4 x, sint4 y, sint4 x2, sint4 y2);
  //@}

  //-------------------------------------------------------------------
  /// @name Movement and position
  //-------------------------------------------------------------------
  //@{
  void init_motion_sectors();

  void move_objects();
  void move_objects(Array<GameObj*> &mobjs, MotionSectors &ms, bool obj_coll);
  void update_sectors();

  void obj_move_to_grid(GameObj &e) const;
  void obj_stop(GameObj &e) const;
  void obj_move_and_stop(GameObj &e, real8 t) const;
  void obj_move(GameObj &e, real8 t) const;

  void handle_component(GameObj **comp, sint4 n);
  real8 collision_time               (GameObj &e1, GameObj &e2, real8 ct, sint4 &coll_type) const;
  real8 collision_time_circle_circle (const GameObj &e1, const GameObj &e2, real8 ct, sint4 &coll_type) const;
  real8 collision_time_circle_line   (const GameObj &e1, const GameObj &e2, real8 ct, sint4 &coll_type) const;
  real8 collision_time_circle_rect   (const GameObj &e1, const GameObj &e2, real8 ct, sint4 &coll_type) const;
  real8 collision_time_line_line     (const GameObj &e1, const GameObj &e2, real8 ct, sint4 &coll_type) const;
  real8 collision_time_line_rect     (const GameObj &e1, const GameObj &e2, real8 ct, sint4 &coll_type) const;
  real8 collision_time_rect_rect     (const GameObj &e1, const GameObj &e2, real8 ct, sint4 &coll_type) const;

  //  real8 collision_time_point_skew    (const GameObj &e1, const GameObj &e2, real8 ct, sint4 &coll_type) const;

  sint4 resolve_collision                (const GameObj &e1, const GameObj &e2, real8 ct, sint4 coll_type) const;
  sint4 resolve_collision_circle_point   (const GameObj &e1, const ScalarPoint &c, const ScalarPoint &h, real8 ct) const;
  sint4 resolve_collision_circle_circle  (const GameObj &e1, const GameObj &e2, real8 ct, sint4 coll_type) const;
  sint4 resolve_collision_circle_line    (const GameObj &e1, const GameObj &e2, real8 ct, sint4 coll_type) const;
  sint4 resolve_collision_circle_rect    (const GameObj &e1, const GameObj &e2, real8 ct, sint4 coll_type) const;
  sint4 resolve_collision_hvline_hvline  (const GameObj &e1, const GameObj &e2, real8 ct, sint4 coll_type) const;

  void compute_ms_footprint(const GameObj &gob, Vector<sint4> &motion_sectors) const;
  //@}

  void update_tile(sint4 pos, Vector<ScriptObj*> *new_boundaries = 0);

  void visit(GameObj &e, GameObj** &comp);
  
  sint4 tile_index(sint4 x, sint4 y) const;
  void inc_magic();

  void clear_tiles();
  //  void line(sint4 x1, sint4 y1, sint4 x2, sint4 y2, sint4 type);
  static sint4 ok_dist2(sint4 k);
};

REGISTER_TYPEOF(62, Vector<PlayerInfo*>::iterator);
REGISTER_TYPEOF(63, Vector<PlayerInfo*>::const_iterator);
#endif
