/** @file GameObj.C
    @see GameObj.H

    $Id: GameObj.C,v 1.77 2006/06/09 21:55:51 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/GameObj.C,v $
*/

// This is an ORTS file
// (c) Tim Furtak
// (c) Michael Buro
// licensed under the GPL

#include "ScriptObj.H"
#include "GameObj.H"
#include "Blueprints.H"
#include "ServerObjData.H"
#include "PlayerInfo.H"
#include "Script.H"
#include "ByteData.H"

using namespace std;

REGISTER_TYPEOF(52, Vector<AttrProp>::iterator);
REGISTER_TYPEOF(53, Vector<AttrProp>::const_iterator);

#define DAMAGE_BITMASK 1

sint4 num_created=0;
sint4 num_deleted=0;

#define PARANOID

set<ScriptObj*> ScriptObj::ptr_list;

//===================================================================

IdObjMap ScriptObj::id2obj_map;
ObjIdMap ScriptObj::obj2id_map;
IdFactory ScriptObj::idf(100000);

Vector<ScriptObj*> ScriptObj::dead_objs;
Vector<ScriptObj*> ScriptObj::destroyed_objs;

//===================================================================

ScriptObj::Ptr::Ptr(ScriptObj *s)
{
  id = ScriptObj::obj2id_map[s];
}

//===================================================================

ScriptObj::ScriptObj()
{
  num_refs = 0;
  dead = false;
  enabled = true;
  parent = this;
#ifdef PARANOID
  //The relevant types cannot be registered due to some unknown reason: FIND(ptr_list, i, this);
  set<ScriptObj*>::iterator i = ptr_list.find(this);
  if (i != ptr_list.end()) ERR("already in set?");
  ptr_list.insert(this);
#endif

  id = idf.next();
  assert(id2obj_map[id] == 0);
  id2obj_map[id] = this;
  obj2id_map[this] = id;
}

//-------------------------------------------------------------------

ScriptObj::~ScriptObj()
{
  //  cout << "deleting scriptobj " << (void*)this << " " << bp_name() << endl;
  idf.free(id);
  id2obj_map.erase(id2obj_map.find(id));
  obj2id_map.erase(obj2id_map.find(this));
#ifdef PARANOID
  //The relevant types cannot be registered due to some unknown reason: FIND(ptr_list, i, this);
  set<ScriptObj*>::iterator it = ptr_list.find(this);
  if (it != ptr_list.end()) ptr_list.erase(it);
#endif
}

//-------------------------------------------------------------------

void ScriptObj::save(ostream &os) const
{
  string s;
  gen_diff(true, true, s, 0);
  os << s;
}

//-------------------------------------------------------------------

void ScriptObj::load(istream &is)
{
  string s;
  is >> s;
  const char *c = s.c_str();
  apply_diff(true, c);
}

//-------------------------------------------------------------------

/** Intended for use in saving/loading the game state in combination with set_ids. */

void ScriptObj::get_ids(Vector<ScriptObj*> &objs, Vector<ObjId> &ids)
{
  ids.clear();
  objs.clear();

  FORALL (id2obj_map, it) {
    ids.push_back(it->first);
    objs.push_back(it->second);
  }
}

//-------------------------------------------------------------------

/** Intended for use in saving/loading the game state in combination with get_ids. */

void ScriptObj::set_ids(const Vector<ScriptObj*> &objs, const Vector<ObjId> &ids)
{
  if (objs.size() != ids.size()) ERR("size mismatch");
  id2obj_map.clear();
  obj2id_map.clear();
  idf.reserve(ids);
  FORU (i, objs.size()) {
    id2obj_map[ids[i]] = objs[i];
    obj2id_map[objs[i]] = ids[i];
  }
}

//-------------------------------------------------------------------

ObjId ScriptObj::get_obj_id(ScriptObj *s)
{
  FIND(obj2id_map, it, s);
  if (it == obj2id_map.end()) return 0;
  return it->second;
}

//-------------------------------------------------------------------

void ScriptObj::kill()
{
  if (!dead) dead_objs.push_back(this);
  dead = true;
}

//-------------------------------------------------------------------

/** Call script destructors for dead objects.
    Delete them if they're not being referenced.
    NOTE: new objects can be added to dead_objs when the destructors
          are called.
*/

void ScriptObj::collect_dead(std::set<ScriptObj*> &dead)
{
  dead.clear();
  // note: if new objects are added to dead_objs when the destructors are called
  //       then the iterator will be invalidated, so do it this way.
  FORU_DYN (i, dead_objs.size()) {
    ScriptObj *g = dead_objs[i];
    dead.insert(g);
    g->destruct();
    if (g->get_refs() == 0) {
      delete g;
    } else {
      destroyed_objs.push_back(g);
    }
  }
  dead_objs.clear();
}

//-------------------------------------------------------------------

/** Delete dead objects that aren't being referenced. */

void ScriptObj::collect_destroyed()
{
  Vector<ScriptObj*>::iterator it;
  for (it = destroyed_objs.begin(); it != destroyed_objs.end();) {
    if ((*it)->get_refs() == 0) {
      delete *it;
      it = destroyed_objs.erase(it);
      continue;
    }
    ++it;
  }
}

//-------------------------------------------------------------------

void ScriptObj::check_ptr(ScriptObj *ptr) {
#ifdef PARANOID
  if (ptr == 0) return;
  if (ptr_list.find(ptr) != ptr_list.end()) return;
  cerr << ptr << endl;
  ERR("no such scriptobj");
#endif
}

//-------------------------------------------------------------------

bool ScriptObj::has_attr(const string &a) const {
  return get_const_val_ptr(a) != 0;
}

//-------------------------------------------------------------------

stype ScriptObj::get_val(const string &a) const {
  const stype *p = get_const_val_ptr(a);
  if (p == 0) {
    cerr << "object type = " << bp_name() << endl;
    ERR2("unknown attribute ", a.c_str());
    return stype();
  }
  return *p;
}

//-------------------------------------------------------------------

void ScriptObj::set_val(const string &a, stype v, bool warn) {
  stype *p =  get_val_ptr(a);
  if (p == 0) {
    if (warn) ERR2("unknown attribute", a.c_str());
    return;
  }
  *p = v;
}

//-------------------------------------------------------------------

sint4 ScriptObj::get_int(const string &a) const {
  const stype *p = get_const_val_ptr(a);
  if (!p) ERR2("unknown attribute", a.c_str());
  return p->to_int();
}

//-------------------------------------------------------------------

sint4 ScriptObj::get_int_prev(const string &a) const {
  stype p = get_val_prev(a);
  return p.to_int();
}

//-------------------------------------------------------------------

sintptr* ScriptObj::get_int_ptr(const string &a) {
  stype *p = get_val_ptr(a);
  if (!p) return 0;
  if (p->type != stype::INT) ERR2("not integer", a.c_str());
  return &p->val.i;
}

//-------------------------------------------------------------------

const sintptr* ScriptObj::get_const_int_ptr(const string &a) const {
  const stype *p = get_const_val_ptr(a);
  if (!p) return 0;
  if (p->type != stype::INT) ERR2("not integer", a.c_str());
  return &p->val.i;
}

//-------------------------------------------------------------------

void ScriptObj::set_int(const string &a, sint4 v, bool warn) {
  sintptr *p = get_int_ptr(a);
  if (!p) {
    if (warn) ERR2("unknown attribute", a.c_str());
    return;
  }
  *p = v;
}

//===================================================================

GameObj::GameObj(GameObjBlueprint &bp, GameObj *p)
{
  ++num_created;
  //  cout << "c,d = " << num_created << " " << num_deleted << endl;
  num_variable = bp.num_variable;
  num_stable = bp.num_stable;
  num_const = bp.num_const;

  if (p == NULL) p = this;

  dir_dmg = 0;
  dir_max_damage = 0;

  blueprint = &bp;
  //  cout << "me = " << (void*)this << "  type = " << bp_name() << endl;

  attr_vals = blueprint->attr_vals;
  prev_attr_vals = attr_vals;

  enabled = blueprint->enabled;
  enable_count = blueprint->enable_count;

  FORU (i, bp.root_action_count) {
    //    assert(bp.action_names[i] != "");
    p->action_enabled_list.push_back(&enable_count);
    p->action2component.push_back(this);
  }

  // Add names to the map, fixing the addresses (in the map) of the action pointers.
  FORALL (bp.action_names, i) last_command_map[*i] = 0;
  
  FORALL (last_command_map, act_it) {
    p->last_command.push_back(&act_it->second); // let the root know where the pointers are
  }

  components.resize(bp.component.size());

  // IMPORTANT: must traverse the components in the same mannner as GameObjBlueprint::lock(...),
  //            this ensures that the action -> component maps are correct.
  FORALL (bp.name2comp, it) {

    name2comp[it->first] = it->second;

    const GameObjBlueprint::BP_Const &bpc = bp.component[it->second];
    if (bpc.bp) {
      components[it->second] = new GameObj(*bpc.bp, p); // create component from blueprint
    } else {
      ScriptObj *s = bpc.c.c();  // create component from constructor
      components[it->second] = s;
      s->parent = this;
    }
  }

  parent = p;
}

//===================================================================
/*
// For object creation from a clean blueprint-object
GameObj::GameObj(const GameObj &gob, GameObj *p)
{
  num_variable = gob.num_variable;
  num_stable = gob.num_stable;
  num_const = gob.num_const;

  parent = p == NULL ? this : p;

  dir_dmg = 0;
  dir_max_damage = 0;

  blueprint = gob.blueprint;

  attr_vals = blueprint->attr_vals;
  prev_attr_vals = attr_vals;

  enabled = blueprint->enabled;
  enable_count = blueprint->enable_count;

  FORALL (blueprint->action_names, i) {
    if (*i == "") continue;
    parent->action_enabled_list.push_back(&enable_count);
    parent->action2component.push_back(this);
  }

  // Add names to the map, fixing the addresses (in the map) of the action pointers.
  FORALL (blueprint->action_names, i) last_command_map[*i] = 0;
  
  FORALL (last_command_map, act_it) {
    parent->last_command.push_back(&act_it->second); // let the root know where the pointers are
  }

  FORALL (blueprint->component_name2bp, it) {
    components[(*it).first] = new GameObj(*(*it).second, parent);
  }

  if (act) {
    Vector<GameObj*> objs;
    Vector<stype> params;
    objs.push_back(this);
    
    FORALL (blueprint->constructor, i) {
      ++num_refs; // automatically decremented when the action terminates
      (*i)->Act(objs, params);
    }
  }
}
*/
//===================================================================

GameObj::~GameObj()
{
  ++num_deleted;
  //  cout << "c,d = " << num_created << " " << num_deleted << endl;

  FORALL (components, it) delete *it;
  components.clear();

  FORALL (last_command_map, it) {
    if (it->second) {
      if (!it->second->queue_ref) { // the queue has abandoned it
        delete it->second; it->second = 0;
      } else {
        it->second->obj_ref = false; // remove our claim
      }
    }
  }
}

//===================================================================

void GameObj::init_sod()
{
  sod.shape     = get_int_ptr("shape");
  sod.x         = get_int_ptr("x");
  sod.y         = get_int_ptr("y");
  sod.x1        = get_int_ptr("x1");
  sod.y1        = get_int_ptr("y1");
  sod.x2        = get_int_ptr("x2");
  sod.y2        = get_int_ptr("y2");
  sod.radius    = get_int_ptr("radius");

  sod.owner     = get_int_ptr("owner");

  sod.max_speed = get_int_ptr("max_speed");
  sod.speed     = get_int_ptr("speed");
  sod.is_moving = get_int_ptr("is_moving");
  sod.zcat      = get_int_ptr("zcat");
  sod.collides  = get_int_ptr("collides");
  sod.ncollides = get_int_ptr("ncollides");

  sod.sight     = get_int_ptr("sight");
  sod.cloaked   = get_int_ptr("cloaked");
  sod.cloaked2  = get_int_ptr("cloaked2");
  sod.detector  = get_int_ptr("detector");

  sod.is_boundary = false;

#define INIT_SOD_PTR(x, y) if (x == 0) { x = new sintptr; *x = (y); }
  INIT_SOD_PTR(sod.shape, Object::SHAPE_UNDEF);
  INIT_SOD_PTR(sod.x, -1);
  INIT_SOD_PTR(sod.y, -1);
  INIT_SOD_PTR(sod.x1, -1);
  INIT_SOD_PTR(sod.y1, -1);
  INIT_SOD_PTR(sod.x2, -1);
  INIT_SOD_PTR(sod.y2, -1);
  INIT_SOD_PTR(sod.radius, -1);

  INIT_SOD_PTR(sod.owner, -1);

  INIT_SOD_PTR(sod.max_speed, 0);
  INIT_SOD_PTR(sod.speed,     0);
  INIT_SOD_PTR(sod.is_moving, 0);
  INIT_SOD_PTR(sod.zcat,      Object::ZCAT_UNDEF);
  INIT_SOD_PTR(sod.collides,  1 << (*sod.zcat));
  INIT_SOD_PTR(sod.ncollides, 0);

  INIT_SOD_PTR(sod.sight,    -1);
  INIT_SOD_PTR(sod.cloaked,   0);
  INIT_SOD_PTR(sod.cloaked2,  0);
  INIT_SOD_PTR(sod.detector,  0);

  sintptr *temp;
  temp = get_int_ptr("heading");  if (temp) *temp = 0;
  *sod.speed = 0;

  update_sod();
}

//-------------------------------------------------------------------

void GameObj::update_sod()
{
  update_points();
}

//-------------------------------------------------------------------

void GameObj::update_points() const
{
  if (*sod.shape == Object::RECTANGLE) {
    sint4 w = get_int("width");
    sint4 h = get_int("height");

    *sod.x1 = *sod.x - w/2;
    *sod.y1 = *sod.y - h/2;
    *sod.x2 = *sod.x1 + w;
    *sod.y2 = *sod.y1 + h;
  }

  if (*sod.shape == Object::LINE) {
    *sod.x = (*sod.x1 + *sod.x2) / 2;
    *sod.y = (*sod.y1 + *sod.y2) / 2;
  }  
}

//===================================================================

void GameObj::attr_names(Vector<string> &names) const
{
  FORALL (blueprint->attr_names, it) names.push_back(*it);
}

//-------------------------------------------------------------------

bool GameObj::attr_changed(const string &s) const
{
  sint4 i = blueprint->attr_index(s);
  if (i < 0) return false;
  return prev_attr_vals[i].val.i != attr_vals[i].val.i;
}

//-------------------------------------------------------------------

stype* GameObj::get_val_ptr(const string &s)
{
  sint4 i = blueprint->attr_index(s);
  if (i < 0) return 0;
  return &attr_vals[i];
}

//-------------------------------------------------------------------

const stype* GameObj::get_const_val_ptr(const string &s) const
{
  sint4 i = blueprint->attr_index(s);
  if (i < 0) return 0;
  return &attr_vals[i];
}

//-------------------------------------------------------------------

stype GameObj::get_val_prev(const string &s) const
{
  sint4 i = blueprint->attr_index(s);
  if (i < 0) {
    cerr << "object type = " << blueprint->name << endl;
    ERR2("unknown attribute ", s.c_str());
  }

  return prev_attr_vals[i];
}

//-------------------------------------------------------------------
/*
sint4 GameObj::get_int2(const string &s, sint4 &prev) const {
  FIND(it, blueprint->attributes, s);
  if (it == blueprint->attributes.end()) { ERR2("unknown attribute ", s.c_str()); }

  sint4 i = (*it).second;
  assert (i >= 0);
  prev = prev_attr_vals[i].val.i;
  return attr_vals[i].val.i;
}
*/
//-------------------------------------------------------------------

/** Assign a value to an integer parameter.
    @param a     Parameter name.
    @param val   Value to assign.
    @param warn  Signal an error if the parameter does not exist.
*/

void GameObj::set_val(const string &a, stype val, bool warn) {
  sint4 i = blueprint->attr_index(a);
  if (i < 0) {
    if (warn) ERR2("unknown attribute", a.c_str());
    return;
  }
  if (blueprint->attr_props[i].constant) { ERR2("overwrite constant attribute ", a.c_str()); }
  attr_vals[i] = stype(val);
}

//-------------------------------------------------------------------

void GameObj::print(ostream &os) const
{
  os << sod.is_boundary << " "
     << *sod.x << " " << *sod.y << " "
     << *sod.x2 << " " << *sod.y2 << " "
     << real8(smd.fine_pos.x)/4096 << " " << real8(smd.fine_pos.y)/4096 << " "
     << real8(smd.fine_heading.x)/4096 << " " << real8(smd.fine_heading.y)/4096 << " "
     << *sod.radius << endl;
}

//-------------------------------------------------------------------

/** Sets "last_direction" (attacking) and "dir_damage" for this object and the target resp. */

void GameObj::notify_attack(GameObj *targ, sint4 damage)
{
  targ = targ->parent->get_GameObj();
  if (!targ) return;

  sint4 dx = *targ->sod.x - *sod.x;
  sint4 dy = *targ->sod.y - *sod.y;

  sint4 attack_angle = GameConst::encode_dir(dx, dy, GameConst::ATTACK_N);
  set_val("last_direction", stype(attack_angle), false); // ignore if the variable doesn't exist

  sint4 damage_dir = GameConst::encode_dir(-dx, -dy, GameConst::DAMAGE_N);

  if (damage >= 0) {
    targ->dir_dmg |= 1 << (damage_dir-1);
    targ->dir_damage[damage_dir] += damage;
    if (targ->dir_damage[damage_dir] > targ->dir_damage[targ->dir_max_damage]) {
      targ->dir_max_damage = damage_dir;
    }
  }
}

//===================================================================

/** Compute a difference vector for an object's attributes.
    @param first  True iff this is the first difference computation (no delta).
    @param allied Includes hidden object information.
    @param s      Append the difference string to s.
*/

void GameObj::gen_diff(bool first, bool allied, string &s, PlayerInfo * /*pi*/) const
{
  s.clear();

  if (first) write_b_int4(s, 1);
  else       write_b_int4(s, 0);

  // damage direction
#if DAMAGE_BITMASK       // encode as a bit for each direction
  //  *u++ = dir_dmg;
  if (GameConst::DAMAGE_N < 32) {
    write_b_int4(s, dir_dmg);
  } else if (GameConst::DAMAGE_N == 32) {
    write_b_flat4(s, dir_dmg);
  } else {
    ERR("DAMAGE_N too large");
  }
#else                    // encode as one direction
  if (dir_dmg) {         // bit vector of all the directions damage has come from
    //    assert(dir_max_damage < 64); // up to 63 is encoded as a byte of the same value
    //    *u++ = dir_max_damage;
    write_b_int4(s, dir_max_damage);
  } else {
    s += char(0); // 0 -> no (directional) damage this turn
  }
#endif

  gen_comp_diff(first, allied, s);
}

//-------------------------------------------------------------------

void GameObj::gen_comp_diff(bool first, bool allied, string &s) const
{
  Vector<AttrProp> &props = blueprint->attr_props;
  sint4 i = 0;

  // enabled status

  write_b_int4(s, enable_count);

  // attributes

  FORALL (props, it) {
    if (i == (num_variable + num_stable)) break;

    if (!it->internal && (allied || it->visible_to_opponents)) {
      
      sint4 d = attr_vals[i].val.i;
      if (!first) { d -= prev_attr_vals[i].val.i; }
      
      write_b_int4(s, d);
      
    } else {
      s += char(0x40); // "-0" (aka not a number)
    }

    ++i;
  }

  FORU (i, blueprint->num_comp) {
    const CompProp &cp = blueprint->comp_props[i];
    if (cp.constant || cp.internal) continue;
    if (allied || cp.visible_to_opponents) {
      ScriptObj *m = components[i];
      m->gen_comp_diff(first, allied, s);
    }
  }
}

//===================================================================

/** Update an object from a difference vector computed by gen_diff.
    @return true iff object changed.

    @todo Increase speed by using runlength encoding for zeros.
*/

bool GameObj::apply_diff(bool allied, const char* &s)
{
  char c = *s++;
  bool first = c == 1;
  if (c != 0 && c != 1) ERR("neither 0 nor 1");

  // damage direction

#if DAMAGE_BITMASK
  if (GameConst::DAMAGE_N < 32) {
    dir_dmg = read_b_int4(s); // *s++;
  } else if (GameConst::DAMAGE_N == 32) {
    dir_dmg = read_b_flat4(s);
  } else {
    ERR("DAMAGE_N too large");
  }
#else
  // up to 63 is encoded as a byte of the same value, assume the value is small enough
  dir_max_damage = read_b_int4(s);// *s++;
#endif

  bool changed = apply_comp_diff(allied, s, first);
  update_sod();
  return changed;
}

//-------------------------------------------------------------------

bool GameObj::apply_comp_diff(bool allied, const char* &s, bool first)
{
  bool changed = false;

  enable_count = read_b_int4(s); // enabled status (-oo, 1]

  // attributes

  Vector<ObjAction*> hooks; // actions to call for changed variables

  FORS (i, num_variable + num_stable) {
    sintptr &val = attr_vals[i].val.i;

    if (*s == 0x40) {
      
      val = stype::UNDEF_INT;
      ++s;
      
    } else {
      
      sint4 q = read_b_int4(s);
      if (first) {
	val = q;
        // fixme: might be more appropriate to call event_changed here
        prev_attr_vals[i] = stype(val); // new object, hasn't really "changed"
      } else {
        prev_attr_vals[i] = stype(val);

        if (q) {
          val += q;
          if (abs(val) > stype::MAX_INT) ERR2("value too big", val);
          changed = true;
          
          ObjAction *hook = blueprint->hooks[i];
          if (hook) hooks.push_back(hook);
        }
      }
    }
  }

  FORU (i, blueprint->num_comp) {
    const CompProp &cp = blueprint->comp_props[i];
    if (cp.constant || cp.internal) continue;
    if (allied || cp.visible_to_opponents) {
      ScriptObj *m = components[i];
      changed |= m->apply_comp_diff(allied, s, first);
    }
  }

  // call the hooks, now that all of the components have been updated
  // fixme: siblings might not be updated (only parents and children are guaranteed)

  FORALL (hooks, i) {
    SActParam p;
    p.obj.push_back(this);
    //          p.val.push_back(q);
    (*i)->Act(p);
  }

  return changed;
}

//===================================================================

/* Clears the special non-scripted attributes. (damage direction, etc) */

void GameObj::reset_frame_data() {
  if (dir_dmg) {
    memset (dir_damage, 0, sizeof(dir_damage)); // array, not pointer!
    dir_dmg = 0;
  }
  set_val("last_direction", stype(0), false); // ignore if the variable doesn't exist
}

/* Sets prev_attr_vals to attr_vals. */

void GameObj::copy_to_prev() {
  if (prev_attr_vals.size())
    memcpy (&prev_attr_vals[0], &attr_vals[0], sizeof(attr_vals[0]) * prev_attr_vals.size());
  //  prev_attr_vals = attr_vals;
  FORALL (components, it) (*it)->copy_to_prev();
}

//==============================================================================

void GameObj::action_names(Vector<std::string> &names) const
{
  FORALL (blueprint->action_names, it) names.push_back(*it);
}

//-------------------------------------------------------------------


bool GameObj::action_already_set() const
{
  return cur_action.is_set();
}

//-------------------------------------------------------------------

/** Sets the outgoing action for this object or its parent. */

uint4 GameObj::set_action(const string &name, const Vector<sint4> &params, sint4 priority)
{
  GameObj *par = parent->get_GameObj();

  if (!par) {
    cout << "blueprint = " << bp_name() << endl;
    cout << "action name = " << name << endl;
    ERR("OBJ HAS NO PARENT!");
    return 0;
  }
  if (!par->sod.in_game) ERR("set_action called for non-game object");


  if (par->cur_action.get_priority() > priority) {
    /*
    cout << "===================================================================" << endl;
    cout << "blueprint = " << bp_name() << endl;
    cout << "action name = " << name << endl;
    cout << "LOWER PRIORITY" << endl;
    */
    return 0;
  }
  /*
  if (par->action_already_set()) {

    // fixme: better error message (relating to object)   
    cout << "===================================================================" << endl;
    cout << "blueprint = " << bp_name() << endl;
    cout << "action name = " << name << endl;
    cout << "ACTION ALREADY SET!" << endl;
  }
  */
  if (name == "move") {
    if (!sod.in_game) ERR("set_action called for non-game object");
    sint4 speed;
    sint4 id;
    Vector<sint4> p;

    if (params.size() == 3) {
      speed = params[2];
      p = params;
      p.pop_back();
    } else if (params.size() == 2) {
      speed = *sod.max_speed;
      p = params;
    } else {
      ERR("wrong number of move parameters");
      return 0;
    }
    id = -1 - speed;
    cur_action.set(id, p, priority);

    return cur_action.client_action_id = GameObj::next_client_action_id();
  }

  FIND (blueprint->action_name2id, it, name);
  if (it == blueprint->action_name2id.end()) {
    cout << "blueprint = " << bp_name() << endl;
    cout << "action name = " << name << endl;
    cout << "ACTION NAME NOT FOUND!" << endl;    
    return 0;
  }

  sint4 cid = it->second;

  uint4 a = par->blueprint->actions[cid]->num_obj_ids;
  uint4 b = par->blueprint->actions[cid]->num_params;
  if ((a + b) != params.size()) {
    cout << "blueprint = " << bp_name() << endl;
    cout << "action name = " << name << endl;
    cout << "action id = " << cid << endl;
    cout << "parameter size mismatch" << endl;
    return 0;
  }

  par->cur_action.set(cid, params, priority);

  return cur_action.client_action_id = GameObj::next_client_action_id();
}

//------------------------------------------------------------------------------

uint4 GameObj::set_action(sint4 id, const Vector<sint4> &params, sint4 priority)
{
  if (cur_action.get_priority() > priority) return 0;

  if (action_already_set()) {
    // fixme: better error message (relating to object)
    cout << "blueprint = " << bp_name() << endl;
    cout << "action id = " << id << endl;
    cout << "ACTION ALREADY SET!" << endl;
  }

  //  if (cur_action.id < 0) return false; 

  uint4 a = blueprint->actions[id]->num_obj_ids;
  uint4 b = blueprint->actions[id]->num_params;
  if ((a + b) != params.size()) {
    cout << "blueprint = " << bp_name() << endl;
    cout << "action id = " << id << endl;
    if (id < static_cast<sint4>(blueprint->action_names.size()))
      cout << "action name = " << blueprint->action_names[id] << endl;
    cout << "PARAMETER SIZE MISMATCH " << a << " " << b << " " << params.size() << endl;
    return false;
  }

  cur_action.set(id, params, priority);

  return cur_action.client_action_id = GameObj::next_client_action_id();
}

//------------------------------------------------------------------------------

void GameObj::clear_action()
{
  cur_action.clear();
}

//------------------------------------------------------------------------------

/** Adds the current action to the list of actions awaiting a response from the server. */

void GameObj::queue_action(sint4 frame_sent)
{
  if (cur_action.is_set())
    pending_actions.push_back(ActInfo(frame_sent, cur_action.client_action_id, this));
  cur_action.clear();
}

//------------------------------------------------------------------------------

string GameObj::write_action(ObjId my_id) const
{
  return cur_action.to_string(my_id);
}

//==============================================================================

ObjAction* GameObj::action(const string &name) const
{
  FIND (blueprint->action_map, t, name);
  if (t == blueprint->action_map.end()) return 0;
  return t->second;
}

//------------------------------------------------------------------------------

sint4 GameObj::action_id(const string &name) const
{
  FIND (blueprint->action_name2id, it, name);
  if (it == blueprint->action_name2id.end()) return -1;
  return it->second;
}

//------------------------------------------------------------------------------

bool GameObj::action_enabled(sint4 id) const
{
  return bool(*action_enabled_list[id]);
}

//------------------------------------------------------------------------------

bool GameObj::action_enabled(const string &name) const
{
  return bool(*action_enabled_list[action_id(name)]);
}

//------------------------------------------------------------------------------

bool GameObj::action_hidden(sint4 id) const
{
  return action(id)->hidden;
}

//------------------------------------------------------------------------------

bool GameObj::action_hidden(const string &name) const
{
  return action(name)->hidden;
}

//==============================================================================

/** Number of things that reference this object and its components. */

sint4 GameObj::get_refs() const
{
  sint4 n_refs = num_refs;
  FORALL(components, it) n_refs += (*it)->get_refs();

  return n_refs;
}

void GameObj::inc_enable_count()
{
  enable_count++;
  FORALL(components, it) {
    GameObj *gob = (*it)->get_GameObj();
    if (gob) gob->inc_enable_count();
  }
}

void GameObj::dec_enable_count()
{
  enable_count--;
  FORALL(components, it) {
    GameObj *gob = (*it)->get_GameObj();
    if (gob) gob->dec_enable_count();
  }
}

void GameObj::enable()
{
  if (!enabled) {
    enabled = true;
    inc_enable_count();
  }
}

void GameObj::disable()
{
  if (enabled) {
    enabled = false;
    dec_enable_count();
  }
}

void GameObj::component_names(Vector<std::string> &names) const
{
  names.clear();
  FORALL (name2comp, it) {
    names.push_back(it->first);
  }
}

ScriptObj* GameObj::component(const string &name) const
{
  FIND (name2comp, it, name);
  if (it != name2comp.end()) return components[it->second];
  return 0;
}

/** Calls ObjAction (scripted, C) constructors. */

void GameObj::construct()
{
  FORALL (components, it) (*it)->construct();

  SActParam p;
  p.obj.push_back(this);

  FORALL (blueprint->constructor, i) {
    (*i)->Act(p);
  }
}

/** Calls ObjAction (scripted, C) destructors, not ~GameObj(). */

void GameObj::destruct()
{
  SActParam p;
  p.obj.push_back(this);

  FORALL (blueprint->destructor, i) {
    (*i)->Act(p);
  }

  FORALL (components, it) (*it)->destruct();
}

std::string GameObj::bp_name() const { return blueprint->name; }

uint4 GameObj::code() const { return blueprint ? blueprint->code : -1; }

ObjAction* GameObj::action(sint4 id) const { return blueprint->actions[id]; }


//===================================================================
// Object virtual functions
//===================================================================

Object::Shape GameObj::get_shape() const { return (Object::Shape)*sod.shape; }

//-------------------------------------------------------------------

sint4 GameObj::get_max_speed() const { return *sod.max_speed; }

//-------------------------------------------------------------------

sint4 GameObj::get_radius() const { assert(*sod.shape == Object::CIRCLE); return *sod.radius; }

//-------------------------------------------------------------------

Object::ZCat GameObj::get_zcat(ZCat *prev) const
{
  if (prev) *prev = (Object::ZCat)get_int_prev("zcat");  
  return (Object::ZCat)*sod.zcat;
}

//-------------------------------------------------------------------

sint4 GameObj::get_speed(sint4 *prev) const
{
  if (prev) *prev = get_int_prev("speed");
  return *sod.speed;
}

//-------------------------------------------------------------------

bool GameObj::get_moving(bool *prev) const
{
  if (prev) *prev = get_int_prev("is_moving");
  return *sod.is_moving;
}

//-------------------------------------------------------------------

void GameObj::get_center(sint4 &x, sint4 &y, sint4 *px, sint4 *py) const
{
  assert(*sod.shape != Object::SHAPE_UNDEF);

  switch (*sod.shape) {
  case Object::CIRCLE:
  case Object::RECTANGLE:

    if (px) *px = get_int_prev("x");
    if (py) *py = get_int_prev("y");
    x = *sod.x;
    y = *sod.y;
    break;

  case Object::LINE:

    if (px) *px = (get_int_prev("x1") + get_int_prev("x2")) / 2;
    if (py) *py = (get_int_prev("y1") + get_int_prev("y2")) / 2;
    x = (*sod.x1 + *sod.x2) / 2;
    y = (*sod.y1 + *sod.y2) / 2;

  default:
    ERR("invalid shape");
  }
}

//-------------------------------------------------------------------

void GameObj::get_p1(sint4 &x, sint4 &y, sint4 *px, sint4 *py) const
{
  assert(*sod.shape == Object::RECTANGLE || *sod.shape == Object::LINE);
  update_points();

  x = *sod.x1;
  y = *sod.y1;

  if (px) *px = get_int_prev("x1");
  if (py) *py = get_int_prev("y1");
}

//-------------------------------------------------------------------

void GameObj::get_p2(sint4 &x, sint4 &y, sint4 *px, sint4 *py) const
{
  assert(*sod.shape == Object::RECTANGLE || *sod.shape == Object::LINE);
  update_points();

  x = *sod.x2;
  y = *sod.y2;

  if (px) *px = get_int_prev("x2");
  if (py) *py = get_int_prev("y2");
}

//-------------------------------------------------------------------

bool GameObj::is_pending_action(void) const
{
  return cur_action.is_set() || !pending_actions.empty();
}

//-------------------------------------------------------------------

real8 GameObj::distance_to(const Object &other) const
{
/*  // fixme: change Game::distance() to work with Object's, and remove this assertion
  const GameObj* gob = dynamic_cast<const GameObj*>(&other);
  assert(gob && "Can only compute distance to another GameObj");*/
  return Game::distance(*this, other);
}

//-------------------------------------------------------------------

uint4 GameObj::next_client_action_id()
{
  static uint4 id = 1;
  return id++;
}
  
//-------------------------------------------------------------------

