/** @file GameObjFactory.C
    @see GameObjFactory.H

    $Id: GameObjFactory.C,v 1.18 2006/04/25 03:08:38 orts_mburo Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/GameObjFactory.C,v $
*/
// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "GameObjFactory.H"
#include "GameObj.H"
#include "Blueprints.H"
#include "PlayerInfo.H"
#include "ScriptParser.H"

using namespace std;

//===================================================================

GameObjFactory::~GameObjFactory()
{
  FORALL (name2blueprint, it) {
    GameObjBlueprint *bp = it->second;
    delete bp;
  }
}

//===================================================================

void comp_enable(GameObjBlueprint *bp, sint4 c)
{
  if (!bp->enabled) c--;
  bp->enable_count = c;
  FORALL (bp->component, it) {
    if ((*it).bp) comp_enable((*it).bp, c);
  }
}

//===================================================================

sint4 GameObjFactory::check_registration(const string &name)
{
  if (blueprint_names.find(name) != blueprint_names.end()) return 1;
  return 0;
}

//===================================================================

void GameObjFactory::register_blueprint(const string &name, const GameObjBlueprint &_bp)
{
  if (check_registration(name)) ERR2("double registration:", name.c_str());

  GameObjBlueprint *bp = new GameObjBlueprint(_bp);
  bp->name = name;
  
  comp_enable(bp, 1);

  // make sure these are blank, they should have elements after binding
  bp->actions.clear();

  if (bp->can_instance) {
    bp->code = code++;
    code2blueprint[bp->code] = bp;
    name2code[name] = bp->code;
  } else {
    bp->code = -1;
  }

  name2blueprint[name] = bp;
}

//===================================================================

void GameObjFactory::register_blueprint(const string &name, ScriptObjConstructor c, sint4 num_actions)
{
  if (check_registration(name)) ERR2("double registration:", name.c_str());

  sint4 mycode = code++;

  BPConstructorData b;
  b.c = c;
  b.num_actions = num_actions;

  name2constructor[name] = b;
  code2constructor[mycode] = b;
  name2code[name] = mycode;
}

//===================================================================

GameObjBlueprint* GameObjFactory::get_blueprint(const string &name)
{
  FIND (name2blueprint, it, name);
  return it != name2blueprint.end() ? it->second : 0;
}

//===================================================================

/** Parse all scripted actions and resolve references to external actions
    for all blueprints.
*/

void GameObjFactory::bind(PlayerInfo &pi, Game &game)
{
  this->pi = &pi;

  // parse (new) global actions
  for (uint4 i = GameObjBlueprint::ext_actions.size(); i < GameObjBlueprint::ext_action_defs.size(); ++i) {
    ObjActionDefinition def = GameObjBlueprint::ext_action_defs[i];
    if (def.act_type == ObjActionDefinition::SCRIPT) {
      ScriptParser parser;
      parser.set_context(pi, game);
      parser.current_action_name = def.name;

      ActionScript *script = parser.Parse(def);
      GameObjBlueprint::ext_actions.push_back(new ObjAction(script, def));
    } else if (def.act_type == ObjActionDefinition::FUNCTION) {
      GameObjBlueprint::ext_actions.push_back(def.func);
    } else {
      ERR("TRYING TO BIND ACTION OF UNKNOWN TYPE");
    }
  }

  // bind and attach actions to each blueprint
  FORALL(name2blueprint, it) {
    GameObjBlueprint *bp = it->second;
    bp->bind(pi, game, it->first);
  }
}

//===================================================================

/** Solidify the layout of blueprint attributes and actions.
*/

void GameObjFactory::lock()
{
  FORALL(name2blueprint, it) {
    it->second->lock();
  }
}

//===================================================================

sint4 GameObjFactory::bp_code(const string &name)
{
  FIND (name2code, it, name);
  if (it == name2code.end()) {
    cout << "could not find code for " << name << endl;
    return -1;
  }
  return it->second;
}

//===================================================================

std::string GameObjFactory::bp_name(sint4 code)
{
  FIND (code2blueprint, it, code);
  if (it == code2blueprint.end()) return "";
  return it->second->name;
}

//===================================================================

ScriptObj* GameObjFactory::create(const string &name, bool construct)
{
  FIND (name2blueprint, it, name);
  FIND (name2constructor, it2, name);
  if (it == name2blueprint.end() &&
      it2 == name2constructor.end()) ERR2("not registered: ", name.c_str());

  ScriptObj *obj=0;
  if (it != name2blueprint.end()) {
    GameObjBlueprint *bp = it->second;
    if (!bp->can_instance) ERR2("cannot create; type is 'class' not 'blueprint':", name.c_str());
    obj = new GameObj(*it->second);
  }
  if (it2 != name2constructor.end()) obj = it2->second.c();
  if (construct) obj->construct();

  GameObj *gob = obj->get_GameObj();
  if (gob != 0) {
    FORS(i, game->get_player_num())
      gob->sod.last_lost.push_back(game->tick()-1);
  }

  obj->creator = pi;

  return obj;
}

//===================================================================

ScriptObj* GameObjFactory::create(sint4 code, bool construct)
{
  FIND (code2blueprint, it, code);
  if (it == code2blueprint.end()) ERR2("not registered: ", code);
  GameObj *gob = new GameObj(*it->second);
  FORS (i, game->get_player_num())
    gob->sod.last_lost.push_back(game->tick()-1);

  if (construct) gob->construct();

  gob->creator = pi;
  
  return gob;
}

//===================================================================
