#ifndef LinAlg_H
#define LinAlg_H

// $Id: LinAlg.H,v 1.1 2006/05/28 17:58:58 orts_furtak Exp $
// This is an ORTS file (c) Keith Yerex, licensed under the GPL

#include "Random.H"

//===================================================================

template <class T> class Vec2
{
public:

  Vec2(T X, T Y) { x = X; y = Y; }
  Vec2(const T *B) { x = B[0]; y = B[1]; }
  Vec2() { x = 0; y = 0; }

  union{
    struct{
      T x, y;
    };
    T A[2];
  };

  const T* ptr() const { return A; }
  T* ptr() { return A; }
  // operator const T *() const { return A; }
  // operator T *() { return A;};
  const T & operator [](int i) const { return A[i]; }
  T & operator [](int i) { return A[i]; }

  Vec2 operator -() const { return Vec2(-x,-y); }

  Vec2 operator +(const Vec2 &v) const { return Vec2(x+v.x, y+v.y); }
  Vec2 operator -(const Vec2 &v) const { return Vec2(x-v.x, y-v.y); }
  Vec2 operator *(const Vec2 &v) const { return Vec2(x*v.x, y*v.y); }
  Vec2 operator /(const Vec2 &v) const { return Vec2(x/v.x, y/v.y); }

  inline void operator +=(const Vec2 &v) { x+=v.x; y+=v.y; }
  inline void operator -=(const Vec2 &v) { x-=v.x; y-=v.y; }
  inline void operator *=(const T &n) { x*=n; y*=n; }
  inline void operator /=(const T &n) { T m = (1.0/n); x*=m; y*=m; }

  Vec2 operator +(const T &n) const { return Vec2(x+n, y+n); }
  Vec2 operator -(const T &n) const { return *this+(-n); }
  Vec2 operator *(const T &n) const { return Vec2(x*n, y*n); }
  Vec2 operator /(const T &n) const { return *this*(1.0/n); }

  T length() const { return sqrt(x*x+y*y); }
  
  void normalize() {T l = 1.0/length(); x *= l; y *= l; }

  T dot(const Vec2 &v) const { return v.x*x + v.y*y; }

  Vec2 rotate(T a) {
    T c = cos(a), s = sin(a);
    return Vec2(x*c - y*s, x*s + y*c);
  }

  void print(std::ostream &os=std::cout) const {
    os << "[" << x << ", " << y << "]" << std::endl;
  }

  static Vec2 vrand(Random &r) { return Vec2(2.0*r.rand_real8()-1.0,
					     2.0*r.rand_real8()-1.0); }
  static Vec2 min(const Vec2 &a, const Vec2 &b) {
    return Vec2(std::min(a.x,b.x), std::min(a.y,b.y)); }

  static Vec2 max(const Vec2 &a, const Vec2 &b) {
    return Vec2(std::max(a.x,b.x), std::max(a.y,b.y)); }
};

//===================================================================

template <class T> class Vec3
{
public:

  Vec3(T X, T Y, T Z) { x = X; y = Y; z = Z; }
  Vec3(const T *B) { x = B[0]; y = B[1]; z = B[2]; }
  Vec3() { x = 0; y = 0; z = 0; }

  union{
    struct{
      T x, y, z;
    };
    struct{
      T r, g, b;
    };
    T A[3];
  };

  const T* ptr() const { return A; }
  T* ptr() { return A; }
  // operator const T *() const { return A; }
  // operator T *() { return A;};
  const T & operator [](int i) const { return A[i]; }
  T & operator [](int i) { return A[i]; }

  Vec3 operator -() const { return Vec3(-x,-y,-z); }

  Vec3 operator +(const Vec3 &v) const { return Vec3(x+v.x, y+v.y, z+v.z); }
  Vec3 operator -(const Vec3 &v) const { return Vec3(x-v.x, y-v.y, z-v.z); }
  Vec3 operator *(const Vec3 &v) const { return Vec3(x*v.x, y*v.y, z*v.z); }
  Vec3 operator /(const Vec3 &v) const { return Vec3(x/v.x, y/v.y, z/v.z); }

  inline void operator +=(const Vec3 &v) { x+=v.x; y+=v.y; z+=v.z; }
  inline void operator -=(const Vec3 &v) { x-=v.x; y-=v.y; z-=v.z; }
  inline void operator *=(const T &n) { x*=n; y*=n; z*=n; }
  inline void operator /=(const T &n) { T m = (1.0/n); x*=m; y*=m; z*=m; }

  Vec3 operator +(const T &n) const { return Vec3(x+n, y+n, z+n); }
  Vec3 operator -(const T &n) const { return *this+(-n); }
  Vec3 operator *(const T &n) const { return Vec3(x*n, y*n, z*n); }
  Vec3 operator /(const T &n) const { return *this*(1.0/n); }

  T length() const { return sqrt(x*x+y*y+z*z); }
  
  void normalize() {T l = 1.0/length(); x *= l; y *= l; z *= l; }

  T dot(const Vec3 &v) const { return v.x*x + v.y*y + v.z*z; }

  Vec3 cross(const Vec3 &v) const {
    return Vec3(y*v.z - z*v.y, z*v.x - x*v.z, x*v.y - y*v.x);
  }

  Vec3 rotate(const Vec3 &axis, T a) {
    T c = cos(a), s = sin(a);
    return (*this)*c + cross(axis)*s + axis*dot(axis)*(1-c);
  }

  void print(std::ostream &os=std::cout) const {
    os << "[" << x << ", " << y << ", " << z << "]" << std::endl;
  }

  // false iff error
  bool write(FILE *f) const { return fwrite(A, sizeof(*this), 1, f) == 1; }
  bool read (FILE *f) { return fread(A, sizeof(*this), 1, f) == 1; }

  static Vec3 vrand(Random &r) { return Vec3(2.0*r.rand_real8()-1.0,
					     2.0*r.rand_real8()-1.0,
					     2.0*r.rand_real8()-1.0); }
  static Vec3 min(const Vec3 &a, const Vec3 &b) {
    return Vec3(std::min(a.x,b.x), std::min(a.y,b.y), std::min(a.z,b.z)); }

  static Vec3 max(const Vec3 &a, const Vec3 &b) {
    return Vec3(std::max(a.x,b.x), std::max(a.y,b.y), std::max(a.z,b.z)); }
};


template <class T> class Vec4 : public Vec3<T>
{
public:
  
  Vec4(T X, T Y, T Z, T W) : Vec3<T>(X, Y, Z) { w = W; }
  Vec4(const T *B) : Vec3<T>(B) { w = B[3]; }
  Vec4(const Vec3<T> &v, T W) { this->x = v.x; this->y = v.y; this->z = v.z; w = W;}

  Vec4() : Vec3<T>() { w = 1; }
    
  union{
    struct{
      T w;
    };
    struct{
      T a;
    };
  };

  Vec4 operator +(const Vec4 &v) const {
    return Vec4(this->x+v.x, this->y+v.y, this->z+v.z, w+v.w);
  }
  Vec4 operator -(const Vec4 &v) const {
    return Vec4(this->x-v.x, this->y-v.y, this->z-v.z, w-v.w);
  }
  Vec4 operator *(const Vec4 &v) const {
    return Vec4(this->x*v.x, this->y*v.y, this->z*v.z, w*v.w);
  }
  Vec4 operator /(const Vec4 &v) const {
    return Vec4(this->x/v.x, this->y/v.y, this->z/v.z, w/v.w);
  }

  Vec4 operator +(const T &n) const { return Vec4(this->x+n, this->y+n, this->z+n, w+n); }
  Vec4 operator -(const T &n) const { return *this+(-n); }
  Vec4 operator *(const T &n) const { return Vec4(this->x*n, this->y*n, this->z*n, w*n); }
  Vec4 operator /(const T &n) const { return *this*(1.0/n); }

  Vec3<T> project() const { return (*this)/w; }
  Vec4 cross(const Vec4 &v) const { return Vec4(Vec3<T>::cross(v), 0); }
  T dot(const Vec4 &v) const { return v.x*this->x + v.y*this->y + v.z*this->z + v.w*w; }

  // IMPORTANT: fix these if you change T
  void print(std::ostream &os=std::cout) const {
    os << "[" << this->x << ", " << this->y << ", " << this->z << ", " << w << "]" << std::endl;
  }

  // false iff error
  bool write(FILE *f) const { return fwrite(this->A, sizeof(*this), 1, f) == 1; }
  bool read(FILE  *f) { return fread(this->A, sizeof(*this), 1, f) == 1; }
};


template <class T> class Mat4
{
public:

  Mat4() {};

  Mat4(Vec4<T> &r0, Vec4<T> &r1, Vec4<T> &r2, Vec4<T> &r3) {
    A[0] = r0; A[1] = r1; A[2] = r2; A[3] = r3;
  }
  
  const T* ptr() const { return &A[0][0]; }
  T* ptr() { return &A[0][0]; }
  // operator const T *() const { return &(A[0][0]); }
  // operator T *() { return &(A[0][0]); }

  Vec4<T> & operator [](int i) { return A[i]; }
  // const Vec4<T> & operator [](int i) const { return A[i]; }

  Vec4<T> col(sint4 i) const { return Vec4<T>(A[0][i], A[1][i], A[2][i], A[3][i]); }

  Mat4 operator *(const Mat4 &c) const;
  Vec4<T> operator *(const Vec4<T> &v) const;
  Vec3<T> operator *(const Vec3<T> &v) const;
  Mat4 inv_quick() const; // inverse of rotation/translation matrix only
  void invert();
  Mat4 inverse() const;
  Mat4 trans() const;
  void identity();

  void print(std::ostream &os=std::cout) const { FORS (i, 4) A[i].print(os); }

  // false iff error
  bool write(FILE *f) const { return fwrite(A, sizeof(A), 1, f) == 1; }
  bool read(FILE *f) { return fread(A, sizeof(A), 1, f) == 1; }

private:

  Vec4<T> A[4];
  //  friend void glGetProjection(Mat4<float> &m);
  //  friend void glGetModelview(Mat4<float> &m);
};

//extern void glGetProjection(Mat4<float> &m);
//extern void glGetModelview(Mat4<float> &m);


// implementation

template <typename T> Mat4<T> Mat4<T>::operator *(const Mat4<T> &c) const
{
  Mat4<T> r;

  FORS (i, 4)
    FORS (j, 4) {
      r[i][j] = 0;
      FORS (k, 4) r[i][j] += A[i][k] * c.A[k][j];
    }
  return r;
};

template <typename T> Vec4<T> Mat4<T>::operator *(const Vec4<T> &v) const
{
  Vec4<T> r;

  FORS (i, 4) {
    r[i] = 0;
    FORS (j, 4) r[i] += A[i][j]*v[j];
  }
  return r;
};

template <typename T> Vec3<T> Mat4<T>::operator *(const Vec3<T> &v) const
{
  Vec3<T> r;
  FORS (i, 3) {
    r[i] = 0;
    FORS (j, 3) r[i] += A[i][j]*v[j];
    r[i] += A[i][4];
  }
  return r;
};

template <typename T> Mat4<T> Mat4<T>::trans() const
{
  Mat4 r;
  r.A[0][0] = A[0][0];
  r.A[0][1] = A[1][0];
  r.A[0][2] = A[2][0];
  r.A[0][3] = A[3][0];

  r.A[1][0] = A[0][1];
  r.A[1][1] = A[1][1];
  r.A[1][2] = A[2][1];
  r.A[1][3] = A[3][1];

  r.A[2][0] = A[0][2];
  r.A[2][1] = A[1][2];
  r.A[2][2] = A[2][2];
  r.A[2][3] = A[3][2];

  r.A[3][0] = A[0][3];
  r.A[3][1] = A[1][3];
  r.A[3][2] = A[2][3];
  r.A[3][3] = A[3][3];

  return r;
};

template <typename T> void Mat4<T>::invert()
{
  T data[16];
  T inv[16];
  T smallest = std::numeric_limits<T>::max();
  
  FORS (i, 16) data[i] = A[0][i];

  memset(inv, 0, sizeof(inv));

  inv[0] = 1;
  inv[5] = 1;
  inv[10] = 1;
  inv[15] = 1;

  FORS (j, 4) {
    T largest = 0;
    int largei = 0;
    for (int i = j; i < 4; i++){
      if (fabs(data[i+4*j]) > largest){
	largei = i;
	largest = fabs(data[i+4*j]);
      }
    }
    // pivot
    FORS (k, 4) {

      swap(data[j+4*k], data[largei+4*k]);
      swap(inv[j+4*k], inv[largei+4*k]);

#if 0      
      T temp = data[j+4*k];
      data[j+4*k] = data[largei+4*k];
      data[largei+4*k] = temp;

      temp = inv[j+4*k];
      inv[j+4*k] = inv[largei+4*k];
      inv[largei+4*k] = temp;
#endif
      
    }
    if (fabs(data[j*4+j]) < smallest)
      smallest = fabs(data[j*4+j]);
    
    T ival = 1.0/data[j+4*j];
    FORS (k, 4) {
      data[j+4*k] *= ival;
      inv[j+4*k] *= ival;
    }
    FORS (i, 4) {
      if (i != j){
	T scale = data[i+4*j];
	FORS (k, 4) {
	  data[i+4*k] = data[i+4*k] - scale*data[j+4*k];
	  inv[i+4*k] = inv[i+4*k] - scale*inv[j+4*k];
	}
      }
    }
  }

  FORS (i, 16) A[0][i] = inv[i];
}

template <typename T> Mat4<T> Mat4<T>::inverse() const
{
  Mat4<T> r(*this);
  r.invert();
  return r;
}


template <typename T> Mat4<T> Mat4<T>::inv_quick() const
{
  Mat4 r;
  r.A[0][0] = A[0][0];
  r.A[0][1] = A[1][0];
  r.A[0][2] = A[2][0];
  r.A[0][3] = -A[0][3];

  r.A[1][0] = A[0][1];
  r.A[1][1] = A[1][1];
  r.A[1][2] = A[2][1];
  r.A[1][3] = -A[1][3];

  r.A[2][0] = A[0][2];
  r.A[2][1] = A[1][2];
  r.A[2][2] = A[2][2];
  r.A[2][3] = -A[2][3];

  r.A[3][0] = 0;
  r.A[3][1] = 0;
  r.A[3][2] = 0;
  r.A[3][3] = 1;

  return r;
};

template <typename T> void Mat4<T>::identity()
{
  A[0][0] = 1; A[0][1] = 0; A[0][2] = 0; A[0][3] = 0;
  A[1][0] = 0; A[1][1] = 1; A[1][2] = 0; A[1][3] = 0;
  A[2][0] = 0; A[2][1] = 0; A[2][2] = 1; A[2][3] = 0;
  A[3][0] = 0; A[3][1] = 0; A[3][2] = 0; A[3][3] = 1;
};

#endif
