#ifndef Script_H
#define Script_H

/** @file Script.H
    Object action scripting language.

    $Id: Script.H,v 1.33 2006/05/19 09:13:23 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/Script.H,v $
*/

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#ifndef NDEBUG
#define SCRIPT_DEBUG
#endif

#ifdef SCRIPT_DEBUG
#define DUMP_CONTEXT()  StmtList::dump_context()
#else
#define DUMP_CONTEXT()
#endif

#include "Global.H"
#include "SType.H"
#include "Game.H"
#include "Action.H"
#include "Token.H"

class StmtList;
class ScriptObj;
class SimpleActions;
class PlayerInfo;
class Game;
class ScriptParser;
class VariableNode;
class ScriptNode;

REGISTER_TYPEOF(40, Vector<VariableNode*>::iterator );
REGISTER_TYPEOF(41, Vector<VariableNode*>::const_iterator );
REGISTER_TYPEOF(42, Vector<ScriptNode*>::iterator );
REGISTER_TYPEOF(43, Vector<ScriptNode*>::const_iterator );

//=============================================================================
class ScriptNode {
public:
  virtual ~ScriptNode() {}
  virtual stype eval() = 0;
};

//=============================================================================

/** Stores the current position and line information within each sub-block. */
struct ScriptParserContext {
  std::string file;
  sint4 line_num;
  std::string line;
};

//=============================================================================

class StmtList : public ScriptNode {
  Vector<ScriptNode*> stmts;
public:
  uint4 PC;
  StmtList() {}
  StmtList(Vector<ScriptNode*> &_stmts) { stmts = _stmts; }
  ~StmtList() { FORALL (stmts, i) { delete *i; *i = 0; } }
  stype eval();
  void init(Vector<ScriptNode*> &_stmts) { stmts = _stmts; }

#if 1 || defined(SCRIPT_DEBUG)
  Vector<ScriptParserContext> stmt_context;
  std::string _function_name;
  sint4 _line_number;
  static std::stack<ScriptParserContext> context;
  static void dump_context();
#else
  static void dump_context() {}
#endif
};

//=============================================================================

/** Variable list (local and parameters) for an action. */

class VarList {
public:
  stype *param, *local, **glob;
  sint4 num_params, num_local, num_glob;
  //  ScriptObj **objs; ///< Array of pointers to existing objects.
  stype *objs; ///< Array of pointers to existing objects.

  sint4 num_objs;   ///< Objects total.
  sint4 num_pobjs;  ///< Objects passed as parameters.

  VarList() {
    param = 0;
    local = 0;
    objs = 0;
    glob = 0;
    num_params = num_local = num_objs = num_glob = 0;
  }

  ~VarList() {
    std::cout << "deleting VarList" << std::endl;
    if (param) { delete[] param; param = 0; }
    if (local) { delete[] local; local = 0; }
    if (objs)  { delete[] objs; objs = 0; }
    if (glob)  { delete[] glob; glob = 0; }
  }
};

//-----------------------------------------------------------------------------

/** A bound script ready to be executed. */

class ActionScript {

protected:
  ScriptNode *root;

  sint4 call_depth;
  std::stack< Vector<stype> > param_stack;
  std::stack< Vector<stype> > obj_stack;
  std::string name;
  std::string fname;
  std::string fpath;

  VarList vlist;

public:

  stype Act(const SActParam &p);

  ActionScript() : call_depth(0) { root = 0; }
  ~ActionScript();

  std::string get_pwd() { return fpath; }

  friend class ScriptParser;
};

//=============================================================================
class InternalNode2 : public ScriptNode {
protected:
  ScriptNode *left_child;
  ScriptNode *right_child;
public:
  InternalNode2() { left_child = right_child = 0; }
  ~InternalNode2() {
    if (left_child)  { delete left_child; left_child = 0; }
    if (right_child) { delete right_child; right_child = 0; }
  }
  void init(ScriptNode *_left_child, ScriptNode *_right_child) {
    left_child = _left_child;
    right_child = _right_child;
  }
};
//=============================================================================
class InternalNode1 : public ScriptNode {
protected:
  ScriptNode *child;
public:
  InternalNode1() { child = 0; }
  ~InternalNode1() { if (child) { delete child; child = 0; } }
  void init(ScriptNode *c) { child = c; }
};

//=============================================================================
//=== binary operators ========================================================
//=============================================================================

#define Op2Node(name,op) \
class name : public InternalNode2 { \
public: \
  stype eval() {\
    stype l = left_child->eval(), r = right_child->eval(); \
    SCRIPT_ASSERT(l.type == stype::INT && r.type == stype::INT && "no 2 int args "#op); \
    return stype(l.val.i op r.val.i);					\
  }\
}

// allow int op int, ptr op 0, 0 op ptr, and ptr op ptr
#define Op2SpecNode(name,op) \
class name : public InternalNode2 { \
public: \
  stype eval() {\
    stype l = left_child->eval(), r = right_child->eval(); \
    bool l_is_int = l.type == stype::INT;\
    bool r_is_int = r.type == stype::INT;\
    if (l_is_int != r_is_int) {\
      SCRIPT_ASSERT(((l_is_int && l.val.i == 0) || (r_is_int && r.val.i == 0)) && "int != 0 in pointer cmp "#op); \
      if (!l_is_int) l.type = r.type;\
      if (!r_is_int) r.type = l.type;\
    }\
    SCRIPT_ASSERT(l.type == r.type && "different types in "#op); \
    return stype(l.val.i op r.val.i);					\
  }\
}

//-----------------------------------------------------------------------------
#define DivOp2Node(name,op)            \
class name : public InternalNode2 {    \
public:                                \
  stype eval() {                       \
    stype r = right_child->eval();    \
    SCRIPT_ASSERT(r.type == stype::INT && "no first int arg "#op);	\
    SCRIPT_ASSERT(r.val.i && "division by 0");		\
    stype l = left_child->eval(); \
    SCRIPT_ASSERT(l.type == stype::INT && "no 2nd int arg "#op);\
    return stype(sint4(l.val.i op r.val.i));			\
  }\
}
//-----------------------------------------------------------------------------
Op2Node(PlusNode,    +);
Op2Node(TimesNode,   *);
Op2Node(MinusNode,   -);
DivOp2Node(DivideNode, /);
DivOp2Node(ModuloNode, %);

Op2Node(LessThanNode,         <);
Op2Node(LessThanEqualNode,    <=);
Op2Node(GreaterThanEqualNode, >=);
Op2Node(GreaterThanNode,      >);

Op2SpecNode(EqualNode,    ==);
Op2SpecNode(NotEqualNode, !=);

Op2Node(AndNode, &);
Op2Node(OrNode,  |);

class LAndNode : public InternalNode2 {
public:
  stype eval() {
    stype l = left_child->eval();
    // SCRIPT_ASSERT(l.type == stype::INT && "no first int arg &&");
    if (!l.val.i) return stype(0);
    stype r = right_child->eval();
    //SCRIPT_ASSERT(r.type == stype::INT && "no second int arg &&");
    return stype(r.val.i != 0);
  }
};

class LOrNode : public InternalNode2 {
public:
  stype eval() {
    stype l = left_child->eval();
    // SCRIPT_ASSERT(l.type == stype::INT && "no first int arg ||");
    if (l.val.i) return stype(1);
    stype r = right_child->eval();
    // SCRIPT_ASSERT(r.type == stype::INT && "no second int arg ||");
    return stype(r.val.i != 0);
  }
};

//Op2Node(LAndNode, &&);
//Op2Node(LOrNode,  ||);

//=============================================================================
//=== unary operators =========================================================
//=============================================================================

#define Op1Node(name,op)                   \
class name : public InternalNode1 {        \
public:                                    \
  name(ScriptNode *c) { child = c; }       \
  stype eval() { \
    stype v = child->eval();\
    SCRIPT_ASSERT(v.type == stype::INT && "no int arg "#op); \
    v.val.i = op v.val.i; \
    return v;\
  } \
}
//-----------------------------------------------------------------------------
Op1Node(NegNode,  -);
Op1Node(NotNode,  ~);

//Op1Node(LNotNode, !);

// allow all arg types
class LNotNode : public InternalNode1 {  
public:      
  LNotNode(ScriptNode *c) { child = c; }

  stype eval() {
    stype v = child->eval();
    return stype(v.val.i == 0);
  }
};


//=============================================================================
//=== atomic values ===========================================================
//=============================================================================

class ConstNode : public ScriptNode {
protected:
  stype val;
public:
  ConstNode(const std::string &constToken) {
    std::stringstream ss(constToken);
    sint4 i;
    ss >> i;
    val.val.i = i;
  }
  ConstNode(const stype v) : val(v) {}
  stype eval() { return val; }
};
//-----------------------------------------------------------------------------
class VariableNode : public ScriptNode {
public:
  VariableNode() {}
  virtual stype eval() { return stype(); }
  virtual stype* addr() { return 0; }
};
//-----------------------------------------------------------------------------
class LocalVarNode : public VariableNode {
protected:
  stype *var;
public:
  LocalVarNode(stype *v) : var(v) {}
  stype eval() { return *var; }
  stype* addr() { return var; }
};
//-----------------------------------------------------------------------------
class ArrayNode : public VariableNode {
protected:
  ScriptNode *ptr; // start of the array
  ScriptNode *index;
public:
  ArrayNode(ScriptNode *p, ScriptNode *i) : ptr(p), index(i) {}
  stype eval();
  stype* addr();
};
//-----------------------------------------------------------------------------
class StringNode : public ScriptNode {
protected:
  const char *addr;
public:
  StringNode(const std::string &s);
  stype eval() { return stype(addr); }
  ~StringNode() { delete [] addr; }
};
//=============================================================================
//=== assignment operators ====================================================
//=============================================================================

class AssignStmt : public ScriptNode {
protected:
  VariableNode *var;
  ScriptNode *expression;
public:
  AssignStmt(VariableNode *var, ScriptNode *expression);
  ~AssignStmt() {
    if (var) { delete var; var = 0; }
    if (expression) { delete expression; expression = 0; }
  }
  stype eval();
};

//=============================================================================
//=== control structures ======================================================
//=============================================================================

class IfStmt : public InternalNode2 {
protected:
  ScriptNode *elseChild;
public:
  IfStmt(ScriptNode *cond, ScriptNode *if_block, ScriptNode *else_block) {
    left_child = cond;
    right_child = if_block;
    elseChild = else_block;
  }
  ~IfStmt() { if (elseChild) { delete elseChild; elseChild = 0; } }
  
  stype eval() {
    if (!!left_child->eval()) {
      if (right_child) return right_child->eval();
      return stype();
    }
    if (elseChild) return elseChild->eval();
    return stype();
  }
};
//-----------------------------------------------------------------------------
class WhileStmt : public InternalNode2 {
public:
  uint4 PC;

  WhileStmt() { PC = 0; }
  void cond_loop(ScriptNode *cond, ScriptNode *loop) {
    left_child = cond;
    right_child = loop;
  }
  stype eval();
};
//-----------------------------------------------------------------------------
class ForLoop : public ScriptNode {
protected:
  ScriptNode *init;
  ScriptNode *cond;
  ScriptNode *inc;
  ScriptNode *loop;

public:
  uint4 PC;

  ForLoop(ScriptNode *init_, ScriptNode *cond_, ScriptNode *inc_) :
    init(init_), cond(cond_), inc(inc_) { loop = 0; PC = 0; }

  void set_body(ScriptNode *body) { loop = body; }

  stype eval();
};

//-----------------------------------------------------------------------------
class ReturnStmt: public ScriptNode {
protected:

  ScriptNode *val;
  Vector<uint4*> pcs;

public:
  ReturnStmt(ScriptNode *v, const std::list<uint4*> &pcs_);
  ~ReturnStmt() { delete val; val = 0; }
  stype eval();
};


//-----------------------------------------------------------------------------
class BreakStmt: public ScriptNode {
protected:
  uint4 *pc;

public:
  BreakStmt(uint4 *x) : pc(x) {}
  stype eval();
};

//=============================================================================
//=== functions ===============================================================
//=============================================================================

class ScriptFunctionDef {
public:
  std::string name;
  sint4 index;

  sint4 num_obj_ids;
  sint4 num_params;

  Vector<std::string> obj_id_names;
  Vector<std::string> param_names;

  enum FUNC_TYPE { STATIC, MEMBER };
};

//=============================================================================

class FunctionCall : public ScriptNode {
protected:
  Vector<VariableNode*> objs;
  Vector<ScriptNode*> params;
public:
  ScriptNode* parse_core(ScriptFunctionDef &def, ScriptParser &p);
  virtual ScriptNode* parse(ScriptFunctionDef &def, ScriptParser &p) = 0;
};

//=============================================================================

class StaticFunctionCall : public FunctionCall {
  ActionFunction func;
public:

  StaticFunctionCall(ActionFunction f) : func(f) {}

  ScriptNode* parse(ScriptFunctionDef &def, ScriptParser &p) {
    return FunctionCall::parse_core(def, p);
  }
  stype eval();
};

//-----------------------------------------------------------------------------

template<class T>
class MemberFunctionCall : public FunctionCall {
  typedef stype(T::*MemberFunc)(const SActParam &);
  MemberFunc func;
public:
  static T *owner;

  MemberFunctionCall(MemberFunc f) : func(f) {}

  ScriptNode* parse(ScriptFunctionDef &def, ScriptParser &p) {
    return FunctionCall::parse_core(def, p);
  }
  stype eval() {
    SCRIPT_ASSERT(owner && "Member function called without base object");
    SActParam p;
#ifndef OS_MAC
    FORALL (objs, i)   { p.obj.push_back((*i)->eval().to_obj()); }
    FORALL (params, i) { p.val.push_back((*i)->eval()); }
#else
    FORU (i, objs.size())   { p.obj.push_back((objs[i])->eval().to_obj()); }
    FORU (i, params.size()) { p.val.push_back((params[i])->eval()); }
#endif
    return (owner->*func)(p);
  }
};

template <typename T> T* MemberFunctionCall<T>::owner;

//=============================================================================

class PreFunction_Base {
public:
  virtual FunctionCall *function_instance() = 0;
  virtual ~PreFunction_Base() { }
};

template<class T>
class PreMemberFunction : public PreFunction_Base {
public:
  typedef stype(T::*MemberFunc)(const SActParam&);  
  MemberFunc func;

  PreMemberFunction(MemberFunc f) : func(f) {}
  FunctionCall* function_instance() { return new MemberFunctionCall<T>(func); }
};  

class PreStaticFunction : public PreFunction_Base {
public:
  ActionFunction func;

  PreStaticFunction(ActionFunction f) : func(f) {}
  FunctionCall* function_instance() { return new StaticFunctionCall(func); }
};

//-----------------------------------------------------------------------------

class ScriptFunctionList {
  Vector<ScriptFunctionDef> function_defs;
  Vector<PreFunction_Base*> pre_functions;
public:
  ScriptNode* create_FunctionCall_node(const std::string &func_name, ScriptParser &p) {
    FORT(i, function_defs.size()) {
      if (function_defs[i].name == func_name) {
        FunctionCall *f = pre_functions[i]->function_instance();
        return f->parse(function_defs[i], p);
      }
    }
    return 0;
  }

  template<class T>
  void add_function(const ScriptFunctionDef &def, stype(T::*f)(const SActParam&) ) {
    function_defs.push_back(def);
    pre_functions.push_back(new PreMemberFunction<T>(f));
  }
  void add_function(const ScriptFunctionDef &def, stype(*f)(const SActParam&) ) {
    function_defs.push_back(def);
    pre_functions.push_back(new PreStaticFunction(f));
  }
};

//=============================================================================

class FunctionNode : public ScriptNode {
public:
  static Game *game; // access to game rng, fixme: shouldn't be static
  virtual ScriptNode* parse(ScriptParser &parser) = 0;
};

//=============================================================================

class RandomNode : public FunctionNode {
protected:
  ScriptNode *min, *max;
public:
  RandomNode() { min = max = 0; }
  ~RandomNode() { delete min; delete max; min = max = 0; }
  ScriptNode* parse(ScriptParser &parser);
  stype eval();
};

//=============================================================================

class QueueAction : public ScriptNode {
protected:
  std::string action_name; // an object action, the object can identify the action by name
  bool is_server;     ///< If this is the client then ignore object enable_count.

  Vector<VariableNode*> objs;
  Vector<ScriptNode*> params;
  ScriptNode *delay;
  SimpleActions *sa;
  Game *game;

  std::stack< Vector<uint4> > pc_stack;
  Vector<uint4*> pcs;

public:
  QueueAction(const std::string &action, Vector<VariableNode*> &o, Vector<ScriptNode*> &p,
              ScriptNode *delay, Game *game, std::list<uint4*> &pcs);
  ~QueueAction();
  stype eval();
  friend class ScriptParser;
};

//=============================================================================

#endif
