#ifndef ScriptObj_H
#define ScriptObj_H

/** @file GameObj.H
    The objects that populate the world.

    $Id: ScriptObj.H,v 1.21 2006/05/27 02:28:02 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/ScriptObj.H,v $
*/

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Global.H"
#include "GameConst.H"
#include "ServerObjData.H"
#include "IdFactory.H"
#include "Attr.H"
#include "SaveLoad.H"

class PlayerInfo;
class ScriptObj;
class ObjAction;
class GameObj;

typedef uint4 ObjId;

// keep outside of class to use with typeof (windows)
typedef std::map<ObjId,ObjId> IdMap;
typedef std::map<ObjId,ScriptObj*> IdObjMap;
typedef std::map<ScriptObj*,ObjId> ObjIdMap;
typedef std::map<std::string, ScriptObj*> NameObjMap;
typedef std::map<std::string, stype*> NameVarMap;

//===================================================================

/** A generic container class that defines script objects. */

class ScriptObj : public SaveLoadAbs {

private:
  ObjId id;
  static IdFactory idf;
  static IdObjMap id2obj_map;
  static ObjIdMap obj2id_map;

  static Vector<ScriptObj*> dead_objs;
  static Vector<ScriptObj*> destroyed_objs; ///< dead objects whose destructor has already
                                            ///< been called, now waiting to be deleted

protected:

  sint4 num_refs;      ///< Number of queued actions that involve this object.

  bool dead;           ///< State of this object, regardless of any parent's state.
  bool enabled;

public:

  /** One of those horrible 'smart pointers'.
      A necessary evil to maintain reasonable saving ability.
      Lets ScriptObjs be referenced by id. */
  class Ptr {
  private:
    ObjId id;
  public:
    
    Ptr(ScriptObj *s=0);
    
    inline ScriptObj* operator->() const { return ScriptObj::id2obj_map[id]; }
    inline ScriptObj& operator*() const { return *ScriptObj::id2obj_map[id]; }
    inline operator ScriptObj* () const { return ScriptObj::id2obj_map[id]; }

    void save(std::ostream &os) const {
      OS_WRITE(id);
    }

    void load(std::istream &is) {
      IS_READ(id);      
    }
  };


  ScriptObj *parent;   ///< The object that this is a node of (can be self referential).
  PlayerInfo *creator;

  ScriptObj();
  virtual ~ScriptObj();

  void save(std::ostream &os) const;
  void load(std::istream &is);
  
  static void get_ids(Vector<ScriptObj*> &objs, Vector<ObjId> &ids);
  static void set_ids(const Vector<ScriptObj*> &objs, const Vector<ObjId> &ids);

  static std::set<ScriptObj*> ptr_list;
  static void check_ptr(ScriptObj*);
  static void collect_dead(std::set<ScriptObj*> &dead);
  static void collect_destroyed();
  static ObjId get_obj_id(ScriptObj *s);

  void add_ref() { ++num_refs; SCRIPT_ASSERT(num_refs <= 10000); } // not a hard assert - only for bug detection
  void del_ref() { --num_refs; SCRIPT_ASSERT(num_refs >= 0); }
  virtual sint4 get_refs() const { return num_refs; }
  virtual void enable() { enabled = true; }
  virtual void disable() { enabled = false; }
  virtual bool is_enabled() { return enabled; }

  void kill();
  bool is_dead() const { return dead; }

  virtual std::string bp_name() const { return ""; }
  virtual uint4 code() const { return 0; }

  virtual void attr_names(Vector<std::string> &) const {}
  virtual bool has_attr    (const std::string &) const;
  virtual bool attr_changed(const std::string &) const { return false; }

  // get value or pointer to value by name
  virtual stype get_val     (const std::string &) const;
  virtual stype get_val_prev(const std::string &) const { return stype(); }
  virtual stype* get_val_ptr(const std::string &)       { return 0; }
  virtual const stype* get_const_val_ptr(const std::string &) const { return 0; }
  virtual void  set_val(const std::string &name, stype val, bool warn=true);

  // for integers
  virtual sint4 get_int     (const std::string &) const;
  virtual sint4 get_int_prev(const std::string &) const;
  virtual sintptr* get_int_ptr(const std::string &);
  virtual const sintptr* get_const_int_ptr(const std::string &) const;
  virtual void  set_int(const std::string &name, sint4 val, bool warn=true);


  virtual void gen_diff(bool, bool, std::string &, PlayerInfo *) const {}
  virtual void gen_comp_diff(bool first, bool allied, std::string &s) const { gen_diff(first, allied, s, 0); }
  virtual bool apply_diff(bool, const char* &) { return false; }
  virtual bool apply_comp_diff(bool allied, const char* &s, bool /*first*/) { return apply_diff(allied, s); }

  virtual void copy_to_prev() {}
  virtual void reset_frame_data() {}

  virtual void action_names(Vector<std::string> &) const {}

  virtual uint4 set_action(const std::string &, const Vector<sint4> &, sint4 =1) {
    ERR("set_action called for non-game object (ScriptObj)");
    return 0;
  }

  virtual uint4 set_action(sint4, const Vector<sint4> &, sint4 =1) {
    ERR("set_action called for non-game object (ScriptObj)");    
    return 0;
  }

  virtual void clear_action() {
    ERR("clear_action called for non-game object (ScriptObj)");
  }
  
  virtual std::string write_action(ObjId) const { return ""; }

  virtual ObjAction* action(const std::string &)   const { return 0; }
  virtual ObjAction* action(sint4)                 const { return 0; }
  virtual sint4 action_id(const std::string &)     const { return -1; }
  virtual bool action_enabled(sint4)               const { return false; }
  virtual bool action_enabled(const std::string &) const { return false; }
  virtual bool action_hidden(sint4)                const { return false; }
  virtual bool action_hidden(const std::string &)  const { return false; }

  virtual void component_names(Vector<std::string> &) const {}
  virtual ScriptObj* component(const std::string &) const { return 0; }
  virtual ScriptObj* action_component(sint4)        const { return 0; }
  
  virtual void construct() {}
  virtual void destruct() {}
  virtual GameObj* get_GameObj() { return 0; }

  friend class GameObjFactory;
  friend class ScriptParser;

};

//===================================================================

REGISTER_TYPEOF(1100, Vector<ScriptObj*>::iterator);
REGISTER_TYPEOF(1101, Vector<ScriptObj*>::const_iterator);
REGISTER_TYPEOF(1102, Vector<ScriptObj::Ptr>::iterator);
REGISTER_TYPEOF(1103, Vector<ScriptObj::Ptr>::const_iterator);
REGISTER_TYPEOF(1104, IdMap::iterator);
REGISTER_TYPEOF(1105, IdMap::const_iterator);
REGISTER_TYPEOF(1106, IdObjMap::iterator);
REGISTER_TYPEOF(1107, IdObjMap::const_iterator);
REGISTER_TYPEOF(1108, ObjIdMap::iterator);
REGISTER_TYPEOF(1109, ObjIdMap::const_iterator);
REGISTER_TYPEOF(1110, NameObjMap::iterator);
REGISTER_TYPEOF(1111, NameObjMap::const_iterator);
REGISTER_TYPEOF(1112, NameVarMap::iterator);
REGISTER_TYPEOF(1113, NameVarMap::const_iterator);

#endif
