/** @file ScriptParser.C
    @see ScriptParser.H.

    $Id: ScriptParser.C,v 1.12 2006/05/19 09:13:23 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/ScriptParser.C,v $
*/

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "ScriptParser.H"
#include "ScriptGame.H"
#include "Action.H"
#include "GameObj.H"
#include "GameObjFactory.H"
#include "PlayerInfo.H"
#include "Game.H"
#include "Random.H"
#include "Blueprints.H"
#include "ScriptVector.H"

using namespace std;

REGISTER_TYPEOF(44, list<uint4*>::iterator);
REGISTER_TYPEOF(45, list<uint4*>::const_iterator);

#define PARSE_ERR(s) {\
  errstr << endl << "Action " << current_action_name\
  << ": parse error at line " << tokenizer.cur_line();\
  ERR(s);\
}

#define PARSE_ERR2(s1, s2) {\
  errstr << endl << "Action " << current_action_name\
  << ": parse error at line " << tokenizer.cur_line();\
  ERR2(s1,s2);\
}

#define PARSE_WARN2(s1, s2) {\
  errstr << endl << "Action " << current_action_name\
  << ": warning at line " << tokenizer.cur_line();\
  REM2(s1,s2);\
}

ScriptFunctionList ScriptParser::functions;

//=============================================================================

/** Converts the ObjActionDefinition into an executable script.
    Requires that all the actions have been read in from the game description
    and context values (PlayerInfo, game, action_name) have been set within the
    ScriptParser making this call.
*/

ActionScript* ScriptParser::Parse(const ObjActionDefinition &def)
{
  param_symbol.clear();
  local_symbol.clear();
  obj_symbol.clear();
  global_symbol.clear();

  param_symbol = def.param_names;
  obj_symbol.push_back("this");

  FORALL(def.obj_id_names, i) obj_symbol.push_back(*i);
  FORALL(def.state_param_names, i) param_symbol.push_back(*i);

  const char *cscript = def.script.c_str();
  tokenizer.init(cscript);

  // variables in the scope of the action
  Vector<stype> local_vars;

  ActionScript *as = new ActionScript;

  as->name = def.name;
  as->vlist.num_params = def.num_params + def.num_state_params;
  as->vlist.num_pobjs = def.num_obj_ids;

  as->vlist.num_local = 0;
  as->vlist.num_objs = 1 + def.num_obj_ids;
  as->vlist.num_glob = 0;

  as->fname = def.file_name;
  as->fpath = def.path;

  // read in variable declarations
  string token;

  token = getToken();
  if (token != "{")  PARSE_ERR("expecting '{' before function body");

  FOREVER {
    token = getToken();
    enum var_type { UNKNOWN, INT, GOB, GLOBAL };
    var_type t = UNKNOWN;

    Vector<string> *sym = 0;
    sint4 *sym_count = 0;

    if (token == "int") { t = INT; sym = &local_symbol; sym_count = &as->vlist.num_local; }
    if (token == "gob") { t = GOB; sym = &obj_symbol;   sym_count = &as->vlist.num_objs; }
    if (token == "global") { t = GLOBAL; }

    if (t == INT || t == GOB) {
      FOREVER {
        token = getToken();
        if (tokenizer.token_type() != Tokenizer::ID) PARSE_ERR("int symbol name expected");
        check_redeclaration(token);

        sym->push_back(token);
        (*sym_count)++;

        token = getToken();
        if (t == INT) {
          if (token == "[") {
            token = getToken();
            sint4 s = 0;
            if (tokenizer.token_type() == Tokenizer::NUMBER) {
              stringstream temp(token);
              temp >> s;
              if (s < 0) PARSE_ERR("negative array size");
              token = getToken();
            }
            if (token != "]") PARSE_ERR("']' expected");
            Vector<stype> *v = new Vector<stype>(s);
            local_vars.push_back(stype(v, ScriptVector::STYPE_ID));
            token = getToken();
          } else {
            local_vars.push_back(stype());
          }
        }
        if (token == ";") break;
        if (token != ",") PARSE_ERR("';' or ',' expected");
      }
    }
    else if (t == GLOBAL) {
      FOREVER {
        token = getToken();
        if (tokenizer.token_type() != Tokenizer::ID) PARSE_ERR("int symbol name expected");
        check_redeclaration(token);

        pi->new_global_var(token, stype());

        global_symbol.push_back(token); // symbol table for this action
        as->vlist.num_glob++;

        token = getToken();
        if (token == ";") break;
        if (token != ",") PARSE_ERR("';' or ',' expected");
      }
    }
    else {
      tokenizer.putback(token);
      // create a fake bracket after variable declarations
      // removes declarations from the StmtList
      tokenizer.putback(string("{"));
      break;
    }
  }

  // allocate space
  if (as->vlist.num_params) as->vlist.param = new stype[as->vlist.num_params];
  if (as->vlist.num_local)  as->vlist.local = new stype[as->vlist.num_local];
  if (as->vlist.num_objs)   as->vlist.objs  = new stype[as->vlist.num_objs];
  if (as->vlist.num_glob)   as->vlist.glob  = new stype*[as->vlist.num_glob];

  // initialize variables to 0
  FORS (i, as->vlist.num_local) as->vlist.local[i] = local_vars[i];
  FORS (i, as->vlist.num_objs)  as->vlist.objs[i] = stype(static_cast<ScriptObj*>(0)); // XXX

  // point symbols to global variables
  FORS(i, as->vlist.num_glob) as->vlist.glob[i] = pi->global_var(global_symbol[i]);

  vlist = &as->vlist;

  token = getToken();
  if (token != "{") PARSE_ERR("expecting '{' at start of action");
  as->root = parseStmtList();

  return as;
}
//=============================================================================
void ScriptParser::check_redeclaration(const string &var)
{
  bool redec = false;
              FORALL (param_symbol, i)  if (*i == var) { redec = true; break; }
  if (!redec) FORALL (local_symbol, i)  if (*i == var) { redec = true; break; }
  if (!redec) FORALL (obj_symbol, i)    if (*i == var) { redec = true; break; }
  if (!redec) FORALL (global_symbol, i) if (*i == var) { redec = true; break; }
  if (redec) PARSE_ERR2("redeclaration of symbol", var.c_str());
}
//=============================================================================
string ScriptParser::getToken()
{
  return tokenizer.next_token();
}
//=============================================================================
void ScriptParser::putbackToken(const string &t)
{
  tokenizer.putback(t);
}
//=============================================================================
// a number, variable, or compound expression
ScriptNode* ScriptParser::getTerm(sint4 &constant)
{
  constant = 0;
  string token = getToken();

  if (token == "")  return 0;
  if (token == "(") return parseExp(")");

  if      (token == "-")  return new NegNode(getTerm(constant));
  else if (token == "!")  return new LNotNode(getTerm(constant));
  else if (token == "~")  return new NotNode(getTerm(constant));


  if (tokenizer.token_type() == Tokenizer::STRING) return new StringNode( token );

  if (!isalpha(token[0])) { constant = 1; return new ConstNode(token); }

  if (token == "npc") { constant = 1; return new ConstNode(stype(game->get_player_num())); }
  if (token == "me")  { constant = 1; return new ConstNode(stype(game->get_client_player())); }

  ScriptNode *t = functions.create_FunctionCall_node(token, *this);
  if (t) return t;


  FunctionNode *func;

  if (token == "id2gob")        { func = new ID_to_GOB(pi); return func->parse(*this); }
  if (token == "type")          { func = new BP_Code;       return func->parse(*this); }
  if (token == "rand")          { func = new RandomNode;    return func->parse(*this); }
  if (token == "num_intersect") { func = new NumIntersect;  return func->parse(*this); }
  if (token == "lost_sight")    { func = new LostSight;     return func->parse(*this); }


  string id = getToken();
  if (id == "(") { // start of a function call

    tokenizer.putback(token);
    ScriptNode *a = getAction(false);
    return a;

  } else {
    tokenizer.putback(id);
    tokenizer.putback(token);
  }

  VariableNode *var = getVariable();
  if (var) return var;

  string temp = getToken();

  PARSE_ERR2("No term", temp.c_str());
  return 0;
}
//=============================================================================
sint4 ScriptParser::precedenceOf(const string &op) const
{
  if (op == "&&" || op == "||")
    return 1;
  if (op == "<" || op == "<=" || op == "==" || op == "!=" || op == ">=" || op == ">")
    return 2;
  if (op == "&" || op == "|")
    return 3;
  if (op == "+" || op == "-")
    return 4;
  if (op == "*" || op == "/" || op == "%")
    return 5;
  return 0;
}
//-----------------------------------------------------------------------------
ScriptNode* ScriptParser::parseTillLowOp(sint4 prec, sint4 *res_is_const, ScriptNode *lnode, sint4 l_is_const)
{
  sint4 lconst;
  ScriptNode *leftOperand;
  if (lnode) { leftOperand = lnode; lconst = l_is_const; }
  else       leftOperand = getTerm(lconst);

  string op;
  while ((op = getToken()) != "") {
    if (precedenceOf(op) <= prec) {
      tokenizer.putback(op);
      break;
    }

    sint4 rconst;
    ScriptNode *rightOperand = parseTillLowOp(precedenceOf(op), &rconst);
    InternalNode2 *node = 0;

    if      (op == "+")  node = new PlusNode;
    else if (op == "-")  node = new MinusNode;
    else if (op == "*")  node = new TimesNode;
    else if (op == "/")  node = new DivideNode;
    else if (op == "%")  node = new ModuloNode;    
    else if (op == "<")  node = new LessThanNode;
    else if (op == "<=") node = new LessThanEqualNode;
    else if (op == "==") node = new EqualNode;
    else if (op == "!=") node = new NotEqualNode;
    else if (op == ">=") node = new GreaterThanEqualNode;
    else if (op == ">")  node = new GreaterThanNode;
    else if (op == "&&") node = new LAndNode;
    else if (op == "||") node = new LOrNode;
    else if (op == "&")  node = new AndNode;
    else if (op == "|")  node = new OrNode;
    else
      PARSE_ERR2 ("Unknown operator: ", op.c_str());

    if (node) {
      node->init(leftOperand, rightOperand);
      if (lconst && rconst) {
        leftOperand = new ConstNode(node->eval());
        delete node;
      } else {
        leftOperand = node;
        lconst = 0;
      }
    }
  }
  if (res_is_const) *res_is_const = lconst;
  return leftOperand;
}
//=============================================================================
ScriptNode* ScriptParser::parseExp(const string &expectedEndMarker, ScriptNode *lnode)
{
  const string test = getToken();
  if (test == "") return 0;
  if (test == expectedEndMarker) return 0;
  tokenizer.putback(test);

  sint4 temp;
  ScriptNode *result = parseTillLowOp(0, &temp, lnode);
  if (expectedEndMarker != "") {
    string endmarker = getToken();
    if (endmarker != "" && endmarker != expectedEndMarker) {
      endmarker = "'" + endmarker + "'";
      PARSE_ERR2 ("Unexpected end of expression marker", endmarker.c_str());
    }
  }
  return result;
}
//-----------------------------------------------------------------------------
ScriptNode* ScriptParser::parseExp(const set<string> &EndMarkers, string &end_marker)
{
  const string test = getToken();
  if (test == "") { cout << "no token" << endl; return 0; }
  if (EndMarkers.find(test) != EndMarkers.end()) {
    end_marker = test;
    return 0;
  }

  tokenizer.putback(test);

  sint4 temp;
  ScriptNode *result = parseTillLowOp(0, &temp);
  if (EndMarkers.size() != 0) {
    end_marker = getToken();
    if (end_marker != "" && (EndMarkers.find(end_marker) == EndMarkers.end())) {
      end_marker = end_marker + "'";
      PARSE_ERR2 ("Unexpected end of expression marker '", end_marker.c_str());
    }
  }
  return result;
}
//-----------------------------------------------------------------------------
VariableNode* ScriptParser::getVariable(bool as_obj)
{
  string token = getToken();

  VariableNode *var = 0;

  FORT (i, param_symbol.size())
    if (param_symbol[i] == token) var = new LocalVarNode(&vlist->param[i]);

  FORT (i, local_symbol.size())
    if (local_symbol[i] == token) var = new LocalVarNode(&vlist->local[i]);

  FORT (i, global_symbol.size())
    if (global_symbol[i] == token) var = new LocalVarNode(vlist->glob[i]);

  string id = getToken();

  // array element
  if (id == "[") {
    if (var) PARSE_ERR2("array operator after variable", token.c_str());
    tokenizer.putback(token);
    var = parseObjAddr(false);
    if (!var) PARSE_ERR2("Unknown array", token.c_str());
    while (id == "[") {
      ScriptNode *index = parseExp("]");
      var = new ArrayNode(var, index);
      id = getToken();
    }
    tokenizer.putback(id);
    return var;
  }
  tokenizer.putback(id);

  if (var) return var;


  string attr_id;

  size_t i = token.find('.', 0);
  if (i != string::npos) { // a scoped variable
    attr_id = token.substr(i+1, token.size() - (i+1));
    token = token.substr(0, i);
  }


  if (token == "BP") {
    
    Vector<string> attrs;
    
    FOREVER { // check for sub-components
      i = attr_id.find('.', 0);
      if (i == string::npos) break;
      string comp = attr_id.substr(0, i);
      attr_id = attr_id.substr(i+1, attr_id.size() - (i+1));
      attrs.push_back(comp);
    }
    attrs.push_back(attr_id);

    if (attrs.empty())     PARSE_ERR2 ("BP without name or attribute:", attr_id.c_str());
    if (attrs.size() == 1) PARSE_ERR2 ("expecting attribute name after blueprint", attr_id.c_str());
    
    return new BP_AttrNode(attrs, &pi->gof);
  }
  

  if (attr_id == "") { // Local game object
    tokenizer.putback(token);
    return parseObjAddr(false);
  }

  token = token + "." + attr_id;
  tokenizer.putback(token);

  if (as_obj) return parseObjAddr(false);
  return parseObjAddr(true);
}
//-----------------------------------------------------------------------------
ScriptNode* ScriptParser::getAction(bool parse_delay)
{
  string token;
  string action_name;
  
  Vector<VariableNode*> obj_params;
  Vector<ScriptNode*> params;
  
  VariableNode *gob_this = parseObjAddr(false, &action_name);
  
  if (action_name == "") PARSE_ERR("Expecting action name");
  
  obj_params.push_back(gob_this); // "this"
  
  // object ids
  token = getToken();
  if (token == ")") goto skip_params;
  if (token != ";") {
    tokenizer.putback(token);
    FOREVER {
      VariableNode *obj = parseObjAddr(false);
      if (obj)
        obj_params.push_back(obj);
      else
        PARSE_ERR("no addr???");
      token = getToken();
      if (token == ")") goto skip_params;
      if (token == ";") break;
      if (token != ",") PARSE_ERR2("unexpected token:", token.c_str());
    }
  }
  
  // parameters
  token = getToken();
  if (token == ")") goto skip_params;
  if (token != ";") {
    tokenizer.putback(token);
    FOREVER {
      ScriptNode *cur = parseExp("");
      if (!cur) PARSE_ERR("expecting parameter");
      params.push_back(cur);
      token = getToken();
      if (token == ")") goto skip_params;
      if (token == ";") break;
      if (token != ",") PARSE_ERR2("unexpected token:", token.c_str());
    }
  }
  
  // state parameters
  token = getToken();
  if (token != ")") {
    tokenizer.putback(token);
    FOREVER {
      ScriptNode *cur = parseExp("");
      params.push_back(cur);
      token = getToken();
      if (token == ")") break;
      if (token != ",") PARSE_ERR2("unexpected token:", token.c_str());
    }
  }

 skip_params:
  
  ScriptNode *delay = 0;
  if (parse_delay) {
    token = getToken();
    if (token != ";") {
      if (token != "in") PARSE_ERR2("expecting 'in' or ';', found:", token.c_str());
      delay = parseExp(";");
    }
  }
  
  return new QueueAction(action_name, obj_params, params, delay, game, PCs);
}
//-----------------------------------------------------------------------------
ScriptNode* ScriptParser::parseReturnStmt()
{
  ScriptNode *x = parseExp(";");
  if (x == 0) PARSE_ERR("must return a value");
  return new ReturnStmt(x, PCs);
}
//-----------------------------------------------------------------------------
ScriptNode* ScriptParser::parseBreakStmt()
{
  if (getToken() != ";") PARSE_ERR("expecting ';'");
  return new BreakStmt(PCs.back());
}
//-----------------------------------------------------------------------------
IfStmt* ScriptParser::parseIfStmt()
{
  string token = getToken();
  if (token != "(") PARSE_ERR("'(' expected after 'if'");

  ScriptNode *condition = parseExp(")");
  if (!condition) PARSE_ERR("Condition expected");

  ScriptNode *action = parseStmt();
  ScriptNode *elseblock = 0;

  token = getToken();
  if (token == "else")
    elseblock = parseStmt();
  else
    tokenizer.putback(token);

  return new IfStmt(condition, action, elseblock);
}
//-----------------------------------------------------------------------------
WhileStmt* ScriptParser::parseWhileStmt()
{
  string token = getToken();
  if (token != "(") PARSE_ERR("'(' expected after 'while'");

  ScriptNode *condition = parseExp(")");
  if (!condition) PARSE_ERR("Condition expected");

  WhileStmt *ws = new WhileStmt;
  PCs.push_back(&ws->PC);

  ScriptNode *action = parseStmt();

  PCs.pop_back();

  ws->cond_loop(condition, action);

  return ws;
}
//-----------------------------------------------------------------------------
ForLoop* ScriptParser::parseForLoop()
{
  string token = getToken();
  if (token != "(") PARSE_ERR("'(' expected after 'for'");

  ScriptNode *init = parseStmt();
  ScriptNode *cond = parseExp(";");
  string e = ")";
  ScriptNode *inc = parseStmt(&e);

  ForLoop *f = new ForLoop(init, cond, inc);
  PCs.push_back(&f->PC);

  ScriptNode *action = parseStmt();

  PCs.pop_back();

  f->set_body(action);

  return f;
}
//-----------------------------------------------------------------------------
ScriptNode* ScriptParser::parseObjLoop()
{
  string token = getToken();
  if (token != "(") PARSE_ERR("'(' expected after forall");
  string id = getToken();
  token = getToken();
  if (token == ")") {
    FORT (i, obj_symbol.size())
      if (obj_symbol[i] == id)
        return new ForAllObj( &vlist->objs[i], parseStmt(), pi->my_objs);
    PARSE_ERR2 ("unknown object variable", id.c_str());
  }
  if (token != ",") PARSE_ERR("',' or ')' expected");
  set<string> bp_types;
  while (token == ",") {
    token = getToken();
    FIND (pi->gof.name2blueprint, it, token);
    if (it == pi->gof.name2blueprint.end()) {
      //      PARSE_WARN2("not registered yet (may still be OK): ", token.c_str());
    }
    bp_types.insert(token);
    token = getToken();
  }
  if (token != ")") PARSE_ERR("')' expected");
  FORT (i, obj_symbol.size())
    if (obj_symbol[i] == id)
      return new ForAllType( &vlist->objs[i], parseStmt(), pi->my_objs, bp_types);
  PARSE_ERR2 ("unknown object variable", id.c_str());
  return 0;
}
//-----------------------------------------------------------------------------
ScriptNode* ScriptParser::parseIntersectLoop()
{
  stype *obj = 0;
  string id;

  if (getToken() != "(") PARSE_ERR("'(' expected after intersect");

  id = getToken();
  FORT (i, obj_symbol.size()) if (obj_symbol[i] == id) obj = &vlist->objs[i];
  if (obj == 0) PARSE_ERR2 ("unknown object variable", id.c_str());

  if (getToken() != ",") PARSE_ERR("',' expected after object variable");

  VariableNode *area = parseObjAddr(false);

  if (getToken() != ")") PARSE_ERR("')' expected after area variable");

  return new ForAllIntersect(obj, area, parseStmt(), game);
}
//-----------------------------------------------------------------------------
// parses an object, descending into components

// ~has_attr &&  _attr -> return an object address node, setting _attr to the last section and ignoring it
//                        eg. "base.foo.action" sets _attr="action"
// ~has_attr && ~_attr -> return an object address node using all sections

//  has_attr &&  _attr -> return an object attribute node (all sections), setting _attr to the last section
//  has_attr && ~_attr -> return an object attribute node

VariableNode* ScriptParser::parseObjAddr(bool has_attr, string *_attr)
{
  stype *gob;
  string token = getToken();

  if (!isalpha(token[0])) {
    tokenizer.putback(token);
    return 0;
    //PARSE_ERR2("expecting object, found:", token.c_str());
  }

  size_t i = token.find('.', 0);
  string tail;

  if (i != string::npos) { // subcomponent
    tail = token.substr(i+1, token.size() - (i+1));
    token = token.substr(0, i);
  }

  GlobalObj *g = 0;
  SharedObj *s = 0;

  if (token == "global") {

    token = tail;

    gob = 0;
    i = tail.find('.', 0);

    if (i != string::npos) {
      token = tail.substr(0, i);
      tail = tail.substr(i+1, tail.size() - (i+1));
    } else {
      if (has_attr) PARSE_ERR ("expecting global object name");
      tail = "";
    }

    g = new GlobalObj(token, pi);
    gob = &g->obj;

  } else if (token == "shared") {

    token = tail;

    gob = 0;
    i = tail.find('.', 0);

    if (i != string::npos) {
      token = tail.substr(0, i);
      tail = tail.substr(i+1, tail.size() - (i+1));
    } else {
      if (has_attr) PARSE_ERR ("expecting shared object name");
      tail = "";
    }

    s = new SharedObj(token);
    gob = &s->obj;

  } else {

    uint4 end = obj_symbol.size();
    FOR (i, end) if (obj_symbol[i] == token) break;
    if (i == end) {
      tokenizer.putback(token);
      PARSE_ERR2 ("variable not found:", token.c_str());
      return 0;
    }
    gob = &vlist->objs[i];

  }

  Vector<string> comps;
  string comp = tail;
  string attr;

  FOREVER {
    i = tail.find('.', 0);
    if (i == string::npos) break;
    comp = tail.substr(0, i); // this isn't the last, so it's a component
    comps.push_back(comp);
    tail = tail.substr(i+1, tail.size() - (i+1));
  }

  if (has_attr) {
    if (tail == "") PARSE_ERR ("expecting component/action/attribute");
    attr = tail;
    if (_attr) *_attr = tail;
    //    if (_attr) cout << "tail = " << tail << endl;
  } else {
    if (_attr)
      *_attr = tail;
    else if (tail != "")
      comps.push_back(tail); // everything is supposed to be a component
  }

  if ((comps.size() > 0) || (g != 0) || (s != 0)) {
    GOB_Component *gob_comp = new GOB_Component(gob, comps);
    gob = &gob_comp->obj;
    if (g) gob_comp->first = g;
    else   gob_comp->first = s; // check for a global or shared parent object
    if (has_attr) {
      return new GOB_AttrNode(attr, gob, gob_comp);
    } else {
      return new GOB_AddrNode(gob, gob_comp);
    }
  } else {
    if (has_attr) {
      return new GOB_AttrNode(attr, gob, 0);
    } else {
      return new GOB_AddrNode(gob, 0);
    }
  }
}
//-----------------------------------------------------------------------------
ScriptNode* ScriptParser::parseStmt(const string *end_marker)
{
  string token = getToken();

  if (token == "")          return 0;
  if (token == "{")         return parseStmtList();
  if (token == "return")    return parseReturnStmt();
  if (token == "break")     return parseBreakStmt();
  if (token == "if")        return parseIfStmt();
  if (token == "while")     return parseWhileStmt();
  if (token == "for")       return parseForLoop();
  if (token == "forall")    return parseObjLoop();
  if (token == "intersect") return parseIntersectLoop();

  ScriptNode *t = functions.create_FunctionCall_node(token, *this);
  string test = end_marker ? *end_marker : ";";
  if (t) {
    if (getToken() != test) PARSE_ERR("expecting ';'");
    return t;
  }

  string id = getToken();
  if (id == "(") { // start of a function call
    tokenizer.putback(token);
    return getAction(true);
  } else {
    tokenizer.putback(id);
  }

  tokenizer.putback(token);
  VariableNode *var = getVariable();
  if (var) {
    token = getToken();
    if (token.length() == 2) { // expand "x ?= y" into "x = x ? y"
      if (token[1] == '=') {
        tokenizer.putback(token.substr(0,1)); // operator
        return new AssignStmt(var, parseExp(test, var));
      }
    }
    if (token == "=") {
      return new AssignStmt(var, parseExp(test));
    }

    tokenizer.putback(token);
  }

  if (!isalpha(token[0])) PARSE_ERR("expecting statement");

  return parseExp(test);
}
//-----------------------------------------------------------------------------
StmtList* ScriptParser::parseStmtList()
{
  string token;
  Vector<ScriptNode*> stmts;

  StmtList *sl = new StmtList;
  PCs.push_back(&sl->PC);

#if 1 || defined(SCRIPT_DEBUG)
  sl->_function_name = current_action_name;
  sl->_line_number = tokenizer.cur_line();
#endif

  FOREVER {
    token = getToken();

    if (token == "") PARSE_ERR("End of statement not found");
    if (token == "}") break;

    tokenizer.putback(token);
#if 1 || defined(SCRIPT_DEBUG)
    ScriptParserContext c;
    c.file = file_name;
    c.line_num = line_num + tokenizer.cur_line();
    c.line = tokenizer.scan_line();
    sl->stmt_context.push_back(c);
#endif
    ScriptNode* stmt = parseStmt();
    if (!stmt) PARSE_ERR("expecting statement");
    stmts.push_back(stmt);
  }

  PCs.pop_back();

  if (stmts.size() <= 0) {
    delete sl;
    sl = 0;
  } else {
    sl->init(stmts);
  }
  return sl;
}
//=============================================================================

/** Provides an interface for one-off evaluations of scripted expressions.
    If the results are to be used later (specifically strings) the script tree should not be deleted.
 */

ScriptNode* ScriptParser::eval_expression(const string &exp)
{
  const char *cexp = exp.c_str();
  tokenizer.init(cexp);

  ScriptNode *s = parseExp("");

  return s;
}

//=============================================================================
