/** @file View2.C
    Creating and applying view differences.

    @see Game.H

    $Id: View2.C,v 1.26 2006/06/09 14:46:08 orts_mburo Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/View2.C,v $
*/

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Global.H"
#include "Game.H"
#include "GameObj.H"
#include "GameChanges.H"
#include "ServerObjData.H"
#include "GameTile.H"
#include "BoolMatrix.H"
#include "PlayerInfo.H"
#include "GameObj.H"
#include "GameObjFactory.H"
#include "ByteData.H"

using namespace std;

#define TEST 0

//===================================================================

/** Compares pointers to game objects by their blueprint code.
    Used for sorting prior to transmission.
*/

struct GameObjCodeCmp {

  bool operator()(GameObj * const &p1, GameObj * const &p2) {
    return p1->code() < p2->code();
  }

};

void Game::gen_view_diffs(Vector<string> &views)
{
  views.clear();
  FORS (i, player_n) views.push_back("");
  
  Vector<sint4> *new_tiles = new Vector<sint4>[player_n];

  // copy previous visibility flags

  FORS (p, player_n+1) {
    FORALL (objs[p], i) {
      GameObj *gob = (*i)->get_GameObj();
      if (gob != 0) {
        gob->sod.prev_seen_by = gob->sod.seen_by;
      }
    }
  }

  // currently visible tiles
  
  FORS (i, player_n) compute_view(i, &new_tiles[i]);

  // current visibility flags

  FORS (p, player_n) {
    PlayerSet seen_by;
    seen_by.set(p);
    FORS (i, player_n) {
      // if (x_shares_view_with_y->get(p,i)) seen_by.set(i);
      if (x_allied_with_y->get(p,i)) seen_by.set(i);
    }
    
    ObjCont &obj = objs[p];
    FORALL (obj, i) {
      GameObj *gob = (*i)->get_GameObj();
      if (gob != 0) {
        compute_visibility(*gob);
        gob->sod.seen_by.unite(seen_by);
      }
    }
  }

  FORALL (objs[player_n], i) { // visibility of NPCs
    GameObj *gob = (*i)->get_GameObj();
    if (gob != 0) {
      compute_visibility(*gob);
    }
  }

  // new visible opponent objects

  ObjCont cobjs;   // pointers to objects known to client
  ObjCont del;     // pointers to objects no longer visible to client
  
  FORS (p, player_n) {

    // compress list of new tiles:
    // 1. sort tile indexes
    // 2. gap: gap-length-1
    // 3. run: <tile-description>+ 
    
    sort(new_tiles[p].begin(), new_tiles[p].end());
    sint4 next = 0;
    string &s = views[p];

    FORALL (new_tiles[p], it) {
      sint4 pos = *it;
      write_b_int4(s, next-pos); // offset from the expected tile position ( value is <= 0 )

      GameTile &t = pf(pos);
      t.write(s);

      next = pos+1;
    }
    s += char(0x40); // end of section
    
    // count new visible objects and create new cids for them after sorting

    Vector<GameObj*> new_objs;
    
    FORS (q, player_n+1) {
      FORALL (objs[q], i) {
        GameObj *gob = (*i)->get_GameObj();
        if (gob) {
          if (!gob->sod.prev_seen_by.is_set(p) && gob->sod.seen_by.is_set(p)) {
            if (!gob->is_dead()) {
              new_objs.push_back(gob);
            }
          }
        }
      }
    }

    // sort new object points w.r.t. blueprint code
    sort(new_objs.begin(), new_objs.end(), GameObjCodeCmp());

    // assign client ids to new objects
    FORALL (new_objs, i) players[p]->new_id(*i);
    

    // create list of new objects in the form <blueprint code, number of those objs>+
    // terminate list by -1
    
    if (new_objs.size() > 0) {
      uint4 last = new_objs[0]->code(), last_index = 0;
      FORU (i, new_objs.size()) {
        if (new_objs[i]->code() != last) {
          write_b_int4(s, last);
          write_b_int4(s, i-last_index);
          last = new_objs[i]->code(); last_index = i;
        }
      }
      // last code 
      write_b_int4(s, last);
      write_b_int4(s, new_objs.size()-last_index);
    }
    s += char(0x41); // -1

    // get sorted object pointers
    players[p]->objs(cobjs);
    del.clear();

    // send updates, collect objects that are no longer visible
    
    //    cout << "num objs = " << cobjs.size() << endl;

    FORALL (cobjs, i) {

      GameObj *gob2 = (*i)->get_GameObj();
      if (gob2 == 0) {
        cout << "non GameObj object" << endl;
        continue;
      }
      GameObj &gob = *gob2;

      bool prev = gob.sod.prev_seen_by.is_set(p);
      bool now = gob.sod.seen_by.is_set(p);

      if (prev) {

        if (gob.is_dead()) {
          s += char(0x03); // 3 (dead)
          del.push_back(&gob);
        } else if (!now) {
          s += char(0x02); // 2 (no longer visible)
          del.push_back(&gob);  
        } else {
          string t;
          // update (line starts with "0")
          gob.gen_diff(false, x_allied_with_y->get(p,*gob.sod.owner),t,players[*gob.sod.owner]);
          s += t; 
        }

      } else if (now) {

        if (gob.is_dead()) {
          // because the object is dead it is no longer in objs[],
          // so seen_by was never copied to prev_seen_by --
          // the object WAS actually seen previously.

          s += char(0x03);
          del.push_back(&gob);

        } else {

          // new (line starts with "1")
          string t;
          gob.gen_diff(true, x_allied_with_y->get(p,*gob.sod.owner),t,players[*gob.sod.owner]);
          s += t;
        }
      }
    }

    // delete cids of invisible or dead objects
    FORALL (del, j) players[p]->delete_obj(*j, tick_count);
    s += char(0x04); // 4 (end of updates)
  }

  // copy object attributes

  FORS (p, player_n+1) {
    ObjCont &obj = objs[p];
    FORALL (obj, i) {
      (*i)->copy_to_prev();
      (*i)->reset_frame_data(); // single frame data such as attack direction
    }
  }

  delete [] new_tiles;
}

//===================================================================

/** Merge changes from the server with the current perceived state.
    Summarize the differences in changes.
    Previous attribute values are maintained by GameObj::apply_diff()
    @bug The call to apply_diff assumes that sod.owner is already set,
         new objects can break this.
*/

void Game::apply_view_diff(const string &view, GameChanges &changes, ActionChanges &act_changes)
{
  ++tick_count; // since there is no simulation_step in the client
  
  const char *s = view.c_str();

  //  cout << "VIEW" << endl << view << endl;

  changes.clear();
  act_changes.clear();
      
  //------------------- frame info --------------------  

  view_frame      = read_b_int4(s); // current simulation frame
  action_id       = read_b_int4(s); // id of action executed (-1 if none)
  skipped_actions = read_b_int4(s); // # of skipped actions
  merged_actions  = read_b_int4(s); // # of merged actions
  save_and_quit   = read_b_int4(s); // save and quit flag
  freq            = read_b_int4(s)/1024.0; // server simulation frequency
  
  if (save_and_quit) {
    cout << "SAVE REQUEST" << endl;
    ostringstream oss;
    save(oss);
    string file = get_savefile();
    ofstream savefile(file.c_str(), ios::binary);
    savefile.write(oss.str().c_str(), oss.str().size());
    cout << "SAVING TO " << file << " " << oss.str().size() << endl;
    return;
  }

  //------------------- new tiles --------------------

  sint4 pos = 0;
  
  FOREVER {
    if (s[0] == 0x40) { s++; break; }

    sint4 index = read_b_int4(s);
    pos -= index;

    pf(pos).read_b(s);
    update_tile(pos, &changes.new_boundaries);
    
    cplayer->explored.set(pf.ind2x(pos), pf.ind2y(pos));

    changes.new_tile_indexes.push_back(pos);
    ++pos;
  }

  // ------------------- new objects --------------------

  FOREVER {

    sint4 c, n;  // code, # of new objects with this code
    
    c = read_b_int4(s);
    if (c < 0) break;

    n = read_b_int4(s);

    if (n < 1) ERR("new object message corrupt");

    // create n new objects with code c

    FORS (i, n) {

      ScriptObj *gob = cplayer->gof.create(c);
      //      GameObj *gob = NPC->gof.create(c);
      changes.new_objs.push_back(gob);
      cplayer->new_id(gob);
      GameObj *g = gob->get_GameObj();
      if (g) {
        g->sod.in_game = true;
        g->init_sod();
      } else {
        cout << "non GameObj object" << endl;
      }
    }
  }

  // ------------------- update objects --------------------

  ObjCont all_objs, remove;

  cplayer->objs(all_objs);

  FORALL (all_objs, it) {

    char q;

    q = *s++; //read_b_int4(s);
    assert(abs(q) < 64);

    if (q == 3) { // object dead
      changes.dead_objs.push_back(*it);
      (*it)->kill();
      remove.push_back(*it);

    } else if (q == 2) { // object no longer visible
      changes.vanished_objs.push_back(*it);
      remove.push_back(*it);

    } else if (q == 0 || q == 1) {
      s--;
      if (q == 0) {
        if ((*it)->get_int("owner") < 0) REM("owner < 0");
        //if ((*it)->sod.owner < 0) REM("owner < 0"); // ScriptObj doesn't have sod!
      }

      if ((*it)->apply_diff(x_allied_with_y->get(opt.cpind, (*it)->get_int("owner")), s))
        changes.changed_objs.push_back(*it);

    } else ERR("corrupt update line");
 
  }

  if (*s++ != 4) { // EOM
    ERR("wrong # of update messages");
  }

  // add new objects to their respective obj vector

  FORALL (changes.new_objs, it) {
    sint4 owner = (*it)->get_int("owner");
    if (owner < 0) owner = player_n;
    objs[owner].push_back(*it);
    //    objs[owner].push_back(const_cast<GameObj*>(*it));
  }
  
  // free ids of removed objects
  // delete removed objects from obj containers

  FORALL (remove, it) {

    // fixme: TIM
    //    if (!(*it)->get_int("static", false)) (*it)->destruct();

    cplayer->delete_obj(*it, tick_count);
    sint4 owner = (*it)->get_int("owner");
    if (owner < 0) owner = player_n;
    // fixme: finding objects should be faster
    FIND (objs[owner], j, *it);
    if (j == objs[owner].end()) ERR("obj not found");
    objs[owner].erase(j);
  }

  update_sectors();
  
  // new: update action queues

  if (get_action_frame() >= 0) {

    // remove all pending actions with frame_sent <= get_action_frame()

    // Example:
    //             | 3 skipped |   4 merged   |
    // msg num:  2   3   4   5   6   7   8   9
    //           ^           ^               ^
    //                                       E = id of last action 
    //                                           msg executed = 
    //                                           get_action_frame() = 9
    //                       S = E - get_merged_actions() = 9 - 4
    //           P = S - get_skipped_actions() = 5 - 3

    // E[xecuted] S[kipped] P[revious]
    const sint4 E = get_action_frame();
    const sint4 S = E - get_merged_actions();
    const sint4 P = S - get_skipped_actions();

#if TEST    
    cout << " vf:" << get_view_frame()    
	 << " af:" << get_action_frame()
	 << " sa:" << get_skipped_actions()
	 << " ma:" << get_merged_actions()
	 << endl;
#endif    
    
    FORALL (objs[opt.cpind], i) {

      GameObj *p = dynamic_cast<GameObj*>(&(**i));
      if (!p) continue;
      
      typeof(p->pending_actions.begin())
	j   = p->pending_actions.begin(),
	end = p->pending_actions.end();

      // sanity check
      while(j != end && j->frame_sent <= P) {
	REM("should not happen! frame_sent <= P ? " << j->frame_sent << " " << P);
	j = p->pending_actions.erase(j);
      }

      // skipped actions
      while (j != end && j->frame_sent <= S) {
	act_changes.skipped.push_back(*j);
	j = p->pending_actions.erase(j);
      }

      // executed actions
      while (j != end && j->frame_sent <= E) {	
	act_changes.executed.push_back(*j);
	j = p->pending_actions.erase(j);
      }
    }
  }

#if TEST  
  cout << "S:" << act_changes.skipped.size() << " "
       << "E:" << act_changes.executed.size()
       << endl;
#endif
  
}
