/** @file Vision.C
    @see Vision.H

    $Id: Vision.C,v 1.10 2006/04/03 21:32:16 orts_mburo Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/Vision.C,v $
*/

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Vision.H"

using namespace std;

// static variable

sint4* MapView::b2t;
sint4* MapView::t2b;

//===================================================================

void PartialTileView::alloc(TileView& x, bool SkipVal) {

  a = z = -1;
  FORS(i, YN) {
    VisionT temp = x.row(i);
    if (SkipVal) {
      if (~temp) { // the row contains a 0
	if (a < 0) a = i;
	z = i;
      }
    } else {
      if (temp) { // the row contains a 1
	if (a < 0) a = i;
	z = i;
      }
    }
  }

  if (a == -1) {
    a = 0;
    z = -1;
    bits = 0;
    return;
  }
  assert(z-a+1 > 0);
  bits = new VisionT[z-a+1];
  
  for (sint4 i = a; i <= z; ++i) {
    bits[i-a] = x.row(i);
  }
}

//===================================================================

/** Dump the state of each pixel to stdout for debugging. */
void TileView::print() const {
  FORS (y, YN) {
    FORS (x, XN) {
      cout << is_set(x,y);
    }
    cout << endl;
  }
}

//--------Row Ops----------------------------------------------------

void TileView::row_mask(VisionT x, sint4 y) {
  assert(y >= 0 && y < YN);
  bits[y] &= x;
}

void TileView::row_mask(const TileView& x, sint4 y) {
  assert(y >= 0 && y < YN);
  bits[y] &= x.bits[y];
}

void TileView::row_set(VisionT x, sint4 y) {
  assert(y >= 0 && y < YN);
  bits[y] = x;
}

VisionT TileView::row_and(const TileView& x, sint4 y) const {
  assert(y >= 0 && y < YN);
  return bits[y] & x.bits[y];
}

VisionT TileView::row(sint4 y) const {
  assert(y >= 0 && y < YN);
  return bits[y];
}

//--------View Ops---------------------------------------------------

void TileView::copy(const TileView& x) {
  if (this == &x) return;
  alloc();  

  NO_OVERLAP(bits, x.bits, YN);
  memcpy(bits, x.bits, sizeof(VisionT)*YN);
}

void TileView::view_and(const TileView& x) {
  assert(bits);
  FORS (i, YN) bits[i] &= x.bits[i];
}

void TileView::view_and(const PartialTileView& x) {
  assert(bits);
  
  for (sint4 i=x.a, j=0; i <= x.z; ++i,++j) bits[i] &= x.bits[j];
}

void TileView::view_and_upper(const TileView& x) {
  assert(bits);
  for (sint4 i = GameConst::VIEW_MAX+1; i < YN; ++i) bits[i] &= x.bits[i];
}

void TileView::view_and_lower(const TileView& x) {
  assert(bits);
  for (sint4 i = 0; i < GameConst::VIEW_MAX; ++i) bits[i] &= x.bits[i];
}

void TileView::view_or(const TileView& x) {
  assert(bits);
  FORS (i, YN) bits[i] |= x.bits[i];
}

//===================================================================

void DualView::alloc(DualView &x)
{
  lshift = x.lshift;
  if (x.dual) {
    if (!bits) {
      bits = new VisionT[2*YN];
      dual = true;
    } else {
      if (!dual) {
        delete [] bits;
        bits = new VisionT[2*YN];
        dual = true;
      }
    }
    memcpy(bits, x.bits, sizeof(bits[0])*2*YN);
  } else {
    dual = false;
    if (!bits) bits = new VisionT[YN];
    memcpy(bits, x.bits, sizeof(bits[0])*YN);
  }
}

/** Set to TileView with lshift as the amount of whitespace added to the left.
    IMPORTANT!!!  The "left" side is stored in the low bits, not the left side as if
    it were written as a binary number.
*/

void DualView::alloc(const TileView &v, sint4 lshift_amount)
{
  lshift = lshift_amount;
  sint4 rshift = XN - lshift;
  bool dual_ = GameConst::VIEW_WIDTH > rshift; // iff crosses a T boundary

  if (dual_) {
    if (!bits) {
      bits = new VisionT[2*YN];
      dual = true;
    } else {
      if (!dual) {
        delete [] bits;
        bits = new VisionT[2*YN];
        dual = true;
      }
    }
    for (sint4 i = 0, j = 0; i < 2*YN; i+=2, ++j) {
      VisionT row = v.row(j);
      bits[i]   = row << lshift;
      bits[i+1] = row >> rshift;
    }
  } else {
    if (!bits) {
      bits = new VisionT[YN];
    }
    dual = false;
    FORS (i, YN) {
      VisionT row = v.row(i);
      bits[i] = row << lshift;
    }
  }
}

/** Create a new TileView extracted from this object. */

TileView* DualView::make_TileView() const
{
  sint4 rshift = XN - lshift;

  TileView *t = new TileView;
  t->alloc();
  t->clear();

  if (dual) { // TileView spans two T words
    for (sint4 i = 0; i < 2*YN; i+=2) {
      VisionT row = (bits[i] >> lshift) | (bits[i+1] << rshift);
      t->row_set(row, i/2);
    }
  } else {
    FORS (i, YN) {
      VisionT row = bits[i] >> lshift;
      t->row_set(row, i);
    }
  }
  return t;
}

//===================================================================

MapView::MapView(uint4 X, uint4 Y): TX(X), TY(Y)
{
  edge_mask = 0;
  sint4 s = TX % BpE;
  if (s == 0) s = BpE;
  FORS (i, s) { edge_mask <<= 1; edge_mask |= 1; }
  
  XN = TX / BpE;
  if (TX % BpE) XN++; // space for a partial block
  
  YN = TY + 2*GameConst::VIEW_MAX + 2; // pad with max view height and a little more
  
  XYN = XN * TY;
  
  bits = new VisionT[XN * YN];
  crop = &bits[XN * (GameConst::VIEW_MAX+1)]; // start at the first actual row
  clear();
}

//-------------------------------------------------------------------

void MapView::save(ostream &os) const
{
  os.write(reinterpret_cast<char*>(bits), sizeof(VisionT)*XN*YN);
}

void MapView::load(istream &is)
{
  is.read(reinterpret_cast<char*>(bits), sizeof(VisionT)*XN*YN);
}

//--------Bit Ops----------------------------------------------------

void MapView::set(sint4 x, sint4 y) {
  assert(bits);
  assert(x >= 0 && x < TX);
  assert(y >= 0 && y < TY);

  crop[y * XN + (x / BpE)] |= (1 << (x & bit_mask));
}

void MapView::unset(sint4 x, sint4 y) {
  assert(bits);
  assert(x >= 0 && x < TX);
  assert(y >= 0 && y < TY);

  crop[y * XN + (x / BpE)] &= ~(1 << (x & bit_mask));
}

#if 0
bool MapView::is_set(sint4 x) const {  // assume TX = k*BpE
  sint4 ind = x / BpE;
  return (crop[ind] & (1 << (x & bit_mask)));
}
#endif

//--------View Ops---------------------------------------------------

void MapView::fill()
{
  assert(bits);
  memset(crop, 0xFF, sizeof(bits[0]) * XYN);
  FORS (i, TY) crop[(i+1)*XN-1] &= edge_mask;
}

/** Merge a tileview (centered at some tile) into the map. */
void MapView::view_or(const DualView& t, sint4 tile_index)
{
  sint4 tind = t2b[tile_index]; // upper left corner of the dualview as a crop index

  if (t.dual) {
    for (sint4 i = 0; i < 2*GameConst::VIEW_WIDTH; i+=2) {
      crop[tind]   |= t.bits[i];
      crop[tind+1] |= t.bits[i+1];
      tind += XN;
    }
  } else {
    FORS(i, GameConst::VIEW_WIDTH) {
      crop[tind] |= t.bits[i];
      tind += XN;
    }
  }
}

/** Check if a tile view (centered at some tile) intersects the map. */
bool MapView::view_intersects(const DualView& t, sint4 tile_index) const
{
  sint4 tind = t2b[tile_index]; // upper left corner of the dualview as a crop index

  if (t.dual) {
    for (sint4 i = 0; i < 2*GameConst::VIEW_WIDTH; i+=2) {
      if (crop[tind]   & t.bits[i]) return true;
      if (crop[tind+1] & t.bits[i+1]) return true;
      tind += XN;
    }
  } else {
    FORS(i, GameConst::VIEW_WIDTH) {
      if (crop[tind] & t.bits[i]) return true;
      tind += XN;
    }
  }

  return false;
}

//--------Map Ops----------------------------------------------------

void MapView::copy(const MapView& x) {
  assert(bits);
  assert(x.bits);
  if (this == &x) return;
  memcpy(crop, x.crop, sizeof(VisionT) * XYN);
}

void MapView::map_and(const MapView& x) {
  assert(bits);
  assert(x.bits);
  if (this == &x) return;
  FORS (i, XYN) crop[i] &= x.crop[i];
}

void MapView::map_or(const MapView& x) {
  assert(bits);
  assert(x.bits);
  if (this == &x) return;
  FORS (i, XYN) crop[i] |= x.crop[i];
}

/** Generate an ordered vector of tiles which are marked in the current map, but not in another.
    @sideeffect v = current map - v
*/
void MapView::map_whatsnew(MapView& v, Vector<sint4> &new_tiles)
{
  assert(bits);
  assert(v.bits);

  new_tiles.clear();
  
  FORS(i, XYN) {
    v.crop[i] = crop[i] & ~(v.crop[i]);
  }

  VisionT block;

  FORS(i, XYN) {
    block = v.crop[i];
    if (block) {
      if (block & 0x0000FFFF) for (sint4 z = 0; z < 16; ++z) {
	if (block & (1<<z)) { new_tiles.push_back(b2t[i]+z); }
      }
      if (block & 0xFFFF0000) for (sint4 z = 16; z < 32; ++z) {
	if (block & (1<<z)) { new_tiles.push_back(b2t[i]+z); }
      }
    }
  }
}

//===================================================================
