#ifndef Vision_H
#define Vision_H

/** @file Vision.H
    Bitmap visibility classes.

    $Id: Vision.H,v 1.11 2006/04/03 00:41:32 orts_mburo Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/Vision.H,v $
*/

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "Global.H"
#include "GameConst.H"
#include "SaveLoad.H"

class TileView;  // bitmap of visible tiles

typedef uint4 VisionT; // data type vision data is stored in


//===================================================================

// unused class
class PartialTileView {  // immutable space/calculation-saving TileView

private:

  // assignment and copy forbidden
  PartialTileView& operator=(const PartialTileView &x);
  PartialTileView(const PartialTileView &x);
  
public:

  static const sint4 XN = GameConst::VIEW_WIDTH;
  static const sint4 YN = GameConst::VIEW_WIDTH;

  VisionT *bits;
  sint4 a, z; // range of rows covered in this construct

  PartialTileView() : bits(0) { assert(static_cast<sint4>(sizeof(VisionT)) * 8 >= XN); a = z = 0; }

  ~PartialTileView() { if (bits) { delete [] bits; bits = 0; } }

  void alloc(TileView &x, bool SkipVal); // ignore boundary rows of value "SkipVal"
};

//===================================================================

/** Parts of the world visible from a tile.
    A 0-1 bitmap representation supporting boolean operations with other views,
    and which can be inserted into larger (similar) data structures.
 */

class TileView {

private:

  // assignment and copy forbidden
  TileView& operator=(const TileView &x);
  TileView(const TileView &x);
  
public:

  static const sint4 XN = GameConst::VIEW_WIDTH;  
  static const sint4 YN = GameConst::VIEW_WIDTH;

  VisionT *bits; ///< The view data, stored as rows of a sufficiently large data type T.

  TileView() : bits(0) { assert(static_cast<sint4>(sizeof(VisionT)) * 8 >= XN); }

  ~TileView() { if (bits) delete [] bits; }

  bool defined() const { return bits != 0; }
  void print() const;
  void alloc() { if (!bits) bits = new VisionT[YN]; }
  void free() { if (bits) delete [] bits; bits = 0; }
  void clear() { assert(bits); memset(bits, 0, sizeof(bits[0]) * YN); } 
  void fill() { assert(bits); memset(bits, 0xFF, sizeof(bits[0]) * YN); }

  // BIT OPERATIONS //------------------------

  inline void set(sint4 x, sint4 y) {
    assert(bits);
    assert(x >= 0 && x < XN);
    assert(y >= 0 && y < YN);
    bits[y] |= (1<<x);
  }

  inline void unset(sint4 x, sint4 y) {
    assert(bits);
    assert(x >= 0 && x < XN);
    assert(y >= 0 && y < YN);
    bits[y] &= ~(1<<x);
  }

  inline bool is_set(sint4 x, sint4 y) const {
    assert(bits);
    assert(x >= 0 && x < XN);
    assert(y >= 0 && y < YN);
    return (bits[y] & (1<<x)) != 0;
  }

  // ROW OPERATIONS //------------------------

  void row_mask(VisionT x, sint4 y);
  void row_mask(const TileView &x, sint4 y);
  void row_set(VisionT x, sint4 y);
  VisionT row_and(const TileView &x, sint4 y) const;
  VisionT row(sint4 y) const;

  // VIEW OPERATIONS //------------------------

  void copy(const TileView &x);             ///< Set this TileView to match x
  void view_and(const TileView &x);         ///< And-combine with x
  void view_and(const PartialTileView &x);  ///< And-combine with x
  void view_and_upper(const TileView &x);   ///< And-combine the upper bitmap lines with x
  void view_and_lower(const TileView &x);   ///< And-combine the lower bitmap lines with x
  void view_or(const TileView &x);          ///< Or-combine with x
};

//===================================================================

/** A T-aligned and offset version of a TileView.
    Normal initialized with from a TileView and an offset, the resulting
    aligned data is able to be quickly combined with MapViews.
 */

class DualView {

private:

  // assignment and copy forbidden
  DualView& operator=(const DualView &x);
  DualView(const DualView &x);
  
public:

  static const sint4 XN = sizeof(VisionT)*8;
  static const sint4 YN = GameConst::VIEW_WIDTH;
  
  VisionT *bits;   ///< Interleaved left and right portions of the view.
  sint4 lshift;
  bool dual;       ///< True iff view crosses a VisionT boundary.

  DualView() : bits(0) { lshift = 0; dual = false; }

  ~DualView() {
    if (bits) { delete [] bits; bits = 0; }
  }

  void alloc(DualView &x);
  void alloc(const TileView &v, sint4 lshift_amount);
  TileView* make_TileView() const;

  bool defined() { return bits != 0; }

  // VIEW OPERATIONS //------------------------

  void view_and(const DualView &x) {
    assert(x.bits && bits);
    assert(x.lshift == lshift);
    assert(x.dual == dual);

    FORS (i, dual ? 2*YN : YN) {
      bits[i] &= x.bits[i];
    }
  }

  void view_or(const DualView &x) {
    assert(x.bits && bits);
    assert(x.lshift == lshift);
    assert(x.dual == dual);

    FORS (i, dual ? 2*YN : YN) {
      bits[i] |= x.bits[i];
    }
  }
};

//===================================================================

/** A 0-1 bitmap of the entire game board.
    Allows for quick boolean operations with DualViews or other MapViews.
 */

class MapView : public SaveLoad {

private:

  // assignment and copy forbidden
  MapView& operator=(const MapView &x);
  MapView(const MapView &x);
  
public:

  void save(std::ostream &os) const;
  void load(std::istream &is);

  /// Maps the starting bit of each element, crop[i], to a tile index (defined by Game).
  static sint4 *b2t;
  static sint4 *t2b;

  static const uint4 BpE = sizeof(VisionT)*8;  ///< # of bits per element.
  /// (# of bits in T) - 1.
  /// Used instead of '% BpE' to determine the bit-position in a VisionT of an x coordinate.
  static const VisionT bit_mask = BpE-1;

  VisionT edge_mask;

  const sint4 TX; ///< Number of tiles in the X direction.
  const sint4 TY; ///< Number of tiles in the Y direction.

  sint4 XN;  ///< # of elements (T) per row.
  sint4 YN;  ///< # of rows (includes padding above and below).
  sint4 XYN; ///< # of non-padding elements (T) in the array.

  VisionT *bits;   ///< Bitmap data.
  VisionT *crop;   ///< A more index-friendly method to access bits (starts at x=0,y=0).
  
  MapView(uint4 X, uint4 Y);
  ~MapView() { if (bits) delete [] bits; }

  //  bool defined() const { return bits != 0; } ///< If space has been allocated.
  
  void clear() { assert(bits); memset(crop, 0, sizeof(bits[0]) * XYN); } 
  void fill();

  // BIT OPERATIONS //-------------------------

  void set(sint4 x, sint4 y);
  void unset(sint4 x, sint4 y);

  inline bool is_set(sint4 x, sint4 y) const {
    assert(bits);
    assert(x >= 0 && x < TX);
    assert(y >= 0 && y < TY);
    
    return (crop[y * XN + (x / BpE)] & (1 << (x & bit_mask))) != 0;
  }

  // VIEW OPERATIONS //------------------------

  //  void view_or(const TileView &t, sint4 x, sint4 y);
  //  bool view_intersect(const TileView &t, sint4 x, sint4 y) const;

  void view_or(const DualView &view, sint4 tile_index);
  bool view_intersects(const DualView &view, sint4 tile_index) const;

  // MAP OPERATIONS //------------------------

  void copy(const MapView &x);
  void map_and(const MapView &m);
  void map_or(const MapView &m);

  void map_whatsnew(MapView &v, Vector<sint4> &new_tiles);
  //  void map_whatsnew(const MapView &v, uint4 *visible, uint4 magic);
  //  void map_tile_list(uint4 *visible, uint4 magic);
  
};

//===================================================================

class BitStrip {
private:
  typedef uint4 B;                       // data type bit strips are made of
  static const uint4 BS = sizeof(B)*8;   // number of bits in a "B"

  // the individual bit functions are only used in precomputing strips
  // masking and intersecting is used within generate tile view

public:

  B *bits;
  uint4 length;
  sint4 nbits;
  
  BitStrip() : bits(0), length(0), nbits(0) {}
  
  BitStrip(const BitStrip &b) {
    bits = new B[b.length];
    length = b.length;
    nbits = b.nbits;
    memcpy(bits, b.bits, b.length*sizeof(B));
  }
  BitStrip &operator=(const BitStrip &b) {
    if (bits) delete [] bits;
    bits = new B[b.length];
    length = b.length;
    nbits = b.nbits;
    memcpy(bits, b.bits, b.length*sizeof(B));
    return *this;
  }
  ~BitStrip() { if (bits) delete [] bits; }

  // Initialization functions, not speed critical

  void alloc(uint4 size) {
    assert(bits==0);
    nbits = size;
    length = size / BS;
    
    if (size % BS) length++;
    bits = new B[length];
    assert(bits);
  }

  void clear() { assert(bits); memset (bits, 0, sizeof(bits[0])*length); }
  void fill()  { assert(bits); memset (bits, 0xFF, sizeof(bits[0])*length); }

  void set(sint4 x) {
    assert(bits);
    assert(x >= 0 && x < nbits);

    uint4 y = x / BS;
    x %= BS; 
    assert(y < length);   
    bits[y] |= (1<<x);
  }

  void unset(sint4 x) {
    assert(bits);
    assert(x >= 0 && x < nbits);

    uint4 y = x / BS;
    x %= BS;
    assert(y < length);
    bits[y] &= ~(1<<x);
  }

  bool is_set(sint4 x) const {
    assert(bits);
    assert(x >= 0 && x < nbits);

    uint4 y = x / BS;
    x %= BS;
    assert(y < length);
    return (bits[y] & (1<<x)) != 0;
  }

  void print() const {
    assert(bits);

    FORS (i, nbits) { //length*BS) {
      if (i == nbits) std::cout << "|"; // mark the end of valid data
      std::cout << (is_set(i) ? "1" : "0");
    }
    std::cout << std::endl;
  }

  bool operator==(const BitStrip &b) const {
    FORU (i, length) {
      if (bits[i] != b.bits[i]) return false;
    }
    return true;
  }

  bool operator!=(const BitStrip &b) const {
    FORU(i, length) {
      if (bits[i] == b.bits[i]) return false;
    }
    return true;
  }

  // Core functions, called many times

  inline void and_strip(const BitStrip &x) {
    assert(bits);
    assert(x.bits);

    FORU(i, length) bits[i] &= x.bits[i];
  }

  inline bool intersects(const BitStrip &x) const {
    assert(bits);
    assert(x.bits);

    FORU (i, length) {
      if (bits[i] & x.bits[i]) return true;
    }
    return false;
  }
};

//===================================================================

#endif
