// $Id: Text_IO_Buffer.C,v 1.6 2006/04/26 01:38:37 orts_mburo Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

/**
 * \class Text_IO_Buffer
 * \author Sami Wagiaalla
 * \brief does text io termintated with a newline
 * Inherits from IO_Buffer.
 * Uses threads to avoid blocking of the main thread
 * unlike IO_Buffer, Text_IO_Buffer deals with newline terminated text messages
 * instead of binary
 * buffered socket I/O
 * maintains messages buffers that are sent/rcvd by threads
 * that do blocking TCP I/O via SDL_net
 */

#include "Text_IO_Buffer.H"
#include "Compress.H"
#include "ShutDown.H"
#include "Random.H"

using namespace std;

/**
* this is the exact same constructor as IO_Buffer
* except that the threads are passed the Text_IO_Buffer_(read/write)_loop
* instead of IO_Buffer_(read/write)_loop
**/
Text_IO_Buffer::Text_IO_Buffer(TCPsocket so_, bool blocking_, RandSrc *rand_delay)
  : IO_Buffer(so_, 0, rand_delay)
{
  REM("Text_IO_Buffer constructor called");
  
  //so = so_; 
  //delay_ms = delay_ms_;  
  //clev = 0; /** cannot compress GGS must understand message g*/
  
  blocking = blocking_;
  
  i_len = o_len = 0;
  i_bytes = o_bytes = 0;
  done = false;
  read_good = write_good = true;

  buf_mut = SDL_CreateMutex();
  i_delay_buf_mut = SDL_CreateMutex();  
  o_buf_sem = SDL_CreateSemaphore(0);
  input_chars = new uint1[MAX_LEN];
  
  if(blocking){
    read_sem = SDL_CreateSemaphore(0);
  }
  
  read_thread  = SDL_CreateThread(Text_IO_Buffer::read_loop, this);  // start the read loop thead
  write_thread = SDL_CreateThread(Text_IO_Buffer::write_loop, this); // start the write loop thread

  if (rand_delay) i_delay_thread = SDL_CreateThread(IO_Buffer::i_delay_loop, this);
  else i_delay_thread = 0;
}


Text_IO_Buffer::~Text_IO_Buffer()
{
  REM("~Text_IO_Buffer called");
  SDL_DestroySemaphore(read_sem);

  SDL_KillThread(read_thread);
  read_thread = 0;

  SDL_KillThread(write_thread);
  write_thread = 0;
}

// remove last message from input buffer
// and return size of buffer (before removal)

sint4 Text_IO_Buffer::read(string &s)
{
  sint4 ret;
  s.clear();

  if (!read_good) return -1; // error

  if(blocking){
    // wait for a message
    sint4 sw;
    while ((sw = SDL_SemWait(read_sem))) {
      REM("Text IO Buffer read awoken prematurely");
    }
  }
  
  // <<<<<<<<<<<<<<<< START CRITICAL SECTION
  if (SDL_mutexP(buf_mut)) ERR("mutex 9");
  ret = i_len;
  if (ret > 0) {
    s = i_buf.front();
    i_buf.pop_front();
    --i_len;
  }
  if (SDL_mutexV(buf_mut)) ERR("mutex 10");
  // >>>>>>>>>>>>>>>> END CRITICAL SECTION

  return ret;
}


sint4 Text_IO_Buffer::probing_blocking_read(void *buffer)
{
  if (!read_good || !write_good) return -1;
 
  sint4 bytes = 0;
  sint4 res=0;
  sint1 ch;

  FOREVER {

    do {
      res = SDLNet_TCP_Recv(so, &ch, sizeof(ch));
    } while ((res < 0) && (!shut_down));
    
    if (shut_down)
      return -1;

    if (res <= 0) {
      cerr << ">>>>>>>> READ: "; perror("read"); cout << flush; cerr << endl;
      //continue;
      return -1;
    }
    
#if 0
    if (res == 0 && errno != 0) {
      cerr << "?????????? READ: "; perror("read"); cout << flush; cerr << endl;
      continue;
    }
#endif
    
    if (ch == '\n' || ch == '\r'){
      return bytes;
    }
    
    (static_cast<uint1*>(buffer))[bytes] = ch;
    bytes++;
    
    if (bytes >= Text_IO_Buffer::MAX_LEN) {
      REM("message too long -> kill connection");
      return -1; // terminate thread
    }
  }
}

int Text_IO_Buffer::write_loop(void *data)
{
  Text_IO_Buffer &io = *static_cast<Text_IO_Buffer*>(data);
  REM("write loop started");

  if (!io.rand_delay) {

    // no delay
    
    while (!io.done) {

      // wait for a message
      sint4 sw;
      while ((sw = SDL_SemWait(io.o_buf_sem))) {
        REM("Text IO Buffer write loop awoken prematurely");
      }

      if (shut_down)
        break;

      // <<<<<<<<<<<<<<<< START CRITICAL SECTION
      if (SDL_mutexP(io.buf_mut)) ERR("mutex 1");
      // get message from o_buf
      string msg;
      if (io.o_len <= 0) ERR("buffer empty!");
      msg = io.o_buf.back();
      io.o_buf.pop_back();
      --io.o_len;
      if (SDL_mutexV(io.buf_mut)) ERR("mutex 2");
      // >>>>>>>>>>>>>>>> END CRITICAL SECTION

      uint4 len = msg.size();
     
      // terminate the messege
      msg += '\n';
      len++;
         
      // send message
      if (io.blocking_write(const_cast<char*>(msg.data()), len) < 0) {
        REM("socket write error 2");
        if (SDLNet_GetError() && strlen(SDLNet_GetError())) // sometimes blank! ????
          REM2("SDLNet_TCP_Send: ", SDLNet_GetError());
        break;
      }

      // <<<<<<<<<<<<<<<< START CRITICAL SECTION
      if (SDL_mutexP(io.buf_mut)) ERR("mutex 3");
      // update output statistics
      io.o_bytes += sizeof(len) + len;
      if (SDL_mutexV(io.buf_mut)) ERR("mutex 4");
      // >>>>>>>>>>>>>>>> END CRITICAL SECTION
    }

  } else {

    // delay messages

    while (!io.done) {

      bool msg_available = false;
      string msg;
      
      // <<<<<<<<<<<<<<<< START CRITICAL SECTION
      if (SDL_mutexP(io.buf_mut)) ERR("mutex 1");
      // get message from o_buf
      if (io.o_len > 1) {
        msg_available = true;
        msg = io.o_buf.back();
        io.o_buf.pop_back();
        --io.o_len;
      }
      if (SDL_mutexV(io.buf_mut)) ERR("mutex 2");
      // >>>>>>>>>>>>>>>> END CRITICAL SECTION

      if (msg_available) {
        
        // put message in delay buffer
	
        io.o_delay_buf.push_front(msg);
        io.o_delay_times.push_front(SDL_GetTicks() + (*io.rand_delay)());
	
      } else SDL_Delay(1); // prevent spinning

      // send messages that are due

      uint4 now = SDL_GetTicks();
	
      while (!shut_down) {

        if (io.o_delay_buf.empty()) break;
        
        if (now <= io.o_delay_times.back()) {
          
          // REM2("send delayed message ", SDL_GetTicks());
          string msg = io.o_delay_buf.back();
          io.o_delay_buf.pop_back();
          io.o_delay_times.pop_back();
	  
          uint4 len = msg.size();
          
          // terminate the messege
          msg+= '\n';
          len++;
					
          // send message
          if (io.blocking_write(const_cast<char*>(msg.data()), len) < 0) {
            REM("socket write error 2");
            if (SDLNet_GetError() && strlen(SDLNet_GetError())) // sometimes blank! ????
              REM2("SDLNet_TCP_Send: ", SDLNet_GetError());
            break;
          }
	  
          // <<<<<<<<<<<<<<<< START CRITICAL SECTION
          if (SDL_mutexP(io.buf_mut)) ERR("mutex 3");
          // update output statistics
          io.o_bytes += sizeof(len) + len;
          if (SDL_mutexV(io.buf_mut)) ERR("mutex 4");
          // >>>>>>>>>>>>>>>> END CRITICAL SECTION
        }
      }
    }
  }

  io.write_good = false;
  REM("write loop stopped");
  return -1;
}

  
int Text_IO_Buffer::read_loop(void *data)
{
  Text_IO_Buffer &io = *static_cast<Text_IO_Buffer*>(data);
  REM("read loop started");

  if (!io.rand_delay) {

    // no delay
  
    while ((!io.done) && (!shut_down)) {

      sint4 len;

      // cerr << "READ len=" << len << endl;

      // read message
      len = io.probing_blocking_read(io.input_chars);
      if (len < 0) {
	REM("socket read error 2");
	if (SDLNet_GetError() && strlen(SDLNet_GetError())) // sometimes blank! ????
	  REM2("SDLNet_TCP_Recv: ", SDLNet_GetError());
	break;
      }

      // REM("READ");

      
      string msg(reinterpret_cast<char*>(io.input_chars), len);
      
      //      cout << msg << endl;
      
      // <<<<<<<<<<<<<<<< START CRITICAL SECTION    
      if (SDL_mutexP(io.buf_mut)) ERR("mutex 5");
      // append message
      io.i_buf.push_back(msg);
      ++io.i_len;
      // update input statistics
      io.i_bytes += sizeof(len) + len;
      if (SDL_mutexV(io.buf_mut)) ERR("mutex 6");
      // >>>>>>>>>>>>>>>> END CRITICAL SECTION

      if(io.blocking){
        SDL_SemPost(io.read_sem); // signal read that there is data available
      }
      
    }
  
  } else {

    // delay

    while ((!io.done) && (!shut_down)) {

      sint4 len;

      // cerr << "READ len=" << len << endl;

      // read message
      len = io.probing_blocking_read(io.input_chars);
      if (len < 0) {
	REM("socket read error 2");
	if (SDLNet_GetError() && strlen(SDLNet_GetError())) // sometimes blank! ????
	  REM2("SDLNet_TCP_Recv: ", SDLNet_GetError());
	break;
      }

      // REM("READ");

      string msg(reinterpret_cast<char*>(io.input_chars), len);
      
      // put message in i_delay_buf

      // <<<<<<<<<<<<<<<< START CRITICAL SECTION    
      if (SDL_mutexP(io.i_delay_buf_mut)) ERR("mutex 5");
      io.i_delay_buf.push_front(msg);
      io.i_delay_times.push_front(SDL_GetTicks() + (*io.rand_delay)());
      if (SDL_mutexV(io.i_delay_buf_mut)) ERR("mutex 6");
      // >>>>>>>>>>>>>>>> END CRITICAL SECTION
    }
  }
  
  io.read_good = false;
  REM("read loop stopped");

  if(io.blocking){ // wake up blocked clients
    SDL_SemPost(io.read_sem);
  }
      
  return -1;
}

#if 0
// delivers read messages delayed
int Text_IO_Buffer_i_delay_loop(void *data)
{
  Text_IO_Buffer &io = *static_cast<Text_IO_Buffer*>(data);
  REM("i_delay loop started");

  while (!io.done) {

    uint4 now = SDL_GetTicks();

    FOREVER {
      
      bool msg_to_be_delivered = false;
      string msg;
      
      // <<<<<<<<<<<<<<<< START CRITICAL SECTION    
      if (SDL_mutexP(io.i_delay_buf_mut)) ERR("mutex 15");
      msg_to_be_delivered = !(io.i_delay_buf.empty() || now < io.i_delay_times.back());
      if (msg_to_be_delivered) {
        msg = io.i_delay_buf.back();
        io.i_delay_buf.pop_back();
        io.i_delay_times.pop_back();
      }
      if (SDL_mutexV(io.i_delay_buf_mut)) ERR("mutex 16");

      if (!msg_to_be_delivered) { SDL_Delay(1); break; }

      // deliver message
      
      // <<<<<<<<<<<<<<<< START CRITICAL SECTION    
      if (SDL_mutexP(io.buf_mut)) ERR("mutex 17");
      // append message
      io.i_buf.push_back(msg);
      ++io.i_len;
      // update input statistics
      io.i_bytes += sizeof(sint4) + msg.size();
      if (SDL_mutexV(io.buf_mut)) ERR("mutex 18");
      // >>>>>>>>>>>>>>>> END CRITICAL SECTION

      REM2("recv delayed message ", SDL_GetTicks());
    }
  }

  REM("i_delay loop stopped");
  return -1;
}
#endif
