#ifndef OImage_H
#define OImage_H

// $Id: OImage.H,v 1.1 2006/05/31 22:30:57 orts_mburo Exp $

// This is an ORTS file
// (c) Tim Furtak
// (c) Erik Benner
// licensed under the GPL

#include "OGfxGlobal.H"

/**
* \class Image
* A TGA image loader.  Supports 8/24/32 bit uncompressed
* or run-length encoded images.
*/
class Image
{
public:
  /**
   * Texture filter quality settings
   */
  enum FILTER {NEAREST, LINEAR, BILINEAR, TRILINEAR};

  Image();
  Image(const std::string &file, real4 gamma=1.0);
  ~Image();

  void copy_rect(const Image &i, sint4 sx, sint4 sy, sint4 w, sint4 h, sint4 dx, sint4 dy);

  /**
   * Load TGA or TARGA image from file
   *
   * \par TGA Format Specification:\n
   * \li 1 byte - Comment Length
   * \li 1 byte - Colour Map Type   (unsupported)
   * \li 1 byte - Data Type Code
   * \li 2 byte - Colour Map Origin (unsupported)
   * \li 2 byte - Colour Map Length (unsupported)
   * \li 1 byte - Colour Map Depth  (unsupported)
   * \li 2 byte - X Origin          (unsupported)
   * \li 2 byte - Y Origin          (unsupported)
   * \li 2 byte - Width
   * \li 2 byte - Height
   * \li 1 byte - Bits Per Pixel
   * \li 1 byte - Image Descriptor
   * \li * byte - Pixel Data
   *
   * \note TGA colours are stored in BGR/A component order.
   * \note TGA data is stored with little-endian byte order.
   */
  bool loadTGA(const std::string &file, real4 gamma=1.0);
  bool saveTGA(const std::string &file);
  
  /**
   * Initialize a blank image
   */
  void empty(uint2 w, uint2 h, uint2 bpp);
  /**
   * Create OpenGL texture from image data
   */
  GLuint createTexture(FILTER filter=LINEAR, sint4 level=0) const;
  /**
   * Delete OpenGL texture
   */
  bool deleteTexture(GLuint id);
  
  // Accessor methods
  uint2 width() const { return w; }
  uint2 height() const { return h; }
  uint2 bpp() const { return channels; }
  uint1 *pixels() { return data; }
  
  static void createGammaTable(real4 gamma, Vector<uint1> &g);
  
  /**
   * Element access operator
   * \param	index of vector component
   * \return	A reference to the specified pixel data
   */
  uint1& operator[](uint4 index) { return data[index]; }
  /**
   * Element access operator
   * \param	index of vector component
   * \return	A reference to the specified pixel data
   */
  const uint1& operator[](uint4 index) const { return data[index]; }

private:
  uint2 w, h, channels;
  uint4 size;
  uint1 *data;

   /**
   * Free image data
   */
  void free();
  
  /**
   * Read in run-length encoded data from the file stream
   */
  bool readRLE(std::ifstream *file);
  /**
   * Flip the image vertically
   */
  void flip();
};

#endif
