// $Id: Lexer.C,v 1.1.1.1 2004/10/11 09:09:20 orts_mburo Exp $ 

// (c) Jonathan Yip & Michael Buro, licensed under the GPL

/* part of code inspired by psl
    PLIB - A Suite of Portable Game Libraries
    Copyright (C) 1998,2002  Steve Baker
*/

#include "Lexer.H"
//#include "OSL_RecStream.H"

using namespace std;

namespace OSL {

  /** insert \ before non-printable characters */

  static string transform_escapes(const string &s)
  {
    string t;
    uint4 l = s.size();
    FORU (i, l) {
      sint4 c = s[i];
      switch (c) {
      case '\a': t += "\\a";  break;
      case '\b': t += "\\b";  break;	
      case '\f': t += "\\f";  break;	
      case '\n': t += "\\n";  break;
      case '\r': t += "\\r";  break;
      case '\t': t += "\\t";  break;
      case '"':  t += "\\\""; break;		
      case '\\': t += "\\";   break;	
      default:   t += c;      break;
      }
    }
    return t;
  }

  
  static sint4 read_digit_sequence(RecStream &rs, string &v) {
    FOREVER {
      sint4 c = rs.peek();
      if (!isdigit(c)) return c;
      v += c;
      rs.get_char();
    }
  }

  
  void Token::write(ostream &os) const
  {
    os << "[";
    
    switch (type) {
      
    case TK_ERROR:     os << "error: " << value; break;
    case TK_EOF:       os << "eof"; break;
    case TK_NIL:       os << "nil: " << value; break;
    case TK_ID:        os << "id: " << value; break;
    case TK_IF:        os << "if"; break;
    case TK_ELSE:      os << "else"; break;
    case TK_WHILE:     os << "while"; break;
    case TK_FOR:       os << "for"; break;
    case TK_BREAK:     os << "break"; break;
    case TK_CONT:      os << "continue"; break;
    case TK_SWITCH:    os << "switch"; break;
    case TK_CASE:      os << "case"; break;
    case TK_DEFAULT:   os << "default"; break;
    case TK_CHARC:     os << "char: '" << transform_escapes(value) << "'"; break;
    case TK_STRINGC:   os << "str: \"" << transform_escapes(value) << '"'; break;
    case TK_INTC:      os << "int: " << value; break;
    case TK_UINTC:     os << "uint: " << value; break;
    case TK_LONGC:     os << "long: " << value; break;
    case TK_ULONGC:    os << "ulong: " << value; break;
    case TK_FLOATC:    os << "float: " << value; break;
    case TK_DOUBLEC:   os << "double: " << value; break;
    case TK_VOID:      os << "void"; break;
    case TK_BOOL:      os << "bool"; break;
    case TK_BYTE:      os << "byte"; break;
    case TK_UBYTE:     os << "ubyte"; break;
    case TK_SHORT:     os << "short"; break;
    case TK_USHORT:    os << "ushort"; break;
    case TK_INT:       os << "int"; break;
    case TK_UINT:      os << "uint"; break;
    case TK_LONG:      os << "long"; break;
    case TK_ULONG:     os << "ulong"; break;
    case TK_FLOAT:     os << "float"; break;
    case TK_DOUBLE:    os << "double"; break;
    case TK_STRING:    os << "string"; break;
    case TK_CLASS:     os << "class"; break;
    case TK_ASSIGN:    os << "="; break; 
    case TK_AND:       os << "&"; break;
    case TK_ANDAND:    os << "&&"; break;
    case TK_OR:        os << "|"; break;
    case TK_OROR:      os << "||"; break;
    case TK_XOR:       os << "^"; break;
    case TK_EQ:        os << "=="; break;
    case TK_NEQ:       os << "!="; break;
    case TK_LT:        os << "<"; break;
    case TK_LEQ:       os << "<="; break;
    case TK_GT:        os << ">"; break;
    case TK_GEQ:       os << ">="; break; 
    case TK_ADD:       os << "+"; break;
    case TK_SUB:       os << "-"; break;
    case TK_MUL:       os << "*"; break;
    case TK_DIV:       os << "/"; break;
    case TK_MOD:       os << "%"; break;
    case TK_COMPLEMENT: os << "~"; break;
    case TK_INVERSE:   os << "!"; break;
    case TK_LSHIFT:    os << "<<"; break;
    case TK_RSHIFT:    os << ">>"; break;
    case TK_COLON:     os << ":"; break;
    case TK_SCOPE:     os << "::"; break;
    case TK_DOT:       os << ";"; break;
    case TK_LPAR:      os << "("; break;
    case TK_RPAR:      os << ")"; break;
    case TK_LBRACE:    os << "{"; break;
    case TK_RBRACE:    os << "}"; break;
    case TK_LBRACKET:  os << "["; break;
    case TK_RBRACKET:  os << "]"; break;
    case TK_COMMA:     os << ","; break;
    case TK_SEMICOLON: os << ";"; break;
      //    case TK_MAIN:      os << "main"; break;
    case TK_NEW:       os << "new"; break;
      //    case TK_DELETE:    os << "delete"; break;
    case TK_RETURN:    os << "return"; break; 
    case TK_EXTERN:    os << "extern"; break;
    case TK_STATIC:    os << "static"; break;
    case TK_PUBLIC:    os << "public"; break;
    case TK_PROTECTED: os << "protected"; break;
    case TK_PRIVATE:   os << "private"; break;
    default: ERR2("unknown token type", type); break;
    }  
    os << "]";
  }


  Lexer::Lexer(RecStream &rstream_) : rstream(rstream_) { reset(); }


  void Lexer::throw_error(const string &msg)
  {
    ostringstream os;

    os << cur_file() << ":" << cur_line_num() << " ";
    error = true;

    throw SyntaxErrorException(os.str() + msg);
  }


  sint4 Lexer::get_char()
  {
    sint4 c ;

    // skip characters so long as the 'skipping' flag is set

    do {
      
      /*
	If we got a newline then we have to test to see whether
	a '#' preprocessor directive is on this line.
      */
    
      c = rstream.get_char();
    
      if (c == '\n' ) {
      
	sint4 d = rstream.get_char();
      
	if (d == '#') {
	  skip_override = true;
	  c = do_preprocessor_command() ;
	  skip_override = false;
	  if (c < 0 || !skipping()) return c ;

	} else put_char(d);
      }

    } while (skipping() && c >= 0) ;

    // all done - return the character

    return c ;
  }


  sint4 Lexer::do_preprocessor_command()
  {
    Token tok = get_token();
    string v = tok.get_value();

    cout << "# command = " << v << " " << endl; tok.write(); 
    
    if      (v == "include") do_include(); 
    else if (v == "undef"  ) do_undef();   
    else if (v == "define" ) do_define();  
    else if (v == "ifdef"  ) do_ifdef(0);  
    else if (v == "ifndef" ) do_ifdef(1);  
    else if (v == "endif"  ) do_endif();   
    else if (v == "else"   ) do_else();    
    else throw_error("unrecognised preprocessor directive " + v);

    return '\n';
  }
  

  void Lexer::do_define()
  {
    if (!skipping()) {
      
      Token tok = get_token();
      
      if (tok.get_type() != TK_ID)
	throw_error("#define token is not a legal identifier: " + tok.get_value());

      defs.insert(tok.get_value());
    }
  }

  
  void Lexer::do_undef()
  {
    if (!skipping()) {
    
      Token tok = get_token();
    
      if (tok.get_type() != TK_ID)
	throw_error("#undef token is not a legal identifier: " + tok.get_value());

      defs.erase(tok.get_value());
    }
    
    skip_to_eol();    
  }


  void Lexer::do_ifdef(bool negate)
  {
    if (next_skipping_level >= MAX_SKIP_LEVEL)
      throw_error("too many levels of #ifdef/#ifndef");

    Token tok = get_token();

    if (tok.get_type() != TK_ID)
      throw_error("#ifdef token is not a legal identifier: " + tok.get_value());
    
    if (defs.find(tok.get_value()) == defs.end() ^ negate)
      skipping_flags |=  (1 << next_skipping_level++);
    else
      skipping_flags &= ~(1 << next_skipping_level++);

    skip_to_eol();
  }


  void Lexer::do_else()
  {
    if (next_skipping_level <= 1)
      throw_error("#else without prior matching #ifdef/#ifndef");

    skipping_flags ^= (1 << (next_skipping_level-1));

    skip_to_eol();
  }


  void Lexer::do_endif()
  {
    if (next_skipping_level <= 1)
      throw_error("#endif without prior matching #ifdef/#ifndef");

    skipping_flags &= ~(1 << (--next_skipping_level));

    skip_to_eol();
  }

  
  void Lexer::do_include()
  {
    if (!skipping()) {

      Token tok = get_token();
      if (tok.get_type() != TK_STRINGC)
	throw_error("string expected after #include");

      FILE *fp = fopen(tok.get_value().c_str(), "r");
      if (!fp) throw_error("can't find #include file: " + tok.get_value());

      rstream.push(fp, tok.get_value());
    }

    skip_to_eol();
  }

  
  Token Lexer::peek_token()
  {
    if (peeked) return current_token;
    current_token = get_token();
    peeked = true;
    return current_token;
  }

  
  void Lexer::reset() {
    next_skipping_level = 1;
    skipping_flags = 0;
    skip_override = false;
    error = false;
    peeked = false;
    defs.clear();
  }

  
  Token Lexer::get_token()
  {
    if (error) throw_error("read after error");

    if (peeked) { peeked = false; return current_token; }
      
    sint4 c;
    string value;
    TokenType type = TK_ERROR;

    // skip whitespace 
    FOREVER {
      c = get_char();

      // skip comments
      if (c == '/' && peek() == '/') {
	FOREVER {
	  c = get_char();
	  if (c < 0) return Token(TK_EOF);
	  if (c == '\n') break;
	}
      }
      else if (c == '/' && peek() == '*') {
	FOREVER {
	  c = get_char();
	  if (c < 0) 
	    return Token(TK_ERROR, "Unexpected EOF inside long comment");
	  if (c == '*' && peek() == '/') {
	    get_char();
	    c = get_char();
	    break;
	  }
	}
	cout << "HELLO " << peek() << endl;
      }

      if (!isspace(c)) break;
    }

    if (c < 0) return Token(TK_EOF); // fixme: deal with read error


    if (isalpha(c) || c == '_') {

      // keyword or identifier

      while (isalnum(c) || c == '_') {
	value += c;
	c = get_char();
      }

      put_char(c);

      if      (value == "void")      return Token(TK_VOID, value);
      else if (value == "bool")      return Token(TK_BOOL, value);
      else if (value == "byte")      return Token(TK_BYTE, value);
      else if (value == "ubyte")     return Token(TK_UBYTE, value);
      else if (value == "short")     return Token(TK_SHORT, value);
      else if (value == "ushort")    return Token(TK_USHORT, value);
      else if (value == "int")       return Token(TK_INT, value);
      else if (value == "uint")      return Token(TK_UINT, value);            
      else if (value == "long")      return Token(TK_LONG, value);
      else if (value == "ulong")     return Token(TK_ULONG, value);
      else if (value == "float")     return Token(TK_FLOAT, value);
      else if (value == "double")    return Token(TK_DOUBLE, value);
      else if (value == "class")     return Token(TK_CLASS);      
      else if (value == "if")        return Token(TK_IF);
      else if (value == "else")      return Token(TK_ELSE, "else"); // also #directive
      else if (value == "while")     return Token(TK_WHILE);
      else if (value == "for")       return Token(TK_FOR);
      else if (value == "break")     return Token(TK_BREAK);
      else if (value == "continue")  return Token(TK_CONT);
      else if (value == "return")    return Token(TK_RETURN);
      //      else if (value == "main")      return Token(TK_MAIN);
      else if (value == "new")       return Token(TK_NEW);
      //      else if (value == "delete")    return Token(TK_DELETE);
      else if (value == "switch")    return Token(TK_SWITCH);
      else if (value == "case")      return Token(TK_CASE);      
      else if (value == "default")   return Token(TK_DEFAULT);
      else if (value == "extern")    return Token(TK_EXTERN);
      else if (value == "static")    return Token(TK_STATIC);
      else if (value == "public")    return Token(TK_PUBLIC);
      else if (value == "protected") return Token(TK_PROTECTED);
      else if (value == "private")   return Token(TK_PRIVATE);
      else 		             return Token(TK_ID, value);
    }
    
    if (isdigit(c)) {

      // number constant
      
      type = TK_INTC;
      value = c;
      c = read_digit_sequence(rstream, value);

      if (c == '.') {

	type = TK_DOUBLEC;
	get_char();
	value += c;

	c = read_digit_sequence(rstream, value);
      }

      if (c == 'e' || c == 'E') {

	type = TK_DOUBLEC;
	get_char();
	value += c;

	c = peek();

	if (c == '+' || c == '-') {
	  get_char(); value += c;
	}
	
	c = get_char();
	if (!isdigit(c)) throw_error("corrupt exponent");
	value += c;
	c = read_digit_sequence(rstream, value);
      }

      if (c == 'f' || c == 'F') {
	type = TK_FLOATC;
	get_char();
	value += c;
	return Token(type, value);
      }

      if (type == TK_INTC) {

	// unsigned and long modifiers (not copied into value)
	
	char mod[3] = { 0, 0, 0 };

	if (isalpha(c)) {
	  mod[0] = toupper(c); get_char(); c = peek();
	  if (isalpha(c)) {
	    mod[1] = toupper(c); get_char(); c = peek();
	  }
	}

	if (!strcmp(mod, "UL") || !strcmp(mod, "LU")) return Token(TK_ULONGC, value);
	if (!strcmp(mod, "L")) return Token(TK_LONGC, value);
	if (!strcmp(mod, "U")) return Token(TK_UINTC, value);
	if (!strcmp(mod, "")) return Token(TK_INTC, value);
	throw_error("corrupt integer constant");
      }

      return Token(type, value);
    }
    
    if (ispunct(c)) {

      switch (c) {

      case '\'':

	// fixme: escaped characters, newline?
	
	c = get_char();
	if (c < 0) throw_error("unexpected end in character constant");
	if (c == '\n') throw_error("\\n must be escaped in character constant");
	if (c == '\'') throw_error("empty character constant");	
	if (c == '\\') {
	  c = get_char();
	  if (c < 0) throw_error("unexpected end in character constant");	  
	  if (!(c == '\\' || c == '\'')) {
	    if      (c == 'a') c = '\a';
	    else if (c == 'b') c = '\b';
	    else if (c == 'f') c = '\f';	      	      
	    else if (c == 'n') c = '\n';
	    else if (c == 't') c = '\t';
	    else if (c == 'r') c = '\r';
	    else throw_error(string("unexpected escape sequence \\") + (char)c);
	  }
	}
	value += c;
	c = get_char();
	if (c != '\'') throw_error("' missing in character constant");
	return Token(TK_CHARC, value);

      case '"':

	// fixme: strings with newline, escape characters
	
	FOREVER {
	  c = get_char();
	  if (c < 0) throw_error("unexpected end in string constant");
	  if (c == '\n') throw_error("\\n must be escaped in string constant");
	  if (c == '"') return Token(TK_STRINGC, value);

	  if (c == '\\') { // escaped character
	    c = get_char();
	    if (c < 0) throw_error("unexpected end in string constant");
	    if (!(c == '\n' || c == '\\' || c == '"')) {
	      if      (c == 'a') c = '\a';
	      else if (c == 'b') c = '\b';
	      else if (c == 'f') c = '\f';	      	      
	      else if (c == 'n') c = '\n';
	      else if (c == 't') c = '\t';
	      else if (c == 'r') c = '\r';
	      else throw_error(string("unexpected escape sequence \\") + char(c));
	    }
	  }
	  value += c;
	}
	
      case '=':
	if (peek() == '=') { get_char(); return Token(TK_EQ); }
	return Token(TK_ASSIGN);

      case '<':
	if (peek() == '=') { get_char(); return Token(TK_LEQ); }
	if (peek() == '<') { get_char(); return Token(TK_LSHIFT); }
	return Token(TK_LT);

      case '>':
	if (peek() == '=') { get_char(); return Token(TK_GEQ); }	
	if (peek() == '>') { get_char(); return Token(TK_RSHIFT); }
	return Token(TK_GT);

      case '(': return Token(TK_LPAR);
      case ')': return Token(TK_RPAR);
      case '{': return Token(TK_LBRACE);
      case '}': return Token(TK_RBRACE);
      case '[': return Token(TK_LBRACKET);
      case ']': return Token(TK_RBRACKET);	
      case '.': return Token(TK_DOT);
      case ',': return Token(TK_COMMA);
      case ':': 
	if (peek() == ':') { get_char(); return Token(TK_SCOPE); }
	return Token(TK_COLON);	
      case ';': return Token(TK_SEMICOLON);
      case '+': return Token(TK_ADD);
      case '-': return Token(TK_SUB);
      case '*': return Token(TK_MUL);
      case '/': return Token(TK_DIV);
      case '%': return Token(TK_MOD);
      case '^': return Token(TK_XOR);
      case '~': return Token(TK_COMPLEMENT);
      case '!': return Token(TK_INVERSE);

      case '&': 
	if (peek() == '&') { get_char(); return Token(TK_ANDAND); }
	return Token(TK_AND);

      case '|': 
	if (peek() == '|') { get_char(); return Token(TK_OROR); }
	return Token(TK_OR);
      
      default:
	throw_error(string("illegal character ") + char(c));
      }
    }

    ERR("here?");
    return Token(TK_ERROR, "here?");
  }

}

