//-----------------------------------------------------------------------------
/** @file astar.h
    A* search algorithm.

    $Id: astar.h,v 1.2 2005/06/12 04:09:19 orts_mburo Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/pathfind/src/astar.h,v $
*/
//-----------------------------------------------------------------------------

#ifndef PATHFIND_ASTAR_H
#define PATHFIND_ASTAR_H

#include <algorithm>
#include <list>
#include <memory>
#include <queue>
#include <set>
#include "marker.h"
#include "search.h"
#include "util.h"

#ifndef PATHFIND_ASTAR_HEAVY_DEBUG
#define PATHFIND_ASTAR_HEAVY_DEBUG 0
#endif

//-----------------------------------------------------------------------------

namespace PathFind
{
    
    using namespace std;    

    /** Node used in implementation of AStar. */
    class AStarNode
    {
    public:
        int m_nodeId;
        
        int m_parent;
        
        int m_g;
        
        int m_h;
        
        int m_f;
        
        AStarNode()
        {
        }
        
        AStarNode(int nodeId, int parent, int g, int h, int w=1)
            : m_nodeId(nodeId),
              m_parent(parent),
              m_g(g),
              m_h(h),
              m_f(g + w*h)
        {
        }
        
        void print(ostream& ostrm) const;
    };
    
    /** Function object for comparing AStarNodes by f-value. */
    class AStarCompareF
    {
    public:
        bool operator()(const AStarNode& node1, const AStarNode& node2)
        {
            int f1 = node1.m_f;
            int f2 = node2.m_f;
            return (f1 < f2);
        }
    };
    
    /** Function object for comparing AStarNodes by f-value and tie-breaking
        by g-value.
    */
    class AStarCompareFG
    {
    public:
        bool operator()(const AStarNode& node1, const AStarNode& node2)
        {
            int f1 = node1.m_f;
            int f2 = node2.m_f;
            if (f1 != f2)
                return (f1 < f2);
            int g1 = node1.m_g;
            int g2 = node2.m_g;
            return (g1 > g2);
        }
    };
    
    /** Closed list implementation using a list. */
    class AStarClosed
    {
    public:
        void add(const AStarNode& node);
        
        vector<int> constructPath(int start, int target);

        void init(int maxNodeId);
        
        void print(ostream& ostrm) const;
        
        void remove(int nodeId);
        
        const AStarNode* search(int nodeId) const;
        
    private:
        list<AStarNode> m_list;            
    };
    
    /** Closed list implementation using lookup table. */
    template<class MARKER>
    class AStarClosedLookup
    {
    public:
        AStarClosedLookup();
        
        void add(const AStarNode& node);
        
        vector<int> constructPath(int start, int target);

        void init(int maxNodeId);
        
        void print(ostream& ostrm) const;
        
        void remove(int nodeId);
        
        const AStarNode* search(int nodeId) const;
        
    private:
        int m_maxNodeId;
        
        /** NodeIds in closed list. */
        MARKER m_marker;
        
        vector<AStarNode> m_nodes;            
    };
    
    /** Open queue for AStar. */
    template<class MARKER, class COMPARE>
    class AStarOpen
    {
    public:
        AStarOpen();
        
        void init(int numberNodes);
        
        void insert(const AStarNode& node);
        
        bool isEmpty() const
        {
            return (m_nodes.size() == 0);
        }
        
        AStarNode pop();
        
        void print(ostream& ostrm) const;
        
        bool  remove(int nodeId);
        
        const AStarNode* search(int nodeId) const;
        
    private:
        typedef multiset<AStarNode, COMPARE> NodeSet;
        
        typedef typename NodeSet::const_iterator NodeSetConstIterator;
        
        typedef typename NodeSet::iterator NodeSetIterator;
        
        int m_numberNodes;
        
        /** NodeIds in open queue. */
        MARKER m_marker;
        
        /** Iterator in m_nodes by nodeId. */
        vector<NodeSetIterator> m_lookupTable;
        
        NodeSet m_nodes;
    };

    /** Open queue for AStar using a binary tree of buckets
        with statically allocated bucket nodes.
        Presently bucketing by f-Value of AStarNode is hardcoded.
    */
    template<class MARKER, class COMPARE>
    class AStarOpenBuckets
    {
    public:
        AStarOpenBuckets();
        
        void init(int numberNodes);
        
        void insert(const AStarNode& node);
        
        bool isEmpty() const
        {
            return (m_head.size() == 0);
        }
        
        AStarNode pop();
        
        void print(ostream& ostrm) const;
        
        bool  remove(int nodeId);
        
        const AStarNode* search(int nodeId) const;
        
    private:
        class BucketNode
        {
        public:
            int m_next;
            
            int m_previous;

            AStarNode m_node;
        };

        class Compare
        {
        public:
            bool operator()(BucketNode* const node1, BucketNode* const node2)
            {
                return m_compare(node1->m_node, node2->m_node);
            }

        private:
            COMPARE m_compare;
        };

        /** Mapping to head node of bucket. */
        typedef multiset<BucketNode*, Compare> Head;
        
        typedef typename Head::iterator HeadIterator;

        typedef typename Head::const_iterator HeadConstIterator;

        int m_numberNodes;
        
        /** NodeIds in open queue. */
        MARKER m_marker;
        
        /** Bucket head notes. */
        Head m_head;

        vector<BucketNode> m_nodes;

        /** Iterator in m_head by nodeId. */
        vector<HeadIterator> m_lookupTable;

        /** Check class invariants in heavy debug mode. */
        void checkInvariants() const;

        /** Check for valid bucket node in debug mode. */
        void checkValidBucketNode(const BucketNode& bucketNode) const;

        bool isValidId(int nodeId) const
        {
            return (nodeId >= 0 && nodeId < m_numberNodes);
        }

        bool isValidIdOrNone(int nodeId) const
        {
            return (nodeId == NO_NODE || isValidId(nodeId));
        }

        void replaceHead(HeadIterator head, BucketNode* bucketNode)
        {
            // std::set iterators don't allow to change the value
            // but we know that the new value would be inserted at the same
            // position and want the iterator for this element remain the same.
            assert(m_head.key_comp()(*head, bucketNode) == 0);
            const_cast<BucketNode*&>(*head) = bucketNode;
        }
    };

    /** A* search engine.

        Pseudo algorithm for basic A* algorithm:
        @verbatim
        priorityqueue Open
        list Closed

        AStarSearch
            s.g = 0 // s is the start node
            s.h = GoalDistEstimate(s)
            s.f = s.g + s.h
            s.parent = null
            push s on Open
            while Open is not empty
                pop node n from Open // n has the lowest f
                if n is a goal node
                    construct path
                    return success
                for each successor n' of n
                    newg = n.g + cost(n,n')
                    if n' is in Open or Closed
                       if n'.g <= newg skip
                       remove n' from Open or Closed
                    n'.parent = n
                    n'.g = newg
                    n'.h = GoalDistEstimate(n')
                    n'.f = n'.g + n'.h
                    push n' on Open
                push n onto Closed
            return failure // if no path found
        @endverbatim

        This class is implemented as a template class taking
        the following template parameters:
        - MARKER Marker class to use (see @ref marker)
        - CLOSED Closed list class to use

        %Statistics (see Search::createStatistics and Search::getStatistics):
        - "cpu_time"
        - "path_cost"
        - "path_length"
        - "nodes_visited"
        - "nodes_visited_last_iter"
        - "branching_factor"
    */
    template<class MARKER = MarkerFastClear,
             class OPEN = AStarOpen<MARKER,AStarCompareFG>,
             class CLOSED = AStarClosedLookup<MARKER> >
    class AStar
        : public Search
    {
    public:
        /** Constructor.
            @param nodeCountForAllTargets
            Count all target nodes on open list. Only useful for getting the
            node statistics. Does not return all paths.
        */
        AStar(bool nodeCountForAllTargets = false, int weight=1);
        
        StatisticsCollection createStatistics();
        
        bool findPath(const Environment& env, int start, int target);

        const vector<int>& getPath() const
        {
            return m_path;
        }

        const StatisticsCollection& getStatistics() const;

        /** Get a vector with 'x' char labels for each visited node. */
        const vector<char>& getVisitedNodes() const
        {
            return m_visitedNodes;
        }

        void setNodesLimit(long long int nodesLimit)
        {
            m_nodesLimit = nodesLimit;
        }

        int getPathCost() const
        {
            return m_pathCost;
        }

    private:
        bool m_nodeCountForAllTargets;

        int m_iteration;

        int m_pathCost;

        int m_target;

        int m_weight;

        const Environment* m_env;

        long long int m_nodesVisited;

        long long int m_nodesVisitedCurrentIter;

        CLOSED m_closed;

        vector<char> m_visitedNodes;

        vector<int> m_path;

        OPEN m_open;

        StatisticsCollection m_statistics;

        Statistics& m_branchingFactor;


        void addStatistics(const string& name, double value)
        {
            m_statistics.get(name).add(value);
        }

        /** Find a node in open or closed lists. */
        const AStarNode* findNode(int nodeId);

        void findPathAStar(int start);

        /** Construct path and set statistics after target node was found. */
        void finishSearch(int start, const AStarNode& node);

        AStarNode getBestNodeFromOpen();
    };

} // namespace PathFind

//-----------------------------------------------------------------------------

namespace PathFind
{
    
  using namespace std;

  template<class MARKER, class OPEN, class CLOSED>
  AStar<MARKER, OPEN, CLOSED>::AStar(bool nodeCountForAllTargets, int weight)
    : m_nodeCountForAllTargets(nodeCountForAllTargets),
      m_weight(weight),
      m_statistics(createStatistics()),
      m_branchingFactor(m_statistics.get("branching_factor"))
  {
  }

  template<class MARKER, class OPEN, class CLOSED>
  StatisticsCollection AStar<MARKER, OPEN, CLOSED>::createStatistics()
  {
    StatisticsCollection collection;
    collection.create("cpu_time");
    collection.create("path_cost");
    collection.create("path_length");
    collection.create("branching_factor");
    collection.create("iterations");
    collection.create("nodes_visited");
    collection.create("nodes_visited_last_iter");
    return collection;
  }
    
  template<class MARKER, class OPEN, class CLOSED>
  const AStarNode* AStar<MARKER, OPEN, CLOSED>::findNode(int nodeId)
  {
    const AStarNode* result = 0;
    result = m_open.search(nodeId);
    if (result != 0)
      return result;
    result = m_closed.search(nodeId);
    return result;
  }

  template<class MARKER, class OPEN, class CLOSED>
  bool AStar<MARKER, OPEN, CLOSED>::findPath(const Environment& env,
					     int start, int target)
  {
    assert(env.isValidNodeId(start));
    assert(env.isValidNodeId(target));
    Clock clock;
    m_statistics.clear();
    m_nodesVisited = 0;
    m_env = &env;
    m_target = target;
    m_path.clear();
    m_visitedNodes.resize(env.getNumberNodes());
    fill(m_visitedNodes.begin(), m_visitedNodes.end(), ' ');
    findPathAStar(start);
    addStatistics("iterations", m_iteration);
    addStatistics("cpu_time", clock.get());
    addStatistics("nodes_visited", m_nodesVisited);
    addStatistics("nodes_visited_last_iter", m_nodesVisitedCurrentIter);
    addStatistics("path_length", m_path.size());
    return true;
  }

  template<class MARKER, class OPEN, class CLOSED>
  void AStar<MARKER, OPEN, CLOSED>::findPathAStar(int start)
  {
    int numberNodes = m_env->getNumberNodes();
    m_closed.init(numberNodes);
    m_open.init(numberNodes);
    int heuristic = m_env->getHeuristic(start, m_target);
    m_pathCost = -1;
    AStarNode startNode(start, NO_NODE, 0, heuristic, m_weight);
    m_open.insert(startNode);
    vector<Environment::Successor> successors;
    int lastFValue = -1;
    m_iteration = 0;
    m_nodesVisitedCurrentIter = 0;
    while (! m_open.isEmpty())
      {
	//m_open.print(cout);
	AStarNode node = getBestNodeFromOpen();
	//cout << '[';  node.print(cout); cout << ']' << endl;
	if (node.m_f != lastFValue)
	  {
	    lastFValue = node.m_f;
	    ++m_iteration;
	    m_nodesVisited += m_nodesVisitedCurrentIter;
	    m_nodesVisitedCurrentIter = 0;
	  }
	++m_nodesVisitedCurrentIter;

	if (node.m_nodeId == m_target)
	  {
	    finishSearch(start, node);
	    m_nodesVisited += m_nodesVisitedCurrentIter;
	    return;
	  }
	m_env->getSuccessors(node.m_nodeId, node.m_parent, successors);
	m_branchingFactor.add(successors.size());
	for (vector<Environment::Successor>::const_iterator i
	       = successors.begin(); i != successors.end(); ++i)
	  {
	    int newg = node.m_g + i->m_cost;
	    int target = i->m_target;
	    const AStarNode* targetAStarNode = findNode(target);
	    if (targetAStarNode != 0)
	      {
		if (newg >= targetAStarNode->m_g)
		  continue;
		if (! m_open.remove(target))
		  m_closed.remove(target);
	      }
	    int newHeuristic = m_env->getHeuristic(target, m_target);
	    AStarNode newAStarNode(target, node.m_nodeId, newg,
				   newHeuristic, m_weight);
	    m_open.insert(newAStarNode);
	  }
	m_closed.add(node);
	//m_closed.print(cout);
      }
  }

  template<class MARKER, class OPEN, class CLOSED>
  void AStar<MARKER, OPEN, CLOSED>::finishSearch(int start,
						 const AStarNode& targetNode)
  {
    if (m_nodeCountForAllTargets)
      {
	while (! m_open.isEmpty())
	  {
	    AStarNode node = getBestNodeFromOpen();
	    ++m_nodesVisitedCurrentIter;
	    if (node.m_f != targetNode.m_f)
	      break;
	  }
      }
    m_closed.add(targetNode);
    m_path = m_closed.constructPath(start, m_target);
    m_pathCost = targetNode.m_f;
    addStatistics("path_cost", m_pathCost);
  }
    
  template<class MARKER, class OPEN, class CLOSED>
  AStarNode AStar<MARKER, OPEN, CLOSED>::getBestNodeFromOpen()
  {
    assert(! m_open.isEmpty());
    AStarNode result = m_open.pop();
    int nodeId = result.m_nodeId;
    assert(nodeId >= 0);
    m_visitedNodes[nodeId] = '+';
    return result;
  }
    
  template<class MARKER, class OPEN, class CLOSED>
  const StatisticsCollection& AStar<MARKER, OPEN, CLOSED>::getStatistics()
    const
  {
    return m_statistics;
  }
    
  template<class MARKER>
  AStarClosedLookup<MARKER>::AStarClosedLookup()
  {
    m_maxNodeId = 0;
    init(0);
  }
    
  template<class MARKER>
  void AStarClosedLookup<MARKER>::add(const AStarNode& node)
  {
    int nodeId = node.m_nodeId;
    assert(nodeId >= 0);
    assert(nodeId < m_maxNodeId);
    assert(m_marker.get(nodeId) == false);
    m_marker.set(nodeId);
    m_nodes[nodeId] = node;
  }
    
  template<class MARKER>
  vector<int> AStarClosedLookup<MARKER>::constructPath(int start, int target)
  {
    vector<int> result;
    int nodeId = target;
    while (true)
      {
	result.push_back(nodeId);
	if (nodeId == start)
	  break;
	const AStarNode* node = search(nodeId);
	assert(node != 0);
	nodeId = node->m_parent;
      }
    assert(result.size() > 0);
    assert(*result.begin() == target);
    assert(*(result.end() - 1) == start);
    return result;
  }

  template<class MARKER>
  void AStarClosedLookup<MARKER>::init(int maxNodeId)
  {
    assert(maxNodeId >= 0);
    m_nodes.resize(maxNodeId);
    m_marker.init(maxNodeId);
    m_maxNodeId = maxNodeId;
  }
    
  template<class MARKER>
  void AStarClosedLookup<MARKER>::print(ostream& ostrm) const
  {
    ostrm << "Closed {\n";
    for (int nodeId = 0; nodeId < m_maxNodeId; ++nodeId)
      if (m_marker.get(nodeId))
	{
	  m_nodes[nodeId].print(ostrm);
	  ostrm << '\n';
	}   
    ostrm << "}\n";
  }

  template<class MARKER>
  void AStarClosedLookup<MARKER>::remove(int nodeId)
  {
    assert(nodeId >= 0);
    assert(nodeId < m_maxNodeId);
    assert(m_marker.get(nodeId) == true);
    m_marker.unset(nodeId);
  }
    
  template<class MARKER>
  const AStarNode* AStarClosedLookup<MARKER>::search(int nodeId) const
  {
    assert(nodeId >= 0);
    assert(nodeId < m_maxNodeId);
    if (! m_marker.get(nodeId))
      return 0;
    const AStarNode* result = &m_nodes[nodeId];
    return result;
  }
    
  template<class MARKER, class COMPARE>
  AStarOpen<MARKER,COMPARE>::AStarOpen()
  {
    init(0);
  }
    
  template<class MARKER, class COMPARE>
  void AStarOpen<MARKER,COMPARE>::init(int numberNodes)
  {
    m_marker.init(numberNodes);
    m_lookupTable.resize(numberNodes);
    m_numberNodes = numberNodes;
    m_nodes.clear();
  }

  template<class MARKER, class COMPARE>
  void AStarOpen<MARKER,COMPARE>::insert(const AStarNode& node)
  {
    int nodeId = node.m_nodeId;
    assert(nodeId < m_numberNodes);
    assert(m_marker.get(nodeId) == false);
    m_marker.set(nodeId);
    NodeSetIterator pos = m_nodes.insert(node);
    m_lookupTable[nodeId] = pos;
  }

  template<class MARKER, class COMPARE>
  AStarNode AStarOpen<MARKER,COMPARE>::pop()
  {
    assert(! isEmpty());
    AStarNode result = *(m_nodes.begin());
    m_nodes.erase(m_nodes.begin());
    int nodeId = result.m_nodeId;
    assert(nodeId < m_numberNodes);
    assert(m_marker.get(nodeId) == true);
    m_marker.unset(nodeId);
    return result;
  }

  template<class MARKER, class COMPARE>
  void AStarOpen<MARKER,COMPARE>::print(ostream& ostrm) const
  {
    ostrm << "Open {\n"; 
    for (NodeSetConstIterator i = m_nodes.begin(); i != m_nodes.end(); ++i)
      {
	i->print(ostrm);
	ostrm << '\n';
      }   
    ostrm << "}\n";
  }

  template<class MARKER, class COMPARE>
  bool AStarOpen<MARKER,COMPARE>::remove(int nodeId)
  {
    assert(nodeId >= 0);
    assert(nodeId < m_numberNodes);
    if (! m_marker.get(nodeId))
      return false;
    NodeSetIterator pos = m_lookupTable[nodeId];
    m_nodes.erase(pos);
    m_marker.unset(nodeId);
    return true;
  }

  template<class MARKER, class COMPARE>
  const AStarNode* AStarOpen<MARKER,COMPARE>::search(int nodeId) const
  {
    assert(nodeId >= 0);
    assert(nodeId < m_numberNodes);
    if (m_marker.get(nodeId))
      return &(*m_lookupTable[nodeId]);
    return 0;
  }

  template<class MARKER, class COMPARE>
  AStarOpenBuckets<MARKER,COMPARE>::AStarOpenBuckets()
  {
    init(0);
  }
    
  template<class MARKER, class COMPARE>
  void AStarOpenBuckets<MARKER,COMPARE>::checkInvariants() const
  {
#if PATHFIND_ASTAR_HEAVY_DEBUG
    for (HeadConstIterator head = m_head.begin(); head != m_head.end();
	 ++head)
      {
	checkValidBucketNode(**head);
	int nodeId = (*head)->m_node.m_nodeId;
	assert((*head) == &m_nodes[nodeId]);
	assert((*head)->m_previous == NO_NODE);
	while (nodeId != NO_NODE)
	  {
	    checkValidBucketNode(m_nodes[nodeId]);
	    nodeId = m_nodes[nodeId].m_next;
	  }
      }
#endif
  }
    
  template<class MARKER, class COMPARE>
  void AStarOpenBuckets<MARKER,COMPARE>
  ::checkValidBucketNode(const BucketNode&
#ifndef NDEBUG			 
			 bucketNode
#endif			 
			 ) const
  {
#ifndef NDEBUG
    int nodeId = bucketNode.m_node.m_nodeId;
    assert(m_marker.get(nodeId));
    assert(&bucketNode == &m_nodes[nodeId]);
    assert(isValidIdOrNone(bucketNode.m_previous));
    assert(isValidIdOrNone(bucketNode.m_next));
#endif
  }

  template<class MARKER, class COMPARE>
  void AStarOpenBuckets<MARKER,COMPARE>::init(int numberNodes)
  {
    m_marker.init(numberNodes);
    m_head.clear();
    m_nodes.resize(numberNodes);
    m_lookupTable.resize(numberNodes);
    m_numberNodes = numberNodes;
  }

  template<class MARKER, class COMPARE>
  void AStarOpenBuckets<MARKER,COMPARE>::insert(const AStarNode& node)
  {
    checkInvariants();
    int nodeId = node.m_nodeId;
    assert(m_marker.get(nodeId) == false);
    m_marker.set(nodeId);
    BucketNode& bucketNode = m_nodes[nodeId];
    bucketNode.m_node = node;
    HeadIterator head = m_head.find(&bucketNode);
    if (head == m_head.end())
      {
	bucketNode.m_next = NO_NODE;
	bucketNode.m_previous = NO_NODE;
	HeadIterator pos = m_head.insert(&bucketNode);
	m_lookupTable[nodeId] = pos;
      }
    else
      {
	m_lookupTable[nodeId] = head;
	int headId = (*head)->m_node.m_nodeId;
	bucketNode.m_next = headId;
	bucketNode.m_previous = NO_NODE;
	BucketNode& oldHead = m_nodes[headId];
	replaceHead(head, &bucketNode);
	oldHead.m_previous = nodeId;
      }
    checkInvariants();
  }

  template<class MARKER, class COMPARE>
  AStarNode AStarOpenBuckets<MARKER,COMPARE>::pop()
  {
    checkInvariants();
    assert(! isEmpty());
    HeadIterator head = m_head.begin();
    checkValidBucketNode(**head);
    AStarNode result = (*head)->m_node;
    int next = (*head)->m_next;
    if (next == NO_NODE)
      m_head.erase(head);
    else
      {
	BucketNode& nextNode = m_nodes[next];
	nextNode.m_previous = NO_NODE;
	replaceHead(head, &nextNode);
      }
    int nodeId = result.m_nodeId;
    assert(m_marker.get(nodeId) == true);
    m_marker.unset(nodeId);
    checkInvariants();
    return result;
  }

  template<class MARKER, class COMPARE>
  void AStarOpenBuckets<MARKER,COMPARE>::print(ostream& ostrm) const
  {
    ostrm << "Open {\n"; 
    for (HeadConstIterator head = m_head.begin(); head != m_head.end();
	 ++head)
      {
	checkValidBucketNode(**head);
	int nodeId = (*head)->m_node.m_nodeId;
	while (nodeId != NO_NODE)
	  {
	    checkValidBucketNode(m_nodes[nodeId]);
	    m_nodes[nodeId].m_node.print(ostrm);
	    ostrm << '\n';
	    nodeId = m_nodes[nodeId].m_next;
	  }
	ostrm << '\n';
      }   
    ostrm << "}\n";
  }

  template<class MARKER, class COMPARE>
  bool AStarOpenBuckets<MARKER,COMPARE>::remove(int nodeId)
  {        
    checkInvariants();
    assert(isValidId(nodeId));
    if (! m_marker.get(nodeId))
      return false;
    BucketNode& bucketNode = m_nodes[nodeId];
    int next = bucketNode.m_next;
    int previous = bucketNode.m_previous;
    if (previous == NO_NODE)
      {
	HeadIterator head = m_lookupTable[nodeId];
	assert(head == m_head.find(&bucketNode));
	if (next == NO_NODE)
	  {
	    m_head.erase(head);
	  }
	else
	  {
	    replaceHead(head, &m_nodes[next]);
	    m_nodes[next].m_previous = NO_NODE;
	  }
      }
    else
      {
	m_nodes[previous].m_next = next;
	if (next != NO_NODE)
	  m_nodes[next].m_previous = previous;

      }
    m_marker.unset(nodeId);
    checkInvariants();
    return true;
  }

  template<class MARKER, class COMPARE>
  const AStarNode* AStarOpenBuckets<MARKER,COMPARE>::search(int nodeId) const
  {
    checkInvariants();
    assert(isValidId(nodeId));
    if (m_marker.get(nodeId))
      return &m_nodes[nodeId].m_node;
    return 0;
  }

} // namespace PathFind

//-----------------------------------------------------------------------------

#endif
