//-----------------------------------------------------------------------------
/** @file fringesearch.h
    IDA* like search keeping a list of the fringe nodes.

    $Id: fringesearch.h,v 1.1.1.1 2004/10/11 09:09:20 orts_mburo Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/pathfind/src/fringesearch.h,v $
*/
//-----------------------------------------------------------------------------

#ifndef PATHFIND_FRINGESEARCH_H
#define PATHFIND_FRINGESEARCH_H

#include <vector>
#include "marker.h"
#include "search.h"
#include "util.h"

//-----------------------------------------------------------------------------

namespace PathFind
{
    using namespace std;

    /** Statically allocated linked list used in FringeSearch. */
    template<class MARKER = MarkerFastClear>
    class Fringe
    {
    public:
        Fringe();
        
        int getCurrentNode() const
        {
            return m_currentNode;
        }
        
        void init(int numberNodes);
        
        void insertNode(int nodeId);
        
        void nextNode();
        
        void print(ostream& o) const;
        
        void removeCurrentNode()
        {
            removeNode(m_currentNode);
        }
        
        void startIteration();
        
    private:
        /** Node used in Fringe. */
        class Node
        {
        public:
            int m_next;
            
            int m_previous;
        };
        
        
        int m_currentNode;
        
        int m_numberNodes;
        
        int m_headNode;
        
        MARKER m_isInList;
        
        vector<Node> m_nodes;
        
        
        Node& getNode(int nodeId)
        {
            assert(nodeId >= 0 && nodeId < m_numberNodes);
            return m_nodes[nodeId];
        }
        
        const Node& getNode(int nodeId) const
        {
            assert(nodeId >= 0 && nodeId < m_numberNodes);
            return m_nodes[nodeId];
        }
        
        void removeNode(int nodeId);
    };

    /** IDA* like search keeping a list of the fringe nodes.
        This avoids re-expansion of interior nodes at each new
        iteration.

        %Statistics (see Search::createStatistics and Search::getStatistics):
        - "cpu_time"
        - "iterations"
        - "nodes_expanded"
        - "nodes_visited"
        - "nodes_expanded_last_iter"
        - "nodes_visited_last_iter"
        - "path_cost"
        - "path_length"
    */
    template<class MARKER = MarkerFastClear>
    class FringeSearch
        : public Search
    {
    public:
        /** Constructor.
            @param nodeCountForAllTargets
            Finish the last iteration. Only useful for getting the node
            statistics. Does not return all paths.
        */
        FringeSearch(bool nodeCountForAllTargets = false);

        StatisticsCollection createStatistics();

        bool findPath(const Environment& env, int start, int target);

        const vector<int>& getPath() const
        {
            return m_path;
        }

        const StatisticsCollection& getStatistics() const;

        const vector<char>& getVisitedNodes() const
        {
            return m_visitedNodes;
        }

    private:   
        /** Cache element for g-value used in FringeSearch. */
        class CacheElement
        {
        public:
            int m_gValue;
            
            int m_parent;
        };


        bool m_nodeCountForAllTargets;

        bool m_targetFound;

        int m_nodesExpanded;

        int m_nodesExpandedCurrentIter;

        int m_nodesVisited;

        int m_nodesVisitedCurrentIter;

        int m_target;

        const Environment* m_env;

        Fringe<MARKER> m_fringe;

        StatisticsCollection m_statistics;

        vector<CacheElement> m_cache;
        
        MARKER m_isCacheValid;
        
        vector<char> m_visitedNodes;

        vector<int> m_path;


        void addCache(int nodeId, int gValue, int parent);

        void addStatistics(const string& name, double value)
        {
            m_statistics.get(name).add(value);
        }

        bool checkCache(int nodeId, int gValue) const;

        void constructPath(int start);

        void getFromCache(int nodeId, int& gValue, int& parent) const;

        void doIteration(int iteration, int fLimit, int& fMin);

        void init(int numberNodes);

        void markVisitedNode(int nodeId, int iteration);
    };
}

//-----------------------------------------------------------------------------

namespace PathFind
{
    using namespace std;

    template<class MARKER>
    Fringe<MARKER>::Fringe()
    {
        m_numberNodes = 0;
    }

    template<class MARKER>
    void Fringe<MARKER>::init(int numberNodes)
    {
        m_nodes.resize(numberNodes);
        m_isInList.init(numberNodes);
        m_numberNodes = numberNodes;
        m_currentNode = NO_NODE;
        m_headNode = NO_NODE;
    }

    template<class MARKER>
    void Fringe<MARKER>::insertNode(int nodeId)
    {
        if (m_isInList.get(nodeId))
            removeNode(nodeId);
        m_isInList.set(nodeId);
        Node& newNode = getNode(nodeId);
        if (m_headNode != NO_NODE)
        {
            Node& current = getNode(m_currentNode);
            int next = current.m_next;
            newNode.m_previous = m_currentNode;
            newNode.m_next = next;
            if (next != NO_NODE)
                getNode(next).m_previous = nodeId;
            current.m_next = nodeId;
        }
        else
        {
            m_headNode = nodeId;
            newNode.m_previous = NO_NODE;
            newNode.m_next = NO_NODE;
        }
    }

    template<class MARKER>
    void Fringe<MARKER>::nextNode()
    {
        assert(m_currentNode != NO_NODE);
        m_currentNode = getNode(m_currentNode).m_next;
        assert(m_currentNode == NO_NODE || m_isInList.get(m_currentNode));
    }

    template<class MARKER>
    void Fringe<MARKER>::print(ostream& o) const
    {
        o << "Fringe {\n";
        int nodeId = m_headNode;
        while (nodeId != NO_NODE)
        {
            o << nodeId << '\n';
            nodeId = getNode(nodeId).m_next;
        }
        o << "}\n";
    }

    template<class MARKER>
    void Fringe<MARKER>::removeNode(int nodeId)
    {
        assert(nodeId != NO_NODE);
        assert(m_isInList.get(nodeId));
        Node& node = getNode(nodeId);
        int previous = node.m_previous;
        int next = node.m_next;
        if (previous != NO_NODE)
            getNode(previous).m_next = next;
        else
            m_headNode = next;    
        if (next != NO_NODE)
            getNode(next).m_previous = previous;
        m_isInList.unset(nodeId);
    }

    template<class MARKER>
    void Fringe<MARKER>::startIteration()
    {
        m_currentNode = m_headNode;
    }

    template<class MARKER>
    FringeSearch<MARKER>::FringeSearch(bool nodeCountForAllTargets)
        : m_nodeCountForAllTargets(nodeCountForAllTargets),
          m_statistics(createStatistics())
    {
    }
    
    template<class MARKER>
    void FringeSearch<MARKER>::addCache(int nodeId, int gValue, int parent)
    {
        CacheElement& cacheElement = m_cache[nodeId];
        cacheElement.m_gValue = gValue;
        cacheElement.m_parent = parent;
        m_isCacheValid.set(nodeId);
    }

    template<class MARKER>
    bool FringeSearch<MARKER>::checkCache(int nodeId, int gValue) const
    {
        if (! m_isCacheValid.get(nodeId))
            return false;
        const CacheElement& cacheElement = m_cache[nodeId];
        int gValueCached = cacheElement.m_gValue;
        return (gValue >= gValueCached);
    }

    template<class MARKER>
    void FringeSearch<MARKER>::constructPath(int start)
    {
        m_path.clear();
        if (! m_isCacheValid.get(m_target))
            return;
        int nodeId = m_target;
        do
        {
            m_path.push_back(nodeId);
            nodeId = m_cache[nodeId].m_parent;
            assert(m_isCacheValid.get(nodeId));
        }
        while (nodeId != start);
        m_path.push_back(nodeId);
    }

    template<class MARKER>
    StatisticsCollection FringeSearch<MARKER>::createStatistics()
    {
        StatisticsCollection collection;
        collection.create("cpu_time");
        collection.create("iterations");
        collection.create("nodes_expanded");
        collection.create("nodes_visited");
        collection.create("nodes_expanded_last_iter");
        collection.create("nodes_visited_last_iter");
        collection.create("path_cost");
        collection.create("path_length");
        return collection;
    }

    template<class MARKER>
    void FringeSearch<MARKER>::doIteration(int iteration, int fLimit,
                                           int& fMin)
    {
        m_fringe.startIteration();
        vector<Environment::Successor> successors;
        while (true)
        {
            int nodeId = m_fringe.getCurrentNode();
            if (nodeId == NO_NODE)
                return;
            ++m_nodesVisitedCurrentIter;
            markVisitedNode(nodeId, iteration);
            int gValue;
            int lastNode;
            getFromCache(nodeId, gValue, lastNode);
            int fValue = gValue + m_env->getHeuristic(nodeId, m_target);
            //cerr << "Visit " << nodeId << " (" << fValue << ")\n";
            if (fValue > fLimit)
            {
                if (fValue < fMin)
                    fMin = fValue;
                m_fringe.nextNode();
                continue;
            }
            if (nodeId == m_target)
            {
                if (! m_targetFound)
                {
                    m_statistics.get("path_cost").add(gValue);
                    m_targetFound = true;
                }
                if (! m_nodeCountForAllTargets)
                    return;
            }
            m_env->getSuccessors(nodeId, lastNode, successors);
            int numberSuccessors = successors.size();
            ++m_nodesExpandedCurrentIter;
            // Reverse loop so that nodes are visited in same order as in IDA*
            for (int i = numberSuccessors; i--; )
            {
                const Environment::Successor& successor = successors[i];
                int targetNodeId = successor.m_target;
                int targetGValue = gValue + successor.m_cost;
                if (checkCache(targetNodeId, targetGValue))
                    continue;
                m_fringe.insertNode(targetNodeId);
                addCache(targetNodeId, targetGValue, nodeId);
            }
            m_fringe.removeCurrentNode();
            m_fringe.nextNode();
        }
    }

    template<class MARKER>
    bool FringeSearch<MARKER>::findPath(const Environment& env, int start,
                                        int target)
    {
        assert(env.isValidNodeId(start));
        assert(env.isValidNodeId(target));
        Clock clock;
        m_env = &env;
        m_target = target;
        int numberNodes = env.getNumberNodes();
        init(numberNodes);
        m_nodesExpanded = 0;
        m_nodesVisited = 0;
        int iteration = 1;
        int fLimit = env.getHeuristic(start, target);
        m_fringe.insertNode(start);
        addCache(start, 0, NO_NODE);
        m_targetFound = false;
        while (true)
        {
            //cerr << "Iteration " << iteration
            //<< " limit = " << fLimit << '\n';
            m_nodesExpandedCurrentIter = 0;
            m_nodesVisitedCurrentIter = 0;
            int fMin = numeric_limits<int>::max();
            doIteration(iteration, fLimit, fMin);
            m_nodesExpanded += m_nodesExpandedCurrentIter;
            m_nodesVisited += m_nodesVisitedCurrentIter;
            if (m_targetFound)
                break;
            fLimit = fMin;
            ++iteration;
            //m_fringe.print(cerr);
        }
        constructPath(start);
        addStatistics("iterations", iteration);
        addStatistics("nodes_visited", m_nodesVisited);
        addStatistics("nodes_expanded", m_nodesExpanded);
        addStatistics("nodes_visited_last_iter", m_nodesVisitedCurrentIter);
        addStatistics("nodes_expanded_last_iter", m_nodesExpandedCurrentIter);
        addStatistics("path_length", m_path.size());
        addStatistics("cpu_time", clock.get());
        return true;
    }

    template<class MARKER>
    void FringeSearch<MARKER>::getFromCache(int nodeId, int& gValue,
                                            int& parent) const
    {
        assert(m_isCacheValid.get(nodeId));
        const CacheElement& cacheElement = m_cache[nodeId];
        parent = cacheElement.m_parent;
        gValue = cacheElement.m_gValue;
    }

    template<class MARKER>
    const StatisticsCollection& FringeSearch<MARKER>::getStatistics() const
    {
        return m_statistics;
    }

    template<class MARKER>
    void FringeSearch<MARKER>::init(int numberNodes)
    {
        m_visitedNodes.resize(numberNodes);
        fill(m_visitedNodes.begin(), m_visitedNodes.end(), ' ');
        m_isCacheValid.init(numberNodes);
        m_cache.resize(numberNodes);
        m_fringe.init(numberNodes);
        m_statistics.clear();
    }

    template<class MARKER>
    void FringeSearch<MARKER>::markVisitedNode(int nodeId, int iteration)
    {
        if (m_visitedNodes[nodeId] != ' ')
            return;
        m_visitedNodes[nodeId] = getVisitedNodeLabel(iteration);
    }

}

//-----------------------------------------------------------------------------

#endif
