// $Id: glxcubes.C,v 1.1.1.1 2004/10/11 09:09:19 orts_mburo Exp $ 

// This is an ORTS file (c) Michael Buro, licensed under the GPL

/* OpenGL test

   draw a rectangular height field, move the viewer location, zoom, and
   pick objects (also reports object intersection point with "click-spaghetti")

   - texture optional
   - uses an octree
   
   compile with:  g++ -O3 glxcubes.C  -L/usr/X11R6/lib -lGL -lGLU -lSDL
*/

#include <cassert>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <GL/gl.h>
#include <GL/glext.h>
#include <GL/glu.h>
#include <iostream>
#include <vector>
#include <SDL/SDL.h>

#include <sys/time.h>
#include <unistd.h>
#include <map>

using namespace std;

typedef   signed char        sint1;
typedef unsigned char        uint1;
typedef   signed short       sint2;
typedef unsigned short       uint2;
typedef   signed long        sint4;
typedef unsigned long        uint4;
typedef               float  real4;
typedef               double real8;
typedef          long double realC;
typedef   signed long long   sint8;
typedef unsigned long long   uint8;

typedef void* vptr;
typedef char* cptr;

typedef const void* cvptr;
typedef const char* ccptr;

typedef const void* const cvptrc;
typedef const char* const ccptrc;

const sint1 min_sint1 = 0x80;
const sint1 max_sint1 = 0x7F;
const uint1 min_uint1 = 0x00;
const uint1 max_uint1 = 0xFF;
const sint2 min_sint2 = 0x8000;
const sint2 max_sint2 = 0x7FFF;
const uint2 min_uint2 = 0x0000;
const uint2 max_uint2 = 0xFFFF;
const sint4 min_sint4 = 0x80000000;
const sint4 max_sint4 = 0x7FFFFFFF;
const uint4 min_uint4 = 0x00000000;
const uint4 max_uint4 = 0xFFFFFFFF;
const sint8 min_sint8 = 0x8000000000000000LL;
const sint8 max_sint8 = 0x7FFFFFFFFFFFFFFFLL;
const uint8 min_uint8 = 0x0000000000000000LL;
const uint8 max_uint8 = 0xFFFFFFFFFFFFFFFFLL;

#define FOREVER    for (;;)
#define FOR(i, n)  for (i = 0; i < (n); ++i)
#define FORS(i, n)  for (sint4 i = 0; i < (n); ++i)
#define FORU(i, n)  for (uint4 i = 0; i < (n); ++i)

#if defined(__GNUC__)
#define FORT(i, n)  for (typeof(n) i = 0; i < (n); ++i)
#else
#define FORT(i, n)  for (sint4 i = 0; i < (n); ++i)
#endif

#define FORALL(CONT,i)  \
  typedef typeof(CONT) i##_t; \
  i##_t::iterator i = CONT.begin();     \
  i##_t::iterator i##_end = CONT.end(); \
  for (; i != i##_end; i++) 

#define FORALL_CONST(CONT,i) \
  typedef typeof(CONT) i##_t; \
  i##_t::const_iterator i = CONT.begin();       \
  i##_t::const_iterator i##_end = CONT.end();   \
  for (; i != i##_end; i++) 

#define errstr cerr
#define ABORT  assert(0==1)

#define ERR(s) {\
  errstr << endl << __FILE__ << " " << __FUNCTION__ << "() (line " << __LINE__ << "): "\
  << s << endl; ABORT; exit(20); \
}

#define ERR2(s1, s2) {\
  errstr << endl << __FILE__ << " " << __FUNCTION__ << "() (line " << __LINE__ << "): "\
  << s1 << " " << s2 << endl; ABORT; exit(20); \
}



#define USE_LIGHT    1
#define USE_TEXTURES 1 
//#define TILE_TEXTURE_FILE "golfrough.bmp"
#define TILE_TEXTURE_FILE "data/flowers.bmp"
#define TILE_TEXTURE_FILE_2 "gfx/water.bmp"
//#define TILE_TEXTURE_FILE "buffalo-grass.bmp"
#define SIZE_X  1000
#define SIZE_Y  1000
#define BPP     16

const sint4 NX = 64;        // world x
const sint4 NY = 64;        // world y
const sint4 NZ = 10;        // world z/2
const sint4 MAX_Z = 2;      // max building height
const float FILL_RATE = 10; // percentage

const float TX = 7.5; // x texture periodicity 

// views (exactly one 1)
#define V1x1     1
#define V2x1     0
#define V2x2     0


struct Col3 {
  float r,g,b;
  Col3(float r_=0, float g_=0, float b_=0) : r(r_), g(g_), b(b_) {}
};

struct Coor3 {
  float x,y,z;
  Coor3(float x_=0, float y_=0, float z_=0) : x(x_), y(y_), z(z_) {}
};

Coor3 operator+(const Coor3 &a, const Coor3 &b) {
  return Coor3(a.x + b.x, a.y + b.y, a.z + b.z);
}

Coor3 operator-(const Coor3 &a, const Coor3 &b) {
  return Coor3(a.x - b.x, a.y - b.y, a.z - b.z);
}

Coor3 operator*(float t, const Coor3 &a) {
  return Coor3(t*a.x, t*a.y, t*a.z);
}

Coor3 operator*(const Coor3 &a, float t) {
  return Coor3(t*a.x, t*a.y, t*a.z);
}

float dot(const Coor3 &a, const Coor3 &b) {
  return a.x*b.x + a.y*b.y + a.z*b.z;
}
 
// intersect line and plane
//
// input: line (l1,l2)
//        plane (point,normal)
//
// returns:
//   0 = intersection (result in intersection)
//   1 = no intersection
//   2 = segment in plane

sint4 intersect_line_plane(const Coor3 &l1, const Coor3 &l2,
			 const Coor3 &point, const Coor3 &normal,
			 Coor3 &intersection)
{
  Coor3 u = l2 - l1;
  Coor3 w = l1 - point;
  
  float d = dot(normal, u);
  float n = -dot(normal, w);
  
  if (fabs(d) < 1e-6) { // segment parallel to plane
    if (n == 0) 
      return 2;         // segment in plane
    else
      return 1;         // no intersection
  }
  
  intersection = l1 + (n/d) * u;    
  return 0;
}

  
float XMIN, XMAX, YMIN, YMAX, ZMIN, ZMAX;

GLuint textures[3];
SDL_Surface *surface;
SDL_Rect mouse_area; // mouse events in this part of the window are handled

float TZ, INV_TX, INV_TZ;



// return current time (in seconds)
static sint4 current_time(void)
{
   struct timeval tv;
   struct timezone tz;
   (void) gettimeofday(&tv, &tz);
   return (int) tv.tv_sec;
}



// simple height field

class World {

public:

  sint4 x, y;
  sint4 n;
  sint4 *height;

  World(sint4 x_, sint4 y_) : x(x_), y(y_) {
    assert(x > 0);
    assert(y > 0);
    n = x*y;
    height = new sint4[n];
    for (sint4 i=0; i<n; ++i) height[i] = 0;
  }

  ~World() { delete height; }
};


World world(NX, NY);
sint4 draw_count = 0;
sint4 aabb_count = 0;

float y_angle   = 45;
float x_angle   = 45;
float lz_angle  = 45;
float lx_angle  = 45;
float scale = 1.0;
float x_off = 0.0;
float y_off = 0.0;

sint4 frames = 0;
timeval start_time;

sint4 window_width, window_height;
uint4 drawn_magic = 1;


void make_sphere(sint4 id, const Coor3 &o, float r, const Col3 &c);
void make_axis_aligned_box(sint4 id, const Coor3 &o, const Coor3 &d,
			   const Col3 &cb, const Col3 &ct);
struct Sector;

class GfxObj {
public:
  uint4 drawn; // magic
  Coor3 origin;
  uint4 id;
  static uint4 id0;
  static map<uint4, GfxObj*> id2obj;   // maps id to object (for picking)
  static const sint4 SUB_ID_LOG = 3;   // at most 2<<SUB_ID_LOG many sub ids per obj
  vector<Sector*> sectors; // octree cubes that intersect with the object
  virtual void draw() = 0;
  virtual void aabb(Coor3 &o, Coor3 &d) = 0; // return axis aligned bounding box
  GfxObj() { id = ++id0; id2obj[id] = this; drawn = 0; }
  virtual ~GfxObj() { id2obj.erase(id); }
};

uint4 GfxObj::id0 = 0;
map<uint4, GfxObj*> GfxObj::id2obj;

class GfxSphere : public GfxObj {
public:
  float radius;
  Col3  col;

  void aabb(Coor3 &o, Coor3 &d) {
    o = origin; o.x -= radius; o.y -= radius; o.z -= radius;
    d = Coor3(2*radius, 2*radius, 2*radius);
  }
  
  void draw() {
    GLfloat col_arr[4] = { col.r, col.g, col.b, 1.0 };
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, col_arr);
    make_sphere(id, origin, radius, col);
  }
};

class GfxBox : public GfxObj {
public:
  Coor3 dim;
  Col3  cb, ct;

  enum FaceID { ID_TOP=0, ID_BOTTOM, ID_LEFT, ID_RIGHT, ID_FRONT, ID_BACK };
  
  // true iff no intersection, shouldn't happen
  bool line_intersection(sint4 sub_id, const Coor3 &l1, const Coor3 &l2, Coor3 &intersec);
  void aabb(Coor3 &o, Coor3 &d) { o = origin; d = dim; }
  void draw() { make_axis_aligned_box(id, origin, dim, cb, ct); }
};


// compute line intersection with face
bool GfxBox::line_intersection(sint4 face_id,
			       const Coor3 &l1, const Coor3 &l2, Coor3 &intersec)
{
  Coor3 point, normal, &o = origin, &d = dim;
  
  switch (FaceID(face_id)) {
  case ID_TOP:
    point  = Coor3(o.x, o.y+d.y, o.z);
    normal = Coor3(0,1,0);
    break;
  case ID_BOTTOM:
    point  = Coor3(o.x, o.y, o.z);
    normal = Coor3(0,-1,0);    
    break;
  case ID_LEFT:
    point  = Coor3(o.x, o.y, o.z);
    normal = Coor3(-1,0,0);    
    break;
  case ID_RIGHT:
    point  = Coor3(o.x+d.x, o.y, o.z);
    normal = Coor3(1,0,0);    
    break;
  case ID_FRONT:
    point  = Coor3(o.x, o.y, o.z+d.z);
    normal = Coor3(0,0,1);    
    break;
  case ID_BACK:
    point  = Coor3(o.x, o.y, o.z);
    normal = Coor3(0,0,-1);
    break;
  }

  return intersect_line_plane(l1, l2, point, normal, intersec) != 0;
}



struct Sector {
  bool used;
  vector<GfxObj*> objs;

  Sector() { used = false; }
};



float frustum[6][4];
float m_Frustum[6][4];

enum { A, B, C, D };
enum { RIGHT, LEFT, BOTTOM, TOP, BACK, FRONT };


// GPL'ed code: Sydengine 3D, author: Frdric Delorme

void normalize_plane(float frustum[6][4], sint4 side)
{
    //PRESERVE:BEGIN
    // Here we calculate the magnitude of the normal to the plane (point A B C)
    // Remember that (A, B, C) is that same thing as the normal's (X, Y, Z).
    // To calculate magnitude you use the equation:  magnitude = sqrt( x^2 + y^2 + z^2)
    float magnitude = (float)sqrt( frustum[side][A] * frustum[side][A] +
                                   frustum[side][B] * frustum[side][B] +
                                   frustum[side][C] * frustum[side][C] );

    // Then we divide the plane's values by it's magnitude.
    // This makes it easier to work with.
    frustum[side][A] /= magnitude;
    frustum[side][B] /= magnitude;
    frustum[side][C] /= magnitude;
    frustum[side][D] /= magnitude;

    //PRESERVE:END
}

void calculate_frustum()
{
  //PRESERVE:BEGIN
  float proj[16];                               // This will hold our projection matrix
  float modl[16];                               // This will hold our modelview matrix
  float clip[16];                               // This will hold the clipping planes

  // glGetFloatv() is used to extract information about our OpenGL world.
  // Below, we pass in GL_PROJECTION_MATRIX to abstract our projection matrix.
  // It then stores the matrix into an array of [16].
  glGetFloatv( GL_PROJECTION_MATRIX, proj );

  // By passing in GL_MODELVIEW_MATRIX, we can abstract our model view matrix.
  // This also stores it in an array of [16].
  glGetFloatv( GL_MODELVIEW_MATRIX, modl );

  // Now that we have our modelview and projection matrix, if we combine these 2 matrices,
  // it will give us our clipping planes.  To combine 2 matrices, we multiply them.

  clip[ 0]= modl[ 0]*proj[ 0] + modl[ 1]*proj[ 4] + modl[ 2]*proj[ 8] + modl[ 3]*proj[12];
  clip[ 1]= modl[ 0]*proj[ 1] + modl[ 1]*proj[ 5] + modl[ 2]*proj[ 9] + modl[ 3]*proj[13];
  clip[ 2]= modl[ 0]*proj[ 2] + modl[ 1]*proj[ 6] + modl[ 2]*proj[10] + modl[ 3]*proj[14];
  clip[ 3]= modl[ 0]*proj[ 3] + modl[ 1]*proj[ 7] + modl[ 2]*proj[11] + modl[ 3]*proj[15];

  clip[ 4]= modl[ 4]*proj[ 0] + modl[ 5]*proj[ 4] + modl[ 6]*proj[ 8] + modl[ 7]*proj[12];
  clip[ 5]= modl[ 4]*proj[ 1] + modl[ 5]*proj[ 5] + modl[ 6]*proj[ 9] + modl[ 7]*proj[13];
  clip[ 6]= modl[ 4]*proj[ 2] + modl[ 5]*proj[ 6] + modl[ 6]*proj[10] + modl[ 7]*proj[14];
  clip[ 7]= modl[ 4]*proj[ 3] + modl[ 5]*proj[ 7] + modl[ 6]*proj[11] + modl[ 7]*proj[15];

  clip[ 8]= modl[ 8]*proj[ 0] + modl[ 9]*proj[ 4] + modl[10]*proj[ 8] + modl[11]*proj[12];
  clip[ 9]= modl[ 8]*proj[ 1] + modl[ 9]*proj[ 5] + modl[10]*proj[ 9] + modl[11]*proj[13];
  clip[10]= modl[ 8]*proj[ 2] + modl[ 9]*proj[ 6] + modl[10]*proj[10] + modl[11]*proj[14];
  clip[11]= modl[ 8]*proj[ 3] + modl[ 9]*proj[ 7] + modl[10]*proj[11] + modl[11]*proj[15];

  clip[12]= modl[12]*proj[ 0] + modl[13]*proj[ 4] + modl[14]*proj[ 8] + modl[15]*proj[12];
  clip[13]= modl[12]*proj[ 1] + modl[13]*proj[ 5] + modl[14]*proj[ 9] + modl[15]*proj[13];
  clip[14]= modl[12]*proj[ 2] + modl[13]*proj[ 6] + modl[14]*proj[10] + modl[15]*proj[14];
  clip[15]= modl[12]*proj[ 3] + modl[13]*proj[ 7] + modl[14]*proj[11] + modl[15]*proj[15];

  // Now we actually want to get the sides of the frustum.  To do this we take
  // the clipping planes we received above and extract the sides from them.

  // This will extract the RIGHT side of the frustum
  m_Frustum[RIGHT][A] = clip[ 3] - clip[ 0];
  m_Frustum[RIGHT][B] = clip[ 7] - clip[ 4];
  m_Frustum[RIGHT][C] = clip[11] - clip[ 8];
  m_Frustum[RIGHT][D] = clip[15] - clip[12];

  // Now that we have a normal (A,B,C) and a distance (D) to the plane,
  // we want to normalize that normal and distance.

  // Normalize the RIGHT side
  normalize_plane(m_Frustum, RIGHT);

  // This will extract the LEFT side of the frustum
  m_Frustum[LEFT][A] = clip[ 3] + clip[ 0];
  m_Frustum[LEFT][B] = clip[ 7] + clip[ 4];
  m_Frustum[LEFT][C] = clip[11] + clip[ 8];
  m_Frustum[LEFT][D] = clip[15] + clip[12];

  // Normalize the LEFT side
  normalize_plane(m_Frustum, LEFT);

  // This will extract the BOTTOM side of the frustum
  m_Frustum[BOTTOM][A] = clip[ 3] + clip[ 1];
  m_Frustum[BOTTOM][B] = clip[ 7] + clip[ 5];
  m_Frustum[BOTTOM][C] = clip[11] + clip[ 9];
  m_Frustum[BOTTOM][D] = clip[15] + clip[13];

  // Normalize the BOTTOM side
  normalize_plane(m_Frustum, BOTTOM);

  // This will extract the TOP side of the frustum
  m_Frustum[TOP][A] = clip[ 3] - clip[ 1];
  m_Frustum[TOP][B] = clip[ 7] - clip[ 5];
  m_Frustum[TOP][C] = clip[11] - clip[ 9];
  m_Frustum[TOP][D] = clip[15] - clip[13];

  // Normalize the TOP side
  normalize_plane(m_Frustum, TOP);

  // This will extract the BACK side of the frustum
  m_Frustum[BACK][A] = clip[ 3] - clip[ 2];
  m_Frustum[BACK][B] = clip[ 7] - clip[ 6];
  m_Frustum[BACK][C] = clip[11] - clip[10];
  m_Frustum[BACK][D] = clip[15] - clip[14];

  // Normalize the BACK side
  normalize_plane(m_Frustum, BACK);

  // This will extract the FRONT side of the frustum
  m_Frustum[FRONT][A] = clip[ 3] + clip[ 2];
  m_Frustum[FRONT][B] = clip[ 7] + clip[ 6];
  m_Frustum[FRONT][C] = clip[11] + clip[10];
  m_Frustum[FRONT][D] = clip[15] + clip[14];

  // Normalize the FRONT side
  normalize_plane(m_Frustum, FRONT);
  //PRESERVE:END
}

// The code below will allow us to make checks within the frustum.  For example,
// if we want to see if a point, a sphere, or a cube lies inside of the frustum.
// Because all of our planes point INWARDS (The normals are all pointing inside the frustum)
// we then can assume that if a point is in FRONT of all of the planes, it's inside.

bool point_in_frustum(float x, float y, float z)
{
  //PRESERVE:BEGIN
  // If you remember the plane equation (A*x + B*y + C*z + D = 0), then the rest
  // of this code should be quite obvious and easy to figure out yourself.
  // In case don't know the plane equation, it might be a good idea to look
  // at our Plane Collision tutorial at www.GameTutorials.com in OpenGL Tutorials.
  // I will briefly go over it here.  (A,B,C) is the (X,Y,Z) of the normal to the plane.
  // They are the same thing... but just called ABC because you don't want to say:
  // (x*x + y*y + z*z + d = 0).  That would be wrong, so they substitute them.
  // the (x, y, z) in the equation is the point that you are testing.  The D is
  // The distance the plane is from the origin.  The equation ends with "= 0" because
  // that is true when the point (x, y, z) is ON the plane.  When the point is NOT on
  // the plane, it is either a negative number (the point is behind the plane) or a
  // positive number (the point is in front of the plane).  We want to check if the point
  // is in front of the plane, so all we have to do is go through each point and make
  // sure the plane equation goes out to a positive number on each side of the frustum.
  // The result (be it positive or negative) is the distance the point is front the plane.

  // Go through all the sides of the frustum
  for(sint4 i = 0; i < 6; i++ )
    {
      // calculate the plane equation and check if the point is behind a side of the frustum
      if(m_Frustum[i][A] * x + m_Frustum[i][B] * y + m_Frustum[i][C] * z + m_Frustum[i][D] <= 0)
        {
	  // The point was behind a side, so it ISN'T in the frustum
	  return false;
        }
    }

  // The point was inside of the frustum (In front of ALL the sides of the frustum)
  return true;
  //PRESERVE:END
}


bool sphere_in_frustum(float x, float y, float z, float radius)
{
  //PRESERVE:BEGIN
  // Now this function is almost identical to the PointInFrustum(), except we
  // now have to deal with a radius around the point.  The point is the center of
  // the radius.  So, the point might be outside of the frustum, but it doesn't
  // mean that the rest of the sphere is.  It could be half and half.  So instead of
  // checking if it's less than 0, we need to add on the radius to that.  Say the
  // equation produced -2, which means the center of the sphere is the distance of
  // 2 behind the plane.  Well, what if the radius was 5?  The sphere is still inside,
  // so we would say, if(-2 < -5) then we are outside.  In that case it's false,
  // so we are inside of the frustum, but a distance of 3.  This is reflected below.

  // Go through all the sides of the frustum
  FORS (i, 6) {
    // If the center of the sphere is farther away from the plane than the radius
    if (m_Frustum[i][A]*x + m_Frustum[i][B]*y + m_Frustum[i][C]*z + m_Frustum[i][D]
	<= -radius ) {
      // The distance was greater than the radius so the sphere is
      // outside of the frustum
      return false;
    }
  }

  // The sphere was inside of the frustum!
  return true;
  //PRESERVE:END
}

bool aabox_in_frustum(const Coor3 &o, const Coor3 &d)
{
  //PRESERVE:BEGIN
  // This test is a bit more work, but not too much more complicated.
  // Basically, what is going on is, that we are given the center of the cube,
  // and half the length.  Think of it like a radius.  Then we checking each point
  // in the cube and seeing if it is inside the frustum.  If a point is found in front
  // of a side, then we skip to the next side.  If we get to a plane that does NOT have
  // a point in front of it, then it will return false.

  // *Note* - This will sometimes say that a cube is inside the frustum when it isn't.
  // This happens when all the corners of the bounding box are not behind any one plane.
  // This is rare and shouldn't effect the overall rendering speed.

  ++aabb_count;

  FORS (i, 6) {

    //++n; if ((n & (16*65536-1)) == 0) cout << "XXX " << n << " " << sec << endl;

    if (m_Frustum[i][A] * (o.x)     +
	m_Frustum[i][B] * (o.y)     +
	m_Frustum[i][C] * (o.z)     + m_Frustum[i][D] > 0) continue;

    //++sec;
    
    if (m_Frustum[i][A] * (o.x+d.x) +
	m_Frustum[i][B] * (o.y)     +
	m_Frustum[i][C] * (o.z)     + m_Frustum[i][D] > 0) continue;
    if (m_Frustum[i][A] * (o.x)     +
	m_Frustum[i][B] * (o.y+d.y) +
	m_Frustum[i][C] * (o.z)     + m_Frustum[i][D] > 0) continue;
    if (m_Frustum[i][A] * (o.x+d.x) +
	m_Frustum[i][B] * (o.y+d.y) +
	m_Frustum[i][C] * (o.z)     + m_Frustum[i][D] > 0) continue;
    if (m_Frustum[i][A] * (o.x)     +
	m_Frustum[i][B] * (o.y)     +
	m_Frustum[i][C] * (o.z+d.z) + m_Frustum[i][D] > 0) continue;
    if (m_Frustum[i][A] * (o.x+d.x) +
	m_Frustum[i][B] * (o.y)     +
	m_Frustum[i][C] * (o.z+d.z) + m_Frustum[i][D] > 0) continue;
    if (m_Frustum[i][A] * (o.x) +
	m_Frustum[i][B] * (o.y+d.y) +
	m_Frustum[i][C] * (o.z+d.z) + m_Frustum[i][D] > 0) continue;
    if (m_Frustum[i][A] * (o.x+d.x) +
	m_Frustum[i][B] * (o.y+d.y) +
	m_Frustum[i][C] * (o.z+d.z) + m_Frustum[i][D] > 0) continue;

    // If we get here, it isn't in the frustum
    return false;
  }

  return true;
  //PRESERVE:END
}



void make_sphere(sint4 id, const Coor3 &o, float r, const Col3 &c)
{
  if (id >= 0) glLoadName(id << GfxObj::SUB_ID_LOG);
  
  glPushMatrix();

  glMatrixMode(GL_MODELVIEW);
  glTranslatef(o.x, o.y, o.z);
  glRotatef(lx_angle/4, 0.0f, 0.0f, -1.0f ); // earth's rotation
  GLint slices = 20, stacks = 14;  
  GLUquadricObj *q = gluNewQuadric();
  glColor3f(c.r, c.g, c.b);
  GLfloat cola[4] = { c.r, c.g, c.b, 1.0 };
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, cola);

#if USE_TEXTURES
  glBindTexture (GL_TEXTURE_2D, textures[0]);
  gluQuadricDrawStyle ( q, GLU_FILL   );
  gluQuadricNormals   ( q, GLU_SMOOTH );
  gluQuadricTexture   ( q, GL_TRUE   );
  glEnable(GL_TEXTURE_2D);
  gluSphere(q, r, slices, stacks);
  glDisable(GL_TEXTURE_2D);  
#else  
  gluSphere(q, r, slices, stacks);
#endif
  

  glPopMatrix();
  gluDeleteQuadric(q);
}


void make_axis_aligned_box(sint4 id, const Coor3 &o, const Coor3 &d,
			   const Col3 &cb, const Col3 &ct)
{
  if (d.y != 0) {
    
    glColor3f(cb.r, cb.g, cb.b);

    GLfloat boxc[4] = { cb.r, cb.g, cb.b, 1.0 };
    // GLfloat boxc[4] = { 1,1,1,1 };    
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, boxc);

    // name all faces so that pick routine can find exact intersection
    glLoadName(((id<<GfxObj::SUB_ID_LOG))+GfxBox::ID_BACK);   

    glBegin(GL_QUADS); 
    glNormal3f(0,0,-1);    
    glVertex3f(o.x,     o.y,     o.z); // back
    glVertex3f(o.x,     o.y+d.y, o.z);
    glVertex3f(o.x+d.x, o.y+d.y, o.z);
    glVertex3f(o.x+d.x, o.y,     o.z);
    glEnd();
    
    glLoadName(((id<<GfxObj::SUB_ID_LOG))+GfxBox::ID_FRONT);
    glBegin(GL_QUADS);    
    glNormal3f(0,0,1);        
    glVertex3f(o.x,     o.y,     o.z+d.z); // front
    glVertex3f(o.x,     o.y+d.y, o.z+d.z);
    glVertex3f(o.x+d.x, o.y+d.y, o.z+d.z);
    glVertex3f(o.x+d.x, o.y,     o.z+d.z);
    glEnd();
    

    glLoadName(((id<<GfxObj::SUB_ID_LOG))+GfxBox::ID_LEFT);
    glBegin(GL_QUADS);    
    glNormal3f(-1,0,0);        
    glVertex3f(o.x, o.y,     o.z); // left
    glVertex3f(o.x, o.y+d.y, o.z);
    glVertex3f(o.x, o.y+d.y, o.z+d.z);
    glVertex3f(o.x, o.y,     o.z+d.z);  
    glEnd();
    
    glLoadName(((id<<GfxObj::SUB_ID_LOG))+GfxBox::ID_RIGHT);
    glBegin(GL_QUADS);    
    glNormal3f(1,0,0);    
    glVertex3f(o.x+d.x, o.y,     o.z); // right
    glVertex3f(o.x+d.x, o.y+d.y, o.z);
    glVertex3f(o.x+d.x, o.y+d.y, o.z+d.z);
    glVertex3f(o.x+d.x, o.y,     o.z+d.z);
    glEnd();
    
#if 0
    // not visible
    glLoadName(((id<<GfxObj::SUB_ID_LOG))+GfxBox::ID_BOTTOM);
    glBegin(GL_QUADS);    
    glNormal3f(0,-1,0);    
    glVertex3f(o.x,     o.y, o.z); // bottom
    glVertex3f(o.x+d.x, o.y, o.z);
    glVertex3f(o.x+d.x, o.y, o.z+d.z);
    glVertex3f(o.x,     o.y, o.z+d.z);
    glEnd();
#endif

  }

  // draw top tile
  
  glColor3f(ct.r, ct.g, ct.b);
  GLfloat topc[4] = { ct.r, ct.g, ct.b, 1.0 };
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, topc);

#if USE_TEXTURES
    
  // with texture
  
  // compute texture points
  // texture periodicity is (TX, TZ)
  
  float tx0 = o.x*INV_TX; tx0 = tx0 - rint(tx0);
  float tz0 = o.z*INV_TZ; tz0 = tz0 - rint(tz0);  
  float lx  = d.x*INV_TX;
  float lz  = d.z*INV_TZ;
  
  glEnable(GL_TEXTURE_2D);

  glLoadName((id<<GfxObj::SUB_ID_LOG)+GfxBox::ID_TOP);
  glBegin(GL_QUADS);
  glNormal3f(0,1,0);
  glTexCoord2f(tx0, tz0);       glVertex3f(o.x,     o.y+d.y, o.z); // top
  glTexCoord2f(tx0+lx, tz0);    glVertex3f(o.x+d.x, o.y+d.y, o.z);
  glTexCoord2f(tx0+lx, tz0+lz); glVertex3f(o.x+d.x, o.y+d.y, o.z+d.z);
  glTexCoord2f(tx0, tz0+lz);    glVertex3f(o.x,     o.y+d.y, o.z+d.z);
  glEnd();

  glDisable(GL_TEXTURE_2D);
  
#else 

  glLoadName((id<<GfxObj::SUB_ID_LOG)+GfxBox::ID_TOP);  
  glBegin(GL_QUADS);
  glNormal3f(0,1,0);  
  glVertex3f(o.x,     o.y+d.y, o.z); // top
  glVertex3f(o.x+d.x, o.y+d.y, o.z);
  glVertex3f(o.x+d.x, o.y+d.y, o.z+d.z);
  glVertex3f(o.x,     o.y+d.y, o.z+d.z);
  glEnd();

  
#endif

}


void enter_2d() {

  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();

  glOrtho( 0, SIZE_X, 0, SIZE_Y, 0, 1 );    
    
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();

  glDisable(GL_LIGHTING);
  glDisable(GL_LIGHT0);

  glViewport(0, 0, SIZE_X, SIZE_Y);

  //  glEnable(GL_ALPHA_TEST);
//   glEnable(GL_BLEND);
//   glBlendFunc(GL_SRC_ALPHA, GL_SRC_COLOR);
  //  glDisable
}

void leave_2d() {

//   glDisable(GL_BLEND);

  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
}

void print_error(GLenum e) {

  if (e == GL_NO_ERROR) cerr << "GL_NO_ERROR" << endl;
  else if (e == GL_INVALID_ENUM) cerr << "GL_INVALID_ENUM" << endl;
  else if (e == GL_INVALID_VALUE) cerr << "GL_INVALID_VALUE" << endl;
  else if (e == GL_INVALID_OPERATION) cerr << "GL_INVALID_OPERATION" << endl;
  else if (e == GL_STACK_OVERFLOW) cerr << "GL_STACK_OVERFLOW" << endl;
  else if (e == GL_STACK_UNDERFLOW) cerr << "GL_STACK_UNDERFLOW" << endl;
  else if (e == GL_OUT_OF_MEMORY) cerr << "GL_OUT_OF_MEMORY" << endl;

}

sint4 *world_tex;


// side-effects:  draws on screen

void create_minimap_tex() {

  sint4 width = (int) pow(2, ceil( (log((double)world.x)/log((double)2)) ) );
  sint4 height = (int) pow(2, ceil(  (log((double)world.y)/log((double)2)) ) );

  cerr << "width = " << width << endl;
  cerr << "height = " << height << endl;

  int len = (world.x+1)*(world.y+1);

  world_tex = (typeof(world_tex)) malloc(len * sizeof(*world_tex));

  FORS (i, len) world_tex[i] = 0xffffff;
  
  glGenTextures(1, &textures[2]);

  // Typical Texture Generation Using Data From The Bitmap
  glBindTexture(GL_TEXTURE_2D, textures[2]);

#if 1

  glTexImage2D(GL_TEXTURE_2D, 0, 4, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, world_tex);                        

  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);

#else

//   // Generate The Texture
  gluBuild2DMipmaps(GL_TEXTURE_2D, 4, world.x,
 		    world.y, GL_RGBA,
 		    GL_UNSIGNED_BYTE, world_tex);

  // Linear Filtering
  //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_NEAREST );
  //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,GL_NEAREST );

#endif

//   leave_2d();

}


void draw_minimap()
{
  return;
  
  sint4 c;

  for (sint4 y=0; y < world.y; ++y) {
    for(sint4 x=0; x < world.x; ++x) {
      switch (world.height[y*world.x + x]) {
      case 0:
	c = (0x0 << 24) + (0x0 << 16) + (0x0 << 8) + 0xff;
	break;
      case 1:
	c = (0x0 << 24) + (0x0 << 16) + (0xf << 8) + 0x0;
	break;
      case 2:
	c = (0x0 << 24) + (0xff << 16) + (0x0 << 8) + 0x0;
	break;
      case 4:
	c = (0x0 << 24) + (0x0 << 16) + (0xff << 8) + 0xff;
	break;
      default:
	c = (0x0 << 24) + (0x0 << 16) + (0x0 << 8) + 0x0;
	break;	
      }
      //      printf("color = %x\n", c);
      printf("h = %d\n", world.height[y*world.x +x]);
      world_tex[(world.y - y-1)*world.x + x] = c;
    }
  }

  glBindTexture(GL_TEXTURE_2D, textures[2]);

#if 0
  sint4 width = (int) pow(2, ceil( (log((double)world.x)/log((double)2)) ) );
  sint4 height = (int) pow(2, ceil(  (log((double)world.y)/log((double)2)) ) );

  glTexImage2D(GL_TEXTURE_2D, 0, 4, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, world_tex);                        
#else
  gluBuild2DMipmaps(GL_TEXTURE_2D, 4, world.x,
 		    world.y, GL_RGBA,
 		    GL_UNSIGNED_BYTE, world_tex);
#endif


}


void draw_2d() {

  // cerr << "," << flush;

  enter_2d();

  glMatrixMode(GL_PROJECTION);
  //  glRotatef(45,0,0,1);  
    
  sint4 w,h,c;
  w = h = 100;
  int* surface;
  surface = (int*) malloc(w*h*4);

  c = (0x0 << 24) + (0xff << 16) + (0xff << 8) + 0xff;
  for (sint4 i=0; i<h*w/2; ++i) {
    surface[i] = c;
  }

  c = (0x80 << 24) + (0x0 << 16) + (0xff << 8) + 0xff;
  for (sint4 i=h*w/2; i<h*w; ++i) {
    surface[i] = c;
  }


  //  glGetError();

#if 0

  glRasterPos2i(100,10);
//   //print_error(glGetError());
  glDrawPixels(w,h,GL_RGBA,GL_UNSIGNED_BYTE,surface);
//   //print_error(glGetError());

#endif

  glTranslatef(100,100,0);
  glRotatef(45,0,0,1);

  draw_minimap();

  glEnable(GL_TEXTURE_2D);

  glBindTexture(GL_TEXTURE_2D, textures[2]);

  double x_fact = 1;
  double y_fact = 1;

  if (world.x < world.y) x_fact = world.x/(world.y+0.0);
  else y_fact = world.y/(world.x+0.0);

  // cerr << "x_fact = " << x_fact << endl;
  // cerr << "y_fact = " << y_fact << endl;

  glColor3f(1,1,1);
  glBegin(GL_QUADS);
  glTexCoord2f(0, 0);  glVertex3f(-50*x_fact,-50*y_fact,0);
  glTexCoord2f(1, 0);  glVertex3f( 50*x_fact,-50*y_fact,0);
  glTexCoord2f(1, 1);  glVertex3f( 50*x_fact, 50*y_fact,0);
  glTexCoord2f(0, 1);  glVertex3f(-50*x_fact, 50*y_fact,0);

  glEnd();

  glDisable(GL_TEXTURE_2D);

  glColor3f(0.6,1,0.6);
  glBegin(GL_LINE_LOOP);
  glVertex3f(-51, -51, 0);
  glVertex3f( 51, -51, 0);
  glVertex3f( 51,  51, 0);
  glVertex3f(-51,  51, 0);
  glEnd();
  
  leave_2d();
}

// Compact octree: sequence of cubic arrays sectors[level][n] (n = 8^level)
// Pointers to objects that intersect with highest level cubes
// are stored in a vector (no objects are stored at lower levels)
// paths to non-empty leaves are marked (used = true)

// When drawing, the current cube is compared against the frustum. If
// there is no intersection, draw_scene returns immediately.
// Otherwise, the children are examined or if it's a leaf, the objects
// are drawn. Because objects can intersect with many highest level
// cubes, a "magic" value is used to keep track what has been drawn.
// As always, "magic" values are used to avoid initializing all
// objects each frame.

const sint4 SEC_LEVELS = 6; // OK for zoom-in
Sector *sectors[SEC_LEVELS];
sint4 sectors_n[SEC_LEVELS];
vector<GfxObj*> gfx_objs;


// o,d:      current cube (origin, dimension)
// level:    at which Octree level are we? starts with 0
// xi,yi,zi: current cube indexes

void draw_scene(const Coor3 &o, const Coor3 &d, sint4 level, sint4 xi, sint4 yi, sint4 zi)
{
  //  cout << level << " " << xi << " " << yi << " " << zi << endl;

  sint4 ind = (((zi << level) + yi) << level) + xi;
  assert(ind >= 0 || ind < sectors_n[level]);

  if (!sectors[level][ind].used) {
    //cout << "unused" << endl;
    return;
  }

  if (level == SEC_LEVELS-1) {

    // leaf: draw all objects

    FORALL (sectors[level][ind].objs, go) {
      if ((*go)->drawn != drawn_magic) {

	// check bounding box against frustum
	Coor3 o, d;
	(*go)->aabb(o, d);
	if (!aabox_in_frustum(o, d)) continue;

	++draw_count;
	(*go)->drawn = drawn_magic;
	(*go)->draw();
      }
    }
    
    return;
  }

  if (!aabox_in_frustum(o, d)) {
    //cout << "outside" << endl;
    return;
  }

  // draw children

  // size of children boxes
  Coor3 h = d;
  h.x *= 0.5;
  h.y *= 0.5;
  h.z *= 0.5;  

  // one level up
  ++level;
  xi *= 2;
  yi *= 2;
  zi *= 2;
  
  {
    Coor3 n = o;
    draw_scene(n, h, level, xi, yi, zi);
  }
  {
    Coor3 n = o; n.x += h.x;
    draw_scene(n, h, level, xi+1, yi, zi);
  }
  {
    Coor3 n = o; n.y += h.y;
    draw_scene(n, h, level, xi, yi+1, zi);
  }
  {
    Coor3 n = o; n.x += h.x; n.y += h.y;
    draw_scene(n, h, level, xi+1, yi+1, zi);
  }
  {
    Coor3 n = o; n.z += h.z;
    draw_scene(n, h, level, xi, yi, zi+1);
  }
  {
    Coor3 n = o; n.z += h.z; n.x += h.x;
    draw_scene(n, h, level, xi+1, yi, zi+1);
  }
  {
    Coor3 n = o; n.z += h.z; n.y += h.y;
    draw_scene(n, h, level, xi, yi+1, zi+1);
  }
  {
    Coor3 n = o; n.z += h.z; n.x += h.x; n.y += h.y;
    draw_scene(n, h, level, xi+1, yi+1, zi+1);
  }
  
}



// draw viewport
	
static void draw(sint4 x0, sint4 y0, sint4 width, sint4 height, sint4 px=-1, sint4 py=-1)
{
  bool select_mode = px >= 0 && py >= 0;

  glViewport(x0, y0, width, height);

  if (select_mode) glRenderMode(GL_SELECT); else glRenderMode(GL_RENDER);

  glInitNames();
  glPushName(0);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  if (select_mode) {
    GLint viewport[4] = { x0, y0, width, height };
    gluPickMatrix(px, height-py, 1, 1, viewport);
  }

  // [-1,1]^2 is mapped to (x,y)-(x+width,y+height)
  // scaling corrects aspect ratio

  if (width > height)
    glScalef(1.0, float(width)/height, 1.0);
  else if (width < height)
    glScalef(float(height)/width, 1.0, 1.0);

  glOrtho(-NX*0.75*scale, +NX*0.75*scale,
	  -NX*0.75*scale, +NX*0.75*scale,
	  -NZ*10, NZ*10);

  glRotatef(x_angle, 1.0, 0.0, 0.0);
  glRotatef(y_angle, 0.0, 1.0, 0.0);

  glTranslatef(x_off, 0, y_off);  

#if 0  
  static GLfloat red[4] = { 0.8, 0.1, 0.0, 1.0 };
  static GLfloat green[4] = { 0.0, 0.8, 0.2, 1.0 };
  static GLfloat blue[4] = { 0.2, 0.2, 1.0, 1.0 };
  static GLfloat yellow[4] = { 1.0, 1.0, 1.0, 1.0 };
#endif  
  static GLfloat white[4] = { 1.0, 1.0, 1.0, 1.0 };
  
  GLfloat light_pos[4] = { 0, 20, 0, 0 };  
  GLfloat mat_specular[] = { 1.0, 1.0, 1.0, 1.0 };
  GLfloat mat_shininess[] = { 20.0 };

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

#if USE_LIGHT

  glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
  glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);

#endif

  {
    // draw moving sphere which indicates light direction
    
    glPushMatrix();
    glRotatef(lx_angle, 0.0, 1.0, 0.0);
    glRotatef(lz_angle, 0.0, 0.0, 1.0);

    glLightfv(GL_LIGHT0, GL_POSITION, light_pos);
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, white);
    Coor3 o(light_pos[0], light_pos[1], light_pos[2]);
    Col3  c(1,1,0);
    make_sphere(-1, o, 2, c);
    glPopMatrix();    
  }
  
  
  // Enable smooth shading
  glShadeModel(GL_SMOOTH);
  
  calculate_frustum(); // for frustum culling

  drawn_magic++;

  if (drawn_magic >= 100000) { // clear draw flags if magic value wraps around
    FORALL (gfx_objs, gob) (*gob)->drawn = 0;
    drawn_magic = 1;
  }

  Coor3 orig(XMIN, YMIN, ZMIN), dim(XMAX-XMIN, YMAX-YMIN, ZMAX-ZMIN);

  // draw all game objects

  draw_scene(orig, dim, 0, 0, 0, 0); 

  glPopName();
}


// return closest object that intersects "click-spaghetti"

sint4 pick(sint4 mx, sint4 my, float &world_x, float &world_z)
{
  world_x = world_z = -1;
  
  if (!(mx >= mouse_area.x && mx < mouse_area.x + mouse_area.w &&
	my >= mouse_area.y && my < mouse_area.y + mouse_area.h)) return 0;

  const sint4 BUFSIZE = 512;
  GLuint select_buffer[BUFSIZE];
  glSelectBuffer(BUFSIZE, select_buffer);

  draw(mouse_area.x, mouse_area.y, mouse_area.w, mouse_area.h, mx, my);
  glFlush();

  sint4 hits = glRenderMode(GL_RENDER);
  printf ("hits = %d\n", hits);
  if (hits == 0) return 0;

#if 0

  // hit details
  {
    sint4 i, j, names;
    GLuint *ptr;
    
    printf ("hits = %d\n", hits);
    ptr = (GLuint *) select_buffer;
    for (i = 0; i < hits; i++) { /*  for each hit  */
      names = (int) *ptr;
      printf (" number of names for hit = %d\n", names); ptr++;
      printf("  z1 is %g;", (float) *ptr/0x7fffffff); ptr++;
      printf(" z2 is %g\n", (float) *ptr/0x7fffffff); ptr++;
      for (j = 0; j < names; j++, ptr++) {      /*  for each name */
	printf ("  hit %d's name is: %d\n", i, *ptr);
      }
    }
  }
#endif
  
  GLint i;
  GLuint j, names, *ptr, minZ, *ptrNames=0, numberOfNames=0;

  // compute nearest intersection object
  
  ptr = (GLuint *) select_buffer;
  minZ = 0xffffffff;
  for (i = 0; i < hits; i++) {	
    names = *ptr;
    ptr++;
    if (*ptr < minZ) {
      numberOfNames = names;
      minZ = *ptr;
      ptrNames = ptr+2;
    }
    ptr += names+2;
  }

  printf ("The closest hit names are ");
  ptr = ptrNames;
  for (j = 0; j < numberOfNames; j++,ptr++) printf ("%d ", *ptr);
  printf ("\n");

  sint4 obj_id = *ptrNames;

  GfxObj *po = GfxObj::id2obj[obj_id >> GfxObj::SUB_ID_LOG];

  GfxBox *pb = dynamic_cast<GfxBox*>(po);

  if (pb) {

    // compute intersection point of "click-spaghetti" and box face
    
    sint4 sub_id = obj_id & ((1 << GfxObj::SUB_ID_LOG)-1);

    Coor3 intersect, l1, l2;

    double ModelMatrix[16]; 
    glGetDoublev(GL_MODELVIEW_MATRIX, ModelMatrix);
    double ProjMatrix[16];
    glGetDoublev(GL_PROJECTION_MATRIX, ProjMatrix);

    GLint viewport[4] = { mouse_area.x, mouse_area.y, mouse_area.w, mouse_area.h };
    GLdouble x, y, z;

    // click-spaghetti (l1,l2)
    
    gluUnProject(mx, mouse_area.h-my, 0.5,
		 ModelMatrix, 
		 ProjMatrix,
		 viewport,
		 &x, &y, &z);

    l1 = Coor3(x,y,z);

    gluUnProject(mx, mouse_area.h-my, 1.5,
		 ModelMatrix, 
		 ProjMatrix,
		 viewport,
		 &x, &y, &z);
    l2 = Coor3(x,y,z);    

    if (pb->line_intersection(sub_id, l1, l2, intersect))
      cout << "OOPS" << endl; // fixme: can happen
    else {
      cout << "Intersection: "
	   << intersect.x << " "
	   << intersect.y << " "
	   << intersect.z << endl;
    }

    world_x = intersect.x;
    world_z = intersect.z;    
  }
  
  return obj_id;
}


// draw all viewports

void draw_all()
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // mouse area: upper left viewport

#if 1
  
#if V1x1

  // 1x1
  draw(0, 0, window_width, window_height);
  mouse_area.x = mouse_area.y = 0;
  mouse_area.w = window_width;
  mouse_area.h = window_height;
  
#elif V2x1

  // 2x1
  draw(0, 0, window_width/2, window_height);
  mouse_area.x = mouse_area.y = 0;
  mouse_area.w = window_width/2;
  mouse_area.h = window_height;
  
  draw(window_width/2, 0, window_width/2, window_height);
#elif V2x2

  // 2x2
  draw(0, 0, window_width/2, window_height/2);
  mouse_area.x = mouse_area.y = 0;
  mouse_area.w = window_width/2;
  mouse_area.h = window_height/2;

  draw(window_width/2, 0, window_width/2, window_height/2);
  draw(0, window_height/2, window_width/2, window_height/2);
  draw(window_width/2, window_height/2, window_width/2, window_height/2);  
#else
#error "how many viewports?"
#endif

#endif

  draw_2d();
  
  SDL_GL_SwapBuffers();
  frames++;
}


// computes equidistant index range in one direction of line (x1,x2)
// in (xmin,xmax) which is covered by nsec segments

static void index_range(float x1, float x2, float xmin, float xmax, sint4 nsec,
			sint4 &xi_min, sint4 &xi_max)
{
  if (x1 > x2) swap(x1, x2);
  if (x1 < xmin-0.1 || x2 > xmax+0.1) {
    cerr << x1 << " " << x2 << endl;
    cerr << xmin << " " << xmax << endl;    
    ERR("x1 or x2 out of range");
  }
  if (x1 < xmin) x1 = xmin;
  if (x2 > xmax) x2 = xmax;

  float inv = nsec/(xmax-xmin);
  
  xi_min = int((x1-xmin)*inv);
  if (xi_min < 0) xi_min = 0; else if (xi_min >= nsec) xi_min = nsec-1;

  xi_max = int((x2-xmin)*inv);
  if (xi_max < 0) xi_max = 0; else if (xi_max >= nsec) xi_max = nsec-1;

  //  cout << "index: " << x1 << " " << x2 << " " << xmin << " " << xmax << " n= " << nsec
  //       << " -> " << xi_min << " " << xi_max << endl;
}

// creates height field and gfx objects

static void init()
{
  glEnable(GL_LIGHT0);
  glEnable(GL_DEPTH_TEST);

  // glEnable(GL_CULL_FACE);

  // Set the polygon mode to fill
  //glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

#if 0
  // initialize world (1x1 blocks, random height)

  for (sint4 x=0; x < world.x; ++x) {
    for (sint4 y=0; y < world.y; ++y) {
      sint4 h = random() % (MAX_Z+1);
      world.height[y*world.x + x] = h;
    }
  }

#else  

  // initialize world (2x2 blocks, random height)

  for (sint4 x=0; x < world.x-1; ++x) {
    for (sint4 y=0; y < world.y-1; ++y) {
      if (random() % 10000 < 100*FILL_RATE) {
	sint4 h = random() % (MAX_Z) + 1;
	world.height[y*world.x + x] = world.height[y*world.x + x + 1] =
	  world.height[(y+1)*world.x + x] = world.height[(y+1)*world.x + x + 1] = h;
      }
    }
  }
#endif
  
  // create all game objects

  World covered(NX, NY);
  
  for (sint4 x=0; x < world.x; ++x) {
    for (sint4 y=0; y < world.y; ++y) {
      sint4 ind0 = y*world.x + x;

      if (covered.height[ind0]) continue;
      
      sint4 h0 = world.height[ind0];
      Coor3 o(x-world.x/2, 0, y-world.y/2);
      
      Col3  cb(127/255.0, 59/255.0, 16/255.0); // box color
      Coor3 d(1, h0, 1);

#if 1
      // combine areas

      const sint4 L = 16; // areas up to L x L
      const sint4 LL = L*L;
      sint4 max_score = -1;
      sint4 mdx = 0, mdy = 0;

      // find maximal equal-height rectangle

      for (sint4 dy=1; dy <= L; ++dy) {
	if (y + dy > world.y) break;
	for (sint4 dx=1; dx <= L; ++dx) {	
	  if (x + dx > world.x) break;
	  sint4 j;
	  FOR (j, dy) {
	    sint4 i;
	    FOR (i, dx) {
	      sint4 ind = ind0 + j*world.x + i;
	      if (covered.height[ind] || (world.height[ind] != h0)) break;
	    }
	    if (i < dx) break;
	  }

	  if (j >= dy && dx*dy*LL-(dx+dy) > max_score) {
	    max_score = dx*dy*LL-(dx+dy);
	    mdx = dx; mdy = dy;
	  }
	}
      }

      if (max_score > 0) { d.x = mdx; d.z = mdy; }

      // mark rectangle as being covered
      FORS (i, d.x) {
	FORS (j, d.z) covered.height[ind0 + j*world.x + i] = 1;
      }
#endif
      
      float f = 0.75 + 0.25 * h0 / MAX_Z;
      Col3  ct(f, f, f);

      GfxBox *go = new GfxBox;

      d.x += 0.01; d.z += 0.01; // enlarge size a bit to avoid gfx artifacts
      
      go->origin = o;
      go->dim    = d;
      go->cb     = cb;
      go->ct     = ct;

      gfx_objs.push_back(go);
    }
  }

  // material determines color if light is on!
  // surface normals are then have to be provided, too!

  GfxSphere *go = new GfxSphere;

  go->origin = Coor3(0, 5, 0);
  go->radius = 2;
  go->col    = Col3(1,0,0);

  gfx_objs.push_back(go);

  // create octree sectors

  FORS (i, SEC_LEVELS) {
    sint4 n = 1;
    FORS (j, i) n *= 8;
    sectors[i] = new Sector[n];
    sectors_n[i] = n;
    cout << i << " " << n << " sectors" << endl;
  }

  // determine sectors for all objects

  // game volume, everything lives in it
  
  XMIN = -NX/2;
  XMAX = +NX/2;
  YMIN = -NZ;
  YMAX = +NZ;
  ZMIN = -NY/2;
  ZMAX = +NY/2;
  
  FORALL (gfx_objs, gob) {
    Coor3 o, d;
    (*gob)->aabb(o, d); // get axis aligned bounding box

    // compute sector ranges in three dimensions

    sint4 shift = SEC_LEVELS-1;    
    sint4 ndim = 1 << shift; 
    
    sint4 xi_min, xi_max;
    index_range(o.x, o.x+d.x, XMIN, XMAX, ndim, xi_min, xi_max);

    sint4 yi_min, yi_max;    
    index_range(o.y, o.y+d.y, YMIN, YMAX, ndim, yi_min, yi_max);

    sint4 zi_min, zi_max;    
    index_range(o.z, o.z+d.z, ZMIN, ZMAX, ndim, zi_min, zi_max);    

    // put pointer to object into all intersecting cubes
    for (sint4 x=xi_min; x <= xi_max; ++x) {
      for (sint4 y=yi_min; y <= yi_max; ++y) {
	for (sint4 z=zi_min; z <= zi_max; ++z) {

	  assert(ndim == 1 << shift);
	  
	  sint4 ind = (((z << shift) + y) << shift) + x;
	  assert(ind >= 0 && ind < sectors_n[SEC_LEVELS-1]);
	  sectors[SEC_LEVELS-1][ind].used = true;
	  sectors[SEC_LEVELS-1][ind].objs.push_back(*gob);

	  // mark path to root as being used
	  sint4 x2 = x, y2 = y, z2 = z, sh2 = shift;;
	  for (sint4 l = SEC_LEVELS-2; l >= 0; --l) {
	    --sh2;
	    x2 >>= 1; y2 >>= 1; z2 >>= 1;
	    sint4 ind2 = (((z2 << sh2) + y2) << sh2) + x2;
	    assert(ind2 >= 0 && ind2 < sectors_n[l]);
	    if (sectors[l][ind2].used) break;
	    sectors[l][ind2].used = true;
	  }
	}
      }
    }
  }
}

#if 0

// Egoboo, Copyright (C) 2000 Aaron Bishop

sint4 load_texture(const char *filename, GLuint &id, sint4 &w, sint4 &h)
{
  SDL_Surface *tempSurface, *imageSurface;
  
  /* Load the bitmap into an SDL_Surface */
  imageSurface = SDL_LoadBMP(filename);

  /* Make sure a valid SDL_Surface was returned */
  if (imageSurface == NULL) return 1;

  /* Generate an OpenGL texture */
  glGenTextures(1, &id);

  /* Set up some parameters for the format of the OpenGL texture */
  /* Bind Our Texture */
  glBindTexture(GL_TEXTURE_2D, id);

  /* Set the original image's size (incase it's not an exact square of a power of two) */
  w = imageSurface->h;
  h = imageSurface->w;

  /* Determine the correct power of two greater than or equal to the original image's size */
  sint4 txDimWidth = 2;
  while(txDimWidth < w) txDimWidth *= 2;

  sint4 txDimHeight = 2;
  while(txDimHeight < h) txDimHeight *= 2;

  txDimWidth = w;
  txDimHeight = h;
  
  // Set the texture's alpha
  //texture->alpha = 1;

  /* Create a blank SDL_Surface (of the smallest size to fit the image) & copy imageSurface into it*/
  if (imageSurface->format->Gmask) {
    tempSurface = SDL_CreateRGBSurface(SDL_SWSURFACE, txDimWidth, txDimHeight, 24,
				       imageSurface->format->Rmask, imageSurface->format->Gmask,
				       imageSurface->format->Bmask, imageSurface->format->Amask );
  } else {
    
#ifdef _LITTLE_ENDIAN

    // Intel-style ordering
    tempSurface = SDL_CreateRGBSurface( SDL_SWSURFACE, txDimWidth, txDimHeight, 24, 0x0000FF, 0x00FF00, 0xFF0000, 0x000000);
#else
    // PowerPC-style ordering
    tempSurface = SDL_CreateRGBSurface( SDL_SWSURFACE, txDimWidth, txDimHeight, 24, 0xFF0000, 0x00FF00, 0x0000FF, 0x000000);
#endif
  }
      
  SDL_BlitSurface(imageSurface, &imageSurface->clip_rect, tempSurface, &imageSurface->clip_rect );

#if 1

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, tempSurface->w, tempSurface->h, 0, GL_RGB, GL_UNSIGNED_BYTE, tempSurface->pixels);

#else

  gluBuild2DMipmaps(GL_TEXTURE_2D, 3, tempSurface->w, tempSurface->h, GL_BGR, GL_UNSIGNED_BYTE, tempSurface->pixels);

#endif
  
  /* get rid of our SDL_Surfaces now that we're done with them */
  SDL_FreeSurface(tempSurface);
  SDL_FreeSurface(imageSurface);
  return 0;
}

#else

// function to load in bitmap as a GL texture

sint4 load_texture(const char *file, GLuint &texture, sint4 &w, sint4 &h)
{
  // Status indicator
  sint4 status = 1;

  // Create storage space for the texture
  SDL_Surface *TextureImage[1]; 

  // Load The Bitmap, Check For Errors, If Bitmap's Not Found Quit
  if ((TextureImage[0] = SDL_LoadBMP(file))) {

    w = TextureImage[0]->w;
    h = TextureImage[0]->h;
    
    // Set the status to true
    status = 0;

    // Create The Texture
    glGenTextures(1, &texture);

    // Typical Texture Generation Using Data From The Bitmap
    glBindTexture(GL_TEXTURE_2D, texture);

    // Generate The Texture

    gluBuild2DMipmaps(GL_TEXTURE_2D, 3, TextureImage[0]->w,
		      TextureImage[0]->h, GL_BGR,
		      GL_UNSIGNED_BYTE, TextureImage[0]->pixels );

    // Linear Filtering
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST );
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
  }

  // Free up any memory we may have used
  if (TextureImage[0]) SDL_FreeSurface(TextureImage[0]);

  return status;
}
#endif

// find delta in world coordinates that lead to window change (dx,dy)

static void change_offset(float dx, float dy)
{
  // dy correction due to x_angle  
  double f = sin((x_angle/180)*M_PI);
  if (f >= 0 && f < 0.5) f = 0.5;    // limit correction
  else if (f >= -0.5 && f < 0) f = -0.5;
  dy /= f; 

  // inverse rotation
  double phi = (y_angle/180)*M_PI;
  double c = cos(phi);
  double s = sin(phi);
  x_off -= dx*c+dy*s;
  y_off += (-dx*s+dy*c);
}


// function to handle key press events

void handleKeyPress(SDL_keysym *keysym)
{
  switch (keysym->sym) {
  case SDLK_ESCAPE:
    // ESC key was pressed
    exit(0);
    break;
  case SDLK_LEFT:
    change_offset(-1,0); break;
  case SDLK_RIGHT:
    change_offset(1,0); break;    
  case SDLK_UP:
    change_offset(0,1); break;
  case SDLK_DOWN:
    change_offset(0,-1); break;
  case SDLK_F1:
    /* F1 key was pressed
     * this toggles fullscreen mode
     */
    SDL_WM_ToggleFullScreen( surface );
    break;

  case ',':
    y_angle -= 1; if (y_angle < 0) y_angle += 360;
    break;

  case '.':    
    y_angle += 1; if (y_angle >= 360) y_angle -= 360;
    break;

  case 'z':
    x_angle -= 1; if (x_angle < 0) x_angle += 360;
    break;

  case 'x':
    x_angle += 1; if (x_angle >= 360) x_angle -= 360;
    break;

  case '1':
    lx_angle -= 1; if (lx_angle < 0) lx_angle += 360;
    break;

  case '2':
    lx_angle += 1; if (lx_angle >= 360) lx_angle -= 360;
    break;

  case '3':
    lz_angle -= 1; if (lz_angle < 0) lz_angle += 360;
    break;

  case '4':
    lz_angle += 1; if (lz_angle >= 360) lz_angle -= 360;
    break;

  case '[':
    scale -= 0.0152625; break;

  case ']':    
    scale += 0.0152625; break;
    
  default:
    return; 
    break;
  }

  // view changed

#if 1	
  cout << int(keysym->sym)
       << " " << x_angle
       << " " << y_angle
       << " " << x_off
       << " " << y_off	      
       << " " << scale
       << endl;
#endif	
  
  return;
}


int main(int, char **)
{
  window_width  = SIZE_X;
  window_height = SIZE_Y;

  sint4 last_id = -1;
  Col3  last_col;
  GfxObj *last_obj = 0;
  
  // Flags to pass to SDL_SetVideoMode
  sint4 videoFlags;
  
  // main loop variable
  bool done = false;
  
  // used to collect events
  SDL_Event event;
  
  // this holds some info about our display
  const SDL_VideoInfo *videoInfo;
  
  // whether or not the window is active
  bool is_active = true;

  // initialize SDL
  if (SDL_Init(SDL_INIT_VIDEO) < 0) ERR("Video initialization failed");

  // Fetch the video info
  videoInfo = SDL_GetVideoInfo( );

  if (!videoInfo) ERR("Video query failed");

  // the flags to pass to SDL_SetVideoMode
  videoFlags  = SDL_OPENGL;          // Enable OpenGL in SDL
  videoFlags |= SDL_GL_DOUBLEBUFFER; // Enable double buffering
  videoFlags |= SDL_HWPALETTE;       // Store the palette in hardware
  videoFlags |= SDL_RESIZABLE;       // Enable window resizing

  // This checks to see if surfaces can be stored in memory
  if (videoInfo->hw_available) 
    videoFlags |= SDL_HWSURFACE;
  else
    videoFlags |= SDL_SWSURFACE;

  // This checks if hardware blits can be done
  if (videoInfo->blit_hw) videoFlags |= SDL_HWACCEL;

  // Sets up OpenGL double buffering
  SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

  // get a SDL surface
  surface = SDL_SetVideoMode(window_width, window_height, BPP, videoFlags);

  // Verify there is a surface
  if (!surface) ERR("Video mode set failed");

  // Enable key press
  if ((SDL_EnableKeyRepeat( 100, SDL_DEFAULT_REPEAT_INTERVAL)))
    ERR("Setting keyboard repeat failed");
  
  SDL_WM_SetCaption("glxcubes", 0);

#if USE_TEXTURES
  
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
 
  sint4 tx, ty;
  if (load_texture(TILE_TEXTURE_FILE, textures[0], tx, ty))
    ERR("can't find bmp file "TILE_TEXTURE_FILE);

  glBindTexture(GL_TEXTURE_2D, textures[0]);

  INV_TX = 1.0/TX;
  TZ = TX*(float(ty)/tx);
  INV_TZ = 1.0/TZ;
  cout << "tx=" << tx << " ty=" << ty << " TX=" << TX << " TZ=" << TZ << endl;

  if (load_texture(TILE_TEXTURE_FILE_2, textures[1], tx, ty))
    ERR("can't find bmp file "TILE_TEXTURE_FILE_2);

  create_minimap_tex();

#endif

  
  cout << "[ ]        - zoom" << endl
       << "a z        - turn left,right" << endl
       << ", .        - turn up,down" << endl
       << "up down    - move up,down" << endl
       << "left right - move left,right" << endl
       << "12 34      - sun motion" << endl;
  
  init();
  
  draw_all();

  // wait for events
  while (!done) {
    
    // handle the events in the queue

    while (SDL_PollEvent(&event)) {

      switch(event.type) {
	
      case SDL_ACTIVEEVENT:
	/* Something's happend with our focus
	 * If we lost focus or we are iconified, we
	 * shouldn't draw the screen
	 */
	is_active = event.active.gain != 0;
	break;                          

      case SDL_VIDEORESIZE:
	// handle resize event
	surface = SDL_SetVideoMode(event.resize.w, event.resize.h, 32, videoFlags);
	if (!surface) ERR("Could not get a surface after resize");
	window_width = event.resize.w;
	window_height = event.resize.h;
	draw_all();
	break;
	
      case SDL_KEYDOWN:
	// handle key presses
	handleKeyPress(&event.key.keysym);
	break;
	
      case SDL_QUIT:
	// handle quit requests
	done = true;
	break;

      case SDL_MOUSEBUTTONDOWN:

	{
	  cout << "mouse down at " << event.button.x << " " << event.button.y << endl;

	  float world_x, world_z;
	  
	  sint4 id = pick(event.button.x, event.button.y, world_x, world_z);

	  if (id > 0) {

	    // clicked on object

	    cout << "picked: "
		 << (id >> GfxObj::SUB_ID_LOG) << ":" << (id & ((1<<GfxObj::SUB_ID_LOG)-1))
		 << " " << world_x << " " << world_z << endl;

	    if (last_id > 0) {

	      // restore color of previous object
	    
	      GfxBox *pb = dynamic_cast<GfxBox*>(last_obj);
	      if (pb) pb->ct = last_col;
	      else {
		GfxSphere *ps = dynamic_cast<GfxSphere*>(last_obj);
		if (ps) ps->col = last_col;
	      }
	    }

	    last_id = id;
	    last_obj = GfxObj::id2obj[id >> GfxObj::SUB_ID_LOG];
	    Col3 new_col(0,1,0);
	  
	    // set new color for clicked object
	    
	    GfxBox *pb = dynamic_cast<GfxBox*>(last_obj);
	    if (pb) { last_col = pb->ct; pb->ct = new_col; }
	    else {
	      GfxSphere *ps = dynamic_cast<GfxSphere*>(last_obj);
	      if (ps) { last_col = ps->col; ps->col = new_col; }
	    }

	    draw_all();
	  }
	}
	break;
	
      default:
	break;
      }
    }

    // draw the scene
    if (1 || is_active) {

      draw_all();

      // sun motion
      lx_angle += 2;
      
      
      // calc frame rate
      {
	static sint4 t0 = -1;
	static sint4 frames = 0;
	sint4 t = current_time();
	
	frames++;
	
	if (t0 < 0) t0 = t;
	
	if (t - t0 >= 3.0) {
	  GLfloat seconds = t - t0;
	  GLfloat fps = frames / seconds;
	  printf("%.1f FPS (%d frames/%.1f sec) %.0f objs/frame %.0f checks\n",
		 fps, frames, seconds, float(draw_count)/frames,
		 float(aabb_count)/frames);
	  draw_count = 0;
	  aabb_count = 0;
	  t0 = t;
	  frames = 0;
	}
      }
    }
  }

  return 0;
}
