// $Id: Game1_SimpleTerrain.C,v 1.6 2006/06/15 20:28:56 ddeutscher Exp $

// This is an ORTS file 
// (c) Michael Buro, Sami Wagiaalla, David Deutscher
// licensed under the GPL

#include "Game1_SimpleTerrain.H"
#include "Game1_ST_PFEngine.H"
#include "Game1_ST_BasicMove.H"
#include "Game1_ST_ForceFieldMove.H"
#include "Options.H"
#include "GameObj.H"
#include <utility>
#include <iterator>
#include <math.h>

using namespace std;

//#define TEST(x)
#define TEST(x) x
#define TEST2(x)
//#define TEST2(x) x


namespace Game1_SimpleTerrain {

  const uint4 ST_Terrain::MAXIMUM_TASK_AGE = 250;          // milliseconds

  //-----------------------------------------------------------------------------
  static Loc get_center_loc(const Object *obj)
  {
    Loc l;
    obj->get_center(l.x, l.y);
    return l;
  }
#if 0
  static ScalarPoint get_center_sp(const Object *obj)
  {
    ScalarPoint l;
    obj->get_center(l.x, l.y);
    return l;
  }
#endif
  static real8 path_length(Path &path)
  {
    if( path.locs.size() < 2 )
      return 0;
    real8 total = 0;
    for( Vector<Loc>::iterator i = ++path.locs.begin() ; i != path.locs.end() ; i++ )
      total += i->distance(*(i-1));
    return total;
  }
    
  // Assuming path is in the usual reversed order, with goal at locs[0]
  static real8 path_length_from(Path &path, const Loc& start_loc)
  {
    if( path.locs.size() < 1 ) 
      return 0;
    return start_loc.distance(path.locs.back()) + path_length(path);
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::add_options()
  {
    ::Options o("SimpleTerrain");
    o.put("-costonly", "represent boundaries only by higher edge cost");
    o.put("-highres", "use a high resolution world map (slower but more accurate)");
    MoveEngine::add_options();
  }

  //-----------------------------------------------------------------------------
  ST_Terrain::ST_Terrain() 
    : currentLevelOfDetail(0), find_path_max_time(0)
  {
    sint4 seed = 0;
    ::Options::get("-seed", seed);
    rand.srand(seed);
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::init(sint4 tiles_x_, sint4 tiles_y_, sint4 tile_points_, sint4 client_pID_, sint4 neutral_pID_)
  {
    TerrainBasicImp<ST_Task>::init(tiles_x_, tiles_y_, tile_points_, client_pID_, neutral_pID_);
    
    // initialize the internal PFEngine 
    // (if changing here, keep "useLevelOfDetail()" appropiately synchronized)
    Vector<sint4> levels_of_detail;
    bool high_resolution_only;
    Options::get("-highres", high_resolution_only);
    if( !high_resolution_only ) {
      levels_of_detail.push_back(4);  // default - use a resolution reduced by a factor of 4.
    }
    levels_of_detail.push_back(1);  // also (or exclusively) maintain a full resolution view.
    pfEngine.reset(new PFEngine(tiles_x_, tiles_y_, tile_points_, levels_of_detail));

    // initialize the internal movement engine
    moveEngine.reset(new MoveEngine(tiles_x_, tiles_y_, tile_points_, client_pID_, neutral_pID_));
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::useLevelOfDetail(LevelOfDetail lod)
  {
    if( lod == MINIMAL_DETAILS || lod < 0 )
      currentLevelOfDetail = 0;
    else {
      bool high_resolution_only;
      Options::get("-highres", high_resolution_only);
      if( !high_resolution_only ) {
        currentLevelOfDetail = 1;
      }
    }
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::add_obj(const Object *obj)
  {
    TerrainBasicImp<ST_Task>::add_obj(obj);
    assert(pfEngine!=NULL);
    pfEngine->insert_object(obj);
    moveEngine->insert_object(obj);
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::update_obj(const Object *obj)
  {
    TerrainBasicImp<ST_Task>::update_obj(obj);
    assert(pfEngine!=NULL);
    pfEngine->update_object(obj);
    moveEngine->update_object(obj);
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::remove_obj(const Object *obj)
  {
    // shorthand
    typedef ObjIsGoalInTaskId::iterator OGIT;

    // remove from world representation
    assert(pfEngine!=NULL);
    pfEngine->remove_object(obj);
    moveEngine->remove_object(obj);

    // If we intended to stop it, we don't need to anymore
    objsToStop.erase(obj);

    // If this is the target of any existing task...
    pair<OGIT,OGIT> range = objIsGoalInTaskId.equal_range(obj);
    for( OGIT i = range.first ; i != range.second ; i++ ) {
      // Change the target of the task to the last known location of the object
      FIND(taskId2task, ti2tIt, i->second);
      assert(ti2tIt != taskId2task.end());
      ti2tIt->second->task.goal.target = Goal::LOCATION;
      ti2tIt->second->task.goal.loc = get_center_loc(ti2tIt->second->task.goal.obj);
    }

    TerrainBasicImp<ST_Task>::remove_obj(obj);
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::add_segments(const Vector<Segment> &segs)
  {
    TerrainBasicImp<ST_Task>::add_segments(segs);
    assert(pfEngine!=NULL);
    FORALL(segs,i) {
      pfEngine->insert_boundary(*i);
      moveEngine->insert_boundary(*i);
    }
    invalidate_paths(segs);
  }

  //----------------------------------------------------------------------------------
  void ST_Terrain::cancel_task(const Object *obj)
  {
    objsToStop.insert(obj);
    TerrainBasicImp<ST_Task>::cancel_task(obj);
  }

  //----------------------------------------------------------------------------------
  void ST_Terrain::remove_task(TaskId tid)
  {
    FIND (taskId2task, it, tid);
    if (it != taskId2task.end() ) {
      Task& task = it->second->task;
      objsToStop.insert(task.objs.begin(), task.objs.end());
    }
    TerrainBasicImp<ST_Task>::remove_task(tid);
  }

  //----------------------------------------------------------------------------------
  real8 ST_Terrain::find_path(const Loc &/*l1*/, const Loc &/*l2*/, sint4 /*radius*/, Path * /*path*/, ConsiderObjects /*consider*/)
  {
    assert(0); // unimplemented yet    
    return -1;
  }

  //----------------------------------------------------------------------------------
  real8 ST_Terrain::find_path(const Object *obj, const Goal &goal, Path *path, ConsiderObjects consider)
  {
    assert( goal.mode == Goal::TOUCH || goal.mode == Goal::VICINITY && 
            "ATTACK mode is currently unimplemented");
    assert( (goal.target == Goal::LOCATION || goal.target == Goal::OBJ) && 
            "Unknown Goal::target type");

    uint4 stop_time = 0;
    if( find_path_max_time > 0 ) {
      stop_time = SDL_GetTicks() + find_path_max_time;
    }

    sint4 distance = (goal.mode == Goal::VICINITY)? goal.distance : 0;
    Path result;
    bool found;
    if (goal.target == Goal::LOCATION) {
      found = pfEngine->find_path(obj, goal.loc, result, distance, currentLevelOfDetail, consider, stop_time);
    } else {
      found = pfEngine->find_path(obj, goal.obj, result, distance, currentLevelOfDetail, consider, stop_time);
    }

    if( !found )
      return -1;

    // need to return waypoints in a 'forward' order
    if( path != NULL ) {
      path->locs.clear();
      reverse_copy(result.locs.begin(), result.locs.end(), back_inserter(path->locs));
    }

    // compute and return the length of the path
    return path_length_from(result, get_center_loc(obj));
  }

  //----------------------------------------------------------------------------------
  void ST_Terrain::block_path(const Object *obj, const Path &path, bool with_endpoints)
  {
    assert(pfEngine!=NULL);
    pfEngine->ir_path(obj, path, true, with_endpoints);
  }

  //----------------------------------------------------------------------------------
  void ST_Terrain::release_path(const Object *obj, const Path &path, bool with_endpoints)
  {
    assert(pfEngine!=NULL);
    pfEngine->ir_path(obj, path, false, with_endpoints);
  }

  //----------------------------------------------------------------------------------
  TaskId ST_Terrain::execute_task(const Task &task, const Path &path)
  {
    TaskId tid = add_task(task);
    TIter ti = taskId2task[tid];
    // mark the task as not pending
    ti->isPending = false;
    eraseFromPending(ti);
    // and insert the requested path to execution
    Vector<Loc> reversed;
    reverse_copy(path.locs.begin(), path.locs.end(), back_inserter(reversed));
    ti->set_path(*task.objs.begin(), reversed);
    return tid;
  }

  //----------------------------------------------------------------------------------
  // fixme: we only check the path of the *center* of the unit vs the new boundaries,
  // and we should account for its size (radius).
  void ST_Terrain::invalidate_paths(const Vector<Segment> &segments)
  {
    FORALL(tasks, ti) {
      if (!ti->isValid() ) continue;
      // fixme: better handling of zcat != ON_LAND ?
      if (ti->path->obj->get_zcat() != Object::ON_LAND ) continue;

      FORALL(segments, si) {
        if (ti->path->is_intersect(*si) ) {
          // this path is invalid, mark as pending for (re)planning
          pendingTasks.push_back(ti);
          ti->isPending = true;
          cout << "Marked an invalid path" << endl;
          break;
        }
      }
    }
  }

  void ST_Terrain::execute_tasks(Vector<MoveCmd> &cmds, Vector<StatusMsg> &msgs)
  {
    moveEngine->execute_tasks(tasks, objsToStop, cmds, msgs);

    // remove finished tasks
    FORALL (msgs, t) {
      clean_task(t->task_id);
    }
  }


  //----------------------------------------------------------------------------------
  bool ST_Terrain::plan_tasks(uint4 max_time)
  {
    TEST2(cout << "plan_tasks - max_time is : " << max_time << endl);
    uint4 stop_time = 0;
    if(max_time > 0){
      stop_time = SDL_GetTicks() + max_time;
    }
   
    // try to plan one failed and one pending tasks
    bool did_something = plan_failed_task(stop_time);
    did_something = plan_pending_task(stop_time) || did_something;

    // nothing more to plan at this time
    if( !did_something )
      return false;

    // Continue planning pending and failed alternately, 
    // until we have reached max_time or nothing more to plan
    bool plan_pending = true, did_pending = true, did_failed = true;
    while( (stop_time == 0 || SDL_GetTicks() < stop_time) && 
           (did_pending || did_failed) )
    {
      if( plan_pending ) { 
        did_pending = plan_pending_task(stop_time); 
      } else {
        did_failed = plan_failed_task(stop_time); 
      }

      TEST( 
        if( (plan_pending && did_pending) || (!plan_pending && did_failed) )
          cout << "Current time is : " << SDL_GetTicks() << " and stop_time = " << stop_time << endl; );

      plan_pending = !plan_pending;
    }

#if 0
    bool did_some_more = plan_pending_task();
    sint4 counter = 0;
    //Continue planning until we have reached the max_time    
    while( (sint4)SDL_GetTicks() < stop_time && (!pendingTasks.empty() || !tasks.empty() )) {
      if(counter++ % 2 == 0)
	did_something = plan_failed_task( stop_time - (sint4)SDL_GetTicks() );
      else
	did_some_more = plan_pending_task( stop_time - (sint4)SDL_GetTicks() );

    TEST2(cout << "Current time is : " << (sint4)SDL_GetTicks() << " and stop_time = " << stop_time << endl);
 
      did_some_more = did_something || did_some_more; // careful not to short-circuit!
    }

    did_something = did_some_more || did_something; 
#endif
    return did_something;
  }

  //----------------------------------------------------------------------------------
  bool ST_Terrain::plan_pending_task(uint4 stop_time)
  {
    if (pendingTasks.empty()) {
      TEST(cout << "No pending task to plan" << endl;);
      return false;
    }

    // TIter tit = pendingTasks.front(); pendingTasks.pop_front();

    std::deque<TIter>::iterator temp = get_best_task();
    TIter tit = *temp;
    pendingTasks.erase(temp);

    assert(tit->task.objs.size() > 0);
    const Object *obj = *(tit->task.objs.begin());

    // if the object is moving, stop it first and come back to it later
/*  problem here: if everybody is moving, we keep checking until max_time
    runs out in plan_tasks(); if we return false here - we don't plan the
    other tasks.
    if (obj->get_speed() > 0 || obj->get_moving()) {
      objsToStop.insert(obj);
      pendingTasks.push_back(tit);
      return true; // report we did something, so we'll get a chance to try another task
    }*/

    // mark the task as handled
    tit->isPending = false;

    // plan
    TEST(cout << obj << " planning pending path" << endl);
    plan_task(tit, stop_time);
    return true;
  }

  //-------------------------------------------------------------------
  struct FindFailed {
    bool operator()(ST_Task &task) { return task.isFailed && !task.noPathFound; }; 
  };

  //-------------------------------------------------------------------
  ST_Terrain::TIter ST_Terrain::get_random_failed_task(void)
  {
    // find a failed path for which we don't yet know that 'noPathFound'
    // (ie we still didn't try to replan and found that there is no path)
    if( tasks.empty() )
      return tasks.end();

    // start the search with in a random position to avoid planning the same
    // path repeatedly (in case it is immediately marked as failed again).
    int skip = rand(tasks.size());
    TIter start_at = tasks.begin();
    advance(start_at, skip);

    // search from the middle to the end
    TIter tit = find_if(start_at, tasks.end(), FindFailed());
    // if found
    if( tit != tasks.end() )
      return tit;
    // else, search from start to middle
    tit = find_if(tasks.begin(), start_at, FindFailed());
    return (tit != start_at)? tit : tasks.end();
  }

  //-------------------------------------------------------------------
  bool ST_Terrain::plan_failed_task(uint4 stop_time)
  {
    assert(pfEngine!=NULL);

    // get a failed task
    TIter tit = get_random_failed_task();
    if( tit == tasks.end() ) {
      TEST(cout << "No failed task to plan" << endl;);
      return false;
    }
  
    assert(tit->task.objs.size() > 0);
    const Object *obj = *(tit->task.objs.begin());

    // if the object is moving, stop it first and come back to it later
/*  problem here: if everybody is moving, we keep checking until max_time
    runs out in plan_tasks(); if we return false here - we don't plan the
    other tasks.
    if( obj->get_speed() > 0 || obj->get_moving() ) {
      objsToStop.insert(obj);
      return true; // report we did something, so we'll get a chance to try another task
    }*/

    tit->replanned++;

    // find-path
    TEST(cout << obj << " replanning path" << endl);
    plan_task(tit, stop_time);
    return true;
  }

  //-------------------------------------------------------------------
  void ST_Terrain::plan_task(TIter tit, sint4 stop_time)
  {
    Path newpath;
    bool found;
    
    // Object to move
    const Object *obj = *(tit->task.objs.begin());
    // distance from target allows
    sint4 distance = (tit->task.goal.mode == Goal::VICINITY)? tit->task.goal.distance : 0;
    if (tit->task.goal.target == Goal::LOCATION) {
      found = pfEngine->find_path(obj, tit->task.goal.loc, newpath, distance, currentLevelOfDetail, CONSIDER_ALL, stop_time);
    } else {
      assert (tit->task.goal.target == Goal::OBJ);
/*      assert (!pushing_on || 
	what_team(tit->task.goal.obj) != client_pID ||
	tit->task.goal.mode != Goal::TOUCH ||
	tit->task.goal.obj->get_max_speed() <= 0); //because you'll push the object out of the way*/
      found = pfEngine->find_path(obj, tit->task.goal.obj, newpath, distance, currentLevelOfDetail, CONSIDER_ALL, stop_time);
    }

    // insert results into data structures
    if( found ) {
      assert(newpath.locs.size() > 0);
      tit->set_path(obj, newpath.locs);
      tit->isFailed = false;
    }
    else {
      // fixme: maybe add to "failed"/sleep etc. ?
      tit->noPathFound = true;
    }
  }

  //-------------------------------------------------------------------
  static bool is_younger(const TerrainBasicImp<ST_Task>::TIter &a, const TerrainBasicImp<ST_Task>::TIter &b) {
    return a->start_time > b->start_time;
  }

  static bool is_shorter(const TerrainBasicImp<ST_Task>::TIter &a, const TerrainBasicImp<ST_Task>::TIter &b) {
    return a->path_dist < b->path_dist; 
  }

  std::deque<ST_Terrain::TIter>::iterator ST_Terrain::get_best_task(void)
  {    
    // Check if the oldest task must be planned now
    std::deque<TIter>::iterator oldest = max_element(pendingTasks.begin(), 
                                                     pendingTasks.end(),
                                                     is_younger);
    if( SDL_GetTicks() - (*oldest)->start_time >= MAXIMUM_TASK_AGE )
      return oldest;

    // else, return the 'shortest' (straight-line) path
    return min_element(pendingTasks.begin(), pendingTasks.end(), is_shorter);
  }

  //----------------------------------------------------------------------------------
  void ST_Terrain::ignore_rectangles(void) { 
    pfEngine->ignore_rectangles(); 
  }

  void ST_Terrain::consider_rectangles(void) { 
    pfEngine->consider_rectangles(); 
  }

  //----------------------------------------------------------------------------------
  //----------------------------------------------------------------------------------

#define EPS 1e-9f


  UnitPath::UnitPath(const Object *obj_, const Vector<Loc> &path_)
    : obj(obj_), path(path_), no_orders_yet(true), 
      stopped_to_avoid_collision(false), collision_step(0), wakeup(0)
  {
    current_target = path.size()-1;
    if (path.size() > 0) {
      goal = path[0];
    }
    spot = Loc(-1, -1);
  }


  bool UnitPath::is_intersect(const Segment &segment) const
  {
    typeof(path.begin()) pj = path.begin();
    typeof(path.begin()) pi = pj++;
    for( ; pj != path.end() ; pi++, pj++ ) {
      if (Segment(*pi, *pj).touches(segment) )
        return true;
    }
    return false;
  }


  //----------------------------------------------------------------------------------
  //----------------------------------------------------------------------------------

  ST_Task::ST_Task(const Task& task_)
    : task(task_), isPending(true), isFailed(false), noPathFound(false), replanned(0),
      start_time(SDL_GetTicks())
  {
    Loc start = get_center_loc(*task.objs.begin());
    if( task.goal.target == Goal::OBJ )
      path_dist = start.distance(get_center_loc(task.goal.obj));
    else if( task.goal.target == Goal::LOCATION )
      path_dist = start.distance(task.goal.loc);
    else
      ERR("Unknown Task::target value");
  }

  void ST_Task::set_path(const Object *obj_, const Vector<Loc> &path_)
  {
    // overwrite an existing path, retaining the collision-monitoring data.
    if( path ) {
      Loc   saved_spot = path->spot;
      int   saved_collision_step = path->collision_step;
      bool  saved_stopped_to_avoid_collision = path->stopped_to_avoid_collision;
      uint4 saved_wakeup = path->wakeup;

      path = boost::shared_ptr<UnitPath>(new UnitPath(obj_, path_));

      path->spot = saved_spot;
      path->collision_step = saved_collision_step;
      path->stopped_to_avoid_collision = saved_stopped_to_avoid_collision;
      path->wakeup = saved_wakeup;
    } 
    // No previous path was planned, so just create a new one.
    else {
       path = boost::shared_ptr<UnitPath>(new UnitPath(obj_, path_));
    }
  }

}   // End of namespace SimpleTerrain
