// Trial client for AIIDE ORTS tournament's game1
// Written by David Deutscher, using sample ORTS code

#define GAME1_NO_GFX // also define GAME1_SIMPLE_TERRAIN_NO_GFX in Game1_SimpleTerrain.H

#include "Global.H"
#include "Options.H"
#include "GameStateModule.H"
#include "ShutDown.H"
#include "SDLinit.H"
#include "Game.H"
#include "ClientState.H"
#include "LoopFunctor.H"

#include "TerrainModule.H"
#include "Game1_SimpleTerrain.H"
#include "AutoGatherAI.H"

#ifndef GAME1_NO_GFX
#include "GfxModule.H"
#include "Game1_SimpleTerrainWidget.H"
#endif // GAME1_NO_GFX

static char *VERSION = "ORTS AIIDE06 tournament - Game1 client by David Deutscher v0.1";

#if 0
boost::shared_ptr<GameStateModule> GlobalState::gsm;
boost::shared_ptr<AutoGatherAI>    GlobalState::ai;
bool GlobalState::quit;
bool GlobalState::error;
#endif

using namespace std;

// what terrain representation to use
typedef Game1_SimpleTerrain::ST_Terrain Terrain;

struct StatisticsViewHandler : public EventHandler
{
  GameStateModule &gsm;
  ofstream stats_file;

  StatisticsViewHandler(GameStateModule &gsm_)
    : gsm(gsm_)
  {
    gsm.add_handler(this);

    string fname;
    ::Options::get("-statsfile", fname);
    if( fname != "" )
      stats_file.open(fname.c_str());
  }

  ~StatisticsViewHandler()
  {
    gsm.remove_handler(this);
    if( stats_file ) stats_file.close();
  }

  bool handle_event(const Event& e)
  {
    if (e.get_who() == GameStateModule::FROM &&
	e.get_what() == GameStateModule::VIEW_MSG) {

      Game &game = gsm.get_game();	

      // frame statistics
      sint4 vf = game.get_view_frame();
      sint4 af = game.get_action_frame();
      sint4 sa = game.get_skipped_actions();
      sint4 ma = game.get_merged_actions();

//    if ((vf % 20) == 0) cout << "[frame " << vf << "]" << endl;
//    if (vf != af || sa || ma > 1) {

        cout << "frame " << vf;
        if (af < 0) cout << " [no action]";
        else if (vf != af) cout << " [behind by " << (vf - af) << " frame(s)]";
        if (sa) 
          cout << " [skipped " << sa << "]";
        if (ma > 1) 
          cout << " [merged " << ma << "]";
        cout << endl;
//    }

      // Report mineral count
      static int minerals = -1;
      PlayerInfo &pi = game.get_cplayer_info();
      ScriptObj *player = pi.global_obj("player");
      int cur_minerals = player->get_int("minerals");
      if( minerals != cur_minerals ) {
        cout << "*** Mineral count is now: " << cur_minerals << endl;
        minerals = cur_minerals;
      }

      // save timing statistics to file
      if( stats_file ) {
        static sint4 last_view_time = SDL_GetTicks();
        sint4 cur_view_time = SDL_GetTicks();
        stats_file << cur_view_time-last_view_time << endl;
        last_view_time = cur_view_time;
      }

      return true;
    }
    return false;
  }
};


#if 0
struct EndOfGameHandler : public EventHandler
{
  GameStateModule &gsm;

  EndOfGameHandler(GameStateModule &gsm_)
    : gsm(gsm_)
  {
    gsm.add_handler(this);
  }

  ~EndOfGameHandler()
  {
    gsm.remove_handler(this);
  }
  
  bool handle_event(const Event& e)
  {
    if (e.get_who() == GameStateModule::FROM) {
      if (e.get_what() == GameStateModule::STOP_MSG ||
	  e.get_what() == GameStateModule::FINISHED_MSG) {
        GlobalState::quit = true;
        return true;
      }
      if (e.get_what() == GameStateModule::READ_ERROR_MSG) {
        GlobalState::quit = GlobalState::error = true;
        return true;
      }
    }
    return false;
  }
};
#endif

#if 0
struct GfxEventHandler : public EventHandler
{
  ClientState &state;

  GfxEventHandler(ClientState &state_) : state(state_) { }

  bool handle_event(const Event& e)
  {
    if (e.get_who() == GameStateModule::FROM) {

      //cout << "Event: " << e.who() << " " << e.what() << endl;

      if (e.get_what() == GameStateModule::VIEW_MSG) {

#if 0
	if (state.gui) {
	  state.gui->event();
	  state.gui->display();
	  if (state.gui->quit) exit(0);
	}
#endif
	state.gfxm->process_changes();
      }
      return true;
    }
    return false;
  }
};
#endif



struct AIState
{
  GameStateModule *gsm;
  AutoGatherAI    *ai;

  bool just_drew;  // just drew a gfx frame
  bool quit;       // true => quit game
  bool error;      // true <=> error occurred

#ifndef GAME1_NO_GFX
  GfxModule       *gfxm; // !=0 : gfx used

  AIState(GameStateModule *gsm_, GfxModule *gfxm_, AutoGatherAI *ai_) :
    gsm(gsm_), ai(ai_), gfxm(gfxm_)
  {
    just_drew = quit = error = false;
  }
#else
  AIState(GameStateModule *gsm_, AutoGatherAI *ai_) :
    gsm(gsm_), ai(ai_)
  {
    just_drew = quit = error = false;
  }
#endif

};


//===================================================================

class Looper : public LoopFunctor
{
private:

  AIState &state;

public:

  Looper(AIState &s) : state(s) { }

  void loop() const
  {
#ifndef GAME1_NO_GFX
    if (state.gfxm) {
      if (state.gfxm && state.gfxm->get_quit()) exit(10);
    }
#endif

    if (state.gsm->error()) exit(20);
    if (state.gsm->exit()) exit(0);

    // server events

    if (!state.gsm->recv_view()) {

      //state.ai->act();
      SDL_Delay(1);
      
    } else {

      //cout << "**************** NOTHING TO READ ****************" << endl;
      //cout << "*";

#ifndef GAME1_NO_GFX
      if (state.gfxm) {
	state.gfxm->process_changes();
	state.gfxm->draw();
      }
#endif

      // Send an actions message (even an empty one) for each game tick, as required.
      state.gsm->send_actions();
    }
  }
};


int main(int argc, char **argv)
{
  try {
    signals_shut_down(true);
    
    Options opt("Game1Client");
#ifdef GAME1_NO_GFX
    opt.put("-seed", (sint4)0, "rng seed (0:time)");
#endif
    opt.put("-statsfile", string(""), "name of timing-statistics file");

    /*    opt.put("-cinfo", "display game state changes");
    opt.put("-tinfo", "display tile info");
    opt.put("-oinfo", "display object info");
    opt.put("-minfo", "display map info");    */


#ifndef GAME1_NO_GFX
    glutInit(&argc, argv);
    opt.put("-gfx",   "display world in 3d");    
    GfxModule::Options::add(); // added by MB
#endif

    // populate opt with options from various modules
    GameStateModule::Options::add();
    Terrain::add_options();
    AutoGatherAI::add_options();

    // process command line arguments
    if (Options::process(argc, argv, cerr, VERSION))
      exit(20);

    // init code copied from sampleai

    // change seed value to time(0) if 0
    sint4 seed;
    Options::get("-seed", seed);
    if (!seed) seed = time(0);
    Options::set("-seed", seed);
    cout << "seed=" << seed << endl;

    srand48(seed); // used by STL, fixme: avoid (global)

    SDLinit::video_init();
    SDLinit::network_init();

    // populate option objects with command line values
    // and create modules

    GameStateModule::Options gsmo;
    GameStateModule gsm(gsmo);

    // create pathfinding module and implementation
    Terrain timp;
    TerrainModule tm(gsm, timp, 110);
    // gsm.add_handler(&tm); done in TerrainModule now

    bool use_gfx = false;

#ifndef GAME1_NO_GFX
    GfxModule::Options gfxmo;
    GfxModule *gfxm = 0;

    Options::get("-gfx", use_gfx);

    if (use_gfx) {
      
      gfxm = new GfxModule;
      
      // hook up terrain widget to gfx module
      gfxm->add_cwidget(*new Game1_SimpleTerrain::SimpleTerrainWidget(timp, gfxm));
    }
#endif

    //=====================================
    // Simul::SimulationAI suicide_ai(gsm);
    // suicide_ai.add_handler(&tm);
    // gsm.add_handler(&suicide_ai);
    // tm.add_handler(&suicide_ai);

    //=====================================
    // GatherAI gather_ai(gsm);
    // gather_ai.add_handler(&tm);
    // gfxm.add_handler(&gather_ai);
    // gsm.add_handler(&gather_ai);
    // tm.add_handler(&gather_ai);

    //=====================================
    // ShooterAI shooter_ai(gsm);
    // shooter_ai.add_handler(&tm);
    // gfxm.add_handler(&shooter_ai);
    // gsm.add_handler(&shooter_ai);
    // tm.add_handler(&shooter_ai);
    // shooter_ai.enable();

    //=====================================
    AutoGatherAI auto_gather_ai(gsm, tm, timp);
    //auto_gather_ai.add_handler(&tm);
    //gfxm.add_handler(&auto_gather_ai);
    //gsm.add_handler(&auto_gather_ai);
    //tm.add_handler(&auto_gather_ai);
    // missing ? gsm.add_handler(terrain.get());             // gsm -> terrain

    // report frame statistics to the screen
    boost::shared_ptr<StatisticsViewHandler> svh(new StatisticsViewHandler(gsm));

    // Handle end-of-game messages (GameStateModule now has error() and quit()
    //boost::shared_ptr<EndOfGameHandler> eogh(new EndOfGameHandler(gsm));

    // connect game state module to server
    if (!gsm.connect()) ERR("connection problems");

    // init gfx and autogather modules
#ifndef GAME1_NO_GFX
    if (gfxm) 
      gfxm->init(gsm, gfxmo);
    AIState state(&gsm, gfxm, &auto_gather_ai);
#else
    AIState state(&gsm, &auto_gather_ai);
#endif


    // Run

    Looper l(state);

    if (use_gfx) {
#ifndef GAME1_NO_GFX
      // game loop (could become a thread)
      gfxm->start_loop(&l);
#endif

    } else {

      // no gfx

      while (!gsm.exit() && !gsm.error()) l.loop();
    }

  } catch( ExitException e ) {
    cerr << "Exit exception was thrown: " << *e.get_reason() << endl;
  } catch( ... ) {
    cerr << "Unhandled exception" << endl;
  }
  return 0;
}
