#ifndef SegmentMerge_H
#define SegmentMerge_H

// $Id: SegmentMerge.H,v 1.1 2006/05/12 04:18:01 demyen Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Global.H"
#include "DCDT.H"

struct SortSegment 
{
  sint4 q, p1;      // sort values: line position, first point on line
  mutable sint4 p2; // last point on line (order does not depend on it)

  mutable DCDT::Segment s; // original segment
  mutable sint4 label;     // < 0 <=> new segment
  
  SortSegment() : q(0), p1(0), p2(0), label(-1) { init(); }

  SortSegment(const DCDT::Segment &s_) : s(s_) { init(); }

  void init() { q = p1 = p2 = 0; label = -1; }
  
  void write(std::ostream &os = std::cout) const {
    s.write(os); 
    os << " " << label;
  }
};


struct SortSegmentCmp {

  bool operator()(const SortSegment &a, const SortSegment &b)
  {
    sint4 r1 = a.q - b.q;
    if (r1) return r1 < 0;
    return (a.p1 - b.p1) < 0;
  }

};


/*

  This class merges touching line segments of identical orientation
  into longer segments. It can be used to minimize the number of
  segments inserted into a dcdt.

  For each line type k (horizontal, vertical, diagonal+, diagonal-) it
  keeps track of all segments in sets all[k], which are sorted
  w.r.t. the line invariant (e.g. y value for horizontal lines) as
  first component, and the lower coordinate on the line as second
  component (e.g. the smaller x value for horizontal lines). Merging a
  new segment with existing ones is then "just" a matter of finding
  its location in all[k] (log time), looking at the predecessor and
  successor and taking the appropriate action.

  Before a sequence of segments can be inserted by calling new_segment
  (multiple times), clear_delta has to be called to reset add[]
  and del[].

  After inserting, add[] contains iterators to new - possibly merged -
  segments to be inserted in the dcdt. del[] contains ids of segements
  to be deleted.

  It is assumed that after adding the add[] segments, their label is
  set to >= 0 to indicate that it is an existing segment (<0
  internally indicates a new segment). It is also assumed that the
  user of this class can remove segments by id.
  
*/

class SegmentMerge
{
public:

  typedef std::set<SortSegment, SortSegmentCmp> SSet;
  typedef std::set<const SortSegment*> ISet;

  // line types: horizontal, vertical, diagonal (x+y constant), diagonal (x-y constant)
  enum Kind { H=0, V=1, DP=2, DM=3 };

  SSet all[4];          // sets of all line segments
  ISet add[4];          // sets of iterators pointing to new/updated segments in all
  Vector<sint4> del[4]; // vectors of ids of segments to be deleted from dcdt

  // has to be called before new_segment is called multiple times
  void clear_delta();

  // insert new segment and update all, add, and del for the
  // appropriate line type
  void new_segment(const DCDT::Segment &seg);
};

#endif
