#ifndef SquadCombatAI_H
#define SquadCombatAI_H

// $Id: SquadCombatAI.H,v 1.12 2006/06/13 23:31:42 anderson Exp $

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "Global.H"
#include "EventHandler.H"
#include "ClientModule.H"
#include "InfluenceMap.H"
#include "VectorField.H"
#include "TerrainVectorField.H"

class GameStateModule;
class GameChanges;
class GameObj;

//===================================================================

class SquadCombatAI : public EventHandler, public ClientModule
{
public:

  static void add_options();

  //------------------------------

  /** Top-level collection of related (allied) objects. */
  // Public only to allow REGISTER_TYPEOF definitions, needed e.g. in MSVC .NET 2003
  class GroupData {
  public:
    std::set<GameObj*> units;
    InfluenceMap attack_influence;
    InfluenceMap target_influence;
    InfluenceMap occupied_influence;

    VectorField attack_field;
    VectorField target_field;
    VectorField occupied_field;
    VectorField motion_field;

    GroupData(sint4 width, sint4 height);
  };

  //------------------------------

protected:

  //------------------------------

  /** Units with a common objective. */

  class Squad {
  public:

    enum Objective { ATTACK_TARGET, ATTACK_LOCATION, FREELANCE, MOVE_TO_POINT };

    std::set<GameObj*> units;
    sint4 x, y;
    GameObj *target;
    Objective objective;

    Squad() : x(0), y(0), target(0), objective(ATTACK_TARGET) {}
  };

  //------------------------------

  class ShooterData {
  public:
    sint4 x, y, r;
    sint4 min_damage;
    sint4 max_damage;
    sint4 avg_damage;
    sint4 range;
    sint4 hp;
    sint4 armor;
    sint4 time_to_shoot;
    Vector<GameObj*> outgoing; ///< potential targets
    Vector<GameObj*> incoming; ///< potential threats
    GameObj *cur_target;
    GameObj *old_target;

    Vec2<real4> dir; ///< Intended direction ignoring negative effects.
    Vec2<real4> ndir; ///< Normalized.
    real4 support;

    enum Mode { ACTIVE, PASSIVE };
    Mode mode;

    Squad *squad;

    ShooterData();
  };

  //------------------------------
public:
  // Public only to allow REGISTER_TYPEOF definitions, needed e.g. in MSVC .NET 2003
  typedef std::map<GameObj*, ShooterData> ShooterMap;

  //------------------------------

protected:

  GameStateModule &gsm;
  uint4 game_tick;     // game time
  sint4 width, height; // map size in tile points

  Vector<GroupData*> groups;
  GroupData *friendlies;
  GroupData *enemies;
  GroupData *neutrals;

  ShooterMap shooter_data;
  Vector<Squad*> squads;

  TerrainVectorField terrain_field;
  VectorField occupied_field;

  void update_units();
  void remove_obj(const GameObj *gob);

  GroupData* get_group(GameObj *gob);


  void compute_influence();
  void compute_influence(GroupData *gd);
  void compute_occupation(GroupData *gd);

  void compute_attack_graph();
  void compute_attack_graph(GroupData *ff, GroupData *ee);

  void compute_support(const std::set<GameObj*> &units);
  void compute_support2(const std::set<GameObj*> &units);

  real4 support(const Vec2<real4> &pos, const Vec2<real4> &dir,
                const Vec2<real4> &spos, const Vec2<real4> &sdir);
  real4 support(const Vec2<real4> &pos, const Vec2<real4> &spos);

  real4 value_at(sint4 x, sint4 y);
  real4 my_value_at(sint4 x, sint4 y);

  Vec2<real4> pro_vec(sint4 x, sint4 y, real4 dx, real4 dy);
  Vec2<real4> spacing_vec(sint4 x, sint4 y, real4 dx, real4 dy);
  Vec2<real4> anti_vec(sint4 x, sint4 y, real4 dx, real4 dy);
  
  void add_to_squad(GameObj *g, Squad *s, ShooterData::Mode m=ShooterData::ACTIVE);
  void update_squads();

  bool enabled_motion;
  bool enabled_attack;

  //  void compute_motion();
  void compute_attack();
  void compute_squad_motion();

  //-------------------------------------
  // Shooter AI
  //-------------------------------------

  void add_friendly(GameObj*);
  void build_range_graph();
  void removeNode(GameObj* node);
  void recompute_targets();
  
  real8 unitRank(GameObj *enemy);
  real8 unitRandom(/*GameObj *enemy*/);
  sint4 unitHealth(GameObj *enemy);
  sint4 unitAttack(GameObj* unit);
  real8 unitImportance(GameObj *enemy);
  real8 avgEnemyImportance(GameObj *enemy);

  bool cmpAttack(GameObj* a, GameObj* b);
  bool cmpEnemyImportance(GameObj* a, GameObj* b);
  Vector<GameObj*> chooseAttackers(GameObj* unit, Vector<GameObj*> possible_attackers);

  static const real8 overkill_percent = 0.0;


  enum Strategy { LOCAL_STRATEGY=1, GLOBAL_STRATEGY };
  Strategy strategy;

  enum TargetSelectionStrategy { SHOOT_NONE=0, SHOOT_RANDOM, SHOOT_WEAKEST,
                                 SHOOT_IMPORTANT, SHOOT_IMPORTANT_2 };
  TargetSelectionStrategy target_strategy;


  //-------------------------------------

public:

  static const sint4 FROM;
  static const sint4 SET_PRIORITY;
  static const sint4 ATTACK_TARGET;
  static const sint4 ATTACK_LOCATION;
  static const sint4 ATTACK_FREELANCE;
  static const sint4 MOVE_LOCATION;
  static const sint4 STOP_ATTACK;

  static const sint4 ENEMY_IN_RANGE;
  static const sint4 TARGET_DESTROYED;
  static const sint4 AT_LOCATION;
  //  static const sint4 TARGET_SECURED;

  SquadCombatAI(GameStateModule &gsm);
  ~SquadCombatAI();

  bool handle_event(const Event &);

  void update(const GameChanges &changes);

  void enable_attack() { enabled_attack = true; }
  void enable_motion() { enabled_motion = true; }
  void disable_attack() { enabled_attack = false; }
  void disable_motion() { enabled_motion = false; }
  bool is_enabled_attack() const { return enabled_attack; }
  bool is_enabled_motion() const { return enabled_motion; }

  friend class SquadCombatAIWidget;
  friend class SquadCombatAIOverlay;
};

//===================================================================
//===================================================================

class ShooterPriorityEvent : public Event
{
public:
  
  enum Priority { ACTIVE, PASSIVE };

  ShooterPriorityEvent(sint4 who, sint4 what, GameObj *o, Priority p) :
    Event(who, what), obj(o), priority(p) {}

  ~ShooterPriorityEvent() {}
  
  GameObj *obj;
  Priority priority;
};

//===================================================================

class SquadCombatEvent : public Event
{
public:
  
  enum Objective { STOP, ATTACK_TARGET, ATTACK_LOCATION }; // MOVE_TO_POINT

  SquadCombatEvent(sint4 who, sint4 what, const Vector<GameObj*> &u, sint4 x, sint4 y);
  SquadCombatEvent(sint4 who, sint4 what, const Vector<GameObj*> &u, GameObj *targ);
  SquadCombatEvent(sint4 who, sint4 what, const Vector<GameObj*> &u);
  ~SquadCombatEvent() {}
  
  Vector<GameObj*> units;
  sint4 x, y;
  GameObj *target;
  Objective objective;
};

//===================================================================

REGISTER_TYPEOF(2881, Vector<SquadCombatAI::GroupData*>::iterator);
REGISTER_TYPEOF(2882, Vector<SquadCombatAI::GroupData*>::const_iterator);
REGISTER_TYPEOF(2883, SquadCombatAI::ShooterMap::iterator);
REGISTER_TYPEOF(2884, SquadCombatAI::ShooterMap::const_iterator);

#endif
