/** @file TerrainVectorField.C
    @see TerrainVectorField.H

    $Id: TerrainVectorField.C,v 1.3 2006/06/12 18:33:30 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvssoftware/orts3loc/libs/ai/squad/src/TerrainVectorField.C,v $
*/

// This is an ORTS file
// (c) Tim Furtak
// licensed under the GPL

#include "Global.H"
#include "TerrainVectorField.H"
#include "DrawOnTerrain.H"
#include "Game.H"
#include "GameChanges.H"
#include "GameStateModule.H"
#include "Profiler.H"

using namespace std;

//===================================================================

TerrainVectorField::TerrainVectorField(GameStateModule &gsm_) : gsm(gsm_)
{
  gsm.add_handler(this);
  marked = 0;
}

//-------------------------------------------------------------------

TerrainVectorField::~TerrainVectorField()
{
  if (marked) delete [] marked;
}

//-------------------------------------------------------------------

bool TerrainVectorField::handle_event(const Event &e)
{
  if (e.get_what() == GameStateModule::VIEW_MSG) {
    update(gsm.get_changes());
    return true;
  }

  return false;
}

//-------------------------------------------------------------------

/** Add new emitters for boundary objects. */

void TerrainVectorField::update(const GameChanges &changes)
{
  const Map<GameTile> &map = gsm.get_game().get_map();
  const sint4 tp = gsm.get_game().get_tile_points();

  mw = map.get_width() * 2 + 1;
  mh = map.get_height() * 2 + 1;

  if (!marked) {
    marked = new sint4[mw * mh];
    memset(marked, 0, sizeof(marked[0])*mw*mh);
  }

  const Vector<ScriptObj*> &boundaries = changes.new_boundaries;
  
  FORALL (boundaries, it) {
    
    const GameObj *g = (*it)->get_GameObj();
    if (g == 0) continue;

    sint4 x1 = (*g->sod.x1 + 8) / tp * 2;
    sint4 y1 = (*g->sod.y1 + 8) / tp * 2;
    sint4 x2 = (*g->sod.x2 + 8) / tp * 2;
    sint4 y2 = (*g->sod.y2 + 8) / tp * 2;

    sint4 t1 = y1 * mw + x1;
    sint4 t2 = y2 * mw + x2;
    marked[t1] = 1;
    marked[t2] = 1;
    marked[(t1+t2) / 2] = 1;

    if (x1 != x2 && y1 != y2) {
      if (x2 < x1) x1 = x2;
      if (y2 < y1) y1 = y2;
      sint4 t3 = (y1+1) * mw + (x1+1);
      marked[t3] = 1;
    }
  }
}

//-------------------------------------------------------------------

Vec2<real4> TerrainVectorField::field(real4 x, real4 y, real4 dx, real4 dy) const
{
  if (!marked) return Vec2<real4>(0,0);

  const sint4 tp = gsm.get_game().get_tile_points();
  const sint4 maxr = 32;

  sint4 x1 = (sint4)floor((x - maxr) / tp);
  sint4 x2 = (sint4)ceil ((x + maxr) / tp);
  sint4 y1 = (sint4)floor((y - maxr) / tp);
  sint4 y2 = (sint4)ceil ((y + maxr) / tp);

  if (x1 < 0) x1 = 0;
  if (y1 < 0) y1 = 0;
  if (x2 > mw) x2 = mw;
  if (y2 > mh) y2 = mh;

  x1 = x1 * 2;
  x2 = x2 * 2;
  y1 = y1 * 2;
  y2 = y2 * 2;

  Vec2<real4> force;

  for (sint4 tx = x1; tx <= x2; ++tx) {
    for (sint4 ty = y1; ty <= y2; ++ty) {
      sint4 t = ty * mw + tx;
      if (marked[t]) {
        real4 px = tx * tp / 2.0;
        real4 py = ty * tp / 2.0;

        VectorField::OrthoEmitter em(px, py, 0, 32, 1.0);
        force += em.field(x, y, dx, dy);
      }
    }
  }

  return force;
}

//-------------------------------------------------------------------

void TerrainVectorField::draw(DrawOnTerrain &dot, const Game &game, const Vec3<real4> &color) const
{
  sint4 tile_points = game.get_tile_points();

  sint4 width = game.get_map().get_width();
  sint4 height = game.get_map().get_height();

  real4 dx = 1.0 / tile_points;
  real4 dy = 1.0 / tile_points;

  {
    dot.begin_lines();

    for (real4 y = 0.0; y <= height; y += 0.25) {
      for (real4 x = 0.0; x <= width; x += 0.25) {

        Vec2<real4> v = field(x*tile_points, y*tile_points);
        if (fabs(v.x) > 0.0001 || fabs(v.y) > 0.0001)
          dot.draw_line(x, y, x+v.x*dx, y+v.y*dy, color, 1.0);

      }
    }
    dot.end_lines();
  }
}

//===================================================================

