/** @file VectorField.C
    @see VectorField.H

    $Id: VectorField.C,v 1.7 2006/06/15 19:47:30 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvssoftware/orts3loc/libs/ai/squad/src/VectorField.C,v $
*/

// This is an ORTS file
// (c) Tim Furtak
// licensed under the GPL

#include "Global.H"
#include "VectorField.H"
#include "DrawOnTerrain.H"
#include "Game.H"
#include "Profiler.H"

using namespace std;

//===================================================================

VectorField::VectorField()
{
}

//-------------------------------------------------------------------

VectorField::~VectorField()
{
  clear();
}

//-------------------------------------------------------------------

void VectorField::add_emitter(Emitter *em)
{
  assert(em);
  emitters.push_back(em);
}

//-------------------------------------------------------------------

void VectorField::clear()
{
  FORALL (emitters, e) {
    delete *e;
  }
  emitters.clear();
}

//-------------------------------------------------------------------

Vec2<real4> VectorField::field(real4 x, real4 y, real4 dx, real4 dy) const
{
  Vec2<real4> vec;
  FORALL (emitters, e) {
    vec += (*e)->field(x, y, dx, dy);
  }
  return vec;
}

//-------------------------------------------------------------------

void VectorField::draw(DrawOnTerrain &dot, const Game &game, const Vec3<real4> &color) const
{
  sint4 tile_points = game.get_tile_points();

  sint4 width = game.get_map().get_width();
  sint4 height = game.get_map().get_height();

  real4 dx = 1.0 / tile_points;
  real4 dy = 1.0 / tile_points;

  {
    PROFILE("draw cells");

    dot.begin_lines();

    for (real4 y = 0.0; y <= height; y += 1.0) {
      for (real4 x = 0.0; x <= width; x += 1.0) {

        //    FORS (y, height+1) {
        //      FORS (x, width+1) {

        Vec2<real4> v = field(x*tile_points, y*tile_points);
        if (fabs(v.x) > 0.0001 || fabs(v.y) > 0.0001)
          dot.draw_line(x, y, x+v.x*dx, y+v.y*dy, color, 1.0);

      }
    }
    dot.end_lines();
  }
}

//===================================================================
//===================================================================

VectorField::PointEmitter::PointEmitter(real4 x, real4 y, real4 radius_, real4 mag_) :
  Emitter(x, y),
  radius(radius_), mag(mag_)
{
  radius_2 = square(radius);
}

//-------------------------------------------------------------------

Vec2<real4> VectorField::PointEmitter::field(real4 px, real4 py, real4, real4) const
{
  Vec2<real4> delta(px - x, py - y);
  real4 d2 = square(delta.x) + square(delta.y);
  if (d2 > radius_2 || d2 <= 1) return Vec2<real4>(0,0);
  delta.normalize();
  return delta * mag;
}

//===================================================================

VectorField::PointEmitter2::PointEmitter2(real4 x, real4 y, real4 r1_, real4 r2_, real4 mag_) :
  Emitter(x, y),
  r1(r1_), r2(r2_), mag(mag_)
{
}

//-------------------------------------------------------------------

Vec2<real4> VectorField::PointEmitter2::field(real4 px, real4 py, real4, real4) const
{
  Vec2<real4> delta(px - x, py - y);
  real4 d = sqrt(square(delta.x) + square(delta.y));
  if (d > r2 || d <= 1) return Vec2<real4>(0,0);
  delta.normalize();
  if (d <= r1) {
    return delta * mag;
  } else {
    return delta * mag * (1.0 - (d - r1) / (r2 - r1));
  }
}


//===================================================================

VectorField::GaussianEmitter::GaussianEmitter
(real4 x, real4 y, real4 r1, real4 r2, real4 mag, real4 decay_) :
  PointEmitter2(x, y, r1, r2, mag),
  decay(decay_)
{
}

//-------------------------------------------------------------------

Vec2<real4> VectorField::GaussianEmitter::field(real4 px, real4 py, real4, real4) const
{
  Vec2<real4> delta(px - x, py - y);
  real4 d = delta.length();
  if (d > r2 || d <= 0.1) return Vec2<real4>(0,0);
  d -= r1;
  if (d < 0.0) d = 0.0;

  real4 d2 = d*d;
  d2 /= 256.0;

  real4 s = decay;
  real4 s2 = s*s;
  const real4 r2pi = 2.506628274;
  real4 v = mag / (s*r2pi) * expf(-d2 / (2*s2));

  delta.normalize();
  return delta * v;
}

//===================================================================

VectorField::DirMultiplier::DirMultiplier(real4 x, real4 y, real4 dx, real4 dy, const Emitter *em_) :
  Emitter(x, y),
  em(em_), dir(dx, dy)
{
  dir.normalize();
}

//-------------------------------------------------------------------

VectorField::DirMultiplier::~DirMultiplier()
{
  if (em) delete em;
}

//-------------------------------------------------------------------

Vec2<real4> VectorField::DirMultiplier::field(real4 px, real4 py, real4, real4) const
{
  Vec2<real4> delta(px - x, py - y);
  if (delta.x == 0 && delta.y == 0) return em->field(px, py);
  delta.normalize();

  real4 dy = dir.dot(delta);
  Vec2<real4> orth = delta - (dir * dy);
  real4 dx = orth.length();

  real4 mult = dx + dy;
  if (mult <= 0.0) return Vec2<real4>(0, 0);

  return em->field(px, py) * mult;
}

//===================================================================

VectorField::OrthoEmitter::OrthoEmitter(real4 x, real4 y, real4 r1_, real4 r2_, real4 mag_) :
  Emitter(x, y),
  r1(r1_), r2(r2_), mag(mag_)
{
}

//-------------------------------------------------------------------

Vec2<real4> VectorField::OrthoEmitter::field(real4 px, real4 py, real4 dx, real4 dy) const
{
  Vec2<real4> delta(px - x, py - y);
  if (fabs(delta.x) <= 0.0001 && fabs(delta.y) <= 0.0001) return Vec2<real4>(0,0);

  Vec2<real4> dir(dx, dy);
  if (fabs(dir.x) <= 0.0001 && fabs(dir.y) <= 0.0001) return Vec2<real4>(0,0);

  real4 dist = delta.length();
  if (dist > r2) return Vec2<real4>(0,0);


  real4 dot = delta.dot(dir);
  if (dot >= 0.0) {
    //    cout << "away" << endl;
    return Vec2<real4>(0,0);
  }

  Vec2<real4> orth(dir.y, -dir.x);



  delta.normalize();
  dir.normalize();

  // determine which side the point is on, to turn it away from the main line
  real4 cross = dir.x * delta.y - dir.y * delta.x;

  //  cout << "cross = " << cross << endl;
  //  cout << "orth = " << orth.x << " " << orth.y << endl;

  if (cross < 0) {
    cross = 1.2 + cross;
  } else {
    cross = -(1.2 - cross);
  }


  dist -= r1;
  if (dist > 0.0) {
    real4 mult = 1.0 - (dist / (r2 - r1));
    orth *= mult;
    //    orth.x -= (1.0 - mult) * dx;
    //    orth.y -= (1.0 - mult) * dy;

    orth *= cross;

  } else {
    //    orth.x -= dx;
    //    orth.y -= dy;
  }

  //  orth *= cross;

  return orth * mag;
}

//===================================================================
