#ifndef SI_SIMULATION_STATE_H
#define SI_SIMULATION_STATE_H

#include <map>
//#include <unistd.h>
#include "timeval.h"

#include "TerrainBase.H"
#include "Vector.H"


#include "SI_Base.h"

#include "SI_Strategies.h"
#include "SI_Player.h"
#include "SI_Group.h"

namespace Simul{

  template <class T> T sgn(const T& x){
    if (x >= 0) return 1;
    return -1;
  }

  // get the current time (in seconds)
  inline real8 getCurrentTime(){
    struct timeval tv;
    gettimeofday(&tv,NULL);
    return (tv.tv_sec + (tv.tv_usec / 1000000.0));
  }

  class SimulationAI;

  struct SavedData{
    bool isDone;
    uint4 ourStrategyIndex;
    uint4 theirStrategyIndex;
    real8 currentValue;
    real8 bestValue;
    TaskCreateFunctionPtr bestStrategy;

    SavedData(){
      isDone = false;
      ourStrategyIndex = 0;
      theirStrategyIndex = 0;
      currentValue = 0.0;
      bestValue = 0.0;
      bestStrategy = NULL;
    }
  };

  typedef SavedData SimulationResult;

  typedef std::map<const uint4,Player*> PlayerMap;

  class SimulationState{
  public:
    SimulationState(SimulationAI* sim){
      simAI = sim;
    }
    ~SimulationState();
    SimulationState(const SimulationState& b);

    SimulationState& operator= (const SimulationState& b);

    SimulationState* clone();
    void cleanup();

    SimulationAI* simAI;


    GameObj* getNearestEnemyBaseOfGroup(uint4 playerID,const ObjectVector& objects);
    GameObj* getNearestEnemyUnitOfGroup(uint4 playerID,const ObjectVector& objects);
    GameObj* getAverageNearestEnemyUnitOfGroup(uint4 playerID,const ObjectVector& objects);

    TerrainBase::Loc getCenterOfObjects(const ObjectVector& objects);
    GameObj* getNearestObject(const ObjectVector& objects, const TerrainBase::Loc& loc);
    GameObj* getNearestObject(const ObjectVector& objects,GameObj* obj);

    // populates X nearest objects to a specified object
    void getNearestObjectsToObject(ObjectVector& closestObjects,ObjectVector originalObjects,GameObj* obj,uint4 numObjects);

    bool removeObjectFromVector(GameObj* obj, ObjectVector& vector);
    bool subtractVector(const ObjectVector& vectorToSubtract,ObjectVector& vector);


    bool enemyInRange(Group* group);
    Player* getPlayer(uint4 playerID);

    void addPlayer(Player* player){
      if(player == NULL || getPlayer(player->getID()) != NULL){
        ERR("error in creating new player");
      }
      players[player->getID()] = player;
    }

    void setOurID(uint4 id){ ourPlayerID = id;}
    uint4 getOurID() { return ourPlayerID;}

    PlayerMap& getPlayers(){return players;}

    void updateObjects(real8 lastMoveUpdateMillis, real8 currUpdateMillis);

    TaskCreateFunctionPtr findBestStrategy(
                                            uint4 playerID,
                                            real8 currentTime,
                                            real8 allowedComputationTime,
                                            real8 timeToSimulate,
                                            SimulationResult& simResult);

    real8 advanceTime(real8 lastTime);

    bool isGameOver();
    bool hasWon(uint4 playerID);

    void updateStrategies();

    uint4 groupSize;
    uint4 smartJoinSize;
    bool smartJoin;
    bool useMoveForAttack;
    bool findspot;

  private:
    SimulationState();

    PlayerMap players;
    uint4 ourPlayerID;

    real8 evaluate(uint4 playerID);
    real8 getNextTime(real8 currentTime);
    void updateOrders();

    void moveObjects(real8 timeElapsed);
    void resolveCollisions(real8 currentTime);
    bool solveQuadratic(real8 a,real8 b,real8 c,real8& t1,real8& t2);
    real8 getCirclePointCollisionTime(const Vector3D& circlePos,
                                       real8 radius,
                                       const Vector3D& pointPos,
                                       const Vector3D& relativeHeading,
                                       real8 startTime);

    real8 getNextCollisionTime(real8 currentTime);
    real8 getNextOrderCompletionTime(real8 currentTime);
    real8 getStrategyValue(
                             uint4 playerID,
                             TaskCreateFunctionPtr ourStrategy,
                             TaskCreateFunctionPtr theirStrategy,
                             real8 currentTime,
                             real8 timeToSimulate);

    real8 simulateForTime(
                            uint4 playerID,
                            real8 currentTime,
                            real8 timeToSimulate);

    void copyPlayerInfo(Player* origPlayer,SimulationState* tempState);

  };


}

REGISTER_TYPEOF(2861, Simul::PlayerMap::iterator);
REGISTER_TYPEOF(2862, Simul::PlayerMap::const_iterator);


#endif



