// $Id: AI_Plan.C,v 1.2 2004/10/14 23:43:36 orts_mburo Exp $

// This is an ORTS file (c) Michael Chung, licensed under the GPL

#include "Global.H"

#include "AI_Board.H"
#include "AI_Common.H"
#include "AI_PathFind.H"
#include "AI_Plan.H"

using namespace std;

//---------------------------------------------------------------------
// AI_Plan Identifiers
//---------------------------------------------------------------------

static const struct {
  char* ai_name;
  sint4 ai_id;
} ai_id_map[] = {
  {"stand", AI_PLAN_STAND},
  {"random", AI_PLAN_RANDOM},
  {"randsim", AI_PLAN_RANDSIM},
  {"flag", AI_PLAN_FLAG},
  {"explore", AI_PLAN_EXPLORE},
  {"all", AI_PLAN_ALL},
  {"mc", AI_PLAN_MC},
  {"group", AI_PLAN_GROUP},
  {"defend", AI_PLAN_DEFEND}
};

string AI_Plan::get_str_from_id(sint4 id)
{
  for (uint4 i = 0; i < sizeof(ai_id_map) / sizeof(ai_id_map[0]); i++) {
    if (id == ai_id_map[i].ai_id) {
      return ai_id_map[i].ai_name;
    }
  }
  
  return "";
}

sint4 AI_Plan::get_id_from_str(const string &s)
{
  for (uint4 i = 0; i < sizeof(ai_id_map) / sizeof(ai_id_map[0]); i++) {
    if (s == ai_id_map[i].ai_name) {
      return ai_id_map[i].ai_id;
    }
  }
  
  return AI_PLAN_UNKNOWN;
}


//---------------------------------------------------------------------
// AI_Path implementation
//---------------------------------------------------------------------

void AI_Path::set(int_vector path, sint4 width)
{
  m_path = path;
  m_width = width;
}

bool AI_Path::get_next_waypoint(Point &pt) const
{
  // return last value = next waypoint
  if (m_path.empty())
    return false;
    
  pt = id_to_point(m_path.back());
  
  return true;
}

// this function is called when a waypoint is reached, to remove the
// waypoint (and so the next call to get_next_waypoint returns the next)
void AI_Path::pop_next_waypoint()
{
  // pop last value = next waypoint
  if (m_path.empty())
    return;
  
  m_path.pop_back();
}


//---------------------------------------------------------------------
// AI_Plan implementation
//---------------------------------------------------------------------

AI_Plan::AI_Plan()
:
  m_paths(),

  m_n_time(-1),
  m_b_completed(false),

  m_plan_id(0)
{
}

void AI_Plan::clear()
{
  m_paths.clear();
  
  m_n_time = -1;
  m_b_completed = false;
}

bool AI_Plan::plan_path(
  sint4 obj_id, 
  const AI_Board &bd, 
  Point pt_start, 
  Point pt_dest)
{
  if (bd.invalid_point(pt_start) || bd.invalid_point(pt_dest)) {
    ERR("invalid parameter");
    return false;
  }
  
  AI_Path path;
  if (!find_path(path, bd, pt_start, pt_dest)) {
    ERR("path not found");
    return false;
  }

  add_path(obj_id, path);
  
  return true;
}

/** get (pointer to) the path for the object id given */
const AI_Path *AI_Plan::get_path_c(sint4 id) const 
{ 
  std::map<sint4, AI_Path>::const_iterator iter = m_paths.find(id);
  
  return (iter == m_paths.end()) ? NULL : &(iter->second);
}

AI_Path *AI_Plan::get_path(sint4 id) 
{ 
  std::map<sint4, AI_Path>::iterator iter = m_paths.find(id);
  
  return (iter == m_paths.end()) ? NULL : &(iter->second);
}

Point AI_Plan::get_dest(sint4 id) const 
{
  Point pt(-1, -1);
  
  const AI_Path *path = get_path_c(id);
  if (NULL != path) {
    path->get_point(pt, 0);
  }
  
  return pt;
}

bool AI_Plan::is_completed(const AI_Board &bd) const
{
  // plan is complete unless there is an object in the plan that has
  // not yet reached its destination
  
  if (is_empty())
    return true;  // if plan is empty - then it's complete
  
  const object_map &objs = bd.get_friendly_objs();
  object_map::const_iterator iter = objs.begin();
  for (; iter != objs.end(); iter++) {
    
    const AI_Object &obj = iter->second;
    if (!obj.is_unit())
      continue;
    
    sint4 obj_id = obj.id();
    if (!contains(obj_id))
      continue;
    
    Point pt_dest = get_dest(obj_id);
    if (pt_dest.x < 0)
      continue;  // ignore invalid destinations...
    
    if (!is_equal(pt_dest, obj.location()))
      return false;  // plan not completed
  }
  
  return true;
}

void AI_Plan::generate_stand_plan()
{
  clear();
  
  // plan is meant to be empty
  
  m_plan_id = AI_PLAN_STAND;
}

void AI_Plan::generate_random_plan(
  const AI_Board &bd, 
  sint4 max_dist, 
  sint4 side)
{
  clear();

  const object_map &objs = bd.get_objs(side);
  
  object_map::const_iterator iter = objs.begin();
  for (; iter != objs.end(); iter++) {
    
    const AI_Object &obj = iter->second;
    if (!obj.is_unit())
      continue;  // only move units
    
    sint4 obj_id = obj.id();
    Point pt_start = obj.location();
    Point pt_dest = get_random_dest(pt_start, bd, max_dist);
    
	//cout << "plan part:" << endl;
	//DB_PT(pt_start);
	//DB_PT(pt_dest);
	
    plan_path(obj_id, bd, pt_start, pt_dest);
  }

  //cout << "random plan generated" << endl;
}

bool AI_Plan::generate_flag_plan(const AI_Board &bd, sint4 side)
{
  clear();
  m_plan_id = AI_PLAN_FLAG;
  
  sint4 enemy_side = OTHER_SIDE(side);

  const object_map &objs = bd.get_objs(side);
  
  Point pt_enemy_flag;
  if (!bd.get_flag_location(pt_enemy_flag, enemy_side)) ERR("!");

  Point pt_flag_loc = pt_enemy_flag;
  
  object_map::const_iterator iter = objs.begin();
  for (; iter != objs.end(); iter++) {
    
    const AI_Object &obj = iter->second;
    if (!obj.is_unit())
      continue;  // only move units
    
    sint4 obj_id = obj.id();
    Point pt_start = obj.location();
    Point pt_dest(-1, -1);
    
    if (!obj.has_flag()) {
      // if we don't have it, go towards the enemy flag
      pt_dest = pt_flag_loc;

    } else {
      // if we are carrying the enemy flag, return to home base
      Point pt_home_base;
      if (bd.get_post_location(pt_home_base, side)) {
        pt_dest = pt_home_base;
      } else {
        // I don't think this ever happens...
        ERR("home base location unknown");
      }
    }
    
    //cout << "adding to plan" << endl;
    
    plan_path(obj_id, bd, pt_start, pt_dest);
    
    // edit pt_flag_loc to point to target location to capture
    // or protect the flag
    const AI_Path *path = get_path_c(obj_id);
    if (NULL != path) {
      // get point near goal... (2 away)
      Point pt;
      path->get_point(pt, 2);
      if (pt.x >= 0) {
        pt_flag_loc = pt;
      }
    }
  }

  return true;
}

bool AI_Plan::generate_explore_plan(const AI_Board &bd)
{
  clear();
  m_plan_id = AI_PLAN_EXPLORE;
  
  const object_map &objs = bd.get_friendly_objs();
  
  object_map::const_iterator iter = objs.begin();
  for (; iter != objs.end(); iter++) {
    
    const AI_Object &obj = iter->second;
    if (!obj.is_unit())
      continue;  // only move units
    
    sint4 obj_id = obj.id();
    Point pt_start = obj.location();
    
    Point pt_dest;
    if (!bd.get_nearby_unexplored_point(pt_dest, pt_start))
      continue;
    
    plan_path(obj_id, bd, pt_start, pt_dest);
  }

  return true;
}

bool AI_Plan::generate_defend_plan(const AI_Board &bd, sint4 side)
{
  clear();
  m_plan_id = AI_PLAN_DEFEND;

  // defend this flag...
  Point pt_flag;
  if (!bd.get_flag_location(pt_flag, side))
    return false;
  
  const object_map &objs = bd.get_objs(side);
  
  object_map::const_iterator iter = objs.begin();
  for (; iter != objs.end(); iter++) {
    
    const AI_Object &obj = iter->second;
    if (!obj.is_unit())
      continue;  // only move units
    
    sint4 obj_id = obj.id();
    Point pt_start = obj.location();
    
    // go to randomized location near the flag
    Point pt_dest;
    if (!bd.get_nearby_empty_point(pt_dest, pt_flag))
      continue;
    
    plan_path(obj_id, bd, pt_start, pt_dest);
  }

  return true;
}

bool AI_Plan::generate_group_plan(const AI_Board &bd)
{
  clear();
  m_plan_id = AI_PLAN_GROUP;

  const object_map &objs = bd.get_friendly_objs();
  
  sint4 leader_id = -1;
  Point leader_loc;
  
  sint4 i = 0;
  
  object_map::const_iterator iter = objs.begin();
  for (; iter != objs.end(); iter++) {
    
    const AI_Object &obj = iter->second;
    if (!obj.is_unit())
      continue;  // only move units
    
    sint4 obj_id = obj.id();
    Point pt_start = obj.location();
    
    if (leader_id < 0) {
      // assign a leader
      leader_id = obj_id;
      leader_loc = pt_start;
      
      // leader doesn't move
      continue;
    }
    
    // move to the leader
    Point pt_dest = leader_loc;
    
    switch (i++) {
    // we must edit the point to avoid "clumping" at the destination
    case 0:
      pt_dest.x--;
      pt_dest.y--;
      break;
    
    case 1:
      pt_dest.x++;
      pt_dest.y--;
      break;

    case 2:
      pt_dest.x--;
      pt_dest.y++;
      break;

    case 3:
      pt_dest.x++;
      pt_dest.y++;
      break;

    default:
      ERR("unexpected case");
      break;
    }
    
    plan_path(obj_id, bd, pt_start, pt_dest);
  }

  return true;
}

bool AI_Plan::generate_plan(sint4 plan_id, const AI_Board &bd, sint4 side)
{
  switch (plan_id) {
    
  case AI_PLAN_FLAG:
    return generate_flag_plan(bd, side);
    
  case AI_PLAN_EXPLORE:
    return generate_explore_plan(bd);
  
  case AI_PLAN_GROUP:
    return generate_group_plan(bd);
  
  case AI_PLAN_STAND:
    generate_stand_plan();
    return true;
  
  case AI_PLAN_DEFEND:
    return generate_defend_plan(bd, side);
  
  default:
    ERR("unhandled plan id");
    break;
  }
  
  return false;
}


//---------------------------------------------------------------------
// Private methods
//---------------------------------------------------------------------

Point AI_Plan::get_random_dest(const Point &pt, const AI_Board &bd, sint4 d)
{
  // generate a nearby random point to move to

  Point ptDest;
  
  const uint4 max_tries = 100;
  uint4 i = 0;
  while (true) {
    // select random point
    ptDest = pt;
    
    // todo: limit manhattan distance to d=max_dist...
    
    ptDest.x += (Rand.rand_uint4() % (2 * d + 1)) - d;
    ptDest.y += (Rand.rand_uint4() % (2 * d + 1)) - d;
    
    // test the point
    if (bd.area_empty(ptDest.x, ptDest.y))
      return ptDest;  // return this point

    if (++i > max_tries) {
      // handle this error by not moving
	  //cout << "warning: max_tries reached" << endl;
      return pt;
    }
  }
  
  ERR("unreachable code");
}


// Local Variables:
// c-file-style:"gnu"
// c-basic-offset:2
// indent-tabs-mode:nil
// tab-width:4
// End:
